// Copyright (C) 2009-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/ff553354.aspx
// https://msdn.microsoft.com/en-us/library/ff563638.aspx
// https://msdn.microsoft.com/en-us/library/ms648409.aspx

// https://msdn.microsoft.com/en-us/library/4141z1cx.aspx

#pragma once

#ifndef __NTDLL_H__
#define __NTDLL_H__

#if 0
#pragma comment(linker, "/DEFAULTLIB:NTDLL.LIB")
#elif 0
#pragma comment(linker, "/DEFAULTLIB:NTDLLP.LIB")
#endif

#ifndef _VC_NODEFAULTLIB
#pragma message("Use '/Zl' compiler switch!")
#endif

#if ANYSIZE_ARRAY != 1
#error ANYSIZE_ARRAY must be defined as 1!
#endif

#ifndef CDECL
#define CDECL		__cdecl
#endif
#define FASTCALL	__fastcall

typedef	LONG		NTSTATUS;
typedef	LONG		SECURITY_STATUS;
typedef	ULONGLONG	QWORD, *LPQWORD;
typedef	ULONGLONG	REGHANDLE;
typedef	ULONGLONG	TRACEHANDLE;

typedef	double		DOUBLE;

// https://msdn.microsoft.com/en-us/library/aa374912.aspx

typedef	struct	_ace
{
	// https://msdn.microsoft.com/en-us/library/aa374919.aspx

	ACE_HEADER	Header;

	// https://msdn.microsoft.com/en-us/library/aa374892.aspx

	ACCESS_MASK	Mask;

	// https://msdn.microsoft.com/en-us/library/aa379594.aspx

	SID		Trustee;
} ACE;

// https://msdn.microsoft.com/en-us/library/aa374931.aspx

#ifndef SECURITY_MAX_ACL_SIZE
#define SECURITY_MAX_ACL_SIZE	65536U
#endif

// https://msdn.microsoft.com/en-us/library/xcb2z8hs.aspx

#ifndef STATUS_CPP_EXCEPTION
#define STATUS_CPP_EXCEPTION			0x406D1388UL
#endif

// https://msdn.microsoft.com/en-us/library/dd941688.aspx

#ifndef STATUS_FAIL_FAST_EXCEPTION
#define STATUS_FAIL_FAST_EXCEPTION		0xC0000602UL
#endif

// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

#ifndef STATUS_CPP_DELAY_LOAD_MODULE_NOT_FOUND
#define STATUS_CPP_DELAY_LOAD_MODULE_NOT_FOUND	0xC06D007EUL
#endif

#ifndef STATUS_CPP_DELAY_LOAD_ENTRY_NOT_FOUND
#define STATUS_CPP_DELAY_LOAD_ENTRY_NOT_FOUND	0xC06D007FL
#endif

// https://msdn.microsoft.com/en-us/library/ff558784.aspx

#ifndef STATUS_APPLICATION_HANG
#define STATUS_APPLICATION_HANG			0xCFFFFFFFUL
#endif

#ifndef STATUS_CLR_EXCEPTION
#define STATUS_CLR_EXCEPTION			0xE0434F4DUL	// = 0xE0000000UL + 'COM'
#endif

// https://support.microsoft.com/en-us/kb/185294

#ifndef STATUS_CPP_EH_EXCEPTION
#define STATUS_CPP_EH_EXCEPTION			0xE06D7363UL	// = 0xE0000000UL + 'msc'
#endif

#define NT_FAILURE(Status)	(((NTSTATUS) (Status)) < 0L)

// https://msdn.microsoft.com/en-us/library/ff565436.aspx

#ifndef NT_SUCCESS
#define NT_SUCCESS(Status)	(((NTSTATUS) (Status)) >= 0L)
#endif

#ifndef NT_INFORMATION
#define NT_INFORMATION(Status)	((((NTSTATUS) (Status)) & 0xC0000000UL) == 0x40000000UL)
#endif

#ifndef NT_WARNING
#define NT_WARNING(Status)	((((NTSTATUS) (Status)) & 0xC0000000UL) == 0x80000000UL)
#endif

#ifndef NT_ERROR
#define NT_ERROR(Status)	((((NTSTATUS) (Status)) & 0xC0000000UL) == 0xC0000000UL)
#endif

#if _WIN32_WINNT >= 0x0600
// NOTE: hash functions missing in NTDLL.LIB, use NTDLLV.LIB instead!

#define A_SHA_DIGEST_LEN	20

typedef	struct	_A_SHA_CTX
{
	BYTE	Buffer[64];
	DWORD	State[5];	// = {0x67452301UL, 0xEFCDAB89UL, 0x98BADCFEUL, 0x10325476UL, 0xC3D2E1F0UL}
	DWORD	Count[2];	// = {0UL, 0UL}
} A_SHA_CTX, *LPA_SHA_CTX, *PA_SHA_CTX;

// https://msdn.microsoft.com/en-us/library/mt806283.aspx
// NOTE: forwarded from ADVAPI32!A_SHAFinal

__declspec(dllimport)
VOID	NTAPI	A_SHAFinal(A_SHA_CTX *Context,
		           LPBYTE    Digest);

// https://msdn.microsoft.com/en-us/library/mt806284.aspx
// NOTE: forwarded from ADVAPI32!A_SHAInit

__declspec(dllimport)
VOID	NTAPI	A_SHAInit(A_SHA_CTX *Context);

// https://msdn.microsoft.com/en-us/library/mt806285.aspx
// NOTE: forwarded from ADVAPI32!A_SHAUpdate

__declspec(dllimport)
VOID	NTAPI	A_SHAUpdate(A_SHA_CTX *Context,
		            LPCVOID   Buffer,
		            DWORD     BufferSize);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff540605.aspx

typedef	struct	_ANSI_STRING
{
	WORD	Length;
	WORD	MaximumLength;
	LPSTR	Buffer;
} ANSI_STRING, *LPANSI_STRING, *PANSI_STRING;

// https://msdn.microsoft.com/en-us/library/ff558741.aspx

typedef	struct	_OEM_STRING
{
	WORD	Length;
	WORD	MaximumLength;
	LPSTR	Buffer;
} OEM_STRING, *LPOEM_STRING, *POEM_STRING;

// https://msdn.microsoft.com/en-us/library/ms648424.aspx

typedef	struct	_STRING
{
	WORD	Length;
	WORD	MaximumLength;
	LPSTR	Buffer;
} STRING, *LPSTRING, *PSTRING;

// https://msdn.microsoft.com/en-us/library/aa380518.aspx
// https://msdn.microsoft.com/en-us/library/ff564879.aspx

typedef	struct	_UNICODE_STRING
{
	WORD	Length;
	WORD	MaximumLength;
	LPWSTR	Buffer;
} UNICODE_STRING, *LPUNICODE_STRING, *PUNICODE_STRING;

#if _WIN32_WINNT >= 0x0A00
// https://msdn.microsoft.com/en-us/library/dn986880.aspx

__declspec(dllimport)
BOOL	WINAPI	ApiSetQueryApiSetPresence(UNICODE_STRING *Namespace,
		                          BOOLEAN        *Present);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0602
typedef	struct	_API_SET_HASH_ENTRY
{
	DWORD	Hash;
	DWORD	Index;
} API_SET_HASH_ENTRY, *LPAPI_SET_HASH_ENTRY, *PAPI_SET_HASH_ENTRY;

typedef	struct	_API_SET_VALUE_ENTRY
{
	DWORD	Flags;
	DWORD	NameOffset;
	DWORD	NameLength;
	DWORD	ValueOffset;
	DWORD	ValueLength;
} API_SET_VALUE_ENTRY, *LPAPI_SET_VALUE_ENTRY, *PAPI_SET_VALUE_ENTRY;

typedef	struct	_API_SET_NAMESPACE_ENTRY
{
	DWORD	Flags;
	DWORD	NameOffset;
	DWORD	NameLength;
	DWORD	HashedLength;
	DWORD	ValueOffset;
	DWORD	ValueCount;
} API_SET_NAMESPACE_ENTRY, *LPAPI_SET_NAMESPACE_ENTRY, *PAPI_SET_NAMESPACE_ENTRY;

typedef	struct	_API_SET_NAMESPACE
{
	DWORD	Version;
	DWORD	Size;
	DWORD	Flags;
	DWORD	Count;
	DWORD	EntryOffset;
	DWORD	HashOffset;
	DWORD	HashFactor;
} API_SET_NAMESPACE, *LPAPI_SET_NAMESPACE, *PAPI_SET_NAMESPACE;

LONG	NTAPI	ApiSetResolveToHost(API_SET_NAMESPACE *ApiSetNameSpace,
		                    UNICODE_STRING    *ApiSetName,
		                    UNICODE_STRING    *ParentName,
		                    BOOLEAN           *Resolved,
		                    UNICODE_STRING    *HostModuleName);
#endif // _WIN32_WINNT

__declspec(dllimport)
DWORD	NTAPI	CsrGetProcessId(VOID);

// https://msdn.microsoft.com/en-us/library/ff543626.aspx

__declspec(dllimport)
VOID	NTAPI	DbgBreakPoint(VOID);

// https://msdn.microsoft.com/en-us/library/ff543632.aspx

__declspec(dllimport)
DWORD	CDECL	DbgPrint(LPCSTR Format,
		         ...);

// https://msdn.microsoft.com/en-us/library/ff543634.aspx

__declspec(dllimport)
DWORD	CDECL	DbgPrintEx(DWORD  ComponentId,
		           DWORD  Level,
		           LPCSTR Format,
		           ...);

__declspec(dllimport)
DWORD	CDECL	DbgPrintReturnControlC(LPCSTR Format,
		                       ...);

// https://msdn.microsoft.com/en-us/library/ff543635.aspx

__declspec(dllimport)
DWORD	NTAPI	DbgPrompt(LPCSTR Prompt,
		          LPSTR  Response,
		          DWORD  MaximumResponseLength);

__declspec(dllimport)
LONG	NTAPI	DbgQueryDebugFilterState(DWORD ComponentId,
		                         DWORD Level);

__declspec(dllimport)
LONG	NTAPI	DbgSetDebugFilterState(DWORD   ComponentId,
		                       DWORD   Level,
		                       BOOLEAN State);

__declspec(dllimport)
LONG	NTAPI	DbgUiConnectToDbg(VOID);

typedef	struct	_CLIENT_ID
{
	DWORD_PTR	UniqueProcess;
	DWORD_PTR	UniqueThread;
} CLIENT_ID, *LPCLIENT_ID, *PCLIENT_ID;

__declspec(dllimport)
LONG	NTAPI	DbgUiContinue(CLIENT_ID *AppClientId,
		              NTSTATUS  ContinueStatus);

typedef	struct	_DBGKM_EXCEPTION
{
	EXCEPTION_RECORD	ExceptionRecord;
	DWORD			FirstChance;
} DBGKM_EXCEPTION, *LPDBGKM_EXCEPTION, *PDBGKM_EXCEPTION;

typedef	struct	_DBGKM_CREATE_THREAD
{
	DWORD	SubSystemKey;
	LPVOID	StartAddress;
} DBGKM_CREATE_THREAD, *LPDBGKM_CREATE_THREAD, *PDBGKM_CREATE_THREAD;

typedef	struct	_DBGKM_CREATE_PROCESS
{
	DWORD			SubSystemKey;
	HANDLE			FileHandle;
	LPVOID			BaseOfImage;
	DWORD			DebugInfoFileOffset;
	DWORD			DebugInfoSize;
	DBGKM_CREATE_THREAD	InitialThread;
} DBGKM_CREATE_PROCESS, *LPDBGKM_CREATE_PROCESS, *PDBGKM_CREATE_PROCESS;

typedef	struct	_DBGKM_EXIT_PROCESS
{
	NTSTATUS	ExitStatus;
} DBGKM_EXIT_PROCESS, *LPDBGKM_EXIT_PROCESS, *PDBGKM_EXIT_PROCESS;

typedef	struct	_DBGKM_EXIT_THREAD
{
	NTSTATUS	ExitStatus;
} DBGKM_EXIT_THREAD, *LPDBGKM_EXIT_THREAD, *PDBGKM_EXIT_THREAD;

typedef	struct	_DBGKM_LOAD_DLL
{
	HANDLE	FileHandle;
	LPVOID	BaseOfDll;
	DWORD	DebugInfoFileOffset;
	DWORD	DebugInfoSize;
	LPVOID	NamePointer;
} DBGKM_LOAD_DLL, *LPDBGKM_LOAD_DLL, *PDBGKM_LOAD_DLL;

typedef	struct	_DBGKM_UNLOAD_DLL
{
	LPVOID	BaseAddress;
} DBGKM_UNLOAD_DLL, *LPDBGKM_UNLOAD_DLL, *PDBGKM_UNLOAD_DLL;

typedef	enum	_DBG_STATE
{
	DbgIdle                     = 0,
	DbgReplyPending,
	DbgCreateThreadStateChange,
	DbgCreateProcessStateChange,
	DbgExitThreadStateChange,
	DbgExitProcessStateChange,
	DbgExceptionStateChange,
	DbgBreakpointStateChange,
	DbgSingleStepStateChange,
	DbgLoadDllStateChange,
	DbgUnloadDllStateChange
} DBG_STATE;

typedef	struct	_DBGUI_CREATE_PROCESS
{
	HANDLE			HandleToProcess;
	HANDLE			HandleToThread;
	DBGKM_CREATE_PROCESS	NewProcess;
} DBGUI_CREATE_PROCESS, *LPDBGUI_CREATE_PROCESS, *PDBGUI_CREATE_PROCESS;

typedef	struct	_DBGUI_CREATE_THREAD
{
	HANDLE			HandleToThread;
	DBGKM_CREATE_THREAD	NewThread;
} DBGUI_CREATE_THREAD, *LPDBGUI_CREATE_THREAD, *PDBGUI_CREATE_THREAD;

typedef	struct	_DBGUI_WAIT_STATE_CHANGE
{
	DBG_STATE	NewState;
	CLIENT_ID	AppClientId;
	union
	{
		DBGKM_EXCEPTION		Exception;
		DBGUI_CREATE_THREAD	CreateThread;
		DBGUI_CREATE_PROCESS	CreateProcessInfo;
		DBGKM_EXIT_THREAD	ExitThread;
		DBGKM_EXIT_PROCESS	ExitProcess;
		DBGKM_LOAD_DLL		LoadDll;
		DBGKM_UNLOAD_DLL	UnloadDll;
	} StateInfo;
} DBGUI_WAIT_STATE_CHANGE, *LPDBGUI_WAIT_STATE_CHANGE, *PDBGUI_WAIT_STATE_CHANGE;

typedef	struct	_DEBUG_EVENT	DEBUG_EVENT, *LPDEBUG_EVENT, *PDEBUG_EVENT;

__declspec(dllimport)
LONG	NTAPI	DbgUiConvertStateChangeStructure(DBGUI_WAIT_STATE_CHANGE *StateChange,
		                                 DEBUG_EVENT             *DebugEvent);

__declspec(dllimport)
LONG	NTAPI	DbgUiDebugActiveProcess(HANDLE ProcessHandle);

__declspec(dllimport)
HANDLE	NTAPI	DbgUiGetThreadDebugObject(VOID);

__declspec(dllimport)
LONG	NTAPI	DbgUiIssueRemoteBreakin(HANDLE ProcessHandle);

__declspec(dllimport)
VOID	NTAPI	DbgUiRemoteBreakin(LPVOID Context);

__declspec(dllimport)
VOID	NTAPI	DbgUiSetThreadDebugObject(HANDLE DebugObject);

__declspec(dllimport)
LONG	NTAPI	DbgUiStopDebugging(HANDLE ProcessHandle);

__declspec(dllimport)
LONG	NTAPI	DbgUiWaitStateChange(DBGUI_WAIT_STATE_CHANGE *StateChange,
		                     LARGE_INTEGER           *Timeout);

__declspec(dllimport)
VOID	NTAPI	DbgUserBreakPoint(VOID);

// https://msdn.microsoft.com/en-us/library/ff556075.aspx

__declspec(dllimport)
DWORD	NTAPI	vDbgPrintEx(DWORD   ComponentId,
		            DWORD   Level,
		            LPCSTR  Format,
		            va_list arglist);

// https://msdn.microsoft.com/en-us/library/ff556076.aspx

__declspec(dllimport)
DWORD	NTAPI	vDbgPrintExWithPrefix(LPCSTR  Prefix,
		                      DWORD   ComponentId,
		                      DWORD   Level,
		                      LPCSTR  Format,
		                      va_list arglist);

// https://msdn.microsoft.com/en-us/library/aa363766.aspx

typedef	struct	_EVENT_INSTANCE_INFO
{
	HANDLE	RegHandle;
	DWORD	InstanceId;
} EVENT_INSTANCE_INFO, *LPEVENT_INSTANCE_INFO, *PEVENT_INSTANCE_INFO;

// https://msdn.microsoft.com/en-us/library/aa363698.aspx
// NOTE: forwarded from ADVAPI32!CreateTraceInstanceId

__declspec(dllimport)
DWORD	NTAPI	EtwCreateTraceInstanceId(HANDLE              RegHandle,
		                         EVENT_INSTANCE_INFO *InstInfo);

// https://msdn.microsoft.com/en-us/library/aa363720.aspx
// NOTE: forwarded from ADVAPI32!EventActivityIdControl

__declspec(dllimport)
DWORD	NTAPI	EtwEventActivityIdControl(DWORD  ControlCode,
		                          LPGUID ActivityId);

// https://msdn.microsoft.com/en-us/library/aa363754.aspx

typedef	struct	_EVENT_DESCRIPTOR
{
	WORD	Id;
	BYTE	Version;
	BYTE	Channel;
	BYTE	Level;
	BYTE	Opcode;
	WORD	Task;
	QWORD	Keyword;
} EVENT_DESCRIPTOR, *LPEVENT_DESCRIPTOR, *PEVENT_DESCRIPTOR;

// https://msdn.microsoft.com/en-us/library/aa363741.aspx
// NOTE: forwarded from ADVAPI32!EventEnabled

__declspec(dllimport)
BOOLEAN	NTAPI	EtwEventEnabled(REGHANDLE        RegHandle,
		                EVENT_DESCRIPTOR *EventDescriptor);

// https://msdn.microsoft.com/en-us/library/aa363742.aspx
// NOTE: forwarded from ADVAPI32!EventProviderEnabled

__declspec(dllimport)
BOOLEAN	NTAPI	EtwEventProviderEnabled(REGHANDLE RegHandle,
		                        BYTE      Level,
		                        QWORD     Keyword);

// https://msdn.microsoft.com/en-us/library/aa363758.aspx

typedef	struct	_EVENT_FILTER_DESCRIPTOR
{
	QWORD	Ptr;
	DWORD	Size;
	DWORD	Type;
} EVENT_FILTER_DESCRIPTOR, *LPEVENT_FILTER_DESCRIPTOR, *PEVENT_FILTER_DESCRIPTOR;

// https://msdn.microsoft.com/en-us/library/aa363707.aspx

typedef	VOID	(NTAPI	ENABLECALLBACK) (LPCGUID                 SourceId,
			                 DWORD                   IsEnabled,
			                 BYTE                    Level,
			                 QWORD                   MatchAnyKeyword,
			                 QWORD                   MatchAllKeywords,
			                 EVENT_FILTER_DESCRIPTOR *FilterData,
			                 LPVOID                  CallbackContext);

// https://msdn.microsoft.com/en-us/library/aa363744.aspx
// NOTE: forwarded from ADVAPI32!EventRegister

__declspec(dllimport)
DWORD	NTAPI	EtwEventRegister(LPCGUID        ProviderId,
		                 ENABLECALLBACK *EnableCallback,
		                 LPVOID         CallbackContext,
		                 REGHANDLE      *RegHandle);

// https://msdn.microsoft.com/en-us/library/hh447257.aspx

typedef	enum	_EVENT_INFO_CLASS
{
	EventProviderBinaryTrackInfo   = 0,
	EventProviderSetTraits         = 1,
	EventProviderUseDescriptorType = 2,
	MaxEventInfo                   = 3
} EVENT_INFO_CLASS;

// https://msdn.microsoft.com/en-us/library/hh447256.aspx
// NOTE: forwarded from ADVAPI32!EventSetInformation

__declspec(dllimport)
DWORD	NTAPI	EtwEventSetInformation(REGHANDLE        RegHandle,
		                       EVENT_INFO_CLASS InformationClass,
		                       LPVOID           EventInformation,
		                       DWORD            InformationLength);

// https://msdn.microsoft.com/en-us/library/aa363749.aspx
// NOTE: forwarded from ADVAPI32!EventUnregister

__declspec(dllimport)
DWORD	NTAPI	EtwEventUnregister(REGHANDLE RegHandle);

// https://msdn.microsoft.com/en-us/library/aa363753.aspx

typedef	struct	_EVENT_DATA_DESCRIPTOR
{
	QWORD	Ptr;
	DWORD	Size;
	DWORD	Type;
} EVENT_DATA_DESCRIPTOR, *LPEVENT_DATA_DESCRIPTOR, *PEVENT_DATA_DESCRIPTOR;

// https://msdn.microsoft.com/en-us/library/aa363752.aspx
// NOTE: forwarded from ADVAPI32!EventWrite

__declspec(dllimport)
DWORD	NTAPI	EtwEventWrite(REGHANDLE             RegHandle,
		              EVENT_DESCRIPTOR      *EventDescriptor,
		              DWORD                 UserDataCount,
		              EVENT_DATA_DESCRIPTOR *UserData);

// NOTE: forwarded from ADVAPI32!EventWriteEndScenario

__declspec(dllimport)
DWORD	NTAPI	EtwEventWriteEndScenario(REGHANDLE             RegHandle,
		                         EVENT_DESCRIPTOR      *EventDescriptor,
		                         DWORD                 UserDataCount,
		                         EVENT_DATA_DESCRIPTOR *UserData);

// https://msdn.microsoft.com/en-us/library/dd392307.aspx

__declspec(dllimport)
DWORD	NTAPI	EtwEventWriteEx(REGHANDLE             RegHandle,
		                EVENT_DESCRIPTOR      *EventDescriptor,
		                QWORD                 Filter,
		                DWORD                 Flags,
		                LPCGUID               ActivityId,
		                LPCGUID               RelatedActivityId,
		                DWORD                 UserDataCount,
		                EVENT_DATA_DESCRIPTOR *UserData);

// NOTE: forwarded from ADVAPI32!EventWriteStartScenario

__declspec(dllimport)
DWORD	NTAPI	EtwEventWriteStartScenario(REGHANDLE             RegHandle,
		                           EVENT_DESCRIPTOR      *EventDescriptor,
		                           DWORD                 UserDataCount,
		                           EVENT_DATA_DESCRIPTOR *UserData);

// https://msdn.microsoft.com/en-us/library/aa363750.aspx
// NOTE: forwarded from ADVAPI32!EventWriteString

__declspec(dllimport)
DWORD	NTAPI	EtwEventWriteString(REGHANDLE RegHandle,
		                    BYTE      Level,
		                    QWORD     Keyword,
		                    LPCWSTR   String);

// https://msdn.microsoft.com/en-us/library/aa363751.aspx
// NOTE: forwarded from ADVAPI32!EventWriteTransfer

__declspec(dllimport)
DWORD	NTAPI	EtwEventWriteTransfer(REGHANDLE             RegHandle,
		                      EVENT_DESCRIPTOR      *EventDescriptor,
		                      LPCGUID               ActivityId,
		                      LPCGUID               RelatedActivityId,
		                      DWORD                 UserDataCount,
		                      EVENT_DATA_DESCRIPTOR *UserData);

// https://msdn.microsoft.com/en-us/library/aa363893.aspx
// NOTE: forwarded from ADVAPI32!GetTraceEnableFlags

__declspec(dllimport)
DWORD	NTAPI	EtwGetTraceEnableFlags(TRACEHANDLE SessionHandle);

// https://msdn.microsoft.com/en-us/library/aa363894.aspx
// NOTE: forwarded from ADVAPI32!GetTraceEnableLevel

__declspec(dllimport)
BYTE	NTAPI	EtwGetTraceEnableLevel(TRACEHANDLE SessionHandle);

// https://msdn.microsoft.com/en-us/library/aa363897.aspx
// NOTE: forwarded from ADVAPI32!GetTraceLoggerHandle

__declspec(dllimport)
TRACEHANDLE	NTAPI	EtwGetTraceLoggerHandle(LPVOID Buffer);

// https://msdn.microsoft.com/en-us/library/aa363776.aspx

#pragma warning(disable: 4201)
typedef	struct	_EVENT_TRACE_HEADER
{
	WORD	Size;
	union
	{
		WORD	FieldTypeFlags;
		struct
		{
			BYTE	HeaderType;
			BYTE	MarkerFlags;
		};
	};
	union
	{
		DWORD	Version;
		struct
		{
			BYTE	Type;
			BYTE	Level;
			WORD	Version;
		} Class;
	};
	DWORD		ThreadId;
	DWORD		ProcessId;
	LARGE_INTEGER	TimeStamp;
	union
	{
		GUID	Guid;
		QWORD	GuidPtr;
	};
	union
	{
		struct
		{
			DWORD	ClientContext;
			DWORD	Flags;
		};
		struct
		{
			DWORD	KernelTime;
			DWORD	UserTime;
		};
		QWORD	ProcessorTime;
	};
} EVENT_TRACE_HEADER, *LPEVENT_TRACE_HEADER, *PEVENT_TRACE_HEADER;
#pragma warning(default: 4201)

// https://msdn.microsoft.com/en-us/library/aa364137.aspx
// NOTE: forwarded from ADVAPI32!TraceEvent and WMI!TraceEvent

__declspec(dllimport)
DWORD	NTAPI	EtwLogTraceEvent(TRACEHANDLE        SessionHandle,
		                 EVENT_TRACE_HEADER *EventTrace);

typedef	enum
{
	WMI_ENABLE_EVENTS  = 4,
	WMI_DISABLE_EVENTS = 5
} WMIDPREQUESTCODE;

// https://msdn.microsoft.com/en-us/library/aa363693.aspx

typedef	DWORD	(NTAPI	WMIDPREQUEST) (WMIDPREQUESTCODE RequestCode,
			               LPVOID           Context,
			               LPDWORD          Reserved,
			               LPVOID           Buffer);

// https://msdn.microsoft.com/en-us/library/aa364144.aspx

typedef	struct	_TRACE_GUID_REGISTRATION
{
	LPCGUID	Guid;
	HANDLE	RegHandle;
} TRACE_GUID_REGISTRATION;

// https://msdn.microsoft.com/en-us/library/aa364105.aspx
// NOTE: forwarded from ADVAPI32!RegisterTraceGuidsA

__declspec(dllimport)
DWORD	NTAPI	EtwRegisterTraceGuidsA(WMIDPREQUEST            *RequestAddress,
		                       LPVOID                  RequestContext,
		                       LPCGUID                 ControlGuid,
		                       DWORD                   GuidCount,
		                       TRACE_GUID_REGISTRATION *TraceGuidReg,
		                       LPCSTR                  MofImagePath,
		                       LPCSTR                  MofResourceName,
		                       TRACEHANDLE             *RegistrationHandle);

// https://msdn.microsoft.com/en-us/library/aa364105.aspx
// NOTE: forwarded from ADVAPI32!RegisterTraceGuidsW

__declspec(dllimport)
DWORD	NTAPI	EtwRegisterTraceGuidsW(WMIDPREQUEST            *RequestAddress,
		                       LPVOID                  RequestContext,
		                       LPCGUID                 ControlGuid,
		                       DWORD                   GuidCount,
		                       TRACE_GUID_REGISTRATION *TraceGuidReg,
		                       LPCWSTR                 MofImagePath,
		                       LPCWSTR                 MofResourceName,
		                       TRACEHANDLE             *RegistrationHandle);

// https://msdn.microsoft.com/en-us/library/aa363762.aspx

#pragma warning(disable: 4201)
typedef	struct	_EVENT_INSTANCE_HEADER
{
	WORD	Size;
	union
	{
		WORD	FieldTypeFlags;
		struct
		{
			BYTE	HeaderType;
			BYTE	MarkerFlags;
		};
	};
	union
	{
		DWORD	Version;
		struct
		{
			BYTE	Type;
			BYTE	Level;
			WORD	Version;
		} Class;
	};
	DWORD		ThreadId;
	DWORD		ProcessId;
	LARGE_INTEGER	TimeStamp;
	QWORD		RegHandle;
	DWORD		InstanceId;
	DWORD		ParentInstanceId;
	union
	{
		struct
		{
			DWORD	ClientContext;
			DWORD	Flags;
		};
		struct
		{
			DWORD	KernelTime;
			DWORD	UserTime;
		};
		QWORD	ProcessorTime;
	};
	QWORD	ParentRegHandle;
} EVENT_INSTANCE_HEADER, *LPEVENT_INSTANCE_HEADER, *PEVENT_INSTANCE_HEADER;
#pragma warning(default: 4201)

// https://msdn.microsoft.com/en-us/library/aa364138.aspx
// NOTE: forwarded from ADVAPI32!TraceEventInstance and WMI!TraceEventInstance

__declspec(dllimport)
DWORD	NTAPI	EtwTraceEventInstance(TRACEHANDLE           SessionHandle,
		                      EVENT_INSTANCE_HEADER *EventTrace,
		                      EVENT_INSTANCE_INFO   *InstInfo,
		                      EVENT_INSTANCE_INFO   *ParentInstInfo);

// https://msdn.microsoft.com/en-us/library/aa364139.aspx
// NOTE: forwarded from ADVAPI32!TraceMessage

__declspec(dllimport)
DWORD	CDECL	EtwTraceMessage(TRACEHANDLE SessionHandle,
		                DWORD       MessageFlags,
		                LPGUID      MessageGuid,
		                WORD        MessageNumber,
		                ...);

// https://msdn.microsoft.com/en-us/library/aa364140.aspx
// NOTE: forwarded from ADVAPI32!TraceMessageVa

__declspec(dllimport)
DWORD	NTAPI	EtwTraceMessageVa(TRACEHANDLE SessionHandle,
		                  DWORD       MessageFlags,
		                  LPGUID      MessageGuid,
		                  WORD        MessageNumber,
		                  va_list     MessageArgList);

// https://msdn.microsoft.com/en-us/library/aa364154.aspx
// NOTE: forwarded from ADVAPI32!UnregisterTraceGuids

__declspec(dllimport)
DWORD	NTAPI	EtwUnregisterTraceGuids(TRACEHANDLE RegistrationHandle);

typedef	enum	_EVENT_TRACE_INFORMATION_CLASS
{
#if _WIN32_WINNT >= 0x0600
	EventTraceKernelVersionInformation      = 0,
	EventTraceGroupMaskInformation          = 1,
	EventTracePerformanceInformation        = 2,
	EventTraceTimeProfileInformation        = 3,
	EventTraceSessionSecurityInformation    = 4,
#if _WIN32_WINNT >= 0x0601
	EventTraceSpinlockInformation           = 5,
	EventTraceStackTracingInformation       = 6,
	EventTraceExecutiveResourceInformation  = 7,
	EventTraceHeapTracingInformation        = 8,
	EventTraceHeapSummaryTracingInformation = 9,
	EventTracePoolTagFilterInformation      = 10,
#if _WIN32_WINNT >= 0x0602
	EventTracePebsTracingInformation        = 11,
	EventTraceProfileConfigInformation      = 12,
	EventTraceProfileSourceListInformation  = 13,
	EventTraceProfileEventListInformation   = 14,
	EventTraceProfileCounterListInformation = 15,
	EventTraceStackCachingInformation       = 16,
	EventTraceObjectTypeFilterInformation   = 17,
#endif
#endif
#endif
	MaxEventTraceInfoClass
} EVENT_TRACE_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff563617.aspx

#define RTL_CONSTANT_STRING(string)	{sizeof(string) - sizeof(*string), sizeof(string), string}

static	const	UNICODE_STRING	RtlAlternateDosPathSeparatorString = RTL_CONSTANT_STRING(L"/");
static	const	UNICODE_STRING	RtlDosPathSeparatorsString = RTL_CONSTANT_STRING(L"\\" L"/");
static	const	UNICODE_STRING	RtlNtPathSeparatorString = RTL_CONSTANT_STRING(L"\\");

// https://msdn.microsoft.com/en-us/library/ff547799.aspx

__forceinline
VOID	NTAPI	InitializeListHead(LIST_ENTRY *ListHead)
{
	ListHead->Flink = ListHead->Blink = ListHead;
}

// https://msdn.microsoft.com/en-us/library/ff547820.aspx

__inline
VOID	NTAPI	InsertHeadList(LIST_ENTRY *ListHead,
		               LIST_ENTRY *Entry)
{
	Entry->Blink = ListHead;
	Entry->Flink = ListHead->Flink;

	ListHead->Flink->Blink = Entry;
	ListHead->Flink = Entry;
}

// https://msdn.microsoft.com/en-us/library/ff547823.aspx

__inline
VOID	NTAPI	InsertTailList(LIST_ENTRY *ListHead,
		               LIST_ENTRY *Entry)
{
	Entry->Flink = ListHead;
	Entry->Blink = ListHead->Blink;

	ListHead->Blink->Flink = Entry;
	ListHead->Blink = Entry;
}

// https://msdn.microsoft.com/en-us/library/ff551789.aspx

__forceinline
BOOL	NTAPI	IsListEmpty(LIST_ENTRY *ListHead)
{
	return (ListHead->Flink == ListHead);
}

#if 0	// NOTE: not exported from NTDLL.DLL!
__declspec(dllimport)
BOOL	NTAPI	IsNodeOnList(LIST_ENTRY *ListHead,
		             LIST_ENTRY *Entry);
#endif

// https://msdn.microsoft.com/en-us/library/ff559712.aspx

__inline
SINGLE_LIST_ENTRY *	NTAPI	PopEntryList(SINGLE_LIST_ENTRY *ListHead)
{
	SINGLE_LIST_ENTRY	*Entry = ListHead->Next;

	if (Entry != NULL)
		ListHead->Next = Entry->Next;

	return Entry;
}

// https://msdn.microsoft.com/en-us/library/ff559964.aspx

__inline
VOID	NTAPI	PushEntryList(SINGLE_LIST_ENTRY *ListHead,
		              SINGLE_LIST_ENTRY *Entry)
{
	Entry->Next = ListHead->Next;
	ListHead->Next = Entry;
}

// https://msdn.microsoft.com/en-us/library/ff561029.aspx

__inline
BOOL	NTAPI	RemoveEntryList(LIST_ENTRY *Entry)
{
	// https://msdn.microsoft.com/en-us/library/ff554296.aspx

	LIST_ENTRY	*Blink = Entry->Blink;
	LIST_ENTRY	*Flink = Entry->Flink;

	Blink->Flink = Flink;
	Flink->Blink = Blink;

	return (Flink == Blink);
}

// https://msdn.microsoft.com/en-us/library/ff561032.aspx

__inline
LIST_ENTRY *	NTAPI	RemoveHeadList(LIST_ENTRY *ListHead)
{
	// https://msdn.microsoft.com/en-us/library/ff554296.aspx

	LIST_ENTRY	*Entry = ListHead->Flink;
	LIST_ENTRY	*Flink = Entry->Flink;

	ListHead->Flink = Flink;
	Flink->Blink = ListHead;

	return Entry;
}

// https://msdn.microsoft.com/en-us/library/ff561036.aspx

__inline
LIST_ENTRY *	NTAPI	RemoveTailList(LIST_ENTRY *ListHead)
{
	// https://msdn.microsoft.com/en-us/library/ff554296.aspx

	LIST_ENTRY	*Entry = ListHead->Blink;
	LIST_ENTRY	*Blink = Entry->Blink;

	ListHead->Blink = Blink;
	Blink->Flink = ListHead;

	return Entry;
}

#define OBJ_INHERIT				0x00000002UL
#define OBJ_PERMANENT				0x00000010UL
#define OBJ_EXCLUSIVE				0x00000020UL
#define OBJ_CASE_INSENSITIVE			0x00000040UL
#define OBJ_OPENIF				0x00000080UL
#define OBJ_OPENLINK				0x00000100UL
#define OBJ_KERNEL_HANDLE			0x00000200UL
#define OBJ_FORCE_ACCESS_CHECK			0x00000400UL
#define OBJ_IGNORE_IMPERSONATED_DEVICEMAP	0x00000800UL
#define OBJ_DONT_REPARSE			0x00001000UL
#define OBJ_VALID_ATTRIBUTES			0x00001FF2UL

// https://msdn.microsoft.com/en-us/library/ff557749.aspx

typedef	struct	_OBJECT_ATTRIBUTES
{
	DWORD		Length;
	HANDLE		RootDirectory;
	UNICODE_STRING	*ObjectName;
	DWORD		Attributes;
	LPVOID		SecurityDescriptor;
	LPVOID		SecurityQualityOfService;
} OBJECT_ATTRIBUTES, *LPOBJECT_ATTRIBUTES, *POBJECT_ATTRIBUTES;

#define RTL_CONSTANT_OBJECT_ATTRIBUTES(objectname, attributes)	{sizeof(OBJECT_ATTRIBUTES), (HANDLE) NULL, objectname, attributes, NULL, NULL}

// https://msdn.microsoft.com/en-us/library/ff547804.aspx

__inline
VOID	NTAPI	InitializeObjectAttributes(OBJECT_ATTRIBUTES   *InitializedAttributes,
		                           UNICODE_STRING      *ObjectName,
		                           DWORD               Attributes,
		                           HANDLE              RootDirectory,
		                           SECURITY_DESCRIPTOR *SecurityDescriptor)
{
	// https://msdn.microsoft.com/en-us/library/ff557749.aspx

	InitializedAttributes->Length = sizeof(OBJECT_ATTRIBUTES);
	InitializedAttributes->ObjectName = ObjectName;
	InitializedAttributes->Attributes = Attributes;
	InitializedAttributes->RootDirectory = RootDirectory;
	InitializedAttributes->SecurityDescriptor = SecurityDescriptor;
	InitializedAttributes->SecurityQualityOfService = NULL;
}

__declspec(dllimport)
LONG	NTAPI	LdrAccessResource(LPVOID                    DllHandle,
		                  IMAGE_RESOURCE_DATA_ENTRY *ResourceDataEntry,
		                  LPVOID                    *Address,
		                  LPDWORD                   Size);

__declspec(dllimport)
LONG	NTAPI	LdrAddRefDll(DWORD  Flags,
		             LPVOID DllHandle);

// NOTE: like KERNEL32!DisableThreadLibraryCalls

__declspec(dllimport)
LONG	NTAPI	LdrDisableThreadCalloutsForDll(LPVOID DllHandle);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/dd347460.aspx

typedef	struct	_LDR_DLL_LOADED_NOTIFICATION_DATA
{
	DWORD		Flags;		// reserved
	UNICODE_STRING	*FullDllName;	// full pathname of DLL module
	UNICODE_STRING	*BaseDllName;	// base filename of DLL module
	LPVOID		DllBase;	// pointer to base address for DLL in memory
	DWORD		SizeOfImage;	// size of DLL image, in bytes
} LDR_DLL_LOADED_NOTIFICATION_DATA, *LPLDR_DLL_LOADED_NOTIFICATION_DATA, *PLDR_DLL_LOADED_NOTIFICATION_DATA;

typedef	struct	_LDR_DLL_UNLOADED_NOTIFICATION_DATA
{
	DWORD		Flags;		// reserved
	UNICODE_STRING	*FullDllName;	// full pathname of DLL module
	UNICODE_STRING	*BaseDllName;	// base filename of DLL module
	LPVOID		DllBase;	// pointer to base address for DLL in memory
	DWORD		SizeOfImage;	// size of DLL image, in bytes
} LDR_DLL_UNLOADED_NOTIFICATION_DATA, *LPLDR_DLL_UNLOADED_NOTIFICATION_DATA, *PLDR_DLL_UNLOADED_NOTIFICATION_DATA;

typedef	union	_LDR_DLL_NOTIFICATION_DATA
{
	LDR_DLL_LOADED_NOTIFICATION_DATA	Loaded;
	LDR_DLL_UNLOADED_NOTIFICATION_DATA	Unloaded;
} LDR_DLL_NOTIFICATION_DATA, *LPLDR_DLL_NOTIFICATION_DATA, *PLDR_DLL_NOTIFICATION_DATA;

#define LDR_DLL_NOTIFICATION_REASON_LOADED	1UL
#define LDR_DLL_NOTIFICATION_REASON_UNLOADED	2UL

typedef	VOID	(NTAPI	LDR_DLL_NOTIFICATION_FUNCTION) (DWORD                     NotificationReason,
			                                LDR_DLL_NOTIFICATION_DATA *NotificationData,
			                                LPVOID                    Context);
#endif // _WIN32_WINNT

typedef	struct	_ACTIVATION_CONTEXT	ACTIVATION_CONTEXT, *LPACTIVATION_CONTEXT, *PACTIVATION_CONTEXT;

#define LDR_STATIC_LINK			0x00000002UL
#define LDR_IMAGE_DLL			0x00000004UL
#define LDR_LOAD_IN_PROGRESS		0x00001000UL
#define LDR_UNLOAD_IN_PROGRESS		0x00002000UL
#define LDR_ENTRY_PROCESSED		0x00004000UL
#define LDR_ENTRY_INSERTED		0x00008000UL
#define LDR_CURRENT_LOAD		0x00010000UL
#define LDR_FAILED_BUILTIN_LOAD		0x00020000UL
#define LDR_DONT_CALL_FOR_THREADS	0x00040000UL
#define LDR_PROCESS_ATTACH_CALLED	0x00080000UL
#define LDR_DEBUG_SYMBOLS_LOADED	0x00100000UL
#define LDR_IMAGE_NOT_AT_BASE		0x00200000UL
#define LDR_WX86_IGNORE_MACHINETYPE	0x00400000UL

// https://msdn.microsoft.com/en-us/library/aa813708.aspx

#pragma warning(disable: 4201)
typedef	struct	_LDR_DATA_TABLE_ENTRY
{
#if 0
	LPVOID		Reserved1[2];
	LIST_ENTRY	InMemoryOrderLinks;
	LPVOID		Reserved2[2];
	LPVOID		DllBase;
	LPVOID		EntryPoint;
	LPVOID		Reserved3;
	UNICODE_STRING	FullDllName;
	BYTE		Reserved4[8];
	LPVOID		Reserved5[3];
	union
	{
		DWORD	CheckSum;
		LPVOID	Reserved6;
	};
	DWORD		TimeDateStamp;
#else
	LIST_ENTRY	InLoadOrderLinks;
	LIST_ENTRY	InMemoryOrderLinks;
	LIST_ENTRY	InInitializationOrderLinks;
	LPVOID		DllBase;
	LPVOID		EntryPoint;
	DWORD		SizeOfImage;
	UNICODE_STRING	FullDllName;
	UNICODE_STRING	BaseDllName;
	DWORD		Flags;
	WORD		LoadCount;
	WORD		TlsIndex;
	union
	{
		LIST_ENTRY	HashLinks;
		struct
		{
			LPVOID	SectionPointer;
			DWORD	CheckSum;
		};
	};
	union
	{
		DWORD		TimeDateStamp;
		LPVOID		LoadedImports;
	};
	ACTIVATION_CONTEXT	*EntryPointActivationContext;
	LPVOID			PatchInformation;
#endif
} LDR_DATA_TABLE_ENTRY, *LPLDR_DATA_TABLE_ENTRY, *PLDR_DATA_TABLE_ENTRY;
#pragma warning(default: 4201)

typedef	VOID	(NTAPI	LDR_ENUM_CALLBACK) (LDR_DATA_TABLE_ENTRY *DataTableEntry,
			                    LPVOID               Context,
			                    BOOLEAN              *StopEnumeration);

__declspec(dllimport)
LONG	NTAPI	LdrEnumerateLoadedModules(DWORD             Flags,
		                          LDR_ENUM_CALLBACK *CallbackFunction,
		                          LPVOID            Context);

#define LDR_MAXIMUM_RESOURCE_PATH_DEPTH	3

#pragma warning(disable: 4201)
typedef	struct	_LDR_ENUM_RESOURCE_ENTRY
{
	union
	{
		DWORD_PTR	NameOrId;
		IMAGE_RESOURCE_DIRECTORY_STRING	*Name;
		struct
		{
			WORD	Id;
			WORD	NameIsPresent;
		};
	} Path[LDR_MAXIMUM_RESOURCE_PATH_DEPTH];
	LPVOID	Data;
	DWORD	Size;
	DWORD	Reserved;
} LDR_ENUM_RESOURCE_ENTRY, *LPLDR_ENUM_RESOURCE_ENTRY, *PLDR_ENUM_RESOURCE_ENTRY;
#pragma warning(default: 4201)

__declspec(dllimport)
LONG	NTAPI	LdrEnumResources(LPVOID                  DllHandle,
		                 DWORD_PTR               *ResourceIdPath,
		                 DWORD                   ResourceIdPathLength,
		                 LPDWORD                 NumberOfResources,
		                 LDR_ENUM_RESOURCE_ENTRY Resource);

#if _WIN32_WINNT >= 0x0A00
// https://msdn.microsoft.com/en-us/library/dn948752.aspx

__declspec(dllimport)
VOID	NTAPI	LdrFastFailInLoaderCallout(VOID);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	LdrFindEntryForAddress(LPVOID               Address,
		                       LDR_DATA_TABLE_ENTRY **DataTableEntry);

__declspec(dllimport)
LONG	NTAPI	LdrFindResource_U(LPVOID                    DllHandle,
		                  DWORD_PTR                 *ResourceIdPath,
		                  DWORD                     ResourceIdPathLength,
		                  IMAGE_RESOURCE_DATA_ENTRY **ResourceDataEntry);

__declspec(dllimport)
LONG	NTAPI	LdrFindResourceDirectory_U(LPVOID                   DllHandle,
		                           DWORD_PTR                *ResourceIdPath,
		                           DWORD                    ResourceIdPathLength,
		                           IMAGE_RESOURCE_DIRECTORY **ResourceDirectory);

#define LDRP_FIND_RESOURCE_DATA				0x00000000UL
#define LDRP_FIND_RESOURCE_DIRECTORY			0x00000002UL

#define LDR_FIND_RESOURCE_LANGUAGE_CAN_FALLBACK		0x00000000UL
#define LDR_FIND_RESOURCE_LANGUAGE_EXACT		0x00000004UL
#define LDR_FIND_RESOURCE_LANGUAGE_REDIRECT_VERSION	0x00000008UL

__declspec(dllimport)
LONG	NTAPI	LdrFindResourceEx_U(DWORD                     Flags,
		                    LPVOID                    DllHandle,
		                    DWORD_PTR                 *ResourceIdPath,
		                    DWORD                     ResourceIdPathLength,
		                    IMAGE_RESOURCE_DATA_ENTRY **ResourceDataEntry);

__declspec(dllimport)
LONG	NTAPI	LdrGetDllHandle(LPCWSTR        DllSearchPath,		// optional PATH variable
		                LPDWORD        DllCharacteristics,
		                UNICODE_STRING *DllName,
		                LPVOID         *BaseAddress);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	LdrGetDllHandleByName(UNICODE_STRING *BaseDllName,
		                      UNICODE_STRING *FullDllName,
		                      LPVOID         *DllHandle);
#endif

__declspec(dllimport)
LONG	NTAPI	LdrGetDllHandleEx(DWORD          Flags,
		                  LPCWSTR        DllSearchPath,		// optional PATH variable
		                  LPDWORD        DllCharacteristics,
		                  UNICODE_STRING *DllName,
		                  LPVOID         *BaseAddress);

#if _WIN32_WINNT >= 0x0601
typedef	struct	_FAILUREDATA
{
	NTSTATUS	Status;
	WCHAR		DllName[32];
	WCHAR		EntryName[32];
} FAILUREDATA, *LPFAILUREDATA, *PFAILUREDATA;

__declspec(dllimport)
FAILUREDATA *	NTAPI	LdrGetFailureData(VOID);
#endif

__declspec(dllimport)
LONG	NTAPI	LdrGetProcedureAddress(LPVOID      BaseAddress,
		                       ANSI_STRING *ProcedureName,
		                       DWORD       ProcedureOrdinal,
		                       LPVOID      *ProcedureAddress);

// https://msdn.microsoft.com/en-us/library/ff550671.aspx

#pragma warning(disable: 4201)
typedef	struct	_IO_STATUS_BLOCK
{
	union
	{
		NTSTATUS	Status;
		LPVOID		Pointer;
	};
	DWORD_PTR	Information;
} IO_STATUS_BLOCK, *LPIO_STATUS_BLOCK, *PIO_STATUS_BLOCK;
#pragma warning(default: 4201)

// https://msdn.microsoft.com/en-us/library/ff540344.aspx

typedef	struct	_FILE_RENAME_INFORMATION
{
#if 0
	BOOLEAN	ReplaceIfExists;
#else			// 'Redstone 1'
	union
	{
		BOOLEAN ReplaceIfExists;	// FileRenameInformation
		DWORD	Flags;			// FileRenameInformationEx
	};
#endif
	HANDLE	RootDirectory;
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_RENAME_INFORMATION, *LPFILE_RENAME_INFORMATION, *PFILE_RENAME_INFORMATION;

#define FLG_HOTPATCH_KERNEL		0x80000000UL
#define FLG_HOTPATCH_RELOAD_NTDLL	0x40000000UL
#define FLG_HOTPATCH_NAME_INFO		0x20000000UL
#define FLG_HOTPATCH_RENAME_INFO	0x10000000UL
#define FLG_HOTPATCH_MAP_ATOMIC_SWAP	0x08000000UL
#define FLG_HOTPATCH_WOW64		0x04000000UL
#define FLG_HOTPATCH_INJECT		0x01000000UL
#define FLG_HOTPATCH_VERIFICATION_ERROR	0x00800000UL
#define FLG_HOTPATCH_ACTIVE		0x00000001UL
#define FLG_HOTPATCH_STATUS_FLAGS	FLG_HOTPATCH_ACTIVE

typedef	struct	_HOTPATCH_HOOK_DESCRIPTOR
{
	DWORD_PTR	TargetAddress;
	LPVOID		MappedAddress;
	DWORD		CodeOffset;
	DWORD		CodeSize;
	DWORD		OrigCodeOffset;
	DWORD		ValidationOffset;
	DWORD		ValidationSize;
} HOTPATCH_HOOK_DESCRIPTOR, *LPHOTPATCH_HOOK_DESCRIPTOR, *PHOTPATCH_HOOK_DESCRIPTOR;

#pragma warning(disable: 4201)
typedef	struct	_SYSTEM_HOTPATCH_CODE_INFORMATION
{
	DWORD	Flags;
	DWORD	InfoSize;
	union
	{
		struct
		{
			DWORD				DescriptorsCount;
			HOTPATCH_HOOK_DESCRIPTOR	CodeDescriptors[ANYSIZE_ARRAY];
		} CodeInfo;
		struct
		{
			WORD	NameOffset;
			WORD	NameLength;
		} KernelModeInfo;
		struct
		{
			WORD	NameOffset;
			WORD	NameLength;
			WORD	TargetNameOffset;
			WORD	TargetNameLength;
		} UserModeInfo;
		struct
		{
			WORD	NameOffset;
			WORD	NameLength;
			WORD	TargetNameOffset;
			WORD	TargetNameLength;
			BOOLEAN	PatchingFinished;
			DWORD	ReturnCode;
			HANDLE	TargetProcess;
		} InjectionInfo;
		struct
		{
			HANDLE			FileHandle1;
			IO_STATUS_BLOCK		*IoStatusBlock1;
			FILE_RENAME_INFORMATION	*RenameInformation1;
			DWORD			RenameInformationLength1;
			HANDLE			FileHandle2;
			IO_STATUS_BLOCK		*IoStatusBlock2;
			FILE_RENAME_INFORMATION	*RenameInformation2;
			DWORD			RenameInformationLength2;
		} RenameInfo;
		struct
		{
			HANDLE	ParentDirectory;
			HANDLE	ObjectHandle1;
			HANDLE	ObjectHandle2;
		} AtomicSwap;
	};
} SYSTEM_HOTPATCH_CODE_INFORMATION, *LPSYSTEM_HOTPATCH_CODE_INFORMATION, *PSYSTEM_HOTPATCH_CODE_INFORMATION;
#pragma warning(default: 4201)

__declspec(dllimport)
LONG	NTAPI	LdrHotPatchRoutine(SYSTEM_HOTPATCH_CODE_INFORMATION *Information);

#define LDR_DONT_RESOLVE_DLL_REFERENCES	0x00000002UL

__declspec(dllimport)
LONG	NTAPI	LdrLoadDll(LPCWSTR        DllSearchPath,	// optional PATH variable
		           LPDWORD        DllCharacteristics,
		           UNICODE_STRING *DllName,
		           LPVOID         *BaseAddress);

#define LDR_LOCK_LOADER_LOCK_FLAG_RAISE_ON_ERRORS	1UL
#define LDR_LOCK_LOADER_LOCK_FLAG_TRY_ONLY		2UL

#define LDR_LOCK_LOADER_LOCK_DISPOSITION_INVALID		0UL
#define LDR_LOCK_LOADER_LOCK_DISPOSITION_LOCK_ACQUIRED		1UL
#define LDR_LOCK_LOADER_LOCK_DISPOSITION_LOCK_NOT_ACQUIRED	2UL

__declspec(dllimport)
LONG	NTAPI	LdrLockLoaderLock(DWORD   Flags,
		                  LPDWORD Disposition,
		                  LPDWORD Cookie);

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	LdrOpenImageFileOptionsKey(UNICODE_STRING *SubKey,
		                           BOOLEAN        Wow64,
		                           HANDLE         *KeyHandle);
#endif

__declspec(dllimport)
LONG	NTAPI	LdrQueryImageFileExecutionOptions(UNICODE_STRING *SubKey,
		                                  LPCWSTR        ValueName,
		                                  DWORD          Type,
		                                  LPVOID         Buffer,
		                                  DWORD          BufferSize,
		                                  LPDWORD        ReturnedLength);

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	LdrQueryImageFileExecutionOptionsEx(UNICODE_STRING *SubKey,
		                                    LPCWSTR        ValueName,
		                                    DWORD          Type,
		                                    LPVOID         Buffer,
		                                    DWORD          BufferSize,
		                                    LPDWORD        ReturnedLength,
		                                    BOOLEAN        Wow64);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	LdrQueryImageFileKeyOption(HANDLE  KeyHandle,
		                           LPCWSTR OptionName,
		                           DWORD   Type,
		                           LPVOID  Buffer,
		                           DWORD   BufferSize,
		                           LPDWORD ResultSize);
#endif

typedef	struct	_RTL_PROCESS_MODULE_INFORMATION
{
	HANDLE	Section;
	LPVOID	MappedBase;
	LPVOID	ImageBase;
	DWORD	ImageSize;
	DWORD	Flags;
	WORD	LoadOrderIndex;
	WORD	InitOrderIndex;
	WORD	LoadCount;
	WORD	OffsetToFileName;
	CHAR	FullPathName[256];
} RTL_PROCESS_MODULE_INFORMATION, *LPRTL_PROCESS_MODULE_INFORMATION, *PRTL_PROCESS_MODULE_INFORMATION;

typedef	struct	_RTL_PROCESS_MODULES
{
	DWORD				NumberOfModules;
	RTL_PROCESS_MODULE_INFORMATION	Modules[ANYSIZE_ARRAY];
} RTL_PROCESS_MODULES, *LPRTL_PROCESS_MODULES, *PRTL_PROCESS_MODULES;

__declspec(dllimport)
LONG	NTAPI	LdrQueryProcessModuleInformation(RTL_PROCESS_MODULES *ModuleInformation,
		                                 DWORD               Size,
		                                 LPDWORD             ReturnedSize);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/dd347461.aspx

__declspec(dllimport)
LONG	NTAPI	LdrRegisterDllNotification(DWORD                         Flags,
		                           LDR_DLL_NOTIFICATION_FUNCTION *NotificationFunction,
		                           LPVOID                        Context,
		                           LPVOID                        *Cookie);
#endif

#if _WIN32_WINNT >= 0x0602
// https://msdn.microsoft.com/en-us/library/hh829881.aspx
// NOTE: forwarded from KERNEL32!ResolveDelayLoadedAPI and KERNELBASE!ResolveDelayLoadedAPI

__declspec(dllimport)
VOID	NTAPI	LdrResolveDelayLoadedAPI(LPVOID                           ParentModuleBase,
		                         IMAGE_DELAYLOAD_DESCRIPTOR       *DelayLoadDescriptor,
		                         DELAYLOAD_FAILURE_DLL_CALLBACK   *FailureDllHook,
		                         DELAYLOAD_FAILURE_SYSTEM_ROUTINE *FailureSystemHook,
		                         IMAGE_THUNK_DATA                 *ThunkAddress,
		                         DWORD                            Flags);

// https://msdn.microsoft.com/en-us/library/hh829882.aspx
// NOTE: forwarded from KERNEL32!ResolveDelayLoadsFromDll and KERNELBASE!ResolveDelayLoadsFromDll

__declspec(dllimport)
LONG	NTAPI	LdrResolveDelayLoadsFromDll(LPVOID ParentBase,
		                            LPCSTR TargetDllName,
		                            DWORD  Flags);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	LdrResSearchResource(LPVOID    ModuleBase,
		                     DWORD_PTR ResourceIds[],
		                     DWORD     ResourceIdCount,
		                     DWORD     Flags,
		                     LPVOID    *Resource,
		                     DWORD_PTR *Size,
		                     WORD      FoundLanguages[],
		                     LPDWORD   FoundLanguagesLength);
#endif

__declspec(dllimport)
VOID	NTAPI	LdrShutdownProcess(VOID);

__declspec(dllimport)
VOID	NTAPI	LdrShutdownThread(VOID);

__declspec(dllimport)
LONG	NTAPI	LdrUnlockLoaderLock(DWORD Flags,
		                    DWORD Cookie);

__declspec(dllimport)
LONG	NTAPI	LdrUnloadDll(LPVOID BaseAddress);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/dd347462.aspx

__declspec(dllimport)
LONG	NTAPI	LdrUnregisterDllNotification(LPVOID Cookie);
#endif

typedef	VOID	(NTAPI	LDR_IMPORT_MODULE_CALLBACK) (LPVOID Parameter,
			                             LPSTR  ModuleName);

__declspec(dllimport)
LONG	NTAPI	LdrVerifyImageMatchesChecksum(HANDLE                     ImageFileHandle,
		                              LDR_IMPORT_MODULE_CALLBACK *ImportCallbackRoutine,
		                              LPVOID                     ImportCallbackParameter,
		                              LPWORD                     ImageCharacteristics);

__declspec(dllimport)
WORD	NlsAnsiCodePage;

__declspec(dllimport)
BOOLEAN	NlsMbCodePageTag;

__declspec(dllimport)
BOOLEAN	NlsMbOemCodePageTag;

#if _WIN32_WINNT >= 0x0600
// NOTE: hash functions missing in NTDLL.LIB, use NTDLLV.LIB instead!

typedef	struct	_MD4_CTX
{
	DWORD	State[4];	// = {0x67452301UL, 0xEFCDAB89UL, 0x98BADCFEUL, 0x10325476UL}
	DWORD	Count[2];	// = {0UL, 0UL}
	BYTE	Buffer[64];
	BYTE	Digest[16];
} MD4_CTX, *LPMD4_CTX, *PMD4_CTX;

// NOTE: forwarded from ADVAPI32!MD4Final

__declspec(dllimport)
VOID	NTAPI	MD4Final(MD4_CTX *Context);

// NOTE: forwarded from ADVAPI32!MD4Init

__declspec(dllimport)
VOID	NTAPI	MD4Init(MD4_CTX *Context);

// NOTE: forwarded from ADVAPI32!MD4Update

__declspec(dllimport)
VOID	NTAPI	MD4Update(MD4_CTX *Context,
		          LPCVOID Buffer,
		          DWORD   BufferSize);

typedef	struct	_MD5_CTX
{
	DWORD	Count[2];	// = {0UL, 0UL}
	DWORD	State[4];	// = {0x67452301UL, 0xEFCDAB89UL, 0x98BADCFEUL, 0x10325476UL}
	BYTE	Buffer[64];
	BYTE	Digest[16];
} MD5_CTX, *LPMD5_CTX, *PMD5_CTX;

// NOTE: forwarded from ADVAPI32!MD5Final

__declspec(dllimport)
VOID	NTAPI	MD5Final(MD5_CTX *Context);

// NOTE: forwarded from ADVAPI32!MD5Init

__declspec(dllimport)
VOID	NTAPI	MD5Init(MD5_CTX *Context);

// NOTE: forwarded from ADVAPI32!MD5Update

__declspec(dllimport)
VOID	NTAPI	MD5Update(MD5_CTX *Context,
		          LPCVOID Buffer,
		          DWORD   BufferSize);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms645450.aspx
// NOTE: forwarded from USER32!DefDlgProcA

__declspec(dllimport)
LRESULT	NTAPI	NtdllDialogWndProc_A(HWND   hDlg,
		                     UINT   Msg,
		                     WPARAM wParam,
		                     LPARAM lParam);

// https://msdn.microsoft.com/en-us/library/ms645450.aspx
// NOTE: forwarded from USER32!DefDlgProcW

__declspec(dllimport)
LRESULT	NTAPI	NtdllDialogWndProc_W(HWND   hDlg,
		                     UINT   Msg,
		                     WPARAM wParam,
		                     LPARAM lParam);

// https://msdn.microsoft.com/en-us/library/ms633572.aspx
// NOTE: forwarded from USER32!DefWindowProcA

__declspec(dllimport)
LRESULT	NTAPI	NtdllDefWindowProc_A(HWND   hWnd,
		                     UINT   Msg,
		                     WPARAM wParam,
		                     LPARAM lParam);

// https://msdn.microsoft.com/en-us/library/ms633572.aspx
// NOTE: forwarded from USER32!DefWindowProcW

__declspec(dllimport)
LRESULT	NTAPI	NtdllDefWindowProc_W(HWND   hWnd,
		                     UINT   Msg,
		                     WPARAM wParam,
		                     LPARAM lParam);
#endif // _WIN32_WINNT

#if 0
// https://msdn.microsoft.com/en-us/library/aa374892.aspx

typedef	struct	_ACCESS_MASK
{
	WORD	SpecificRights;
	BYTE	StandardRights;
	BYTE	AccessSystemAcl : 1;
	BYTE	Reserved        : 3;
	BYTE	GenericAll      : 1;
	BYTE	GenericExecute  : 1;
	BYTE	GenericWrite    : 1;
	BYTE	GenericRead     : 1;
} ACCESS_MASK, *LPACCESS_MASK, *PACCESS_MASK;
#endif

__declspec(dllimport)
LONG	NTAPI	NtAccessCheck(SECURITY_DESCRIPTOR *SecurityDescriptor,
		              HANDLE              ClientToken,
		              ACCESS_MASK         DesiredAccess,
		              GENERIC_MAPPING     *GenericMapping,
		              PRIVILEGE_SET       *PrivilegeSet,
		              LPDWORD             PrivilegeSetLength,
		              ACCESS_MASK         *GrantedAccess,
		              NTSTATUS            *AccessStatus);

__declspec(dllimport)
LONG	NTAPI	NtAccessCheckAndAuditAlarm(UNICODE_STRING      *SubSystemName,
		                           LPVOID              HandleId,
		                           UNICODE_STRING      *ObjectTypeName,
		                           UNICODE_STRING      *ObjectName,
		                           SECURITY_DESCRIPTOR *SecurityDescriptor,
		                           ACCESS_MASK         DesiredAccess,
		                           GENERIC_MAPPING     *GenericMapping,
		                           BOOLEAN             ObjectCreation,
		                           ACCESS_MASK         *GrantedAccess,
		                           NTSTATUS            *AccessStatus,
		                           BOOLEAN             *GenerateOnClose);

__declspec(dllimport)
LONG	NTAPI	NtAccessCheckByType(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                    SID                 *PrincipalSelfSid,
		                    HANDLE              ClientToken,
		                    ACCESS_MASK         DesiredAccess,
		                    OBJECT_TYPE_LIST    *ObjectTypeList,
		                    DWORD               ObjectTypeListLength,
		                    GENERIC_MAPPING     *GenericMapping,
		                    PRIVILEGE_SET       *PrivilegeSet,
		                    LPDWORD             PrivilegeSetLength,
		                    ACCESS_MASK         *GrantedAccess,
		                    NTSTATUS            *AccessStatus);

__declspec(dllimport)
LONG	NTAPI	NtAccessCheckByTypeAndAuditAlarm(UNICODE_STRING      *SubSystemName,
		                                 LPVOID              HandleId,
		                                 UNICODE_STRING      *ObjectTypeName,
		                                 UNICODE_STRING      *ObjectName,
		                                 SECURITY_DESCRIPTOR *SecurityDescriptor,
		                                 SID                 *PrincipalSelfSid,
		                                 ACCESS_MASK         DesiredAccess,
		                                 AUDIT_EVENT_TYPE    AuditType,
		                                 DWORD               Flags,
		                                 OBJECT_TYPE_LIST    *ObjectTypeList,
		                                 DWORD               ObjectTypeListLength,
		                                 GENERIC_MAPPING     *GenericMapping,
		                                 BOOLEAN             ObjectCreation,
		                                 ACCESS_MASK         *GrantedAccess,
		                                 NTSTATUS            *AccessStatus,
		                                 BOOLEAN             *GenerateOnClose);


__declspec(dllimport)
LONG	NTAPI	NtAccessCheckByTypeResultList(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                              SID                 *PrincipalSelfSid,
		                              HANDLE              ClientToken,
		                              ACCESS_MASK         DesiredAccess,
		                              OBJECT_TYPE_LIST    *ObjectTypeList,
		                              DWORD               ObjectTypeListLength,
		                              GENERIC_MAPPING     *GenericMapping,
		                              PRIVILEGE_SET       *PrivilegeSet,
		                              LPDWORD             PrivilegeSetLength,
		                              ACCESS_MASK         *GrantedAccess,
		                              NTSTATUS            *AccessStatus);

__declspec(dllimport)
LONG	NTAPI	NtAccessCheckByTypeResultListAndAuditAlarm(UNICODE_STRING      *SubSystemName,
		                                           LPVOID              HandleId,
		                                           UNICODE_STRING      *ObjectTypeName,
		                                           UNICODE_STRING      *ObjectName,
		                                           SECURITY_DESCRIPTOR *SecurityDescriptor,
		                                           SID                 *PrincipalSelfSid,
		                                           ACCESS_MASK         DesiredAccess,
		                                           AUDIT_EVENT_TYPE    AuditType,
		                                           DWORD               Flags,
		                                           OBJECT_TYPE_LIST    *ObjectTypeList,
		                                           DWORD               ObjectTypeListLength,
		                                           GENERIC_MAPPING     *GenericMapping,
		                                           BOOLEAN             ObjectCreation,
		                                           ACCESS_MASK         *GrantedAccess,
		                                           NTSTATUS            *AccessStatus,
		                                           BOOLEAN             *GenerateOnClose);

typedef	WORD	RTL_ATOM;

__declspec(dllimport)
LONG	NTAPI	NtAddAtom(LPCWSTR  AtomName,
		          DWORD    Length,
		          RTL_ATOM *Atom);

#define FILE_PATH_VERSION		1UL

#define FILE_PATH_TYPE_ARC		1UL
#define FILE_PATH_TYPE_ARC_SIGNATURE	2UL
#define FILE_PATH_TYPE_NT		3UL
#define FILE_PATH_TYPE_EFI		4UL

#define FILE_PATH_TYPE_MIN		FILE_PATH_TYPE_ARC
#define FILE_PATH_TYPE_MAX		FILE_PATH_TYPE_EFI

typedef	struct	_FILE_PATH
{
	DWORD	Version;
	DWORD	Length;
	DWORD	Type;
	UCHAR	FilePath[ANYSIZE_ARRAY];
} FILE_PATH, *LPFILE_PATH, *PFILE_PATH;

#define BOOT_ENTRY_VERSION			1UL

#define BOOT_ENTRY_ATTRIBUTE_ACTIVE		0x00000001UL
#define BOOT_ENTRY_ATTRIBUTE_DEFAULT		0x00000002UL
#define BOOT_ENTRY_ATTRIBUTE_WINDOWS		0x00000004UL
#define BOOT_ENTRY_ATTRIBUTE_REMOVABLE_MEDIA	0x00000008UL

typedef	struct	_BOOT_ENTRY
{
	DWORD	Version;
	DWORD	Length;
	DWORD	Id;
	DWORD	Attributes;
	DWORD	FriendlyNameOffset;
	DWORD	BootFilePathOffset;
	DWORD	OsOptionsLength;
	UCHAR	OsOptions[ANYSIZE_ARRAY];
//	WCHAR	FriendlyName[ANYSIZE_ARRAY];
//	FILE_PATH	BootFilePath;
} BOOT_ENTRY, *LPBOOT_ENTRY, *PBOOT_ENTRY;

__declspec(dllimport)
LONG	NTAPI	NtAddBootEntry(BOOT_ENTRY *BootEntry,
		               LPDWORD    Id);

// NOTE: like ADVAPI32!AdjustTokenGroups

__declspec(dllimport)
LONG	NTAPI	NtAdjustGroupsToken(HANDLE       TokenHandle,
		                    BOOLEAN      ResetToDefault,
		                    TOKEN_GROUPS *NewGroups,
		                    DWORD        PreviousGroupsLength,
		                    TOKEN_GROUPS *PreviousGroups,
		                    LPDWORD      ReturnLength);

// NOTE: like ADVAPI32!AdjustTokenPrivileges

__declspec(dllimport)
LONG	NTAPI	NtAdjustPrivilegesToken(HANDLE           TokenHandle,
		                        BOOLEAN          DisableAllPrivileges,
		                        TOKEN_PRIVILEGES *NewPrivileges,
		                        DWORD            PreviousPrivilegesLength,
		                        TOKEN_PRIVILEGES *PreviousPrivileges,
		                        LPDWORD          ReturnLength);


__declspec(dllimport)
LONG	NTAPI	NtAlertResumeThread(HANDLE  ThreadHandle,
		                    LPDWORD PreviousSuspendCount);

__declspec(dllimport)
LONG	NTAPI	NtAlertThread(HANDLE ThreadHandle);

// https://msdn.microsoft.com/en-us/library/ff556436.aspx
// https://msdn.microsoft.com/en-us/library/ff566415.aspx

__declspec(dllimport)
LONG	NTAPI	NtAllocateLocallyUniqueId(LUID *Luid);


#if _WIN32_WINNT >= 0x0601
typedef	enum	_MEMORY_RESERVE_TYPE
{
	USER_APC_RESERVE_TYPE      = 0,
	IO_COMPLETION_RESERVE_TYPE = 1
} MEMORY_RESERVE_TYPE;

__declspec(dllimport)
LONG	NTAPI	NtAllocateReserveObject(HANDLE              *MemoryReserveHandle,
		                        OBJECT_ATTRIBUTES   *ObjectAttributes,
		                        MEMORY_RESERVE_TYPE Type);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtAllocateUserPhysicalPages(HANDLE    ProcessHandle,
		                            DWORD_PTR *NumberOfPages,
		                            DWORD_PTR *UserPfnArray);

__declspec(dllimport)
LONG	NTAPI	NtAllocateUuids(ULARGE_INTEGER *Time,
		                LPDWORD        Range,
		                LPDWORD        Sequence,
		                LPBYTE         Seed);

// https://msdn.microsoft.com/en-us/library/ff556440.aspx
// https://msdn.microsoft.com/en-us/library/ff566416.aspx

__declspec(dllimport)
LONG	NTAPI	NtAllocateVirtualMemory(HANDLE    ProcessHandle,
		                        LPVOID    *BaseAddress,
		                        DWORD_PTR ZeroBits,
		                        SIZE_T    *RegionSize,
		                        DWORD     AllocationType,
		                        DWORD     Protect);

__declspec(dllimport)
LONG	NTAPI	NtAreMappedFilesTheSame(LPVOID File1MappedAsAnImage,
		                        LPVOID File2MappedAsFile);

// NOTE: like KERNEL32!AssignProcessToJobObject

__declspec(dllimport)
LONG	NTAPI	NtAssignProcessToJobObject(HANDLE JobHandle,
		                           HANDLE ProcessHandle);

__declspec(dllimport)
LONG	NTAPI	NtCancelTimer(HANDLE  TimerHandle,
		              BOOLEAN *CurrentState);

__declspec(dllimport)
LONG	NTAPI	NtClearEvent(HANDLE EventHandle);

// https://msdn.microsoft.com/en-us/library/ms648410.aspx
// https://msdn.microsoft.com/en-us/library/ff566417.aspx

__declspec(dllimport)
LONG	NTAPI	NtClose(HANDLE Handle);

__declspec(dllimport)
LONG	NTAPI	NtCloseObjectAuditAlarm(UNICODE_STRING *SubSystemName,
		                        LPVOID         HandleId,
		                        BOOLEAN        GenerateOnClose);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff566418.aspx

__declspec(dllimport)
LONG	NTAPI	NtCommitComplete(HANDLE        EnlistmentHandle,
		                 LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff566419.aspx

__declspec(dllimport)
LONG	NTAPI	NtCommitEnlistment(HANDLE        EnlistmentHandle,
		                   LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff566420.aspx

__declspec(dllimport)
LONG	NTAPI	NtCommitTransaction(HANDLE  TransactionHandle,
		                    BOOLEAN Wait);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtCompactKeys(DWORD  Count,
		              HANDLE KeyArray[]);

// https://msdn.microsoft.com/en-us/library/aa379285.aspx

__declspec(dllimport)
LONG	NTAPI	NtCompareTokens(HANDLE  FirstTokenHandle,
		                HANDLE  SecondTokenHandle,
		                BOOLEAN *Equal);

__declspec(dllimport)
LONG	NTAPI	NtCompressKey(HANDLE KeyHandle);

__declspec(dllimport)
LONG	NTAPI	NtContinue(CONTEXT *ContextRecord,
		           BOOLEAN TestAlert);

#if 0	// NOTE: only present in x86 emulator of Microsoft anti-malware engine!
__declspec(dllimport)
LONG	NTAPI	NtControlChannel(DWORD  ControlCode,
		                 LPVOID Data);
#endif

#define DEBUG_READ_EVENT		0x00000001UL
#define DEBUG_PROCESS_ASSIGN		0x00000002UL
#define DEBUG_SET_INFORMATION		0x00000004UL
#define DEBUG_QUERY_INFORMATION		0x00000008UL
#define DEBUG_ALL_ACCESS		(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | DEBUG_READ_EVENT | DEBUG_PROCESS_ASSIGN | DEBUG_SET_INFORMATION | DEBUG_QUERY_INFORMATION)

__declspec(dllimport)
LONG	NTAPI	NtCreateDebugObject(HANDLE            *DebugObjectHandle,
		                    ACCESS_MASK       DesiredAccess,
		                    OBJECT_ATTRIBUTES *ObjectAttributes,
		                    DWORD             Flags);

// https://msdn.microsoft.com/en-us/library/ff556456.aspx
// https://msdn.microsoft.com/en-us/library/ff566421.aspx

#define DIRECTORY_QUERY			0x00000001UL
#define DIRECTORY_TRAVERSE		0x00000002UL
#define DIRECTORY_CREATE_OBJECT		0x00000004UL
#define DIRECTORY_CREATE_SUBDIRECTORY	0x00000008UL
#define DIRECTORY_ALL_ACCESS		(STANDARD_RIGHTS_REQUIRED | DIRECTORY_QUERY | DIRECTORY_TRAVERSE | DIRECTORY_CREATE_OBJECT | DIRECTORY_CREATE_SUBDIRECTORY)

__declspec(dllimport)
LONG	NTAPI	NtCreateDirectoryObject(HANDLE            *DirectoryHandle,
		                        ACCESS_MASK       DesiredAccess,
		                        OBJECT_ATTRIBUTES *ObjectAttributes);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff566422.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateEnlistment(HANDLE            *EnlistmentHandle,
		                   ACCESS_MASK       DesiredAccess,
		                   HANDLE            ResourceManagerHandle,
		                   HANDLE            TransactionHandle,
		                   OBJECT_ATTRIBUTES *ObjectAttributes,
		                   DWORD             CreateOptions,
		                   NOTIFICATION_MASK NotificationMask,
		                   LPVOID            EnlistmentKey);
#endif

// https://msdn.microsoft.com/en-us/library/ff566423.aspx

#define EVENT_QUERY_STATE	0x00000001UL
#ifndef _WINNT_
#define EVENT_MODIFY_STATE	0x00000002UL
#define EVENT_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | EVENT_QUERY_STATE | EVENT_MODIFY_STATE)
#endif

typedef	enum	_EVENT_TYPE
{
	NotificationEvent    = 0,
	SynchronizationEvent = 1
} EVENT_TYPE, *LPEVENT_TYPE, *PEVENT_TYPE;

__declspec(dllimport)
LONG	NTAPI	NtCreateEvent(HANDLE            *EventHandle,
		              ACCESS_MASK       DesiredAccess,
		              OBJECT_ATTRIBUTES *ObjectAttributes,
		              EVENT_TYPE        EventType,
		              BOOLEAN           InitialState);

#define FILE_SUPERSEDED		0UL
#define FILE_OPENED		1UL
#define FILE_CREATED		2UL
#define FILE_OVERWRITTEN	3UL
#define FILE_EXISTS		4UL
#define FILE_DOES_NOT_EXIST	5UL

#define FILE_SUPERSEDE			0x00000000UL
#define FILE_OPEN			0x00000001UL
#define FILE_CREATE			0x00000002UL
#define FILE_OPEN_IF			0x00000003UL
#define FILE_OVERWRITE			0x00000004UL
#define FILE_OVERWRITE_IF		0x00000005UL
#define FILE_MAXIMUM_DISPOSITION	0x00000005UL

#define FILE_DIRECTORY_FILE		0x00000001UL
#define FILE_WRITE_THROUGH		0x00000002UL
#define FILE_SEQUENTIAL_ONLY		0x00000004UL
#define FILE_NO_INTERMEDIATE_BUFFERING	0x00000008UL
#define FILE_SYNCHRONOUS_IO_ALERT	0x00000010UL
#define FILE_SYNCHRONOUS_IO_NONALERT	0x00000020UL
#define FILE_NON_DIRECTORY_FILE		0x00000040UL
#define FILE_CREATE_TREE_CONNECTION	0x00000080UL
#define FILE_COMPLETE_IF_OPLOCKED	0x00000100UL
#define FILE_NO_EA_KNOWLEDGE		0x00000200UL
#define FILE_OPEN_FOR_RECOVERY		0x00000400UL
#define FILE_RANDOM_ACCESS		0x00000800UL
#define FILE_DELETE_ON_CLOSE		0x00001000UL
#define FILE_OPEN_BY_FILE_ID		0x00002000UL
#define FILE_OPEN_FOR_BACKUP_INTENT	0x00004000UL
#define FILE_NO_COMPRESSION		0x00008000UL
#define FILE_OPEN_REQUIRING_OPLOCK	0x00010000UL
#define FILE_DISALLOW_EXCLUSIVE		0x00020000UL
#define FILE_SESSION_AWARE		0x00040000UL
#define FILE_RESERVE_OPFILTER		0x00100000UL
#define FILE_OPEN_REPARSE_POINT		0x00200000UL
#define FILE_OPEN_NO_RECALL		0x00400000UL
#define FILE_OPEN_FOR_FREE_SPACE_QUERY	0x00800000UL

#define FILE_VALID_OPTION_FLAGS			0x00FFFFFFUL
#define FILE_VALID_PIPE_OPTION_FLAGS		0x00000032UL
#define FILE_VALID_MAILSLOT_OPTION_FLAGS	0x00000032UL
#define FILE_VALID_SET_FLAGS			0x00000036UL

#define FILE_WRITE_TO_END_OF_FILE	0xFFFFFFFFUL
#define FILE_USE_FILE_POINTER_POSITION	0xFFFFFFFEUL

#define FILE_BYTE_ALIGNMENT		0x00000000UL
#define FILE_WORD_ALIGNMENT		0x00000001UL
#define FILE_LONG_ALIGNMENT		0x00000003UL
#define FILE_QUAD_ALIGNMENT		0x00000007UL
#define FILE_OCTA_ALIGNMENT		0x0000000FUL
#define FILE_32_BYTE_ALIGNMENT		0x0000001FUL
#define FILE_64_BYTE_ALIGNMENT		0x0000003FUL
#define FILE_128_BYTE_ALIGNMENT		0x0000007FUL
#define FILE_256_BYTE_ALIGNMENT		0x000000FFUL
#define FILE_512_BYTE_ALIGNMENT		0x000001FFUL

// https://msdn.microsoft.com/en-us/library/bb432380.aspx
// https://msdn.microsoft.com/en-us/library/ff566424.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateFile(HANDLE            *FileHandle,
		             ACCESS_MASK       DesiredAccess,
		             OBJECT_ATTRIBUTES *ObjectAttributes,
		             IO_STATUS_BLOCK   *IoStatusBlock,
		             LARGE_INTEGER     *AllocationSize,
		             DWORD             FileAttributes,
		             DWORD             ShareAccess,
		             DWORD             CreateDisposition,
		             DWORD             CreateOptions,
		             LPVOID            ExtendedAttributesBuffer,
		             DWORD             ExtendedAttributesLength);

#define IO_COMPLETION_QUERY_STATE	0x00000001UL
#ifndef _WINNT_
#define IO_COMPLETION_MODIFY_STATE	0x00000002UL
#define IO_COMPLETION_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | IO_COMPLETION_QUERY_STATE | IO_COMPLETION_MODIFY_STATE)
#endif

__declspec(dllimport)
LONG	NTAPI	NtCreateIoCompletion(HANDLE            *IoCompletionPortHandle,
		                     ACCESS_MASK       DesiredAccess,
		                     OBJECT_ATTRIBUTES *ObjectAttributes,
	                             DWORD             NumberOfConcurrentThreads);

__declspec(dllimport)
LONG	NTAPI	NtCreateJobObject(HANDLE            *JobHandle,
		                  ACCESS_MASK       DesiredAccess,
		                  OBJECT_ATTRIBUTES *ObjectAttributes);

#ifndef _WINNT_
typedef	struct	_JOB_SET_ARRAY
{
	HANDLE	JobHandle;
	DWORD	MemberLevel;
	DWORD	Flags;
} JOB_SET_ARRAY, *LPJOB_SET_ARRAY, *PJOB_SET_ARRAY;
#endif

__declspec(dllimport)
LONG	NTAPI	NtCreateJobSet(DWORD         NumberOfJobs,
		               JOB_SET_ARRAY *UserJobSet,
		               DWORD         Flags);

// https://msdn.microsoft.com/en-us/library/ff556468.aspx
// https://msdn.microsoft.com/en-us/library/ff566425.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateKey(HANDLE            *KeyHandle,
		            ACCESS_MASK       DesiredAccess,
		            OBJECT_ATTRIBUTES *ObjectAttributes,
		            DWORD             TitleIndex,
		            UNICODE_STRING    *Class,
		            DWORD             CreateOptions,
		            LPDWORD           Disposition);

#ifndef _WINNT_
#define MUTANT_QUERY_STATE	0x00000001UL
#define MUTANT_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | MUTANT_QUERY_STATE)
#endif

__declspec(dllimport)
LONG	NTAPI	NtCreateMailslotFile(HANDLE            *FileHandle,
		                     DWORD             DesiredAccess,
		                     OBJECT_ATTRIBUTES *ObjectAttributes,
		                     IO_STATUS_BLOCK   *IoStatusBlock,
		                     DWORD             CreateOptions,
		                     DWORD             MailslotQuota,
		                     DWORD             MaximumMessageSize,
		                     LARGE_INTEGER     *ReadTimeout);

__declspec(dllimport)
LONG	NTAPI	NtCreateMutant(HANDLE            *MutantHandle,
		               ACCESS_MASK       DesiredAccess,
		               OBJECT_ATTRIBUTES *ObjectAttributes,
		               BOOLEAN           InitialOwner);

#define FILE_PIPE_BYTE_STREAM_TYPE	0x00000000UL
#define FILE_PIPE_MESSAGE_TYPE		0x00000001UL
#define FILE_PIPE_ACCEPT_REMOTE_CLIENTS	0x00000000UL
#define FILE_PIPE_REJECT_REMOTE_CLIENTS	0x00000002UL
#define FILE_PIPE_TYPE_VALID_MASK	0x00000003UL

#define FILE_PIPE_QUEUE_OPERATION	0x00000000UL
#define FILE_PIPE_COMPLETE_OPERATION	0x00000001UL

#define FILE_PIPE_BYTE_STREAM_MODE	0x00000000UL
#define FILE_PIPE_MESSAGE_MODE		0x00000001UL

__declspec(dllimport)
LONG	NTAPI	NtCreateNamedPipeFile(HANDLE            *FileHandle,
		                      DWORD             DesiredAccess,
		                      OBJECT_ATTRIBUTES *ObjectAttributes,
		                      IO_STATUS_BLOCK   *IoStatusBlock,
		                      DWORD             ShareAccess,
		                      DWORD             CreateDisposition,
		                      DWORD             CreateOptions,
		                      DWORD             NamedPipeType,
		                      DWORD             ReadMode,
		                      DWORD             CompletionMode,
		                      DWORD             MaximumInstances,
		                      DWORD             InboundQuota,
		                      DWORD             OutboundQuota,
		                      LARGE_INTEGER     *DefaultTimeout);

__declspec(dllimport)
LONG	NTAPI	NtCreatePagingFile(UNICODE_STRING *PageFileName,
		                   LARGE_INTEGER  *MinimumSize,
		                   LARGE_INTEGER  *MaximumSize,
		                   DWORD          Flags);

#if _WIN32_WINNT >= 0x0600
#pragma warning(disable: 4201)
typedef	struct	_BOUNDARY_DESCRIPTOR
{
	DWORD	Version;
	DWORD	Count;			// number of entries
	DWORD	Size;			// size of boundary descriptor
	DWORD	Flags;
	struct
	{
		DWORD	Type;		// 1: name
		DWORD	Size;		// size of entry
		union
		{
			WCHAR	Name[ANYSIZE_ARRAY];
			BYTE	Data[ANYSIZE_ARRAY];
		};
	} Entries[ANYSIZE_ARRAY];
} BOUNDARY_DESCRIPTOR, *LPBOUNDARY_DESCRIPTOR, *PBOUNDARY_DESCRIPTOR;
#pragma warning(default: 4201)

__declspec(dllimport)
LONG	NTAPI	NtCreatePrivateNamespace(HANDLE            *NamespaceHandle,
		                         ACCESS_MASK       DesiredAccess,
		                         OBJECT_ATTRIBUTES *ObjectAttributes,
		                         LPVOID            BoundaryDescriptor);
#endif // _WIN32_WINNT

typedef	enum	_KPROFILE_SOURCE
{
	ProfileTime                 = 0,
	ProfileAlignmentFixup,
	ProfileTotalIssues,
	ProfilePipelineDry,
	ProfileLoadInstructions,
	ProfilePipelineFrozen,
	ProfileBranchInstructions,
	ProfileTotalNonissues,
	ProfileDcacheMisses,
	ProfileIcacheMisses,
	ProfileCacheMisses,
	ProfileBranchMispredictions,
	ProfileStoreInstructions,
	ProfileFpInstructions,
	ProfileIntegerInstructions,
	Profile2Issue,
	Profile3Issue,
	Profile4Issue,
	ProfileSpecialInstructions,
	ProfileTotalCycles,
	ProfileIcacheIssues,
	ProfileDcacheAccesses,
	ProfileMemoryBarrierCycles,
	ProfileLoadLinkedIssues,
	ProfileMaximum
} KPROFILE_SOURCE;

typedef	DWORD_PTR	KAFFINITY;

__declspec(dllimport)
LONG	NTAPI	NtCreateProfile(HANDLE          *ProfileHandle,
		                HANDLE          ProcessHandle,
		                LPVOID          ProfileBase,
		                SIZE_T          ProfileSize,
		                DWORD           BucketSize,
		                LPDWORD         Buffer,
		                DWORD           BufferSize,
		                KPROFILE_SOURCE ProfileSource,
		                KAFFINITY       Affinity);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	NtCreateProfileEx(HANDLE          *ProfileHandle,
		                  HANDLE          ProcessHandle,
		                  LPVOID          ProfileBase,
		                  SIZE_T          ProfileSize,
		                  DWORD           BucketSize,
		                  LPDWORD         Buffer,
		                  DWORD           BufferSize,
		                  KPROFILE_SOURCE ProfileSource,
		                  WORD            GroupCount,
		                  GROUP_AFFINITY  *AffinityArray);
#endif

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff566427.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateResourceManager(HANDLE            *ResourceManagerHandle,
		                        ACCESS_MASK       DesiredAccess,
		                        HANDLE            TmHandle,
		                        GUID              *ResourceManagerGuid,
		                        OBJECT_ATTRIBUTES *ObjectAttributes,
		                        DWORD             CreateOptions,
		                        UNICODE_STRING    *Description);
#endif

// https://msdn.microsoft.com/en-us/library/ff556473.aspx
// https://msdn.microsoft.com/en-us/library/ff566428.aspx

#ifndef _WINNT_
#define SECTION_QUERY			0x00000001UL
#define SECTION_MAP_WRITE		0x00000002UL
#define SECTION_MAP_READ		0x00000004UL
#define SECTION_MAP_EXECUTE		0x00000008UL
#define SECTION_EXTEND_SIZE		0x00000010UL
#define SECTION_MAP_EXECUTE_EXPLICIT	0x00000020UL	// not set in SECTION_ALL_ACCESS
#define SECTION_ALL_ACCESS		(STANDARD_RIGHTS_REQUIRED | SECTION_QUERY | SECTION_MAP_WRITE | SECTION_MAP_READ | SECTION_MAP_EXECUTE | SECTION_EXTEND_SIZE)
#endif

__declspec(dllimport)
LONG	NTAPI	NtCreateSection(HANDLE            *SectionHandle,
		                ACCESS_MASK       DesiredAccess,
		                OBJECT_ATTRIBUTES *ObjectAttributes,
		                LARGE_INTEGER     *MaximumSize,
		                DWORD             SectionPageProtection,
		                DWORD             AllocationAttributes,
		                HANDLE            FileHandle);

#define SEMAPHORE_QUERY_STATE	0x00000001UL
#ifndef _WINNT_
#define SEMAPHORE_MODIFY_STATE	0x00000002UL
#define SEMAPHORE_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | SEMAPHORE_QUERY_STATE | SEMAPHORE_MODIFY_STATE)
#endif

__declspec(dllimport)
LONG	NTAPI	NtCreateSemaphore(HANDLE            *SemaphoreHandle,
		                  ACCESS_MASK       DesiredAccess,
		                  OBJECT_ATTRIBUTES *ObjectAttributes,
		                  DWORD             InitialCount,
		                  DWORD             MaximumCount);

#define SYMBOLIC_LINK_QUERY		0x00000001UL
#define SYMBOLIC_LINK_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYMBOLIC_LINK_QUERY)

__declspec(dllimport)
LONG	NTAPI	NtCreateSymbolicLinkObject(HANDLE            *LinkHandle,
		                           ACCESS_MASK       DesiredAccess,
		                           OBJECT_ATTRIBUTES *ObjectAttributes,
		                           UNICODE_STRING    *LinkTarget);

#ifndef _WINNT_
#define TIMER_QUERY_STATE	0x00000001UL
#define TIMER_MODIFY_STATE	0x00000002UL
#define TIMER_ALL_ACCESS	(STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | TIMER_QUERY_STATE | TIMER_MODIFY_STATE)
#endif

typedef	enum	_TIMER_TYPE
{
	NotificationTimer    = 0,
	SynchronizationTimer
} TIMER_TYPE;

__declspec(dllimport)
LONG	NTAPI	NtCreateTimer(HANDLE            *TimerHandle,
		              ACCESS_MASK       DesiredAccess,
		              OBJECT_ATTRIBUTES *ObjectAttributes,
		              TIMER_TYPE        TimerType);

typedef	struct	_INITIAL_TEB
{
	struct
	{
		LPVOID	OldStackBase;
		LPVOID	OldStackLimit;
	} OldInitialTeb;
	LPVOID	StackBase;
	LPVOID	StackLimit;
	LPVOID	StackAllocationBase;
} INITIAL_TEB, *LPINITIAL_TEB, *PINITIAL_TEB;

__declspec(dllimport)
LONG	NTAPI	NtCreateThread(HANDLE            *ThreadHandle,
		               ACCESS_MASK       DesiredAccess,
		               OBJECT_ATTRIBUTES *ObjectAttributes,
		               HANDLE            ProcessHandle,
		               CLIENT_ID         *ClientId,
		               CONTEXT           *ThreadContext,
		               INITIAL_TEB       *InitialTeb,
		               BOOLEAN           CreateSuspended);

#pragma warning(disable: 4201)
typedef	struct	_PS_ATTRIBUTE
{
	DWORD	Attribute;
	SIZE_T	Size;
	union
	{
		DWORD	Value;
		LPVOID	ValuePtr;
	};
	SIZE_T	*ReturnLength;
} PS_ATTRIBUTE, *LPPS_ATTRIBUTE, *PPS_ATTRIBUTE;
#pragma warning(default: 4201)

typedef	struct	_PS_ATTRIBUTE_LIST
{
	SIZE_T	TotalLength;
	PS_ATTRIBUTE	Attributes[ANYSIZE_ARRAY];
} PS_ATTRIBUTE_LIST, *LPPS_ATTRIBUTE_LIST, *PPS_ATTRIBUTE_LIST;

typedef	LONG	(NTAPI	THREAD_START_ROUTINE) (LPVOID ThreadParameter);

#define THREAD_CREATE_FLAGS_CREATE_SUSPENDED		0x00000001UL
#define THREAD_CREATE_FLAGS_SKIP_THREAD_ATTACH		0x00000002UL
#define THREAD_CREATE_FLAGS_HIDE_FROM_DEBUGGER		0x00000004UL
#define THREAD_CREATE_FLAGS_HAS_SECURITY_DESCRIPTOR	0x00000010UL
#define THREAD_CREATE_FLAGS_ACCESS_CHECK_IN_TARGET	0x00000020UL
#define THREAD_CREATE_FLAGS_INITIAL_THREAD		0x00000080UL

__declspec(dllimport)
LONG	NTAPI	NtCreateThreadEx(HANDLE               *ThreadHandle,
		                 ACCESS_MASK          DesiredAccess,
		                 OBJECT_ATTRIBUTES    *ObjectAttributes,
		                 HANDLE               ProcessHandle,
		                 THREAD_START_ROUTINE *StartAddress,
		                 LPVOID               Parameter,
		                 DWORD                CreateFlags,
		                 SIZE_T               StackZeroBits,
		                 SIZE_T               SizeOfStackCommit,
		                 SIZE_T               SizeOfStackReserve,
		                 PS_ATTRIBUTE_LIST    *AttributeList);

__declspec(dllimport)
LONG	NTAPI	NtCreateToken(HANDLE              *TokenHandle,
		              ACCESS_MASK         DesiredAccess,
		              OBJECT_ATTRIBUTES   *ObjectAttributes,
		              TOKEN_TYPE          TokenType,
		              LUID                *AuthenticationId,
		              LARGE_INTEGER       *ExpirationTime,
		              TOKEN_USER          *TokenUser,
		              TOKEN_GROUPS        *TokenGroups,
		              TOKEN_PRIVILEGES    *TokenPrivileges,
		              TOKEN_OWNER         *TokenOwner,
		              TOKEN_PRIMARY_GROUP *TokenPrimaryGroup,
		              TOKEN_DEFAULT_DACL  *TokenDefaultDacl,
		              TOKEN_SOURCE        *TokenSource);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff566429.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateTransaction(HANDLE            *TransactionHandle,
		                    ACCESS_MASK       DesiredAccess,
		                    OBJECT_ATTRIBUTES *ObjectAttributes,
		                    GUID              *Uow,
		                    HANDLE            TmHandle,
		                    DWORD             CreateOptions,
		                    DWORD             IsolationLevel,
		                    DWORD             IsolationFlags,
		                    LARGE_INTEGER     *Timeout,
		                    UNICODE_STRING    *Description);

// https://msdn.microsoft.com/en-us/library/ff566430.aspx

__declspec(dllimport)
LONG	NTAPI	NtCreateTransactionManager(HANDLE            *TmHandle,
		                           ACCESS_MASK       DesiredAccess,
		                           OBJECT_ATTRIBUTES *ObjectAttributes,
		                           UNICODE_STRING    LogFileName,
		                           DWORD             CreateOptions,
		                           DWORD             CommitStrength);
#endif // _WIN32_WINNT

#define DOS_MAX_COMPONENT_LENGTH	255
#define DOS_MAX_PATH_LENGTH		(DOS_MAX_COMPONENT_LENGTH + 5)

typedef	struct	_CURDIR
{
	UNICODE_STRING	DosPath;
	HANDLE		Handle;
} CURDIR, *LPCURDIR, *PCURDIR;

typedef	struct	_RTL_DRIVE_LETTER_CURDIR
{
	WORD	Flags;
	WORD	Length;
	DWORD	TimeStamp;
	STRING	DosPath;
} RTL_DRIVE_LETTER_CURDIR, *LPRTL_DRIVE_LETTER_CURDIR, *PRTL_DRIVE_LETTER_CURDIR;

#define RTL_MAX_DRIVE_LETTERS	32
#define RTL_DRIVE_LETTER_VALID	0x0001

// https://msdn.microsoft.com/en-us/library/aa813741.aspx

typedef	struct	_RTL_USER_PROCESS_PARAMETERS
{
#if 1
	DWORD			MaximumLength;
	DWORD			Length;
	DWORD			Flags;
	DWORD			DebugFlags;
	HANDLE			ConsoleHandle;
	DWORD			ConsoleFlags;
	HANDLE			StandardInput;
	HANDLE			StandardOutput;
	HANDLE			StandardError;
	CURDIR			CurrentDirectory;
	UNICODE_STRING		DllPath;
#else
	BYTE			Reserved1[16];
	LPVOID			Reserved2[10];
#endif
	UNICODE_STRING		ImagePathName;
	UNICODE_STRING		CommandLine;
#if 1
	LPVOID			Environment;
	DWORD			StartingX;
	DWORD			StartingY;
	DWORD			CountX;
	DWORD			CountY;
	DWORD			CountCharsX;
	DWORD			CountCharsY;
	DWORD			FillAttribute;
	DWORD			WindowFlags;
	DWORD			ShowWindowFlags;
	UNICODE_STRING		WindowTitle;
	UNICODE_STRING		DesktopInfo;
	UNICODE_STRING		ShellInfo;
	UNICODE_STRING		RuntimeData;
	RTL_DRIVE_LETTER_CURDIR	CurrentDirectories[RTL_MAX_DRIVE_LETTERS];
#if _WIN32_WINNT >= 0x0600
	DWORD			EnvironmentSize;
#if _WIN32_WINNT >= 0x0601
	DWORD			EnvironmentVersion;
#if _WIN32_WINNT >= 0x0602
	LPVOID			PackageDependencyData;
	DWORD			ProcessGroupId;
#if _WIN32_WINNT >= 0x0A00
	DWORD			LoaderThreads;
	UNICODE_STRING		RedirectionDllName;
	UNICODE_STRING		HeapPartitionName;
	QWORD			*DefaultThreadpoolCpuSetMasks;
	DWORD			DefaultThreadpoolCpuSetMaskCount;
	DWORD			DefaultThreadpoolThreadMaximum;
#endif
#endif
#endif
#endif
#endif
} RTL_USER_PROCESS_PARAMETERS, *LPRTL_USER_PROCESS_PARAMETERS, *PRTL_USER_PROCESS_PARAMETERS;

#define PROCESS_CREATE_FLAGS_LARGE_PAGE_SYSTEM_DLL	0x00000020UL
#define PROCESS_CREATE_FLAGS_PROTECTED_PROCESS		0x00000040UL
#define PROCESS_CREATE_FLAGS_CREATE_SESSION		0x00000080UL
#define PROCESS_CREATE_FLAGS_INHERIT_FROM_PARENT	0x00000100UL

typedef	enum	_PS_CREATE_STATE
{
	PsCreateInitialState        = 0,
	PsCreateFailOnFileOpen,
	PsCreateFailOnSectionCreate,
	PsCreateFailExeFormat,
	PsCreateFailMachineMismatch,
	PsCreateFailExeName,		// debugger specified
	PsCreateSuccess,
	PsCreateMaximumStates
} PS_CREATE_STATE;

#pragma warning(disable: 4201 4214)
typedef	struct	_PS_CREATE_INFO
{
	SIZE_T		Length;
	PS_CREATE_STATE	State;
	union
	{
		struct		// PsCreateInitialState
		{
			union
			{
				DWORD	InitFlags;
				struct
				{
					BYTE	WriteOutputOnExit              : 1;
					BYTE	DetectManifest                 : 1;
					BYTE	IFEOSkipDebugger               : 1;
					BYTE	IFEODoNotPropagateKeyState     : 1;
					BYTE	SpareBits1                     : 4;
					BYTE	SpareBits2                     : 8;
					WORD	ProhibitedImageCharacteristics : 16;
				};
			};
			ACCESS_MASK	AdditionalFileAccess;
		} InitState;

		struct		// PsCreateFailOnSectionCreate
		{
			HANDLE	FileHandle;
		} FailSection;

		struct		// PsCreateFailExeFormat
		{
			WORD	DllCharacteristics;
		} ExeFormat;

		struct		// PsCreateFailExeName
		{
			HANDLE	IFEOKey;
		} ExeName;

		struct		// PsCreateSuccess
		{
			union
			{
				DWORD	OutputFlags;
				struct
				{
					BYTE	ProtectedProcess      : 1;
					BYTE	AddressSpaceOverride  : 1;
					BYTE	DevOverrideEnabled    : 1;	// from ImageFileExecutionOptions
					BYTE	ManifestDetected      : 1;
					BYTE	ProtectedProcessLight : 1;
					BYTE	SpareBits1            : 3;
					BYTE	SpareBits2            : 8;
					WORD	SpareBits3            : 16;
				};
			};
			HANDLE	FileHandle;
			HANDLE	SectionHandle;
			QWORD	UserProcessParametersNative;
			DWORD	UserProcessParametersWow64;
			DWORD	CurrentParameterFlags;
			QWORD	PebAddressNative;
			DWORD	PebAddressWow64;
			QWORD	ManifestAddress;
			DWORD	ManifestSize;
		} SuccessState;
	};
} PS_CREATE_INFO, *LPPS_CREATE_INFO, *PPS_CREATE_INFO;
#pragma warning(default: 4201 4214)

__declspec(dllimport)
LONG	NTAPI	NtCreateUserProcess(HANDLE                      *ProcessHandle,
		                    HANDLE                      *ThreadHandle,
		                    ACCESS_MASK                 ProcessDesiredAccess,
		                    ACCESS_MASK                 ThreadDesiredAccess,
		                    OBJECT_ATTRIBUTES           *ProcessObjectAttributes,
		                    OBJECT_ATTRIBUTES           *ThreadObjectAttributes,
		                    DWORD                       ProcessCreateFlags,
		                    DWORD                       ThreadCreateFlags,
		                    RTL_USER_PROCESS_PARAMETERS *ProcessParameters,
		                    PS_CREATE_INFO              *CreateInfo,
		                    PS_ATTRIBUTE_LIST           *AttributeList);

#if 1
#define NtCurrentPeb()	(NtCurrentTeb()->ProcessEnvironmentBlock)
#else
__forceinline
PEB *	NTAPI	NtCurrentPeb(VOID)
{
	return NtCurrentTeb()->ProcessEnvironmentBlock;
}
#endif

// https://msdn.microsoft.com/en-us/library/ff556482.aspx
// https://msdn.microsoft.com/en-us/library/ff566431.aspx

__forceinline
HANDLE	NTAPI	NtCurrentProcess(VOID)
{
#ifdef _WIN64
	return (HANDLE) -1LL;
#else
	return (HANDLE) -1L;
#endif
}

#if _WIN32_WINNT >= 0x0602
// NOTE: the pseudo handle returned from NtCurrentProcessToken()
//       can only be used with NtGetInformationToken() and
//       NtSetInformationToken()!

__forceinline
HANDLE	NTAPI	NtCurrentProcessToken(VOID)
{
#ifdef _WIN64
	return (HANDLE) -4LL;
#else
	return (HANDLE) -4L;
#endif
}

__forceinline
HANDLE	NTAPI	NtCurrentSession(VOID)
{
#ifdef _WIN64
	return (HANDLE) -3LL;
#else
	return (HANDLE) -3L;
#endif
}
#endif // _WIN32_WINNT

#ifndef _WINNT_
typedef	enum	_EXCEPTION_DISPOSITION
{
	ExceptionContinueExecution = 0,
	ExceptionContinueSearch,
	ExceptionNestedException,
	ExceptionCollidedUnwind
} EXCEPTION_DISPOSITION;
#endif

// https://msdn.microsoft.com/en-us/library/ssa62fwe.aspx

typedef	enum	_UNWIND_OP_CODES
{
	UWOP_PUSH_NONVOL = 0,	// info == register number
	UWOP_ALLOC_LARGE,	// no info, allocation size in next 2 slots
	UWOP_ALLOC_SMALL,	// info == size of allocation / 8 - 1
	UWOP_SET_FPREG,		// no info, FP = RSP + UNWIND_INFO.FPRegOffset * 16
	UWOP_SAVE_NONVOL,	// info == register number, offset in next slot
	UWOP_SAVE_NONVOL_FAR,	// info == register number, offset in next 2 slots
	UWOP_SAVE_XMM128,	// info == XMM reg number, offset in next slot
	UWOP_SAVE_XMM128_FAR,	// info == XMM reg number, offset in next 2 slots
	UWOP_PUSH_MACHFRAME	// info == 0: no error-code, 1: error-code
} UNWIND_CODE_OPS;

#pragma warning(disable: 4201 4214)
// https://msdn.microsoft.com/en-us/library/ck9asaa9.aspx

typedef	union	_UNWIND_CODE
{
	struct
	{
		BYTE	CodeOffset;
		BYTE	UnwindOp : 4;
		BYTE	OpInfo   : 4;
	};
	WORD	FrameOffset;
} UNWIND_CODE, *LPUNWIND_CODE, *PUNWIND_CODE;

// https://msdn.microsoft.com/en-us/library/ddssxxy8.aspx

#define UNW_FLAG_NHANDLER	0x00
#define UNW_FLAG_EHANDLER	0x01
#define UNW_FLAG_UHANDLER	0x02
#define UNW_FLAG_CHAININFO	0x04

typedef	struct	_UNWIND_INFO
{
	BYTE		Version       : 3;
	BYTE		Flags         : 5;
	BYTE		SizeOfProlog;
	BYTE		CountOfCodes;
	BYTE		FrameRegister : 4;
	BYTE		FrameOffset   : 4;
	UNWIND_CODE	UnwindCode[ANYSIZE_ARRAY];
//	UNWIND_CODE	MoreUnwindCode[((CountOfCodes + 1) & ~1) - 1];
//	union
//	{
//		DWORD	ExceptionHandler;
//		DWORD	FunctionEntry;
//	};
//	DWORD	ExceptionData[];
} UNWIND_INFO, *LPUNWIND_INFO, *PUNWIND_INFO;
#pragma warning(default: 4201 4214)

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ft9x1kdx.aspx
// https://msdn.microsoft.com/en-us/library/ssa62fwe.aspx

typedef	struct	_RUNTIME_FUNCTION
{
	DWORD	BeginAddress;
	DWORD	EndAddress;
	DWORD	UnwindData;
} RUNTIME_FUNCTION, *LPRUNTIME_FUNCTION, *PRUNTIME_FUNCTION;

// https://msdn.microsoft.com/en-us/library/b6sf5kbd.aspx
// https://msdn.microsoft.com/en-us/library/ms680617.aspx

typedef	struct	_DISPATCHER_CONTEXT	DISPATCHER_CONTEXT;

typedef	EXCEPTION_DISPOSITION	(CDECL	EXCEPTION_ROUTINE) (EXCEPTION_RECORD   *ExceptionRecord,
					                    LPVOID             EstablisherFrame,
					                    CONTEXT            *ContextRecord,
					                    DISPATCHER_CONTEXT *DispatcherContext);
#endif // _WINNT_

typedef	struct	_DISPATCHER_CONTEXT
{
	LPVOID			ControlPc;
	LPVOID			ImageBase;
#ifdef _WIN64
	RUNTIME_FUNCTION	*FunctionEntry;
#else
	LPVOID			FunctionEntry;
#endif
	LPVOID			EstablisherFrame;
	LPVOID			TargetIp;
	CONTEXT			*ContextRecord;
	EXCEPTION_ROUTINE	*LanguageHandler;
	LPVOID			HandlerData;
} DISPATCHER_CONTEXT, *LPDISPATCHER_CONTEXT, *PDISPATCHER_CONTEXT;

typedef	struct	_EXCEPTION_REGISTRATION_RECORD
{
	struct	_EXCEPTION_REGISTRATION_RECORD	*Next;
	EXCEPTION_ROUTINE			*Handler;
} EXCEPTION_REGISTRATION_RECORD, *LPEXCEPTION_REGISTRATION_RECORD, *PEXCEPTION_REGISTRATION_RECORD;

// https://msdn.microsoft.com/en-us/library/aa813708.aspx

typedef	struct	_PEB_LDR_DATA
{
#if 0
	BYTE		Reserved1[8];
	LPVOID		Reserved2[3];
	LIST_ENTRY	InMemoryOrderModuleList;
#else
	DWORD		Length;
	BOOLEAN		Initialized;
	LPVOID		SsHandle;
	LIST_ENTRY	InLoadOrderModuleList;
	LIST_ENTRY	InMemoryOrderModuleList;
	LIST_ENTRY	InInitializationOrderModuleList;
	LPVOID		EntryInProgress;
#if _WIN32_WINNT >= 0x0600
	BOOLEAN		ShutdownInProgress;
	HANDLE		ShutdownThreadId;
#endif
#endif
} PEB_LDR_DATA, *LPPEB_LDR_DATA, *PPEB_LDR_DATA;

typedef	VOID	(NTAPI	PS_POST_PROCESS_INIT_ROUTINE) (VOID);

// AppCompatFlags

#define KACF_OLDGETSHORTPATHNAME			0x00000001UL
#define KACF_VERSIONLIE_NOT_USED			0x00000002UL
#define KACF_GETDISKFREESPACE				0x00000008UL
#define KACF_FTMFROMCURRENTAPT				0x00000020UL
#define KACF_DISALLOWORBINDINGCHANGES			0x00000040UL
#define KACF_OLE32VALIDATEPTRS				0x00000080UL
#define KACF_DISABLECICERO				0x00000100UL
#define KACF_OLE32ENABLEASYNCDOCFILE			0x00000200UL
#define KACF_OLE32ENABLELEGACYEXCEPTIONHANDLING		0x00000400UL
#define KACF_RPCDISABLENDRCLIENTHARDENING		0x00000800UL
#define KACF_RPCDISABLENDRMAYBENULL_SIZEIS		0x00001000UL
#define KACF_DISABLEALLDDEHACK_NOT_USED			0x00002000UL
#define KACF_RPCDISABLENDR61_RANGE			0x00004000UL
#define KACF_RPC32ENABLELEGACYEXCEPTIONHANDLING		0x00008000UL
#define KACF_OLE32DOCFILEUSELEGACYNTFSFLAGS		0x00010000UL
#define KACF_RPCDISABLENDRCONSTIIDCHECK			0x00020000UL
#define KACF_USERDISABLEFORWARDERPATCH			0x00040000UL
#define KACF_OLE32DISABLENEW_WMPAINT_DISPATCH		0x00100000UL
#define KACF_ADDRESTRICTEDSIDINCOINITIALIZESECURITY	0x00200000UL
#define KACF_ALLOCDEBUGINFOFORCRITSECTIONS		0x00400000UL
#define KACF_OLEAUT32ENABLEUNSAFELOADTYPELIBRELATIVE	0x00800000UL
#define KACF_ALLOWMAXIMIZEDWINDOWGAMMA			0x01000000UL
#define KACF_DONOTADDTOCACHE				0x80000000UL

// https://msdn.microsoft.com/en-us/library/aa813706.aspx

#ifndef _WIN64
typedef	struct	_PEB
{
#if 1
	BOOLEAN				InheritedAddressSpace;
	BOOLEAN				ReadImageFileExecOptions;
#else
	BYTE				Reserved1[2];
#endif
	BOOLEAN				BeingDebugged;
	BYTE				Reserved2[1];
#if 1
	HANDLE				Mutant;
	LPVOID				ImageBaseAddress;
#else
	LPVOID				Reserved3[2];
#endif
	PEB_LDR_DATA			*Ldr;
	RTL_USER_PROCESS_PARAMETERS	*ProcessParameters;
#if 1
	LPVOID				SubSystemData;
	HANDLE				ProcessHeap;
	RTL_CRITICAL_SECTION		*FastPebLock;
	LPVOID				FastPebLockRoutine;
	LPVOID				FastPebUnlockRoutine;
	BYTE				Reserved4[48];
	LPVOID				AnsiCodePageData;
	LPVOID				OemCodePageData;
	LPVOID				UnicodeCaseTableData;
	DWORD				NumberOfProcessors;
	DWORD				NtGlobalFlag;
	LARGE_INTEGER			CriticalSectionTimeout;
	DWORD				HeapSegmentReserve;
	DWORD				HeapSegmentCommit;
	DWORD				HeapDeCommitTotalFreeThreshold;
	DWORD				HeapDeCommitFreeBlockThreshold;
	DWORD				NumberOfHeaps;
	DWORD				MaximumNumberOfHeaps;
	LPVOID				*ProcessHeaps;
	LPVOID				GdiSharedHandleTable;
	LPVOID				ProcessStarterHelper;
	DWORD				GdiDCAttributeList;
	RTL_CRITICAL_SECTION		*LoaderLock;
	DWORD				OSMajorVersion;
	DWORD				OSMinorVersion;
	WORD				OSBuildNumber;
	WORD				OSCSDVersion;
	DWORD				OSPlatformId;
	DWORD				ImageSubsystem;
	DWORD				ImageSubsystemMajorVersion;
	DWORD				ImageSubsystemMinorVersion;
	DWORD				ActiveProcessAffinityMask;
	DWORD				GdiHandleBuffer[34];
#else
	LPVOID				Reserved4[3];
	LPVOID				AtlThunkSListPtr;
	LPVOID				Reserved5;
	DWORD				Reserved6;
	LPVOID				Reserved7;
	DWORD				Reserved8;
	DWORD				AtlThunkSListPtr32;
	LPVOID				Reserved9[45];
	BYTE				Reserved10[96];
#endif
	PS_POST_PROCESS_INIT_ROUTINE	*PostProcessInitRoutine;
	BYTE				Reserved11[128];
	LPVOID				Reserved12[1];
	DWORD				SessionId;
#if 1
	ULARGE_INTEGER			AppCompatFlags;
	ULARGE_INTEGER			AppCompatFlagsUser;
	LPVOID				ShimData;
	LPVOID				AppCompatInfo;
	UNICODE_STRING			CSDVersion;
#endif
} PEB, *LPPEB, *PPEB;
#else // _WIN64
typedef	struct	_PEB
{
#if 1
	BOOLEAN				InheritedAddressSpace;
	BOOLEAN				ReadImageFileExecOptions;
#else
	BYTE				Reserved1[2];
#endif
	BOOLEAN				BeingDebugged;
	BYTE				Reserved2[21];
	PEB_LDR_DATA			*LoaderData;
	RTL_USER_PROCESS_PARAMETERS	*ProcessParameters;
#if 1
	LPVOID				SubSystemData;
	HANDLE				ProcessHeap;
	RTL_CRITICAL_SECTION		*FastPebLock;
	LPVOID				FastPebLockRoutine;
	LPVOID				FastPebUnlockRoutine;
	BYTE				Reserved3[80];
	LPVOID				AnsiCodePageData;
	LPVOID				OemCodePageData;
	LPVOID				UnicodeCaseTableData;
	DWORD				NumberOfProcessors;
	DWORD				NtGlobalFlag;
	LARGE_INTEGER			CriticalSectionTimeout;
	QWORD				HeapSegmentReserve;
	QWORD				HeapSegmentCommit;
	QWORD				HeapDeCommitTotalFreeThreshold;
	QWORD				HeapDeCommitFreeBlockThreshold;
	DWORD				NumberOfHeaps;
	DWORD				MaximumNumberOfHeaps;
	LPVOID				*ProcessHeaps;
	LPVOID				GdiSharedHandleTable;
	LPVOID				ProcessStarterHelper;
	DWORD				GdiDCAttributeList;
//	DWORD				Padding1;
	RTL_CRITICAL_SECTION		*LoaderLock;
	DWORD				OSMajorVersion;
	DWORD				OSMinorVersion;
	WORD				OSBuildNumber;
	WORD				OSCSDVersion;
	DWORD				OSPlatformId;
	DWORD				ImageSubsystem;
	DWORD				ImageSubsystemMajorVersion;
	DWORD				ImageSubsystemMinorVersion;
//	DWORD				Padding2;
	QWORD				ActiveProcessAffinityMask;
	DWORD				GdiHandleBuffer[60];
#else
	BYTE				Reserved3[520];
#endif
	PS_POST_PROCESS_INIT_ROUTINE	*PostProcessInitRoutine;
	BYTE				Reserved4[136];
	DWORD				SessionId;
#if 1
	ULARGE_INTEGER			AppCompatFlags;
	ULARGE_INTEGER			AppCompatFlagsUser;
	LPVOID				ShimData;
	LPVOID				AppCompatInfo;
	UNICODE_STRING			CSDVersion;
#endif
} PEB, *LPPEB, *PPEB;
#endif // _WIN64

typedef	struct	_RTL_ACTIVATION_CONTEXT_STACK_FRAME
{
	struct	_RTL_ACTIVATION_CONTEXT_STACK_FRAME	*Previous;
	ACTIVATION_CONTEXT				*ActivationContext;
	DWORD						Flags;
} RTL_ACTIVATION_CONTEXT_STACK_FRAME, *LPRTL_ACTIVATION_CONTEXT_STACK_FRAME, *PRTL_ACTIVATION_CONTEXT_STACK_FRAME;

typedef	struct	_ACTIVATION_CONTEXT_STACK
{
	RTL_ACTIVATION_CONTEXT_STACK_FRAME	*ActiveFrame;
	LIST_ENTRY				FrameListCache;
	DWORD					Flags;
	DWORD					NextCookieSequenceNumber;
	DWORD					StackId;
} ACTIVATION_CONTEXT_STACK, *LPACTIVATION_CONTEXT_STACK, *PACTIVATION_CONTEXT_STACK;

typedef	struct	_GDI_TEB_BATCH
{
	DWORD	Offset;
	DWORD	HDC;
	DWORD	Buffer[310];
} GDI_TEB_BATCH, *LPGDI_TEB_BATCH, *PGDI_TEB_BATCH;

// https://msdn.microsoft.com/en-us/library/ms686708.aspx

typedef	struct	_TEB
{
#if 0
	LPVOID				Reserved1[12];
	PEB				*ProcessEnvironmentBlock;
	LPVOID				Reserved2[399];
	BYTE				Reserved3[1952];
#else
	NT_TIB				NtTib;
	LPVOID				EnvironmentPointer;
	CLIENT_ID			ClientId;
	LPVOID				ActiveRpcHandle;
	LPVOID				ThreadLocalStoragePointer;
	PEB				*ProcessEnvironmentBlock;
	DWORD				LastErrorValue;
	DWORD				CountOfOwnedCriticalSections;
	LPVOID				CsrClientThread;
	LPVOID				Win32ThreadInfo;
	DWORD				User32Reserved[26];
	DWORD				UserReserved[5];
	LPVOID				WOW32Reserved;
	LCID				CurrentLocale;
	DWORD				FpSoftwareStatusRegister;
	LPVOID				SystemReserved1[54];
	DWORD				ExceptionCode;
	ACTIVATION_CONTEXT_STACK	*ActivationContextStackPointer;
	BYTE				SpareBytes1[36];
	DWORD				TxFsContext;
	GDI_TEB_BATCH			GdiTebBatch;
	CLIENT_ID			RealClientId;
	LPVOID				GdiCachedProcessHandle;
	DWORD				GdiClientPID;
	DWORD				GdiClientTID;
	LPVOID				GdiThreadLocalInfo;
	DWORD_PTR			Win32ClientInfo[62];
	LPVOID				glDispatchTable[233];
	DWORD_PTR			glReserved1[29];
	LPVOID				glReserved2;
	LPVOID				glSectionInfo;
	LPVOID				glSection;
	LPVOID				glTable;
	LPVOID				glCurrentRC;
	LPVOID				glContext;
	DWORD				LastStatusValue;
	UNICODE_STRING			StaticUnicodeString;
	WCHAR				StaticUnicodeBuffer[261];
	LPVOID				DeallocationStack;	// bottom of stack
#endif
	LPVOID				TlsSlots[64];
#if 0
	BYTE				Reserved4[8];
#else
	LIST_ENTRY			TlsLinks;
#endif
#if 0
	LPVOID				Reserved5[26];
#else
	LPVOID				Vdm;
	LPVOID				ReservedForNtRpc;
	LPVOID				DbgSsReserved[2];
	DWORD				HardErrorMode;
	LPVOID				Instrumentation[14];
	LPVOID				SubProcessTag;
	LPVOID				EtwTraceData;
	LPVOID				WinSockData;
	DWORD				GdiBatchCount;
	BOOLEAN				InDbgPrint;
	BOOLEAN				FreeStackOnTermination;
	BOOLEAN				HasFiberData;
	DWORD				GuaranteedStackBytes;
	LPVOID				ReservedForPerf;
#endif
	LPVOID				ReservedForOle;
#if 0
	LPVOID				Reserved6[4];
#else
	DWORD				WaitingOnLoaderLock;
	DWORD_PTR			SparePointer1;
	DWORD_PTR			SoftPatchPtr1;
	DWORD_PTR			SoftPatchPtr2;
#endif
	LPVOID				*TlsExpansionSlots;
} TEB, *LPTEB, *PTEB;

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/hh285210.aspx

__declspec(dllimport)
TEB *	NTAPI	NtCurrentTeb(VOID);
#endif

// https://msdn.microsoft.com/en-us/library/ff556485.aspx
// https://msdn.microsoft.com/en-us/library/ff566434.aspx

__forceinline
HANDLE	NTAPI	NtCurrentThread(VOID)
{
#ifdef _WIN64
	return (HANDLE) -2LL;
#else
	return (HANDLE) -2L;
#endif
}

#if _WIN32_WINNT >= 0x0602
// NOTE: the pseudo handles returned from NtCurrentThreadToken()
//       and NtCurrentThreadEffectiveToken() can only be used with
//       NtGetInformationToken() and NtSetInformationToken()!

__forceinline
HANDLE	NTAPI	NtCurrentThreadToken(VOID)
{
#ifdef _WIN64
	return (HANDLE) -5LL;
#else
	return (HANDLE) -5L;
#endif
}

__forceinline
HANDLE	NTAPI	NtCurrentThreadEffectiveToken(VOID)
{
#ifdef _WIN64
	return (HANDLE) -6LL;
#else
	return (HANDLE) -6L;
#endif
}
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtDebugActiveProcess(HANDLE ProcessHandle,
		                     HANDLE DebugObjectHandle);

__declspec(dllimport)
LONG	NTAPI	NtDebugContinue(HANDLE    DebugObjectHandle,
		                CLIENT_ID *ClientId,
		                NTSTATUS  ContinueStatus);

__declspec(dllimport)
LONG	NTAPI	NtDelayExecution(BOOLEAN       Alertable,
		                 LARGE_INTEGER *DelayInterval);

__declspec(dllimport)
LONG	NTAPI	NtDeleteAtom(RTL_ATOM Atom);

__declspec(dllimport)
LONG	NTAPI	NtDeleteBootEntry(DWORD Id);

// https://msdn.microsoft.com/en-us/library/ff566435.aspx

__declspec(dllimport)
LONG	NTAPI	NtDeleteFile(OBJECT_ATTRIBUTES *ObjectAttributes);

// https://msdn.microsoft.com/en-us/library/ff556490.aspx
// https://msdn.microsoft.com/en-us/library/ff566437.aspx

__declspec(dllimport)
LONG	NTAPI	NtDeleteKey(HANDLE KeyHandle);

__declspec(dllimport)
LONG	NTAPI	NtDeleteObjectAuditAlarm(UNICODE_STRING *SubSystemName,
		                         LPVOID         HandleId,
		                         BOOLEAN        GenerateOnClose);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtDeletePrivateNamespace(HANDLE NamespaceHandle);
#endif

// https://msdn.microsoft.com/en-us/library/ff556495.aspx
// https://msdn.microsoft.com/en-us/library/ff566439.aspx

__declspec(dllimport)
LONG	NTAPI	NtDeleteValueKey(HANDLE         KeyHandle,
		                 UNICODE_STRING *ValueName);

typedef	VOID	(NTAPI	IO_APC_ROUTINE) (LPVOID          ApcContext,
			                 IO_STATUS_BLOCK *IoStatusBlock,
			                 DWORD           Reserved);

// https://msdn.microsoft.com/en-us/library/ms648411.aspx
// https://msdn.microsoft.com/en-us/library/ff566441.aspx

__declspec(dllimport)
LONG	NTAPI	NtDeviceIoControlFile(HANDLE          FileHandle,
		                      HANDLE          EventHandle,
		                      IO_APC_ROUTINE  *ApcRoutine,
		                      LPVOID          ApcContext,
		                      IO_STATUS_BLOCK *IoStatusBlock,
		                      DWORD           IoControlCode,
		                      LPVOID          InputBuffer,
		                      DWORD           InputBufferLength,
		                      LPVOID          OutputBuffer,
		                      DWORD           OutputBufferLength);

__declspec(dllimport)
LONG	NTAPI	NtDisplayString(UNICODE_STRING *String);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	NtDrawText(UNICODE_STRING *Text);
#endif

// https://msdn.microsoft.com/en-us/library/ff556502.aspx
// https://msdn.microsoft.com/en-us/library/ff566445.aspx

#ifndef _WINNT_
#define DUPLICATE_CLOSE_SOURCE		0x00000001UL
#define DUPLICATE_SAME_ACCESS		0x00000002UL
#endif
#define DUPLICATE_SAME_ATTRIBUTES	0x00000004UL

__declspec(dllimport)
LONG	NTAPI	NtDuplicateObject(HANDLE      SourceProcessHandle,
		                  HANDLE      SourceHandle,
		                  HANDLE      TargetProcessHandle,
		                  HANDLE      *TargetHandle,
		                  ACCESS_MASK DesiredAccess,
		                  DWORD       HandleAttributes,
		                  DWORD       Options);

// https://msdn.microsoft.com/en-us/library/ff556505.aspx
// https://msdn.microsoft.com/en-us/library/ff566446.aspx

__declspec(dllimport)
LONG	NTAPI	NtDuplicateToken(HANDLE            ExistingTokenHandle,
		                 ACCESS_MASK       DesiredAccess,
		                 OBJECT_ATTRIBUTES *ObjectAttributes,
		                 BOOLEAN           EffectiveOnly,
		                 TOKEN_TYPE        TokenType,
		                 HANDLE            *NewTokenHandle);

typedef	struct	_BOOT_ENTRY_LIST
{
	DWORD		NextEntryOffset;
	BOOT_ENTRY	BootEntry;
} BOOT_ENTRY_LIST, *LPBOOT_ENTRY_LIST, *PBOOT_ENTRY_LIST;

__declspec(dllimport)
LONG	NTAPI	NtEnumerateBootEntries(LPVOID  Buffer,
		                       LPDWORD BufferLength);

// https://msdn.microsoft.com/en-us/library/ff553373.aspx

typedef	enum	_KEY_INFORMATION_CLASS
{
	KeyBasicInformation          = 0,
	KeyNodeInformation           = 1,
	KeyFullInformation           = 2,
	KeyNameInformation           = 3,
	KeyCachedInformation         = 4,
	KeyFlagsInformation          = 5,
	KeyVirtualizationInformation = 6,
	KeyHandleTagsInformation     = 7,
	MaxKeyInfoClass              = 8
} KEY_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff553355.aspx

typedef	struct	_KEY_BASIC_INFORMATION
{
	LARGE_INTEGER	LastWriteTime;
	DWORD		TitleIndex;
	DWORD		NameLength;
	WCHAR		Name[ANYSIZE_ARRAY];
} KEY_BASIC_INFORMATION, *LPKEY_BASIC_INFORMATION, *PKEY_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff553392.aspx

typedef	struct	_KEY_NODE_INFORMATION
{
	LARGE_INTEGER	LastWriteTime;
	DWORD		TitleIndex;
	DWORD		ClassOffset;
	DWORD		ClassLength;
	DWORD		NameLength;
	WCHAR		Name[ANYSIZE_ARRAY];
} KEY_NODE_INFORMATION, *LPKEY_NODE_INFORMATION, *PKEY_NODE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff553367.aspx

typedef	struct	_KEY_FULL_INFORMATION
{
	LARGE_INTEGER	LastWriteTime;
	DWORD		TitleIndex;
	DWORD		ClassOffset;
	DWORD		ClassLength;
	DWORD		SubKeys;
	DWORD		MaxNameLen;
	DWORD		MaxClassLen;
	DWORD		Values;
	DWORD		MaxValueNameLen;
	DWORD		MaxValueDataLen;
	WCHAR		Class[ANYSIZE_ARRAY];
} KEY_FULL_INFORMATION, *LPKEY_FULL_INFORMATION, *PKEY_FULL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff553381.aspx

typedef	struct	_KEY_NAME_INFORMATION
{
	DWORD	NameLength;
	WCHAR	Name[ANYSIZE_ARRAY];
} KEY_NAME_INFORMATION, *LPKEY_NAME_INFORMATION, *PKEY_NAME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff553358.aspx

typedef	struct	_KEY_CACHED_INFORMATION
{
	LARGE_INTEGER	LastWriteTime;
	DWORD		TitleIndex;
	DWORD		SubKeys;
	DWORD		MaxNameLen;
	DWORD		Values;
	DWORD		MaxValueNameLen;
	DWORD		MaxValueDataLen;
	DWORD		NameLength;
} KEY_CACHED_INFORMATION, *LPKEY_CACHED_INFORMATION, *PKEY_CACHED_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff554221.aspx

#pragma warning(disable: 4214)
typedef	struct	_KEY_VIRTUALIZATION_INFORMATION
{
	DWORD	VirtualizationCandidate : 1;
	DWORD	VirtualizationEnabled   : 1;
	DWORD	VirtualTarget           : 1;
	DWORD	VirtualStore            : 1;
	DWORD	VirtualSource           : 1;
	DWORD	Reserved                : 27;
} KEY_VIRTUALIZATION_INFORMATION, *LPKEY_VIRTUALIZATION_INFORMATION, *PKEY_VIRTUALIZATION_INFORMATION;
#pragma warning(default: 4214)

// https://msdn.microsoft.com/en-us/library/ff556510.aspx
// https://msdn.microsoft.com/en-us/library/ff566447.aspx

__declspec(dllimport)
LONG	NTAPI	NtEnumerateKey(HANDLE                KeyHandle,
		               DWORD                 Index,
		               KEY_INFORMATION_CLASS KeyInformationClass,
		               LPVOID                KeyInformation,
		               DWORD                 KeyInformationLength,
		               LPDWORD               ReturnLength);

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff554263.aspx

typedef	enum	_KTMOBJECT_TYPE
{
	KTMOBJECT_TRANSACTION         = 0,
	KTMOBJECT_TRANSACTION_MANAGER = 1,
	KTMOBJECT_RESOURCE_MANAGER    = 2,
	KTMOBJECT_ENLISTMENT          = 3,
	KTMOBJECT_INVALID             = 4
} KTMOBJECT_TYPE, *LPKTMOBJECT_TYPE, *PKTMOBJECT_TYPE;

// https://msdn.microsoft.com/en-us/library/ff554259.aspx

typedef	struct	_KTMOBJECT_CURSOR
{
	GUID	LastQuery;
	DWORD	ObjectIdCount;
	GUID	ObjectIds[ANYSIZE_ARRAY];
} KTMOBJECT_CURSOR, *LPKTMOBJECT_CURSOR, *PKTMOBJECT_CURSOR;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff566450.aspx

__declspec(dllimport)
LONG	NTAPI	NtEnumerateTransactionObject(HANDLE           RootObjectHandle,
		                             KTMOBJECT_TYPE   QueryType,
		                             KTMOBJECT_CURSOR *ObjectCursor,
		                             DWORD            ObjectCursorLength,
		                             LPDWORD          ReturnLength);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff554218.aspx

typedef	enum	_KEY_VALUE_INFORMATION_CLASS
{
	KeyValueBasicInformation          = 0,
	KeyValueFullInformation,
	KeyValuePartialInformation,
	KeyValueFullInformationAlign64,
	KeyValuePartialInformationAlign64,
	MaxKeyValueInfoClass
} KEY_VALUE_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff553410.aspx

typedef	struct	_KEY_VALUE_BASIC_INFORMATION
{
	DWORD	TitleIndex;
	DWORD	Type;
	DWORD	NameLength;
	WCHAR	Name[ANYSIZE_ARRAY];
} KEY_VALUE_BASIC_INFORMATION, *LPKEY_VALUE_BASIC_INFORMATION, *PKEY_VALUE_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff554217.aspx

typedef	struct	_KEY_VALUE_FULL_INFORMATION
{
	DWORD	TitleIndex;
	DWORD	Type;
	DWORD	DataOffset;
	DWORD	DataLength;
	DWORD	NameLength;
	WCHAR	Name[ANYSIZE_ARRAY];
} KEY_VALUE_FULL_INFORMATION, *LPKEY_VALUE_FULL_INFORMATION, *PKEY_VALUE_FULL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff554220.aspx

typedef	struct	_KEY_VALUE_PARTIAL_INFORMATION
{
	DWORD	TitleIndex;
	DWORD	Type;
	DWORD	DataLength;
	BYTE	Data[ANYSIZE_ARRAY];
} KEY_VALUE_PARTIAL_INFORMATION, *LPKEY_VALUE_PARTIAL_INFORMATION, *PKEY_VALUE_PARTIAL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff556514.aspx
// https://msdn.microsoft.com/en-us/library/ff566453.aspx

__declspec(dllimport)
LONG	NTAPI	NtEnumerateValueKey(HANDLE                      KeyHandle,
		                    DWORD                       Index,
		                    KEY_VALUE_INFORMATION_CLASS KeyValueInformationClass,
		                    LPVOID                      KeyValueInformation,
		                    DWORD                       KeyValueInformationLength,
		                    LPDWORD                     ReturnLength);

__declspec(dllimport)
LONG	NTAPI	NtExtendSection(HANDLE        SectionHandle,
		                LARGE_INTEGER *NewSectionSize);

__declspec(dllimport)
LONG	NTAPI	NtFilterToken(HANDLE           ExistingTokenHandle,
		              DWORD            Flags,
		              TOKEN_GROUPS     *SidsToDisable,
		              TOKEN_PRIVILEGES *PrivilegesToDelete,
		              TOKEN_GROUPS     *RestrictedSids,
		              HANDLE           *NewTokenHandle);

__declspec(dllimport)
LONG	NTAPI	NtFindAtom(LPCWSTR  AtomName,
		           DWORD    Length,
		           RTL_ATOM *Atom);

// https://msdn.microsoft.com/en-us/library/ff566454.aspx

__declspec(dllimport)
LONG	NTAPI	NtFlushBuffersFile(HANDLE          FileHandle,
		                   IO_STATUS_BLOCK *IoStatusBlock);

#if _WIN32_WINNT >= 0x0602
// https://msdn.microsoft.com/en-us/library/hh967720.aspx

__declspec(dllimport)
LONG	NTAPI	NtFlushBuffersFileEx(HANDLE          FileHandle,
		                     DWORD           Flags,
		                     IO_STATUS_BLOCK *IoStatusBlock);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtFlushInstallUILanguage(LANGID InstallUILanguageId,
		                         DWORD  SetComittedFlag);
#endif

__declspec(dllimport)
LONG	NTAPI	NtFlushInstructionCache(HANDLE ProcessHandle,
		                        LPVOID BaseAddress,
		                        DWORD  NumberOfBytesToFlush);

// https://msdn.microsoft.com/en-us/library/ff566457.aspx

__declspec(dllimport)
LONG	NTAPI	NtFlushKey(HANDLE KeyHandle);

// https://msdn.microsoft.com/en-us/library/ms683148.aspx
// NOTE: forwarded from KERNEL32!FlushProcessWriteBuffers and KERNELBASE!FlushProcessWriteBuffers

__declspec(dllimport)
VOID	NTAPI	NtFlushProcessWriteBuffers(VOID);

// https://msdn.microsoft.com/en-us/library/ff556527.aspx
// https://msdn.microsoft.com/en-us/library/ff566458.aspx

__declspec(dllimport)
LONG	NTAPI	NtFlushVirtualMemory(HANDLE          ProcessHandle,
		                     LPVOID          *BaseAddress,
		                     SIZE_T          *RegionSize,
		                     IO_STATUS_BLOCK *IoStatusBlock);

__declspec(dllimport)
LONG	NTAPI	NtFreeUserPhysicalPages(HANDLE    ProcessHandle,
		                        DWORD_PTR *NumberOfPages,
		                        DWORD_PTR *UserPfnArray);

// https://msdn.microsoft.com/en-us/library/ff556528.aspx
// https://msdn.microsoft.com/en-us/library/ff566460.aspx

__declspec(dllimport)
LONG	NTAPI	NtFreeVirtualMemory(HANDLE ProcessHandle,
		                    LPVOID *BaseAddress,
		                    SIZE_T *RegionSize,
		                    DWORD  FreeType);

// https://msdn.microsoft.com/en-us/library/ff556532.aspx
// https://msdn.microsoft.com/en-us/library/ff566462.aspx

__declspec(dllimport)
LONG	NTAPI	NtFsControlFile(HANDLE          FileHandle,
		                HANDLE          EventHandle,
		                IO_APC_ROUTINE  *ApcRoutine,
		                LPVOID          ApcContext,
		                IO_STATUS_BLOCK *IoStatusBlock,
		                DWORD           FsControlCode,
		                LPVOID          InputBuffer,
		                DWORD           InputBufferLength,
		                LPVOID          OutputBuffer,
		                DWORD           OutputBufferLength);

__declspec(dllimport)
LONG	NTAPI	NtGetContextThread(HANDLE  ThreadHandle,
		                   CONTEXT *ContextRecord);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms684278.aspx
// NOTE: like KERNEL32!GetCurrentProcessorNumber

__declspec(deprecated("obsolete function, may be altered or unavailable in future versions"))
__declspec(dllimport)
DWORD	NTAPI	NtGetCurrentProcessorNumber(VOID);

__declspec(dllimport)
LONG	NTAPI	NtGetMUIRegistryInfo(DWORD   Flags,
		                     LPDWORD BufferLength,
		                     LPVOID  Buffer);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtGetNextProcess(HANDLE      ProcessHandle,	// pass NULL in first call
		                 ACCESS_MASK DesiredAccess,
		                 DWORD       HandleAttributes,
		                 DWORD       Flags,		// must be 0!
		                 HANDLE      *NextProcessHandle);

__declspec(dllimport)
LONG	NTAPI	NtGetNextThread(HANDLE      ProcessHandle,
		                HANDLE      ThreadHandle,	// pass NULL in first call
		                ACCESS_MASK DesiredAccess,
		                DWORD       HandleAttributes,
		                DWORD       Flags,		// must be 0!
		                HANDLE      *NextThreadHandle);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff564813.aspx

typedef	struct	_TRANSACTION_NOTIFICATION
{
	LPVOID		TransactionKey;
	DWORD		TransactionNotification;
	LARGE_INTEGER	TmVirtualClock;
	DWORD		ArgumentLength;
} TRANSACTION_NOTIFICATION, *LPTRANSACTION_NOTIFICATION, *PTRANSACTION_NOTIFICATION;
#endif

// https://msdn.microsoft.com/en-us/library/ff566467.aspx

__declspec(dllimport)
LONG	NTAPI	NtGetNotificationResourceManager(HANDLE                   ResourceManagerHandle,
		                                 TRANSACTION_NOTIFICATION *TransactionNotification,
		                                 DWORD                    NotificationLength,
		                                 LARGE_INTEGER            *Timeout,
		                                 LPDWORD                  ReturnLength,
		                                 DWORD                    Asynchronous,
		                                 DWORD_PTR                AsynchronousContext);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
DWORD	NTAPI	NtGetTickCount(VOID);
#endif

// NOTE: like ADVAPI32!ImpersonateAnonymousToken and KERNELBASE!ImpersonateAnonymousToken

__declspec(dllimport)
LONG	NTAPI	NtImpersonateAnonymousToken(HANDLE ThreadHandle);

typedef	struct	_SECURITY_QUALITY_OF_SERVICE	SECURITY_QUALITY_OF_SERVICE;

__declspec(dllimport)
LONG	NTAPI	NtImpersonateThread(HANDLE                      ThreadHandle,
		                    HANDLE                      ThreadToImpersonate,
		                    SECURITY_QUALITY_OF_SERVICE *Qos);

#define REG_INIT_BOOT_SM		0x0000
#define REG_INIT_BOOT_SETUP		0x0001
#define REG_INIT_BOOT_ACCEPTED_BASE	0x0002
#define REG_INIT_BOOT_ACCEPTED_MAX	(REG_INIT_BOOT_ACCEPTED_BASE + 999)

__declspec(dllimport)
LONG	NTAPI	NtInitializeRegistry(WORD BootCondition);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtIsUILanguageComitted(VOID);
#endif

__declspec(dllimport)
LONG	NTAPI	NtIsProcessInJob(HANDLE ProcessHandle,
		                 HANDLE JobHandle);

// https://msdn.microsoft.com/en-us/library/ff556539.aspx
// https://msdn.microsoft.com/en-us/library/ff566470.aspx

__declspec(dllimport)
LONG	NTAPI	NtLoadDriver(UNICODE_STRING *DriverServiceName);

__declspec(dllimport)
LONG	NTAPI	NtLoadKey(OBJECT_ATTRIBUTES *TargetKey,
		          OBJECT_ATTRIBUTES *SourceFile);

__declspec(dllimport)
LONG	NTAPI	NtLoadKey2(OBJECT_ATTRIBUTES *TargetKey,
		           OBJECT_ATTRIBUTES *SourceFile,
		           DWORD             Flags);

#if _WIN32_WINNT >= 0x0A00
typedef	enum	_KEY_LOAD_HANDLE_TYPE
{
	KeyLoadTrustKey = 1,
	KeyLoadEvent,
	KeyLoadToken
} KEY_LOAD_HANDLE_TYPE;

typedef	struct	_KEY_LOAD_HANDLE
{
	KEY_LOAD_HANDLE_TYPE	Type;
	HANDLE Handle;
} KEY_LOAD_HANDLE, *LPKEY_LOAD_HANDLE, *PKEY_LOAD_HANDLE;

__declspec(dllimport)
LONG	NTAPI	NtLoadKey3(OBJECT_ATTRIBUTES *TargetKey,
		           OBJECT_ATTRIBUTES *SourceFile,
		           DWORD             Flags,
		           KEY_LOAD_HANDLE   *LoadEntries,
		           DWORD             LoadEntryCount,
		           ACCESS_MASK       DesiredAccess,
		           HANDLE            *RootHandle,
		           LPVOID            Unused);
#endif

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	NtLoadKeyEx(OBJECT_ATTRIBUTES *TargetKey,
		            OBJECT_ATTRIBUTES *SourceFile,
		            DWORD             Flags,
		            HANDLE            TrustClassKey);
#endif

// https://msdn.microsoft.com/en-us/library/ff556543.aspx
// https://msdn.microsoft.com/en-us/library/ff566474.aspx

__declspec(dllimport)
LONG	NTAPI	NtLockFile(HANDLE          FileHandle,
		           HANDLE          EventHandle,
		           IO_APC_ROUTINE  *ApcRoutine,
		           LPVOID          ApcContext,
		           IO_STATUS_BLOCK *IoStatusBlock,
		           LARGE_INTEGER   *ByteOffset,
		           LARGE_INTEGER   *Length,
		           DWORD           Key,
		           BOOLEAN         FailImmediately,
		           BOOLEAN         ExclusiveLock);


__declspec(dllimport)
LONG	NTAPI	NtLockProductActivationKeys(LPDWORD pPrivateVer,
		                            LPDWORD pSafeMode);

__declspec(dllimport)
LONG	NTAPI	NtLockRegistryKey(HANDLE KeyHandle);

#define UNLOCK_TYPE_NON_PRIVILEGED	0x00000001UL
#define UNLOCK_TYPE_PRIVILEGED		0x00000002UL

__declspec(dllimport)
LONG	NTAPI	NtLockVirtualMemory(HANDLE ProcessHandle,
		                    LPVOID *BaseAddress,
		                    SIZE_T *RegionSize,
		                    DWORD  MapType);

__declspec(dllimport)
LONG	NTAPI	NtMakePermanentObject(HANDLE Handle);

// https://msdn.microsoft.com/en-us/library/ff556546.aspx
// https://msdn.microsoft.com/en-us/library/ff566477.aspx

__declspec(dllimport)
LONG	NTAPI	NtMakeTemporaryObject(HANDLE Handle);

__declspec(dllimport)
LONG	NTAPI	NtMapUserPhysicalPages(LPVOID    VirtualAddress,
		                       DWORD_PTR NumberOfPages,
		                       DWORD_PTR *UserPfnArray);

__declspec(dllimport)
LONG	NTAPI	NtMapUserPhysicalPagesScatter(LPVOID    *VirtualAddresses,
		                              DWORD_PTR NumberOfPages,
		                              DWORD_PTR *UserPfnArray);

typedef	enum	_SECTION_INHERIT
{
	ViewShare = 1,
	ViewUnmap = 2
} SECTION_INHERIT;

// https://msdn.microsoft.com/en-us/library/ff556551.aspx
// https://msdn.microsoft.com/en-us/library/ff566481.aspx

__declspec(dllimport)
LONG	NTAPI	NtMapViewOfSection(HANDLE          SectionHandle,
		                   HANDLE          ProcessHandle,
		                   LPVOID          *BaseAddress,
		                   DWORD_PTR       ZeroBits,
		                   SIZE_T          CommitSize,
		                   LARGE_INTEGER   *SectionOffset,
		                   SIZE_T          *ViewSize,
		                   SECTION_INHERIT InheritDisposition,
		                   DWORD           AllocationType,
		                   DWORD           Win32Protect);

__declspec(dllimport)
LONG	NTAPI	NtModifyBootEntry(BOOT_ENTRY *BootEntry);

__declspec(dllimport)
LONG	NTAPI	NtNotifyChangeDirectoryFile(HANDLE          FileHandle,
		                            HANDLE          EventHandle,
		                            IO_APC_ROUTINE  *ApcRoutine,
		                            LPVOID          ApcContext,
		                            IO_STATUS_BLOCK *IoStatusBlock,
		                            LPVOID          Buffer,
		                            DWORD           BufferLength,
		                            DWORD           CompletionFilter,
		                            BOOLEAN         WatchTree);

// https://msdn.microsoft.com/en-us/library/ff556554.aspx
// https://msdn.microsoft.com/en-us/library/ff566488.aspx

__declspec(dllimport)
LONG	NTAPI	NtNotifyChangeKey(HANDLE          KeyHandle,
		                  HANDLE          EventHandle,
		                  IO_APC_ROUTINE  *ApcRoutine,
		                  LPVOID          ApcContext,
		                  IO_STATUS_BLOCK *IoStatusBlock,
		                  DWORD           CompletionFilter,
		                  BOOLEAN         WatchTree,
		                  LPVOID          Buffer,
		                  DWORD           BufferSize,
		                  BOOLEAN         Asynchronous);

// https://msdn.microsoft.com/en-us/library/cc512136.aspx

__declspec(dllimport)
LONG	NTAPI	NtNotifyChangeMultipleKeys(HANDLE            MasterKeyHandle,
		                           DWORD             Count,
		                           OBJECT_ATTRIBUTES *SubordinateObjects,
		                           HANDLE            EventHandle,
		                           IO_APC_ROUTINE    *ApcRoutine,
		                           LPVOID            ApcContext,
		                           IO_STATUS_BLOCK   *IoStatusBlock,
		                           DWORD             CompletionFilter,
		                           BOOLEAN           WatchTree,
		                           LPVOID            Buffer,
		                           DWORD             BufferSize,
		                           BOOLEAN           Asynchronous);

// https://msdn.microsoft.com/en-us/library/bb470234.aspx

#ifndef DIRECTORY_ALL_ACCESS
#define DIRECTORY_QUERY			0x00000001UL
#define DIRECTORY_TRAVERSE		0x00000002UL
#define DIRECTORY_CREATE_OBJECT		0x00000004UL
#define DIRECTORY_CREATE_SUBDIRECTORY	0x00000008UL
#define DIRECTORY_ALL_ACCESS		(STANDARD_RIGHTS_REQUIRED | DIRECTORY_QUERY | DIRECTORY_TRAVERSE | DIRECTORY_CREATE_OBJECT | DIRECTORY_CREATE_SUBDIRECTORY)
#endif

// https://msdn.microsoft.com/en-us/library/bb470234.aspx
// https://msdn.microsoft.com/en-us/library/ff566492.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenDirectoryObject(HANDLE            *DirectoryHandle,
		                      ACCESS_MASK       DesiredAccess,
		                      OBJECT_ATTRIBUTES *ObjectAttributes);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567008.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenEnlistment(HANDLE           *EnlistmentHandle,
		                 ACCESS_MASK       DesiredAccess,
		                 HANDLE            RmHandle,
		                 GUID              *EnlistmentGuid,
		                 OBJECT_ATTRIBUTES *ObjectAttributes);
#endif

// https://msdn.microsoft.com/en-us/library/ff567009.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenEvent(HANDLE            *EventHandle,
		            ACCESS_MASK       DesiredAccess,
		            OBJECT_ATTRIBUTES *ObjectAttributes);

// https://msdn.microsoft.com/en-us/library/bb432381.aspx
// https://msdn.microsoft.com/en-us/library/ff567011.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenFile(HANDLE            *FileHandle,
		           ACCESS_MASK       DesiredAccess,
		           OBJECT_ATTRIBUTES *ObjectAttributes,
		           IO_STATUS_BLOCK   *IoStatusBlock,
		           DWORD             ShareAccess,
		           DWORD             OpenOptions);

__declspec(dllimport)
LONG	NTAPI	NtOpenIoCompletion(HANDLE            *IoCompletionPortHandle,
		                   ACCESS_MASK       DesiredAccess,
		                   OBJECT_ATTRIBUTES *ObjectAttributes);

__declspec(dllimport)
LONG	NTAPI	NtOpenJobObject(HANDLE            *JobHandle,
		                ACCESS_MASK       DesiredAccess,
		                OBJECT_ATTRIBUTES *ObjectAttributes);

#if _WIN32_WINNT >= 0x0602
__declspec(dllimport)
LONG	NTAPI	NtOpenJobObjectToken(HANDLE      JobHandle,
		                     ACCESS_MASK DesiredAccess,
		                     HANDLE      *TokenHandle);
#endif

// https://msdn.microsoft.com/en-us/library/ff556570.aspx
// https://msdn.microsoft.com/en-us/library/ff567014.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenKey(HANDLE            *KeyHandle,
		          ACCESS_MASK       DesiredAccess,
		          OBJECT_ATTRIBUTES *ObjectAttributes);

__declspec(dllimport)
LONG	NTAPI	NtOpenMutant(HANDLE            *MutantHandle,
		             ACCESS_MASK       DesiredAccess,
		             OBJECT_ATTRIBUTES *ObjectAttributes);

__declspec(dllimport)
LONG	NTAPI	NtOpenObjectAuditAlarm(UNICODE_STRING      *SubsystemName,
		                       LPVOID              HandleId,
		                       UNICODE_STRING      *ObjectTypeName,
		                       UNICODE_STRING      *ObjectName,
		                       SECURITY_DESCRIPTOR *SecurityDescriptor,
		                       HANDLE              TokenClient,
		                       ACCESS_MASK         DesiredAccess,
		                       ACCESS_MASK         GrantedAccess,
		                       PRIVILEGE_SET       *PrivilegeSet,
		                       BOOLEAN             ObjectCreation,
		                       BOOLEAN             AccessGranted,
		                       BOOLEAN             *GenerateOnClose);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtOpenPrivateNamespace(HANDLE            *NamespaceHandle,
		                       ACCESS_MASK       DesiredAccess,
		                       OBJECT_ATTRIBUTES *ObjectAttributes,
		                       LPVOID            BoundaryDescriptor);
#endif

// https://msdn.microsoft.com/en-us/library/ff556571.aspx
// https://msdn.microsoft.com/en-us/library/ff567022.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenProcess(HANDLE            *ProcessHandle,
		              ACCESS_MASK       DesiredAccess,
		              OBJECT_ATTRIBUTES *ObjectAttributes,
		              CLIENT_ID         *ClientId);

__declspec(dllimport)
LONG	NTAPI	NtOpenProcessToken(HANDLE      ProcessHandle,
		                   ACCESS_MASK DesiredAccess,
		                   HANDLE      *TokenHandle);

// https://msdn.microsoft.com/en-us/library/ff556574.aspx
// https://msdn.microsoft.com/en-us/library/ff567024.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenProcessTokenEx(HANDLE      ProcessHandle,
		                     ACCESS_MASK DesiredAccess,
		                     DWORD       HandleAttributes,
		                     HANDLE      *TokenHandle);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567026.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenResourceManager(HANDLE            ResourceManagerHandle,
		                      ACCESS_MASK       DesiredAccess,
		                      HANDLE            TmHandle,
		                      GUID              *ResourceManagerGuid,
		                      OBJECT_ATTRIBUTES *ObjectAttributes);
#endif

// https://msdn.microsoft.com/en-us/library/ff556580.aspx
// https://msdn.microsoft.com/en-us/library/ff567029.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenSection(HANDLE            *SectionHandle,
		              ACCESS_MASK       DesiredAccess,
		              OBJECT_ATTRIBUTES *ObjectAttributes);

__declspec(dllimport)
LONG	NTAPI	NtOpenSemaphore(HANDLE            *SemaphoreHandle,
		                ACCESS_MASK       DesiredAccess,
		                OBJECT_ATTRIBUTES *ObjectAttributes);

// https://msdn.microsoft.com/en-us/library/bb470236.aspx
// https://msdn.microsoft.com/en-us/library/ff567030.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenSymbolicLinkObject(HANDLE            *LinkHandle,
		                         ACCESS_MASK       DesiredAccess,
		                         OBJECT_ATTRIBUTES *ObjectAttributes);

__declspec(dllimport)
LONG	NTAPI	NtOpenTimer(HANDLE            *TimerHandle,
		            ACCESS_MASK       DesiredAccess,
		            OBJECT_ATTRIBUTES *ObjectAttributes);

// https://msdn.microsoft.com/en-us/library/bb432382.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenThread(HANDLE            *ThreadHandle,
		             ACCESS_MASK       DesiredAccess,
		             OBJECT_ATTRIBUTES *ObjectAttributes,
		             CLIENT_ID         *ClientId);

__declspec(dllimport)
LONG	NTAPI	NtOpenThreadToken(HANDLE      ThreadHandle,
		                  ACCESS_MASK DesiredAccess,
		                  BOOLEAN     OpenAsSelf,
		                  HANDLE      *TokenHandle);

// https://msdn.microsoft.com/en-us/library/ff556591.aspx
// https://msdn.microsoft.com/en-us/library/ff567032.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenThreadTokenEx(HANDLE      ThreadHandle,
		                    ACCESS_MASK DesiredAccess,
		                    BOOLEAN     OpenAsSelf,
		                    DWORD       HandleAttributes,
		                    HANDLE      *TokenHandle);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567033.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenTransaction(HANDLE            *TransactionHandle,
		                  ACCESS_MASK       DesiredAccess,
		                  OBJECT_ATTRIBUTES *ObjectAttributes,
		                  GUID              *Uow,
		                  HANDLE            TmHandle);

// https://msdn.microsoft.com/en-us/library/ff567035.aspx

__declspec(dllimport)
LONG	NTAPI	NtOpenTransactionManager(HANDLE            *TmHandle,
		                         ACCESS_MASK       DesiredAccess,
		                         OBJECT_ATTRIBUTES *ObjectAttributes,
		                         UNICODE_STRING    *LogFileName,
		                         GUID              *TmIdentity,
		                         DWORD             OpenOptions);
#endif // _WIN32_WINNT

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/mt807499.aspx

typedef	enum	_POWER_INFORMATION_LEVEL
{
	SystemPowerPolicyAc                = 0,
	SystemPowerPolicyDc,
	VerifySystemPolicyAc,
	VerifySystemPolicyDc,
	SystemPowerCapabilities,
	SystemBatteryState,
	SystemPowerStateHandler,
	ProcessorStateHandler,
	SystemPowerPolicyCurrent,
	AdministratorPowerPolicy,
	SystemReserveHiberFile,
	ProcessorInformation,
	SystemPowerInformation,
	ProcessorStateHandler2,
	LastWakeTime,
	LastSleepTime,
	SystemExecutionState,
	SystemPowerStateNotifyHandler,
	ProcessorPowerPolicyAc,
	ProcessorPowerPolicyDc,
	VerifyProcessorPowerPolicyAc,
	VerifyProcessorPowerPolicyDc,
	ProcessorPowerPolicyCurrent,
	SystemPowerStateLogging,
	SystemPowerLoggingEntry,
	SetPowerSettingValue,
	NotifyUserPowerSetting,
	PowerInformationLevelUnused0,
	SystemMonitorHiberBootPowerOff,
	SystemVideoState,
	TraceApplicationPowerMessage,
	TraceApplicationPowerMessageEnd,
	ProcessorPerfStates,
	ProcessorIdleStates,
	ProcessorCap,
	SystemWakeSource,
	SystemHiberFileInformation,
	TraceServicePowerMessage,
	ProcessorLoad,
	PowerShutdownNotification,
	MonitorCapabilities,
	SessionPowerInit,
	SessionDisplayState,
	PowerRequestCreate,
	PowerRequestAction,
	GetPowerRequestList,
	ProcessorInformationEx,
	NotifyUserModeLegacyPowerEvent,
	GroupPark,
	ProcessorIdleDomains,
	WakeTimerList,
	SystemHiberFileSize,
	ProcessorIdleStatesHv,
	ProcessorPerfStatesHv,
	ProcessorPerfCapHv,
	ProcessorSetIdle,
	LogicalProcessorIdling,
	UserPresence,
	PowerSettingNotificationName,
	GetPowerSettingValue,
	IdleResiliency,
	SessionRITState,
	SessionConnectNotification,
	SessionPowerCleanup,
	SessionLockState,
	SystemHiberbootState,
	PlatformInformation,
	PdcInvocation,
	MonitorInvocation,
	FirmwareTableInformationRegistered,
	SetShutdownSelectedTime,
	SuspendResumeInvocation,
	PlmPowerRequestCreate,
	ScreenOff,
	CsDeviceNotification,
	PlatformRole,
	LastResumePerformance,
	DisplayBurst,
	ExitLatencySamplingPercentage,
	RegisterSpmPowerSettings,
	PlatformIdleStates,
	ProcessorIdleVeto,
	PlatformIdleVeto,
	SystemBatteryStatePrecise,
	ThermalEvent,
	PowerRequestActionInternal,
	BatteryDeviceState,
	PowerInformationInternal,
	ThermalStandby,
	SystemHiberFileType,
	PhysicalPowerButtonPress,
	QueryPotentialDripsConstraint,
	EnergyTrackerCreate,
	EnergyTrackerQuery,
	UpdateBlackBoxRecorder,
	PowerInformationLevelMaximum
} POWER_INFORMATION_LEVEL;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/dn957451.aspx
// https://msdn.microsoft.com/en-us/library/dn957454.aspx
// NOTE: like POWRPROF!CallNtPowerInformation

__declspec(dllimport)
LONG	NTAPI	NtPowerInformation(POWER_INFORMATION_LEVEL InformationLevel,
		                   LPVOID                  InputBuffer,
		                   DWORD                   InputBufferLength,
		                   LPVOID                  OutputBuffer,
		                   DWORD                   OutputBufferLength);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567037.aspx

__declspec(dllimport)
LONG	NTAPI	NtPrepareComplete(HANDLE        EnlistmentHandle,
		                  LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567039.aspx

__declspec(dllimport)
LONG	NTAPI	NtPrepareEnlistment(HANDLE        EnlistmentHandle,
		                    LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567040.aspx

__declspec(dllimport)
LONG	NTAPI	NtPrePrepareComplete(HANDLE        EnlistmentHandle,
		                     LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567044.aspx

__declspec(dllimport)
LONG	NTAPI	NtPrePrepareEnlistment(HANDLE        EnlistmentHandle,
		                       LARGE_INTEGER *TmVirtualClock);
#endif // _WIN32_WINNT

// NOTE: like ADVAPI32!PrivilegeCheck

__declspec(dllimport)
LONG	NTAPI	NtPrivilegeCheck(HANDLE        ClientToken,
		                 PRIVILEGE_SET *RequiredPrivileges,
		                 BOOLEAN       *Result);

__declspec(dllimport)
LONG	NTAPI	NtPrivilegeObjectAuditAlarm(UNICODE_STRING *SubsystemName,
		                            LPVOID         HandleId,
		                            HANDLE         TokenHandle,
		                            ACCESS_MASK    DesiredAccess,
		                            PRIVILEGE_SET  *PrivilegeSet,
		                            BOOLEAN        AccessGranted);

__declspec(dllimport)
LONG	NTAPI	NtPrivilegedServiceAuditAlarm(UNICODE_STRING *SubsystemName,
		                              UNICODE_STRING *ServiceName,
		                              HANDLE         TokenHandle,
		                              PRIVILEGE_SET  *PrivilegeSet,
		                              BOOLEAN        AccessGranted);

// NOTE: like KERNEL32!VirtualProtectEx

__declspec(dllimport)
LONG	NTAPI	NtProtectVirtualMemory(HANDLE  ProcessHandle,
		                       LPVOID  *BaseAddress,
		                       SIZE_T  *RegionSize,
		                       DWORD   NewProtection,
		                       LPDWORD PreviousProtection);

__declspec(dllimport)
LONG	NTAPI	NtPulseEvent(HANDLE EventHandle,
		             LPLONG PreviousState);

typedef	VOID	(NTAPI	KNORMAL_ROUTINE) (LPVOID NormalContext,
			                  LPVOID SystemArgument1,
			                  LPVOID SystemArgument2);

__declspec(dllimport)
LONG	NTAPI	NtQueueApcThread(HANDLE          ThreadHandle,
		                 KNORMAL_ROUTINE *ApcRoutine,
		                 LPVOID          NormalContext,
		                 LPVOID          SystemArgument1,
		                 LPVOID          SystemArgument2);

#if _WIN32_WINNT >= 0x0601
typedef	VOID	(NTAPI	PS_APC_ROUTINE) (LPVOID ApcArgument1,
			                 LPVOID ApcArgument2,
			                 LPVOID ApcArgument3);

__declspec(dllimport)
LONG	NTAPI	NtQueueApcThreadEx(HANDLE         ThreadHandle,
		                   HANDLE         UserApcReserveHandle,
		                   PS_APC_ROUTINE *ApcRoutine,
		                   LPVOID         ApcArgument1,
		                   LPVOID         ApcArgument2,
		                   LPVOID         ApcArgument3);
#endif

// https://msdn.microsoft.com/en-us/library/ff545762.aspx

typedef	struct	_FILE_BASIC_INFORMATION
{
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	DWORD		FileAttributes;
} FILE_BASIC_INFORMATION, *LPFILE_BASIC_INFORMATION, *PFILE_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/cc512135.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryAttributesFile(OBJECT_ATTRIBUTES      *ObjectAttributes,
		                      FILE_BASIC_INFORMATION *FileInformation);

__declspec(dllimport)
LONG	NTAPI	NtQueryBootEntryOrder(LPDWORD Ids,
		                      LPDWORD Count);

#define BOOT_OPTIONS_VERSION			1UL

#define BOOT_OPTIONS_FIELD_TIMEOUT		0x00000001UL
#define BOOT_OPTIONS_FIELD_NEXT_BOOT_ENTRY_ID	0x00000002UL
#define BOOT_OPTIONS_FIELD_HEADLESS_REDIRECTION	0x00000004UL

typedef	struct	_BOOT_OPTIONS
{
	DWORD	Version;
	DWORD	Length;
	DWORD	Timeout;
	DWORD	CurrentBootEntryId;
	DWORD	NextBootEntryId;
	WCHAR	HeadlessRedirection[ANYSIZE_ARRAY];
} BOOT_OPTIONS, *LPBOOT_OPTIONS, *PBOOT_OPTIONS;

__declspec(dllimport)
LONG	NTAPI	NtQueryBootOptions(BOOT_OPTIONS *BootOptions,
		                   LPDWORD      BootOptionsLength);

__declspec(dllimport)
LONG	NTAPI	NtQueryDebugFilterState(DWORD ComponentId,
		                        DWORD Level);

__declspec(dllimport)
LONG	NTAPI	NtQueryDefaultLocale(BOOLEAN UserProfile,
		                     LCID    *DefaultLocaleId);

__declspec(dllimport)
LONG	NTAPI	NtQueryDefaultUILanguage(LANGID *LanguageId);

// https://msdn.microsoft.com/en-us/library/ff728840.aspx

typedef	enum	_FILE_INFORMATION_CLASS
{
	FileDirectoryInformation                     = 1,
	FileFullDirectoryInformation                 = 2,
	FileBothDirectoryInformation                 = 3,
	FileBasicInformation                         = 4,
	FileStandardInformation                      = 5,
	FileInternalInformation                      = 6,
	FileEaInformation                            = 7,
	FileAccessInformation                        = 8,
	FileNameInformation                          = 9,
	FileRenameInformation                        = 10,
	FileLinkInformation                          = 11,
	FileNamesInformation                         = 12,
	FileDispositionInformation                   = 13,
	FilePositionInformation                      = 14,
	FileFullEaInformation                        = 15,
	FileModeInformation                          = 16,
	FileAlignmentInformation                     = 17,
	FileAllInformation                           = 18,
	FileAllocationInformation                    = 19,
	FileEndOfFileInformation                     = 20,
	FileAlternateNameInformation                 = 21,
	FileStreamInformation                        = 22,
	FilePipeInformation                          = 23,
	FilePipeLocalInformation                     = 24,
	FilePipeRemoteInformation                    = 25,
	FileMailslotQueryInformation                 = 26,
	FileMailslotSetInformation                   = 27,
	FileCompressionInformation                   = 28,
	FileObjectIdInformation                      = 29,
	FileCompletionInformation                    = 30,
	FileMoveClusterInformation                   = 31,
	FileQuotaInformation                         = 32,
	FileReparsePointInformation                  = 33,
	FileNetworkOpenInformation                   = 34,
	FileAttributeTagInformation                  = 35,
	FileTrackingInformation                      = 36,
	FileIdBothDirectoryInformation               = 37,
	FileIdFullDirectoryInformation               = 38,
	FileValidDataLengthInformation               = 39,
	FileShortNameInformation                     = 40,
	FileIoCompletionNotificationInformation      = 41,
	FileIoStatusBlockRangeInformation            = 42,
	FileIoPriorityHintInformation                = 43,
	FileSfioReserveInformation                   = 44,
	FileSfioVolumeInformation                    = 45,
	FileHardLinkInformation                      = 46,
	FileProcessIdsUsingFileInformation           = 47,
	FileNormalizedNameInformation                = 48,
	FileNetworkPhysicalNameInformation           = 49,
	FileIdGlobalTxDirectoryInformation           = 50,
	FileIsRemoteDeviceInformation                = 51,
	FileUnusedInformation                        = 52,
	FileNumaNodeInformation                      = 53,
	FileStandardLinkInformation                  = 54,
	FileRemoteProtocolInformation                = 55,
#if 0 // NOTE: only for kernel mode drivers!
	FileRenameInformationBypassAccessCheck       = 56,
	FileLinkInformationBypassAccessCheck         = 57,
#endif
	FileVolumeNameInformation                    = 58,
	FileIdInformation                            = 59,
	FileIdExtdDirectoryInformation               = 60,
	FileReplaceCompletionInformation             = 61,
	FileHardLinkFullIdInformation                = 62,
	FileIdExtdBothDirectoryInformation           = 63,
	FileDispositionInformationEx                 = 64,
	FileRenameInformationEx                      = 65,
	FileRenameInformationExBypassAccessCheck     = 66,
	FileDesiredStorageClassInformation           = 67,
	FileStatInformation                          = 68,
	FileMemoryPartitionInformation               = 69,
	FileStatLxInformation                        = 70,
	FileCaseSensitiveInformation                 = 71,
	FileLinkInformationEx                        = 72,
	FileLinkInformationExBypassAccessCheck       = 73,
	FileStorageReserveIdInformation              = 74,
	FileCaseSensitiveInformationForceAccessCheck = 75,
	FileKnownFolderInformation                   = 76,
	FileMaximumInformation
} FILE_INFORMATION_CLASS, *LPFILE_INFORMATION_CLASS, *PFILE_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff545733.aspx

typedef	struct	_FILE_ACCESS_INFORMATION
{
	ACCESS_MASK	AccessFlags;
} FILE_ACCESS_INFORMATION, *LPFILE_ACCESS_INFORMATION, *PFILE_ACCESS_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545740.aspx

typedef	struct	_FILE_ALIGNMENT_INFORMATION
{
	DWORD	AlignmentRequirement;
} FILE_ALIGNMENT_INFORMATION, *LPFILE_ALIGNMENT_INFORMATION, *PFILE_ALIGNMENT_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540232.aspx

typedef	struct	_FILE_ALLOCATION_INFORMATION
{
	LARGE_INTEGER	AllocationSize;
} FILE_ALLOCATION_INFORMATION, *LPFILE_ALLOCATION_INFORMATION, *PFILE_ALLOCATION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/aa364216.aspx
// https://msdn.microsoft.com/en-us/library/ff545750.aspx

typedef	struct	_FILE_ATTRIBUTE_TAG_INFORMATION
{
	DWORD	FileAttributes;
	DWORD	ReparseTag;
} FILE_ATTRIBUTE_TAG_INFORMATION, *LPFILE_ATTRIBUTE_TAG_INFORMATION, *PFILE_ATTRIBUTE_TAG_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540235.aspx

typedef	struct	_FILE_BOTH_DIR_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	DWORD		EaSize;
	BYTE		ShortNameLength;
	WCHAR		ShortName[12];
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_BOTH_DIR_INFORMATION, *LPFILE_BOTH_DIR_INFORMATION, *PFILE_BOTH_DIR_INFORMATION;

#define FILE_CS_FLAG_CASE_SENSITIVE_DIR	0x00000001UL

typedef	struct	_FILE_CASE_SENSITIVE_INFORMATION
{
	DWORD	Flags;
} FILE_CASE_SENSITIVE_INFORMATION, *LPFILE_CASE_SENSITIVE_INFORMATION, *PFILE_CASE_SENSITIVE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/dn369252.aspx

typedef	struct	_FILE_COMPLETION_INFORMATION
{
	HANDLE	Port;
	LPVOID	Key;
} FILE_COMPLETION_INFORMATION, *LPFILE_COMPLETION_INFORMATION, *PFILE_COMPLETION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540239.aspx

typedef	struct	_FILE_COMPRESSION_INFORMATION
{
	LARGE_INTEGER	CompressedFileSize;
	WORD		CompressionFormat;
	BYTE		CompressionUnitShift;
	BYTE		ChunkShift;
	BYTE		ClusterShift;
	BYTE		Reserved[3];
} FILE_COMPRESSION_INFORMATION, *LPFILE_COMPRESSION_INFORMATION, *PFILE_COMPRESSION_INFORMATION;

#if _WIN32_WINNT >= 0x0A00
typedef	enum	_FILE_STORAGE_TIER_CLASS
{
	FileStorageTierClassUnspecified,
	FileStorageTierClassCapacity,
	FileStorageTierClassPerformance,
	FileStorageTierClassMax
} FILE_STORAGE_TIER_CLASS, *LPFILE_STORAGE_TIER_CLASS, *PFILE_STORAGE_TIER_CLASS;

typedef	struct	_FILE_DESIRED_STORAGE_CLASS_INFORMATION
{
	FILE_STORAGE_TIER_CLASS	Class;
	DWORD			Flags;
} FILE_DESIRED_STORAGE_CLASS_INFORMATION, *LPFILE_DESIRED_STORAGE_CLASS_INFORMATION, *PFILE_DESIRED_STORAGE_CLASS_INFORMATION;
#endif

// https://msdn.microsoft.com/en-us/library/ff540248.aspx

typedef	struct	_FILE_DIRECTORY_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_DIRECTORY_INFORMATION, *LPFILE_DIRECTORY_INFORMATION, *PFILE_DIRECTORY_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545765.aspx
// https://msdn.microsoft.com/en-us/library/aa364221.aspx

typedef	struct	_FILE_DISPOSITION_INFORMATION
{
	BOOLEAN	DeleteFile;
} FILE_DISPOSITION_INFORMATION, *LPFILE_DISPOSITION_INFORMATION, *PFILE_DISPOSITION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545773.aspx

typedef	struct	_FILE_EA_INFORMATION
{
	DWORD	EaSize;
} FILE_EA_INFORMATION, *LPFILE_EA_INFORMATION, *PFILE_EA_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545780.aspx

typedef	struct	_FILE_END_OF_FILE_INFORMATION
{
	LARGE_INTEGER	EndOfFile;
} FILE_END_OF_FILE_INFORMATION, *LPFILE_END_OF_FILE_INFORMATION, *PFILE_END_OF_FILE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540289.aspx

typedef	struct	_FILE_FULL_DIR_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	DWORD		EaSize;
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_FULL_DIR_INFORMATION, *LPFILE_FULL_DIR_INFORMATION, *PFILE_FULL_DIR_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545793.aspx

typedef	struct	_FILE_FULL_EA_INFORMATION
{
	DWORD	NextEntryOffset;
	BYTE	Flags;
	BYTE	EaNameLength;
	WORD	EaValueLength;
	CHAR	EaName[ANYSIZE_ARRAY];
} FILE_FULL_EA_INFORMATION, *LPFILE_FULL_EA_INFORMATION, *PFILE_FULL_EA_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540303.aspx

typedef	struct	_FILE_ID_BOTH_DIR_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	DWORD		EaSize;
	BYTE		ShortNameLength;
	WCHAR		ShortName[12];
	LARGE_INTEGER	FileId;
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_ID_BOTH_DIR_INFORMATION, *LPFILE_ID_BOTH_DIR_INFORMATION, *PFILE_ID_BOTH_DIR_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540310.aspx

typedef	struct	_FILE_ID_FULL_DIR_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	DWORD		EaSize;
	LARGE_INTEGER	FileId;
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_ID_FULL_DIR_INFORMATION, *LPFILE_ID_FULL_DIR_INFORMATION, *PFILE_ID_FULL_DIR_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728839.aspx

typedef	struct	_FILE_ID_GLOBAL_TX_DIR_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		FileIndex;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	EndOfFile;
	LARGE_INTEGER	AllocationSize;
	DWORD		FileAttributes;
	DWORD		FileNameLength;
	LARGE_INTEGER	FileId;
	GUID		LockingTransactionId;
	DWORD		TxInfoFlags;
	WCHAR		FileName[ANYSIZE_ARRAY];
} FILE_ID_GLOBAL_TX_DIR_INFORMATION, *LPFILE_ID_GLOBAL_TX_DIR_INFORMATION, *PFILE_ID_GLOBAL_TX_DIR_INFORMATION;

typedef	struct	_FILE_ID_128
{
	BYTE	Identifier[16];
} FILE_ID_128, *LPFILE_ID_128, *PFILE_ID_128;

typedef	struct	_FILE_ID_INFORMATION
{
	QWORD		VolumeSerialNumber;
	FILE_ID_128	FileId;
} FILE_ID_INFORMATION, *LPFILE_ID_INFORMATION, *PFILE_ID_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540318.aspx

typedef	struct	_FILE_INTERNAL_INFORMATION
{
	LARGE_INTEGER	IndexNumber;
} FILE_INTERNAL_INFORMATION, *LPFILE_INTERNAL_INFORMATION, *PFILE_INTERNAL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff550594.aspx

typedef	enum	_IO_PRIORITY_HINT
{
	IoPriorityVeryLow  = 0,
	IoPriorityLow      = 1,
	IoPriorityNormal   = 2,
	IoPriorityHigh     = 3,
	IoPriorityCritical = 4,
	MaxIoPriorityTypes = 5
} IO_PRIORITY_HINT;

// https://msdn.microsoft.com/en-us/library/ff545802.aspx

typedef	struct	_FILE_IO_PRIORITY_HINT_INFORMATION
{
	IO_PRIORITY_HINT	PriorityHint;
} FILE_IO_PRIORITY_HINT_INFORMATION, *LPFILE_IO_PRIORITY_HINT_INFORMATION, *PFILE_IO_PRIORITY_HINT_INFORMATION;

typedef	struct	_FILE_IS_REMOTE_DEVICE_INFORMATION
{
	BOOLEAN	IsRemote;
} FILE_IS_REMOTE_DEVICE_INFORMATION, *LPFILE_IS_REMOTE_DEVICE_INFORMATION, *PFILE_IS_REMOTE_DEVICE_INFORMATION;

#if _WIN32_WINNT >= 0x0A00
typedef	enum	_FILE_KNOWN_FOLDER_TYPE
{
	KnownFolderNone,
	KnownFolderDesktop,
	KnownFolderDocuments,
	KnownFolderDownloads,
	KnownFolderMusic,
	KnownFolderPictures,
	KnownFolderVideos,
	KnownFolderOther,
	KnownFolderMax
} FILE_KNOWN_FOLDER_TYPE;

typedef	struct	_FILE_KNOWN_FOLDER_INFORMATION
{
	FILE_KNOWN_FOLDER_TYPE	Type;
} FILE_KNOWN_FOLDER_INFORMATION, *LPFILE_KNOWN_FOLDER_INFORMATION, *PFILE_KNOWN_FOLDER_INFORMATION;
#endif

// https://msdn.microsoft.com/en-us/library/ff540324.aspx

typedef	struct	_FILE_LINK_INFORMATION
{
	BOOLEAN	ReplaceIfExists;
	HANDLE	RootDirectory;
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_LINK_INFORMATION, *LPFILE_LINK_INFORMATION, *PFILE_LINK_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728842.aspx

typedef	struct	_FILE_LINK_ENTRY_INFORMATION
{
	DWORD		NextEntryOffset;
	LONGLONG	ParentFileId;
	DWORD		FileNameLength;
	WCHAR		FileName;
} FILE_LINK_ENTRY_INFORMATION, *LPFILE_LINK_ENTRY_INFORMATION, *PFILE_LINK_ENTRY_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728841.aspx

typedef	struct	_FILE_LINKS_INFORMATION
{
	DWORD				BytesNeeded;
	DWORD				EntriesReturned;
	FILE_LINK_ENTRY_INFORMATION	Entry;
} FILE_LINKS_INFORMATION, *LPFILE_LINKS_INFORMATION, *PFILE_LINKS_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728843.aspx

typedef	struct	_FILE_MAILSLOT_QUERY_INFORMATION
{
	DWORD		MaximumMessageSize;
	DWORD		MailslotQuota;
	DWORD		NextMessageSize;
	DWORD		MessagesAvailable;
	LARGE_INTEGER	ReadTimeout;
} FILE_MAILSLOT_QUERY_INFORMATION, *LPFILE_MAILSLOT_QUERY_INFORMATION, *PFILE_MAILSLOT_QUERY_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728844.aspx

typedef	struct	_FILE_MAILSLOT_SET_INFORMATION
{
	LARGE_INTEGER	*ReadTimeout;
} FILE_MAILSLOT_SET_INFORMATION, *LPFILE_MAILSLOT_SET_INFORMATION, *PFILE_MAILSLOT_SET_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545809.aspx

typedef	struct	_FILE_MODE_INFORMATION
{
	DWORD	Mode;
} FILE_MODE_INFORMATION, *LPFILE_MODE_INFORMATION, *PFILE_MODE_INFORMATION;

typedef	struct	_FILE_MOVE_CLUSTER_INFORMATION
{
	DWORD	ClusterCount;
	HANDLE	RootDirectory;
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_MOVE_CLUSTER_INFORMATION, *LPFILE_MOVE_CLUSTER_INFORMATION, *PFILE_MOVE_CLUSTER_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545817.aspx

typedef	struct	_FILE_NAME_INFORMATION
{
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_NAME_INFORMATION, *LPFILE_NAME_INFORMATION, *PFILE_NAME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540329.aspx

typedef	struct	_FILE_NAMES_INFORMATION
{
	DWORD	NextEntryOffset;
	DWORD	FileIndex;
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_NAMES_INFORMATION, *LPFILE_NAMES_INFORMATION, *PFILE_NAMES_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545822.aspx

typedef	struct	_FILE_NETWORK_OPEN_INFORMATION
{
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	AllocationSize;
	LARGE_INTEGER	EndOfFile;
	DWORD		FileAttributes;
} FILE_NETWORK_OPEN_INFORMATION, *LPFILE_NETWORK_OPEN_INFORMATION, *PFILE_NETWORK_OPEN_INFORMATION;

// https://msdn.microsoft.com/en-us/library/dn602488.aspx

typedef	struct	_FILE_NETWORK_PHYSICAL_NAME_INFORMATION
{
	DWORD	FileNameLength;
	WCHAR	FileName[ANYSIZE_ARRAY];
} FILE_NETWORK_PHYSICAL_NAME_INFORMATION, *LPFILE_NETWORK_PHYSICAL_NAME_INFORMATION, *PFILE_NETWORK_PHYSICAL_NAME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540335.aspx

#pragma warning(disable: 4201)
typedef	struct	_FILE_OBJECTID_INFORMATION
{
	LONGLONG	FileReference;
	BYTE		ObjectId[16];
	union
	{
		struct
		{
			BYTE	BirthVolumeId[16];
			BYTE	BirthObjectId[16];
			BYTE	DomainId[16];
		};
		BYTE		ExtendedInfo[48];
	};
} FILE_OBJECTID_INFORMATION, *LPFILE_OBJECTID_INFORMATION, *PFILE_OBJECTID_INFORMATION;
#pragma warning(default: 4201)

// https://msdn.microsoft.com/en-us/library/ff728845.aspx

typedef	struct	_FILE_PIPE_INFORMATION
{
	DWORD	ReadMode;
	DWORD	CompletionMode;
} FILE_PIPE_INFORMATION, *LPFILE_PIPE_INFORMATION, *PFILE_PIPE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728846.aspx

#define FILE_PIPE_INBOUND		0x00000000UL
#define FILE_PIPE_OUTBOUND		0x00000001UL
#define FILE_PIPE_FULL_DUPLEX		0x00000002UL

#define FILE_PIPE_DISCONNECTED_STATE	0x00000001UL
#define FILE_PIPE_LISTENING_STATE	0x00000002UL
#define FILE_PIPE_CONNECTED_STATE	0x00000003UL
#define FILE_PIPE_CLOSING_STATE		0x00000004UL

#define FILE_PIPE_CLIENT_END		0x00000000UL
#define FILE_PIPE_SERVER_END		0x00000001UL

typedef	struct	_FILE_PIPE_LOCAL_INFORMATION
{
	DWORD	NamedPipeType;
	DWORD	NamedPipeConfiguration;
	DWORD	MaximumInstances;
	DWORD	CurrentInstances;
	DWORD	InboundQuota;
	DWORD	ReadDataAvailable;
	DWORD	OutboundQuota;
	DWORD	WriteQuotaAvailable;
	DWORD	NamedPipeState;
	DWORD	NamedPipeEnd;
} FILE_PIPE_LOCAL_INFORMATION, *LPFILE_PIPE_LOCAL_INFORMATION, *PFILE_PIPE_LOCAL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff728847.aspx

typedef	struct	_FILE_PIPE_REMOTE_INFORMATION
{
	LARGE_INTEGER	CollectDataTime;
	DWORD		MaximumCollectionCount;
} FILE_PIPE_REMOTE_INFORMATION, *LPFILE_PIPE_REMOTE_INFORMATION, *PFILE_PIPE_REMOTE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545848.aspx

typedef	struct	_FILE_POSITION_INFORMATION
{
	LARGE_INTEGER	CurrentByteOffset;
} FILE_POSITION_INFORMATION, *LPFILE_POSITION_INFORMATION, *PFILE_POSITION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540342.aspx

typedef	struct	_FILE_QUOTA_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		SidLength;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	QuotaUsed;
	LARGE_INTEGER	QuotaThreshold;
	LARGE_INTEGER	QuotaLimit;
	SID		Sid;
} FILE_QUOTA_INFORMATION, *LPFILE_QUOTA_INFORMATION, *PFILE_QUOTA_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540354.aspx

typedef	struct	_FILE_REPARSE_POINT_INFORMATION
{
	LONGLONG	FileReference;
	DWORD		Tag;
} FILE_REPARSE_POINT_INFORMATION, *LPFILE_REPARSE_POINT_INFORMATION, *PFILE_REPARSE_POINT_INFORMATION;

typedef	struct	_FILE_SFIO_RESERVE_INFORMATION
{
	DWORD	RequestsPerPeriod;
	DWORD	Period;
	BOOLEAN	RetryFailures;
	BOOLEAN	Discardable;
	DWORD	RequestSize;
	DWORD	NumOutstandingRequests;
} FILE_SFIO_RESERVE_INFORMATION, *LPFILE_SFIO_RESERVE_INFORMATION, *PFILE_SFIO_RESERVE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545855.aspx

typedef	struct	_FILE_STANDARD_INFORMATION
{
	LARGE_INTEGER	AllocationSize;
	LARGE_INTEGER	EndOfFile;
	DWORD		NumberOfLinks;
	BOOLEAN		DeletePending;
	BOOLEAN		Directory;
} FILE_STANDARD_INFORMATION, *LPFILE_STANDARD_INFORMATION, *PFILE_STANDARD_INFORMATION;

typedef	struct	_FILE_STANDARD_LINK_INFORMATION
{
	DWORD	NumberOfAccessibleLinks;
	DWORD	TotalNumberOfLinks;
	BOOLEAN	DeletePending;
	BOOLEAN	Directory;
} FILE_STANDARD_LINK_INFORMATION, *LPFILE_STANDARD_LINK_INFORMATION, *PFILE_STANDARD_LINK_INFORMATION;

typedef	struct	_FILE_STAT_INFORMATION
{
	LARGE_INTEGER	FileId;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	AllocationSize;
	LARGE_INTEGER	EndOfFile;
	DWORD		FileAttributes;
	DWORD		ReparseTag;
	DWORD		NumberOfLinks;
	ACCESS_MASK	EffectiveAccess;
} FILE_STAT_INFORMATION, *LPFILE_STAT_INFORMATION, *PFILE_STAT_INFORMATION;

#if _WIN32_WINNT >= 0x0A00
#define LX_FILE_METADATA_HAS_UID	0x00000001UL
#define LX_FILE_METADATA_HAS_GID	0x00000002UL
#define LX_FILE_METADATA_HAS_MODE	0x00000004UL
#define LX_FILE_METADATA_HAS_DEVICE_ID	0x00000008UL
#define LX_FILE_CASE_SENSITIVE_DIR	0x00000010UL

typedef	struct	_FILE_STAT_LX_INFORMATION
{
	LARGE_INTEGER	FileId;
	LARGE_INTEGER	CreationTime;
	LARGE_INTEGER	LastAccessTime;
	LARGE_INTEGER	LastWriteTime;
	LARGE_INTEGER	ChangeTime;
	LARGE_INTEGER	AllocationSize;
	LARGE_INTEGER	EndOfFile;
	DWORD		FileAttributes;
	DWORD		ReparseTag;
	DWORD		NumberOfLinks;
	ACCESS_MASK	EffectiveAccess;
	DWORD		LxFlags;
	DWORD		LxUid;
	DWORD		LxGid;
	DWORD		LxMode;
	DWORD		LxDeviceIdMajor;
	DWORD		LxDeviceIdMinor;
} FILE_STAT_LX_INFORMATION, *LPFILE_STAT_LX_INFORMATION, *PFILE_STAT_LX_INFORMATION;

typedef	enum	_STORAGE_RESERVE_ID
{
	StorageReserveIdNone,
	StorageReserveIdHard,
	StorageReserveIdSoft,
	StorageReserveIdUpdateScratch,
	StorageReserveIdMax
} STORAGE_RESERVE_ID, *LPSTORAGE_RESERVE_ID, *PSTORAGE_RESERVE_ID;

typedef	struct	_FILE_STORAGE_RESERVE_ID_INFORMATION
{
	STORAGE_RESERVE_ID	StorageReserveId;
} FILE_STORAGE_RESERVE_ID_INFORMATION, *LPFILE_STORAGE_RESERVE_ID_INFORMATION, *PFILE_STORAGE_RESERVE_ID_INFORMATION;
#endif

// https://msdn.microsoft.com/en-us/library/ff540364.aspx

typedef	struct	_FILE_STREAM_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		StreamNameLength;
	LARGE_INTEGER	StreamSize;
	LARGE_INTEGER	StreamAllocationSize;
	WCHAR		StreamName[ANYSIZE_ARRAY];
} FILE_STREAM_INFORMATION, *LPFILE_STREAM_INFORMATION, *PFILE_STREAM_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545873.aspx

typedef	struct	_FILE_VALID_DATA_LENGTH_INFORMATION
{
	LARGE_INTEGER	ValidDataLength;
} FILE_VALID_DATA_LENGTH_INFORMATION, *LPFILE_VALID_DATA_LENGTH_INFORMATION, *PFILE_VALID_DATA_LENGTH_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff545743.aspx

typedef	struct	_FILE_ALL_INFORMATION
{
	FILE_BASIC_INFORMATION		BasicInformation;
	FILE_STANDARD_INFORMATION	StandardInformation;
	FILE_INTERNAL_INFORMATION	InternalInformation;
	FILE_EA_INFORMATION		EaInformation;
	FILE_ACCESS_INFORMATION		AccessInformation;
	FILE_POSITION_INFORMATION	PositionInformation;
	FILE_MODE_INFORMATION		ModeInformation;
	FILE_ALIGNMENT_INFORMATION	AlignmentInformation;
	FILE_NAME_INFORMATION		NameInformation;
} FILE_ALL_INFORMATION, *LPFILE_ALL_INFORMATION, *PFILE_ALL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff556633.aspx
// https://msdn.microsoft.com/en-us/library/ff567047.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryDirectoryFile(HANDLE                 FileHandle,
		                     HANDLE                 EventHandle,
		                     IO_APC_ROUTINE         *ApcRoutine,
		                     LPVOID                 ApcContext,
		                     IO_STATUS_BLOCK        *IoStatusBlock,
		                     LPVOID                 FileInformation,
		                     DWORD                  FileInformationLength,
		                     FILE_INFORMATION_CLASS FileInformationClass,
		                     BOOLEAN                ReturnSingleEntry,
		                     UNICODE_STRING         *FileName,
		                     BOOLEAN                RestartScan);

// https://msdn.microsoft.com/en-us/library/bb470238.aspx

typedef	struct	_OBJECT_DIRECTORY_INFORMATION
{
	UNICODE_STRING	Name;
	UNICODE_STRING	TypeName;
} OBJECT_DIRECTORY_INFORMATION, *LPOBJECT_DIRECTORY_INFORMATION, *POBJECT_DIRECTORY_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQueryDirectoryObject(HANDLE  DirectoryHandle,
		                       LPVOID  Buffer,
		                       DWORD   BufferLength,
		                       BOOLEAN ReturnSingleEntry,
		                       BOOLEAN RestartScan,
		                       LPDWORD Context,
		                       LPDWORD ReturnLength);

// https://msdn.microsoft.com/en-us/library/ff625894.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryEaFile(HANDLE          FileHandle,
		              IO_STATUS_BLOCK *IoStatusBlock,
		              LPVOID          Buffer,
		              DWORD           BufferLength,
		              BOOLEAN         ReturnSingleEntry,
		              LPVOID          EaList,
		              DWORD           EaListLength,
		              LPDWORD         EaIndex,
		              BOOLEAN         RestartScan);

typedef	enum	_EVENT_INFORMATION_CLASS
{
	EventBasicInformation = 0
} EVENT_INFORMATION_CLASS, *LPEVENT_INFORMATION_CLASS, *PEVENT_INFORMATION_CLASS;

typedef	struct	_EVENT_BASIC_INFORMATION
{
	EVENT_TYPE	EventType;
	LONG		EventState;
} EVENT_BASIC_INFORMATION, *LPEVENT_BASIC_INFORMATION, *PEVENT_BASIC_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQueryEvent(HANDLE                  EventHandle,
		             EVENT_INFORMATION_CLASS EventInformationClass,
		             LPVOID                  EventInformation,
		             DWORD                   EventInformationLength,
		             LPDWORD                 ReturnLength);

// https://msdn.microsoft.com/en-us/library/ff567049.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryFullAttributesFile(OBJECT_ATTRIBUTES             *ObjectAttributes,
		                          FILE_NETWORK_OPEN_INFORMATION *FileInformation);

typedef	struct	_ATOM_BASIC_INFORMATION
{
	WORD	UsageCount;
	WORD	Flags;
	WORD	NameLength;
	WCHAR	Name[ANYSIZE_ARRAY];
} ATOM_BASIC_INFORMATION, *LPATOM_BASIC_INFORMATION, *PATOM_BASIC_INFORMATION;

typedef	struct	_ATOM_TABLE_INFORMATION
{
	DWORD		NumberOfAtoms;
	RTL_ATOM	Atoms[ANYSIZE_ARRAY];
} ATOM_TABLE_INFORMATION, *LPATOM_TABLE_INFORMATION, *PATOM_TABLE_INFORMATION;

typedef	enum	_ATOM_INFORMATION_CLASS
{
	AtomBasicInformation = 0,
	AtomTableInformation = 1
} ATOM_INFORMATION_CLASS;

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationAtom(RTL_ATOM               Atom,
		                       ATOM_INFORMATION_CLASS AtomInformationClass,
		                       LPVOID                 AtomInformation,
		                       DWORD                  AtomInformationLength,
		                       LPDWORD                ReturnLength);

#if _WIN32_WINNT >= 0x0A00
__declspec(dllimport)
LONG	NTAPI	NtQueryInformationByName(OBJECT_ATTRIBUTES      *ObjectAttributes,
		                         IO_STATUS_BLOCK        *IoStatusBlock,
		                         LPVOID                 FileInformation,
		                         DWORD                  FileInformationLength,
		                         FILE_INFORMATION_CLASS FileInformationClass);
#endif

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff544254.aspx

typedef	struct	_ENLISTMENT_BASIC_INFORMATION
{
	GUID	EnlistmentId;
	GUID	TransactionId;
	GUID	ResourceManagerId;
} ENLISTMENT_BASIC_INFORMATION, *LPENLISTMENT_BASIC_INFORMATION, *PENLISTMENT_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff544260.aspx

typedef	enum	_ENLISTMENT_INFORMATION_CLASS
{
	EnlistmentBasicInformation    = 0,
	EnlistmentRecoveryInformation,
	EnlistmentCrmInformation
} ENLISTMENT_INFORMATION_CLASS;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff567051.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationEnlistment(HANDLE                       EnlistmentHandle,
		                             ENLISTMENT_INFORMATION_CLASS EnlistmentInformationClass,
		                             LPVOID                       EnlistmentInformation,
		                             DWORD                        EnlistmentInformationLength,
		                             LPDWORD                      ReturnLength);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff556646.aspx
// https://msdn.microsoft.com/en-us/library/ff567052.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationFile(HANDLE                 FileHandle,
		                       IO_STATUS_BLOCK        *IoStatusBlock,
		                       LPVOID                 FileInformation,
		                       DWORD                  FileInformationLength,
		                       FILE_INFORMATION_CLASS FileInformationClass);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms684925.aspx
// https://msdn.microsoft.com/en-us/library/ms686216.aspx

typedef	enum	_JOBOBJECTINFOCLASS
{
	JobObjectBasicAccountingInformation         = 1,
	JobObjectBasicLimitInformation              = 2,
	JobObjectBasicProcessIdList                 = 3,
	JobObjectBasicUIRestrictions                = 4,
	JobObjectSecurityLimitInformation           = 5,
	JobObjectEndOfJobTimeInformation            = 6,
	JobObjectAssociateCompletionPortInformation = 7,
	JobObjectBasicAndIoAccountingInformation    = 8,
	JobObjectExtendedLimitInformation           = 9,
	JobObjectJobSetInformation                  = 10,
#if _WIN32_WINNT >= 0x0601
	JobObjectGroupInformation                   = 11,
#if _WIN32_WINNT >= 0x0602
	JobObjectNotificationLimitInformation       = 12,
	JobObjectGroupInformationEx                 = 14,
	JobObjectCpuRateControlInformation          = 15,
#if _WIN32_WINNT >= 0x0A00
	JobObjectNetRateControlInformation          = 32,
	JobObjectNotificationLimitInformation2      = 34,
	JobObjectLimitViolationInformation2         = 35,
#endif
#endif
#endif
	MaxJobObjectInfoClass
} JOBOBJECTINFOCLASS;

// https://msdn.microsoft.com/en-us/library/ms684143.aspx

typedef	struct	_JOBOBJECT_BASIC_ACCOUNTING_INFORMATION
{
	LARGE_INTEGER	TotalUserTime;
	LARGE_INTEGER	TotalKernelTime;
	LARGE_INTEGER	ThisPeriodTotalUserTime;
	LARGE_INTEGER	ThisPeriodTotalKernelTime;
	DWORD		TotalPageFaultCount;
	DWORD		TotalProcesses;
	DWORD		ActiveProcesses;
	DWORD		TotalTerminatedProcesses;
} JOBOBJECT_BASIC_ACCOUNTING_INFORMATION, *LPJOBOBJECT_BASIC_ACCOUNTING_INFORMATION, *PJOBOBJECT_BASIC_ACCOUNTING_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ms684147.aspx

#define JOB_OBJECT_LIMIT_WORKINGSET			0x00000001UL
#define JOB_OBJECT_LIMIT_PROCESS_TIME			0x00000002UL
#define JOB_OBJECT_LIMIT_JOB_TIME			0x00000004UL
#define JOB_OBJECT_LIMIT_ACTIVE_PROCESS			0x00000008UL
#define JOB_OBJECT_LIMIT_AFFINITY			0x00000010UL
#define JOB_OBJECT_LIMIT_PRIORITY_CLASS			0x00000020UL
#define JOB_OBJECT_LIMIT_PRESERVE_JOB_TIME		0x00000040UL
#define JOB_OBJECT_LIMIT_SCHEDULING_CLASS		0x00000080UL
#define JOB_OBJECT_LIMIT_PROCESS_MEMORY			0x00000100UL
#define JOB_OBJECT_LIMIT_JOB_MEMORY			0x00000200UL
#define JOB_OBJECT_LIMIT_DIE_ON_UNHANDLED_EXCEPTION	0x00000400UL
#define JOB_OBJECT_LIMIT_BREAKAWAY_OK			0x00000800UL
#define JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK		0x00001000UL
#define JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE		0x00002000UL
#define JOB_OBJECT_LIMIT_SUBSET_AFFINITY		0x00004000UL

typedef	struct	_JOBOBJECT_BASIC_LIMIT_INFORMATION
{
	LARGE_INTEGER	PerProcessUserTimeLimit;
	LARGE_INTEGER	PerJobUserTimeLimit;
	DWORD		LimitFlags;
	SIZE_T		MinimumWorkingSetSize;
	SIZE_T		MaximumWorkingSetSize;
	DWORD		ActiveProcessLimit;
	DWORD_PTR	Affinity;
	DWORD		PriorityClass;
	DWORD		SchedulingClass;
} JOBOBJECT_BASIC_LIMIT_INFORMATION, *LPJOBOBJECT_BASIC_LIMIT_INFORMATION, *PJOBOBJECT_BASIC_LIMIT_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ms684150.aspx

typedef	struct	_JOBOBJECT_BASIC_PROCESS_ID_LIST
{
	DWORD		NumberOfAssignedProcesses;
	DWORD		NumberOfProcessIdsInList;
	DWORD_PTR	ProcessIdList[ANYSIZE_ARRAY];
} JOBOBJECT_BASIC_PROCESS_ID_LIST, *LPJOBOBJECT_BASIC_PROCESS_ID_LIST, *PJOBOBJECT_BASIC_PROCESS_ID_LIST;

// https://msdn.microsoft.com/en-us/library/ms684152.aspx

#define JOB_OBJECT_UILIMIT_HANDLES		0x00000001UL
#define JOB_OBJECT_UILIMIT_READCLIPBOARD	0x00000002UL
#define JOB_OBJECT_UILIMIT_WRITECLIPBOARD	0x00000004UL
#define JOB_OBJECT_UILIMIT_SYSTEMPARAMETERS	0x00000008UL
#define JOB_OBJECT_UILIMIT_DISPLAYSETTINGS	0x00000010UL
#define JOB_OBJECT_UILIMIT_GLOBALATOMS		0x00000020UL
#define JOB_OBJECT_UILIMIT_DESKTOP		0x00000040UL
#define JOB_OBJECT_UILIMIT_EXITWINDOWS		0x00000080UL

typedef	struct	_JOBOBJECT_BASIC_UI_RESTRICTIONS
{
	DWORD	UIRestrictionsClass;
} JOBOBJECT_BASIC_UI_RESTRICTIONS, *LPJOBOBJECT_BASIC_UI_RESTRICTIONS, *PJOBOBJECT_BASIC_UI_RESTRICTIONS;

// https://msdn.microsoft.com/en-us/library/ms684159.aspx

#define JOB_OBJECT_SECURITY_NO_ADMIN		0x00000001UL
#define JOB_OBJECT_SECURITY_RESTRICTED_TOKEN	0x00000002UL
#define JOB_OBJECT_SECURITY_ONLY_TOKEN		0x00000004UL
#define JOB_OBJECT_SECURITY_FILTER_TOKENS	0x00000008UL

typedef	struct	_JOBOBJECT_SECURITY_LIMIT_INFORMATION
{
	DWORD			SecurityLimitFlags;
	HANDLE			JobToken;
	TOKEN_GROUPS		*SidsToDisable;
	TOKEN_PRIVILEGES	*PrivilegesToDelete;
	TOKEN_GROUPS		*RestrictedSids;
} JOBOBJECT_SECURITY_LIMIT_INFORMATION, *LPJOBOBJECT_SECURITY_LIMIT_INFORMATION, *PJOBOBJECT_SECURITY_LIMIT_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ms684155.aspx

#define JOB_OBJECT_TERMINATE_AT_END_OF_JOB	0UL
#define JOB_OBJECT_POST_AT_END_OF_JOB		1UL

typedef	struct	_JOBOBJECT_END_OF_JOB_TIME_INFORMATION
{
	DWORD	EndOfJobTimeAction;
} JOBOBJECT_END_OF_JOB_TIME_INFORMATION, *LPJOBOBJECT_END_OF_JOB_TIME_INFORMATION, *PJOBOBJECT_END_OF_JOB_TIME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ms684141.aspx

typedef	struct	_JOBOBJECT_ASSOCIATE_COMPLETION_PORT
{
	LPVOID	CompletionKey;
	HANDLE	CompletionPort;
} JOBOBJECT_ASSOCIATE_COMPLETION_PORT, *LPJOBOBJECT_ASSOCIATE_COMPLETION_PORT, *PJOBOBJECT_ASSOCIATE_COMPLETION_PORT;

// https://msdn.microsoft.com/en-us/library/ms684144.aspx

typedef	struct	JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION
{
	JOBOBJECT_BASIC_ACCOUNTING_INFORMATION	BasicInfo;
	IO_COUNTERS				IoInfo;
} JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION, *LPJOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION, *PJOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ms684156.aspx

typedef	struct	_JOBOBJECT_EXTENDED_LIMIT_INFORMATION
{
	JOBOBJECT_BASIC_LIMIT_INFORMATION	BasicLimitInformation;
	IO_COUNTERS				IoInfo;
	SIZE_T					ProcessMemoryLimit;
	SIZE_T					JobMemoryLimit;
	SIZE_T					PeakProcessMemoryUsed;
	SIZE_T					PeakJobMemoryUsed;
} JOBOBJECT_EXTENDED_LIMIT_INFORMATION, *LPJOBOBJECT_EXTENDED_LIMIT_INFORMATION, *PJOBOBJECT_EXTENDED_LIMIT_INFORMATION;

// https://msdn.microsoft.com/en-us/library/hh448384.aspx

#define JOB_OBJECT_CPU_RATE_CONTROL_ENABLE		0x00000001UL
#define JOB_OBJECT_CPU_RATE_CONTROL_WEIGHT_BASED	0x00000002UL
#define JOB_OBJECT_CPU_RATE_CONTROL_HARD_CAP		0x00000004UL
#define JOB_OBJECT_CPU_RATE_CONTROL_NOTIFY		0x00000008UL
#define JOB_OBJECT_ CPU_RATE_CONTROL_MIN_MAX_RATE	0x00000010UL

typedef	struct	_JOBOBJECT_CPU_RATE_CONTROL_INFORMATION
{
	DWORD	ControlFlags;
	union
	{
		DWORD	CpuRate;
		DWORD	Weight;
		struct
		{
			WORD	MinRate;
			WORD	MaxRate;
		};
	};
} JOBOBJECT_CPU_RATE_CONTROL_INFORMATION, *LPJOBOBJECT_CPU_RATE_CONTROL_INFORMATION, *PJOBOBJECT_CPU_RATE_CONTROL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/hh448386.aspx

typedef	enum	_JOBOBJECT_RATE_CONTROL_TOLERANCE
{
	ToleranceLow    = 1,
	ToleranceMedium = 2,
	ToleranceHigh   = 3
} JOBOBJECT_RATE_CONTROL_TOLERANCE;

typedef	enum	_JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL
{
	ToleranceIntervalShort  = 1,
	ToleranceIntervalMedium = 2,
	ToleranceIntervalLong   = 3
} JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL;

#define JOB_OBJECT_LIMIT_JOB_TIME		0x00000004UL
#define JOB_OBJECT_LIMIT_JOB_MEMORY		0x00000200UL
#define JOB_OBJECT_LIMIT_JOB_MEMORY_HIGH	0x00000200UL
#define JOB_OBJECT_LIMIT_JOB_MEMORY_LOW		0x00008000UL
#define JOB_OBJECT_LIMIT_JOB_READ_BYTES		0x00010000UL
#define JOB_OBJECT_LIMIT_JOB_WRITE_BYTES	0x00020000UL
#define JOB_OBJECT_LIMIT_RATE_CONTROL		0x00040000UL
#define JOB_OBJECT_LIMIT_CPU_RATE_CONTROL	0x00040000UL
#define JOB_OBJECT_LIMIT_IO_RATE_CONTROL	0x00080000UL
#define JOB_OBJECT_LIMIT_NET_RATE_CONTROL	0x00100000UL

typedef	struct	_JOBOBJECT_NOTIFICATION_LIMIT_INFORMATION
{
	QWORD						IoReadBytesLimit;
	QWORD						IoWriteBytesLimit;
	LARGE_INTEGER					PerJobUserTimeLimit;
	QWORD						JobMemoryLimit;
	JOBOBJECT_RATE_CONTROL_TOLERANCE		RateControlTolerance;
	JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL	RateControlToleranceInterval;
	DWORD						LimitFlags;
} JOBOBJECT_NOTIFICATION_LIMIT_INFORMATION, *LPJOBOBJECT_NOTIFICATION_LIMIT_INFORMATION, *PJOBOBJECT_NOTIFICATION_LIMIT_INFORMATION;

// https://msdn.microsoft.com/en-us/library/hh448385.aspx

typedef	struct	_JOBOBJECT_LIMIT_VIOLATION_INFORMATION
{
	DWORD					LimitFlags;
	DWORD					ViolationLimitFlags;
	QWORD					IoReadBytes;
	QWORD					IoReadBytesLimit;
	QWORD					IoWriteBytes;
	QWORD					IoWriteBytesLimit;
	LARGE_INTEGER				PerJobUserTime;
	LARGE_INTEGER				PerJobUserTimeLimit;
	QWORD					JobMemory;
	QWORD					JobMemoryLimit;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	RateControlTolerance;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	RateControlToleranceLimit;
} JOBOBJECT_LIMIT_VIOLATION_INFORMATION, *LPJOBOBJECT_LIMIT_VIOLATION_INFORMATION, *PJOBOBJECT_LIMIT_VIOLATION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/mt280126.aspx

typedef	enum	_JOB_OBJECT_NET_RATE_CONTROL_FLAGS
{
	JOB_OBJECT_NET_RATE_CONTROL_ENABLE       = 0x1,
	JOB_OBJECT_NET_RATE_CONTROL_MAX_BANDWITH = 0x2,
	JOB_OBJECT_NET_RATE_CONTROL_DSCP_TAG     = 0x4,
	JOB_OBJECT_NET_RATE_CONTROL_VALID_FLAGS  = 0x7
} JOB_OBJECT_NET_RATE_CONTROL_FLAGS;

// https://msdn.microsoft.com/en-us/library/mt280124.aspx

typedef	struct	_JOBOBJECT_NET_RATE_CONTROL_INFORMATION
{
	QWORD					MaxBandwidth;
	JOB_OBJECT_NET_RATE_CONTROL_FLAGS	ControlFlags;
	BYTE					DscpTag;
} JOBOBJECT_NET_RATE_CONTROL_INFORMATION, *LPJOBOBJECT_NET_RATE_CONTROL_INFORMATION, *PJOBOBJECT_NET_RATE_CONTROL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/mt280125.aspx

typedef	struct	_JOBOBJECT_NOTIFICATION_LIMIT_INFORMATION_2
{
	QWORD		IoReadBytesLimit;
	QWORD		IoWriteBytesLimit;
	LARGE_INTEGER	PerJobUserTimeLimit;
	union
	{
		QWORD	JobHighMemoryLimit;
		QWORD	JobMemoryLimit;
	};
	union
	{
		JOBOBJECT_RATE_CONTROL_TOLERANCE	RateControlTolerance;
		JOBOBJECT_RATE_CONTROL_TOLERANCE	CpuRateControlTolerance;
	};
	union
	{
		JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL	RateControlToleranceInterval;
		JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL	CpuRateControlToleranceInterval;
	};
	DWORD						LimitFlags;
	JOBOBJECT_RATE_CONTROL_TOLERANCE		IoRateControlTolerance;
	QWORD						JobLowMemoryLimit;
	JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL	IoRateControlToleranceInterval;
	JOBOBJECT_RATE_CONTROL_TOLERANCE		NetRateControlTolerance;
	JOBOBJECT_RATE_CONTROL_TOLERANCE_INTERVAL	NetRateControlToleranceInterval;
} JOBOBJECT_NOTIFICATION_LIMIT_INFORMATION_2, *LPJOBOBJECT_NOTIFICATION_LIMIT_INFORMATION_2, *PJOBOBJECT_NOTIFICATION_LIMIT_INFORMATION_2;

// https://msdn.microsoft.com/en-us/library/mt280123.aspx

typedef	struct	_JOBOBJECT_LIMIT_VIOLATION_INFORMATION_2
{
	DWORD		LimitFlags;
	DWORD		ViolationLimitFlags;
	QWORD		IoReadBytes;
	QWORD		IoReadBytesLimit;
	QWORD		IoWriteBytes;
	QWORD		IoWriteBytesLimit;
	LARGE_INTEGER	PerJobUserTime;
	LARGE_INTEGER	PerJobUserTimeLimit;
	QWORD		JobMemory;
	union
	{
		QWORD	JobHighMemoryLimit;
		QWORD	JobMemoryLimit;
	};
	union
	{
		JOBOBJECT_RATE_CONTROL_TOLERANCE	RateControlTolerance;
		JOBOBJECT_RATE_CONTROL_TOLERANCE	CpuRateControlTolerance;
	};
	union
	{
		JOBOBJECT_RATE_CONTROL_TOLERANCE	RateControlToleranceLimit;
		JOBOBJECT_RATE_CONTROL_TOLERANCE	CpuRateControlToleranceLimit;
	};
	QWORD					JobLowMemoryLimit;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	IoRateControlTolerance;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	IoRateControlToleranceLimit;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	NetRateControlTolerance;
	JOBOBJECT_RATE_CONTROL_TOLERANCE	NetRateControlToleranceLimit;
} JOBOBJECT_LIMIT_VIOLATION_INFORMATION_2, *LPJOBOBJECT_LIMIT_VIOLATION_INFORMATION_2, *PJOBOBJECT_LIMIT_VIOLATION_INFORMATION_2;
#endif // _WINNT_

// NOTE: like KERNEL32!QueryInformationJobObject

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationJobObject(HANDLE             JobHandle,
		                            JOBOBJECTINFOCLASS JobObjectInformationClass,
		                            LPVOID             JobObjectInformation,
		                            DWORD              JobObjectInformationLength,
		                            LPDWORD            ReturnLength);

// https://msdn.microsoft.com/en-us/library/ms684280.aspx
// https://msdn.microsoft.com/en-us/library/ms687420.aspx

typedef	enum	_PROCESSINFOCLASS
{
#if 0
	ProcessBasicInformation                     = 0,
	ProcessDebugPort                            = 7,
	ProcessWow64Information                     = 26,
	ProcessImageFileName                        = 27,
	ProcessBreakOnTermination                   = 29,
	ProcessProtectionInformation                = 61,
	ProcessSubsystemInformation                 = 75,
#else
	ProcessBasicInformation                     = 0,
	ProcessQuotaLimits                          = 1,
	ProcessIoCounters                           = 2,
	ProcessVmCounters                           = 3,
	ProcessTimes                                = 4,
	ProcessBasePriority                         = 5,
	ProcessRaisePriority                        = 6,
	ProcessDebugPort                            = 7,
	ProcessExceptionPort                        = 8,
	ProcessAccessToken                          = 9,
	ProcessLdtInformation                       = 10,
	ProcessLdtSize                              = 11,
	ProcessDefaultHardErrorMode                 = 12,
	ProcessIoPortHandlers                       = 13,
	ProcessPooledUsageAndLimits                 = 14,
	ProcessWorkingSetWatch                      = 15,
	ProcessUserModeIOPL                         = 16,
	ProcessEnableAlignmentFaultFixup            = 17,
	ProcessPriorityClass                        = 18,
	ProcessWx86Information                      = 19,
	ProcessHandleCount                          = 20,
	ProcessAffinityMask                         = 21,
	ProcessPriorityBoost                        = 22,
	ProcessDeviceMap                            = 23,
	ProcessSessionInformation                   = 24,
	ProcessForegroundInformation                = 25,
	ProcessWow64Information                     = 26,
	ProcessImageFileName                        = 27,
	ProcessLUIDDeviceMapsEnabled                = 28,
	ProcessBreakOnTermination                   = 29,
	ProcessDebugObjectHandle                    = 30,
	ProcessDebugFlags                           = 31,
	ProcessHandleTracing                        = 32,
	ProcessIoPriority                           = 33,
	ProcessExecuteFlags                         = 34,
	ProcessTlsInformation                       = 35,
	ProcessCookie                               = 36,
	ProcessImageInformation                     = 37,
#if _WIN32_WINNT >= 0x0600
	ProcessCycleTime                            = 38,
	ProcessPagePriority                         = 39,
	ProcessInstrumentationCallback              = 40,
	ProcessThreadStackAllocation                = 41,
	ProcessWorkingSetWatchEx                    = 42,
	ProcessImageFileNameWin32                   = 43,
	ProcessImageFileMapping                     = 44,
	ProcessAffinityUpdateMode                   = 45,
	ProcessMemoryAllocationMode                 = 46,
	ProcessGroupInformation                     = 47,
	ProcessTokenVirtualizationEnabled           = 48,
	ProcessConsoleHostProcess                   = 49,
	ProcessWindowInformation                    = 50,
#if _WIN32_WINNT >= 0x0602
	ProcessHandleInformation                    = 51,
	ProcessMitigationPolicy                     = 52,
	ProcessDynamicFunctionTableInformation      = 53,
	ProcessHandleCheckingMode                   = 54,
	ProcessKeepAliveCount                       = 55,
	ProcessRevokeFileHandles                    = 56,
	ProcessWorkingSetControl                    = 57,
#if _WIN32_WINNT >= 0x0603
	ProcessHandleTable                          = 58,
	ProcessCheckStackExtentsMode                = 59,
	ProcessCommandLineInformation               = 60,
	ProcessProtectionInformation                = 61,
#if _WIN32_WINNT >= 0x0A00	// 'Threshold 1' (1507)
	ProcessMemoryExhaustion                     = 62,
	ProcessFaultInformation                     = 63,
	ProcessTelemetryIdInformation               = 64,
	ProcessCommitReleaseInformation             = 65,
	ProcessDefaultCpuSetsInformation            = 66,
	ProcessAllowedCpuSetsInformation            = 67,
	ProcessSubsystemProcess                     = 68,
	ProcessJobMemoryInformation                 = 69,
				// 'Threshold 2' (1511)
	ProcessInPrivate                            = 70,
	ProcessRaiseUMExceptionOnInvalidHandleClose = 71,
	ProcessIumChallengeResponse                 = 72,
	ProcessChildProcessInformation              = 73,
	ProcessHighGraphicsPriorityInformation      = 74,
				// 'Redstone 1' alias 'Anniversary Update' (1607)
				// 'Redstone 2' alias 'Creators Update' (1703)
	ProcessSubsystemInformation                 = 75,
	ProcessEnergyValues                         = 76,
	ProcessActivityThrottleState                = 77,
	ProcessActivityThrottlePolicy               = 78,
	ProcessWin32kSyscallFilterInformation       = 79,
	ProcessDisableSystemAllowedCpuSets          = 80,
	ProcessWakeInformation                      = 81,
	ProcessEnergyTrackingState                  = 82,
				// 'Redstone 3' alias 'Fall Creators Update' (1709)
	ProcessManageWritesToExecutableMemory       = 83,
	ProcessCaptureTrustletLiveDump              = 84,
	ProcessTelemetryCoverage                    = 85,
	ProcessEnclaveInformation                   = 86,
	ProcessEnableReadWriteVmLogging             = 87,
	ProcessUptimeInformation                    = 88,
	ProcessImageSection                         = 89,
				// 'Redstone 4' alias 'April 2018 Update' (1803)
	ProcessDebugAuthInformation                 = 90,
	ProcessSystemResourceManagement             = 91,
	ProcessSequenceNumber                       = 92,
				// 'Redstone 5' alias 'October 2018 Update' (1809)
	ProcessLoaderDetour                         = 93,
	ProcessSecurityDomainInformation            = 94,
	ProcessCombineSecurityDomainsInformation    = 95,
	ProcessEnableLogging                        = 96,
	ProcessLeapSecondInformation                = 97,
				// 'Vanadium 1' alias 'May 2019 Update' (1903)
	ProcessFiberShadowStackAllocation           = 98,
	ProcessFreeFiberShadowStackAllocation       = 99,
				// 'Vanadium 2' alias 'November 2019 Update' (1909)
				// 'Vibranium 1' alias 'May 2020 Update' (2004/20H1)
	ProcessAltSystemCallInformation             = 100,
	ProcessDynamicEHContinuationTargets         = 101,
				// 'Vibranium 2' alias 'October 2020 Update' (2010/20H2)
	ProcessDynamicEnforcedCetCompatibleRanges   = 102,
				// 'Vibranium 3' alias 'May 2021 Update' (2105/21H1)
				// 'Vibranium 4' alias 'November 2021 Update' (2111/21H2)
#endif
#endif
#endif
#endif
#endif
	MaxProcessInfoClass
} PROCESSINFOCLASS;

typedef	DWORD_PTR	KPRIORITY;

typedef	struct	_PROCESS_ACCESS_TOKEN
{
	HANDLE	Token;
	HANDLE	Thread;
} PROCESS_ACCESS_TOKEN, *LPPROCESS_ACCESS_TOKEN, *PPROCESS_ACCESS_TOKEN;

typedef	struct	_PROCESS_BASIC_INFORMATION
{
	LONG		ExitStatus;
	PEB		*PebBaseAddress;
	KAFFINITY	AffinityMask;
	KPRIORITY	BasePriority;
	DWORD_PTR	UniqueProcessId;
	DWORD_PTR	InheritedFromUniqueProcessId;
#endif
} PROCESS_BASIC_INFORMATION, *LPPROCESS_BASIC_INFORMATION, *PPROCESS_BASIC_INFORMATION;

typedef	struct	_PROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGE
{
	LPVOID	BaseAddress;
	SIZE_T	Size;
	DWORD	Flags;
} PROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGE, *LPPROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGE, *PPROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGE;

typedef	struct	_PROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGES_INFORMATION
{
	WORD	NumberOfRanges;
	WORD	Reserved;
	DWORD	Reserved2;
	PROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGE	*Ranges;
} PROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGES_INFORMATION, *LPPROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGES_INFORMATION, *PPROCESS_DYNAMIC_ENFORCED_ADDRESS_RANGES_INFORMATION;

// https://msdn.microsoft.com/en-us/library/mt809132.aspx

#pragma warning(disable: 4201 4214)
typedef	struct	_PROCESS_EXTENDED_BASIC_INFORMATION
{
	SIZE_T	Size;	// ignored on input, written with structure size on output
	PROCESS_BASIC_INFORMATION	BasicInfo;
	union
	{
		DWORD	Flags;
		struct
		{
			DWORD	IsProtectedProcess   : 1;
			DWORD	IsWow64Process       : 1;
			DWORD	IsProcessDeleting    : 1;
			DWORD	IsCrossSessionCreate : 1;
			DWORD	IsFrozen             : 1;
			DWORD	IsBackground         : 1;
			DWORD	IsStronglyNamed      : 1;
			DWORD	IsSecureProcess      : 1;
			DWORD	IsSubsystemProcess   : 1;
			DWORD	SpareBits            : 23;
		};
	};
} PROCESS_EXTENDED_BASIC_INFORMATION, *LPPROCESS_EXTENDED_BASIC_INFORMATION, *PPROCESS_EXTENDED_BASIC_INFORMATION;
#pragma warning(default: 4201 4214)

typedef	struct	_PROCESS_INSTRUMENTATION_CALLBACK_INFORMATION
{
	DWORD	Version;	// 0UL = x64, 1UL = x86
	DWORD	Reserved;	// 0UL
	LPVOID	Callback;
} PROCESS_INSTRUMENTATION_CALLBACK_INFORMATION, *LPPROCESS_INSTRUMENTATION_CALLBACK_INFORMATION, *PPROCESS_INSTRUMENTATION_CALLBACK_INFORMATION;

#define PROCESS_LEAP_SECOND_INFO_FLAG_ENABLE_SIXTY_SECOND	1UL

typedef	struct	_PROCESS_LEAP_SECOND_INFO
{
	DWORD	Flags;
	DWORD	Reserved;
} PROCESS_LEAP_SECOND_INFO, *PPROCESS_LEAP_SECOND_INFO, *PPROCESS_LEAP_SECOND_INFO;

typedef	enum	_PROCESS_MEMORY_EXHAUSTION_TYPE
{
	PMETypeFailFastOnCommitFailure,
	PMETypeMax
} PROCESS_MEMORY_EXHAUSTION_TYPE, *LPPROCESS_MEMORY_EXHAUSTION_TYPE, *PPROCESS_MEMORY_EXHAUSTION_TYPE;

typedef	struct	_PROCESS_MEMORY_EXHAUSTION_INFO
{
	WORD				Version;
	WORD				Reserved;
	PROCESS_MEMORY_EXHAUSTION_TYPE	Type;
	DWORD_PTR			Value;
} PROCESS_MEMORY_EXHAUSTION_INFO, *LPPROCESS_MEMORY_EXHAUSTION_INFO, *PPROCESS_MEMORY_EXHAUSTION_INFO;

// https://msdn.microsoft.com/en-us/library/mt823702.aspx

#define SECURITY_PROCESS_PROTECTION_LEVEL_WINTCB	0x00002000UL
#define SECURITY_PROCESS_PROTECTION_LEVEL_WINDOWS	0x00001000UL
#define SECURITY_PROCESS_PROTECTION_LEVEL_APP		0x00000800UL
#define SECURITY_PROCESS_PROTECTION_LEVEL_ANTIMALWARE	0x00000600UL
#define SECURITY_PROCESS_PROTECTION_LEVEL_AUTHENTICODE	0x00000400UL
#define SECURITY_PROCESS_PROTECTION_LEVEL_NONE		0x00000000UL

typedef	struct	PROCESS_PROTECTION_LEVEL_INFORMATION
{
	DWORD	ProtectionLevel;
} PROCESS_PROTECTION_LEVEL_INFORMATION, *LPPROCESS_PROTECTION_LEVEL_INFORMATION, *PPROCESS_PROTECTION_LEVEL_INFORMATION;

typedef	enum	_SUBSYSTEM_INFORMATION_TYPE
{
	SubsystemInformationTypeWin32,
	SubsystemInformationTypeWSL,
	MaxSubsystemInformationType
} SUBSYSTEM_INFORMATION_TYPE, *LPSUBSYSTEM_INFORMATION_TYPE, *PSUBSYSTEM_INFORMATION_TYPE;

// NOTE: like KERNEL32!GetProcessInformation

__declspec(deprecated("obsolete function, may be altered or unavailable in future versions"))
__declspec(dllimport)
LONG	NTAPI	NtQueryInformationProcess(HANDLE           ProcessHandle,
		                          PROCESSINFOCLASS ProcessInformationClass,
		                          LPVOID           ProcessInformation,
		                          DWORD            ProcessInformationLength,
		                          LPDWORD          ReturnLength);

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff561090.aspx

typedef	enum	_RESOURCEMANAGER_INFORMATION_CLASS
{
	ResourceManagerBasicInformation      = 0,
	ResourceManagerCompletionInformation = 1
} RESOURCEMANAGER_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff561084.aspx

typedef	struct	_RESOURCEMANAGER_BASIC_INFORMATION
{
	GUID	ResourceManagerId;
	DWORD	DescriptionLength;
	WCHAR	Description[ANYSIZE_ARRAY];
} RESOURCEMANAGER_BASIC_INFORMATION, *LPRESOURCEMANAGER_BASIC_INFORMATION, *PRESOURCEMANAGER_BASIC_INFORMATION;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff567054.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationResourceManager(HANDLE                            ResourceManagerHandle,
		                                  RESOURCEMANAGER_INFORMATION_CLASS ResourceManagerInformationClass,
		                                  LPVOID                            ResourceManagerInformation,
		                                  DWORD                             ResourceManagerInformationLength,
		                                  LPDWORD                           ReturnLength);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms684283.aspx

typedef	enum	_THREADINFOCLASS
{
#if 1
	ThreadBasicInformation               = 0,
	ThreadTimes                          = 1,
	ThreadPriority                       = 2,
	ThreadBasePriority                   = 3,
	ThreadAffinityMask                   = 4,
	ThreadImpersonationToken             = 5,
	ThreadDescriptorTableEntry           = 6,
	ThreadEnableAlignmentFaultFixup      = 7,
	ThreadEventPair_Reusable             = 8,
	ThreadQuerySetWin32StartAddress      = 9,
	ThreadZeroTlsCell                    = 10,
	ThreadPerformanceCount               = 11,
	ThreadAmILastThread                  = 12,
	ThreadIdealProcessor                 = 13,
	ThreadPriorityBoost                  = 14,
	ThreadSetTlsArrayAddress             = 15,
#endif
	ThreadIsIoPending                    = 16,
#if 1
	ThreadHideFromDebugger               = 17,
#if _WIN32_WINNT >= 0x0502
	ThreadBreakOnTermination             = 18,
	ThreadSwitchLegacyState              = 19,
	ThreadIsTerminated                   = 20,
#if _WIN32_WINNT >= 0x0600
	ThreadLastSystemCall                 = 21,
	ThreadIoPriority                     = 22,
	ThreadCycleTime                      = 23,
	ThreadPagePriority                   = 24,
	ThreadActualBasePriority             = 25,
	ThreadTebInformation                 = 26,
	ThreadCSwitchMon                     = 27,
#if _WIN32_WINNT >= 0x0601
	ThreadCSwitchPmu                     = 28,
	ThreadWow64Context                   = 29,
	ThreadGroupInformation               = 30,
	ThreadUmsInformation                 = 31,
	ThreadCounterProfiling               = 32,
	ThreadIdealProcessorEx               = 33,
#if _WIN32_WINNT >= 0x0602
	ThreadCpuAccountingInformation       = 34,
#if _WIN32_WINNT >= 0x0603
	ThreadSuspendCount                   = 35,
#if _WIN32_WINNT >= 0x0A00	// 'Threshold 1' (1507)
	ThreadHeterogeneousCpuPolicy         = 36,
	ThreadContainerId                    = 37,
	ThreadNameInformation                = 38,
	ThreadSelectedCpuSets                = 39,
	ThreadSystemThreadInformation        = 40,
				// 'Threshold 2' (1511)
	ThreadActualGroupAffinity            = 41,
	ThreadDynamicCodePolicyInfo          = 42,
	ThreadExplicitCaseSensitivity        = 43,
	ThreadWorkOnBehalfTicket             = 44,
	ThreadSubsystemInformation           = 45,
				// 'Redstone 2' alias 'Creators Update' (1703)
	ThreadDbgkWerReportActive            = 46,
	ThreadAttachContainer                = 47,
				// 'Redstone 3' alias 'Fall Creators Update' (1709)
	ThreadManageWritesToExecutableMemory = 48,
	ThreadPowerThrottlingState           = 49,
				// 'Redstone 5' alias 'October 2018 Update' (1809)
	ThreadWorkloadClass                  = 50,
#endif
#endif
#endif
#endif
#endif
#endif
#endif
	MaxThreadInfoClass
} THREADINFOCLASS;

typedef	struct	_THREAD_BASIC_INFORMATION
{
	NTSTATUS	ExitStatus;
	TEB		*TebBaseAddress;
	CLIENT_ID	ClientId;
	KAFFINITY	AffinityMask;
	KPRIORITY	BasePriority;
	KPRIORITY	Priority;
} THREAD_BASIC_INFORMATION, *LPTHREAD_BASIC_INFORMATION, *PTHREAD_BASIC_INFORMATION;

typedef	struct	_THREAD_AM_I_LAST_THREAD
{
	BOOLEAN	AmILastThread;
} THREAD_AM_I_LAST_THREAD, *LPTHREAD_AM_I_LAST_THREAD, *PTHREAD_AM_I_LAST_THREAD;

typedef	struct	_THREAD_TEB_INFORMATION
{
	LPVOID	OutputBuffer;
	DWORD	TebOffset;
	DWORD	BytesToRead;
} THREAD_TEB_INFORMATION, *LPTHREAD_TEB_INFORMATION, *PTHREAD_TEB_INFORMATION;

// NOTE: like KERNEL32!GetThreadInformation

__declspec(deprecated("obsolete function, may be altered or unavailable in future versions"))
__declspec(dllimport)
LONG	NTAPI	NtQueryInformationThread(HANDLE          ThreadHandle,
		                         THREADINFOCLASS ThreadInformationClass,
		                         LPVOID          ThreadInformation,
		                         DWORD           ThreadInformationLength,
		                         LPDWORD         ReturnLength);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff556838.aspx

typedef	enum	_TOKEN_INFORMATION_CLASS
{
	TokenUser                             = 1,
	TokenGroups                           = 2,
	TokenPrivileges                       = 3,
	TokenOwner                            = 4,
	TokenPrimaryGroup                     = 5,
	TokenDefaultDacl                      = 6,
	TokenSource                           = 7,
	TokenType                             = 8,
	TokenImpersonationLevel               = 9,
	TokenStatistics                       = 10,
	TokenRestrictedSids                   = 11,
	TokenSessionId                        = 12,
	TokenGroupsAndPrivileges              = 13,
	TokenSessionReference                 = 14,
	TokenSandBoxInert                     = 15,
	TokenAuditPolicy                      = 16,
	TokenOrigin                           = 17,
	TokenElevationType                    = 18,
	TokenLinkedToken                      = 19,
	TokenElevation                        = 20,
	TokenHasRestrictions                  = 21,
	TokenAccessInformation                = 22,
	TokenVirtualizationAllowed            = 23,
	TokenVirtualizationEnabled            = 24,
	TokenIntegrityLevel                   = 25,
	TokenUIAccess                         = 26,
	TokenMandatoryPolicy                  = 27,
	TokenLogonSid                         = 28,
	TokenIsAppContainer                   = 29,
	TokenCapabilities                     = 30,
	TokenAppContainerSid                  = 31,
	TokenAppContainerNumber               = 32,
	TokenUserClaimAttributes              = 33,
	TokenDeviceClaimAttributes            = 34,
	TokenRestrictedUserClaimAttributes    = 35,
	TokenRestrictedDeviceClaimAttributes  = 36,
	TokenDeviceGroups                     = 37,
	TokenRestrictedDeviceGroups           = 38,
	TokenSecurityAttributes               = 39,
	TokenIsRestricted                     = 40,
	TokenProcessTrustLevel                = 41,
	TokenPrivateNameSpace                 = 42,
	TokenSingletonAttributes              = 43,
	TokenBnoIsolation                     = 44,
	TokenChildProcessFlags                = 45,
	TokenIsLessPrivilegedAppContainer     = 46,
	TokenIsSandboxed                      = 47,
#if 0
	TokenOriginatingProcessTrustLevel     = 48,
#else
	TokenIsAppSilo                        = 48,
#endif
	MaxTokenInfoClass                     = 49
} TOKEN_INFORMATION_CLASS;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff556653.aspx
// https://msdn.microsoft.com/en-us/library/ff567055.aspx
// NOTE: like ADVAPI32!GetTokenInformation and KERNEL32!GetTokenInformation or KERNELBASE!GetTokenInformation

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationToken(HANDLE                  TokenHandle,
		                        TOKEN_INFORMATION_CLASS TokenInformationClass,
		                        LPVOID                  TokenInformation,
		                        DWORD                   TokenInformationLength,
		                        LPDWORD                 ReturnLength);

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff564800.aspx

typedef	enum	_TRANSACTION_INFORMATION_CLASS
{
	TransactionBasicInformation              = 0,
	TransactionPropertiesInformation,
	TransactionEnlistmentInformation,
	TransactionSuperiorEnlistmentInformation
} TRANSACTION_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff564849.aspx

typedef	enum	_TRANSACTION_STATE
{
	TransactionStateNormal          = 1,
	TransactionStateIndoubt         = 2,
	TransactionStateCommittedNotify = 3
} TRANSACTION_STATE;

// https://msdn.microsoft.com/en-us/library/ff564835.aspx

typedef	enum	_TRANSACTION_OUTCOME
{
	TransactionOutcomeUndetermined = 1,
	TransactionOutcomeCommitted    = 2,
	TransactionOutcomeAborted      = 3
} TRANSACTION_OUTCOME;

// https://msdn.microsoft.com/en-us/library/ff564781.aspx

typedef	struct	_TRANSACTION_BASIC_INFORMATION
{
	GUID	TransactionId;
	DWORD	State;
	DWORD	Outcome;
} TRANSACTION_BASIC_INFORMATION, *LPTRANSACTION_BASIC_INFORMATION, *PTRANSACTION_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff564844.aspx

typedef	struct	_TRANSACTION_PROPERTIES_INFORMATION
{
	DWORD		IsolationLevel;
	DWORD		IsolationFlags;
	LARGE_INTEGER	Timeout;
	DWORD		Outcome;
	DWORD		DescriptionLength;
	WCHAR		Description[ANYSIZE_ARRAY];
} TRANSACTION_PROPERTIES_INFORMATION, *LPTRANSACTION_PROPERTIES_INFORMATION, *PTRANSACTION_PROPERTIES_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff564793.aspx

typedef	struct	_TRANSACTION_ENLISTMENT_PAIR
{
	GUID	EnlistmentId;
	GUID	ResourceManagerId;
} TRANSACTION_ENLISTMENT_PAIR, *LPTRANSACTION_ENLISTMENT_PAIR, *PTRANSACTION_ENLISTMENT_PAIR;

// https://msdn.microsoft.com/en-us/library/ff564787.aspx

typedef	struct	_TRANSACTION_ENLISTMENTS_INFORMATION
{
	DWORD				NumberOfEnlistments;
	TRANSACTION_ENLISTMENT_PAIR	EnlistmentPair[ANYSIZE_ARRAY];
} TRANSACTION_ENLISTMENTS_INFORMATION, *LPTRANSACTION_ENLISTMENTS_INFORMATION, *PTRANSACTION_ENLISTMENTS_INFORMATION;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff567057.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationTransaction(HANDLE                        TransactionHandle,
		                              TRANSACTION_INFORMATION_CLASS TransactionInformationClass,
		                              LPVOID                        TransactionInformation,
		                              DWORD                         TransactionInformationLength,
		                              LPDWORD                       ReturnLength);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff564762.aspx

typedef	enum	_TRANSACTIONMANAGER_INFORMATION_CLASS
{
	TransactionManagerBasicInformation    = 0,
	TransactionManagerLogInformation      = 1,
	TransactionManagerLogPathInformation  = 2,
	TransactionManagerRecoveryInformation = 4
} TRANSACTIONMANAGER_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff564753.aspx

typedef	struct	_TRANSACTIONMANAGER_BASIC_INFORMATION
{
	GUID	TmIdentity;
	LARGE_INTEGER	VirtualClock;
} TRANSACTIONMANAGER_BASIC_INFORMATION, *LPTRANSACTIONMANAGER_BASIC_INFORMATION, *PTRANSACTIONMANAGER_BASIC_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff564767.aspx

typedef	struct	_TRANSACTIONMANAGER_LOG_INFORMATION
{
	GUID	LogIdentity;
} TRANSACTIONMANAGER_LOG_INFORMATION, *LPTRANSACTIONMANAGER_LOG_INFORMATION, *PTRANSACTIONMANAGER_LOG_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff564764.aspx

typedef	struct	_TRANSACTIONMANAGER_LOGPATH_INFORMATION
{
	DWORD	LogPathLength;
	WCHAR	LogPath[ANYSIZE_ARRAY];
} TRANSACTIONMANAGER_LOGPATH_INFORMATION, *LPTRANSACTIONMANAGER_LOGPATH_INFORMATION, *PTRANSACTIONMANAGER_LOGPATH_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff564776.aspx

typedef	struct	_TRANSACTIONMANAGER_RECOVERY_INFORMATION
{
	QWORD	LastRecoveredLsn;
} TRANSACTIONMANAGER_RECOVERY_INFORMATION, *LPTRANSACTIONMANAGER_RECOVERY_INFORMATION, *PTRANSACTIONMANAGER_RECOVERY_INFORMATION;
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/ff567058.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryInformationTransactionManager(HANDLE                               TransactionManagerHandle,
		                                     TRANSACTIONMANAGER_INFORMATION_CLASS TransactionManagerInformationClass,
		                                     LPVOID                               TransactionManagerInformation,
		                                     DWORD                                TransactionManagerInformationLength,
		                                     LPDWORD                              ReturnLength);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtQueryInstallUILanguage(LANGID *LanguageId);

__declspec(dllimport)
LONG	NTAPI	NtQueryIntervalProfile(KPROFILE_SOURCE ProfileSource,
		                       LPDWORD         Interval);

typedef	enum	_IOCOMPLETION_INFORMATION_CLASS
{
	IoCompletionInfo = 0
} IOCOMPLETION_INFORMATION_CLASS, *LPIOCOMPLETION_INFORMATION_CLASS, *PIOCOMPLETION_INFORMATION_CLASS;

typedef	struct	_IOCOMPLETION_INFO
{
	DWORD	QueueLength;
} IOCOMPLETION_INFO, *LPIOCOMPLETION_INFO, *PIOCOMPLETION_INFO;

__declspec(dllimport)
LONG	NTAPI	NtQueryIoCompletion(HANDLE                         IoCompletionPortHandle,
		                    IOCOMPLETION_INFORMATION_CLASS IoCompletionInformationClass,
		                    LPVOID                         IoCompletionInformation,
		                    DWORD                          IoCompletionInformationLength,
		                    LPDWORD                        ReturnLength);

// https://msdn.microsoft.com/en-us/library/ff556680.aspx
// https://msdn.microsoft.com/en-us/library/ff567060.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryKey(HANDLE                KeyHandle,
		           KEY_INFORMATION_CLASS KeyInformationClass,
		           LPVOID                KeyInformation,
		           DWORD                 KeyInformationLength,
		           LPDWORD               ReturnLength);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtQueryLicenseValue(UNICODE_STRING *Name,
		                    LPDWORD        Type,	// REG_SZ, REG_BINARY or REG_DWORD
		                    LPVOID         Buffer,
		                    DWORD          BufferLength,
		                    LPDWORD        ReturnLength);
#endif

// https://msdn.microsoft.com/en-us/library/ff554214.aspx

typedef	struct	_KEY_VALUE_ENTRY
{
	UNICODE_STRING	*ValueName;
	DWORD		DataLength;
	DWORD		DataOffset;
	DWORD		Type;
} KEY_VALUE_ENTRY, *LPKEY_VALUE_ENTRY, *PKEY_VALUE_ENTRY;

// https://msdn.microsoft.com/en-us/library/cc512137.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryMultipleValueKey(HANDLE          KeyHandle,
		                        KEY_VALUE_ENTRY *ValueEntries,
		                        DWORD           EntryCount,
		                        LPVOID          ValueBuffer,
		                        LPDWORD         BufferLength,
		                        LPDWORD         RequiredBufferLength);

typedef	enum	_MUTANT_INFORMATION_CLASS
{
	MutantBasicInformation = 0,
	MutantOwnerInformation = 1
} MUTANT_INFORMATION_CLASS, *LPMUTANT_INFORMATION_CLASS, *PMUTANT_INFORMATION_CLASS;

typedef	struct	_MUTANT_BASIC_INFORMATION
{
	LONG	CurrentCount;
	BOOLEAN	OwnedByCaller;
	BOOLEAN	AbandonedState;
} MUTANT_BASIC_INFORMATION, *LPMUTANT_BASIC_INFORMATION, *PMUTANT_BASIC_INFORMATION;

typedef	struct	_MUTANT_OWNER_INFORMATION
{
	CLIENT_ID ClientId;
} MUTANT_OWNER_INFORMATION, *LPMUTANT_OWNER_INFORMATION, *PMUTANT_OWNER_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQueryMutant(HANDLE                   MutantHandle,
		              MUTANT_INFORMATION_CLASS MutantInformationClass,
		              LPVOID                   MutantInformation,
		              DWORD                    MutantInformationLength,
		              LPDWORD                  ReturnLength);

// https://msdn.microsoft.com/en-us/library/ff550964.aspx

typedef	enum	_OBJECT_INFORMATION_CLASS
{
	ObjectBasicInformation         = 0,
	ObjectNameInformation          = 1,
	ObjectTypeInformation          = 2,
	ObjectTypesInformation         = 3,
	ObjectHandleFlagInformation    = 4,
	ObjectSessionInformation       = 5,
				// 'Redstone 2' alias 'Creators Update' (1703)
	ObjectSessionObjectInformation = 6,
	MaxObjectInfoClass
} OBJECT_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff551944.aspx

typedef	struct	_OBJECT_BASIC_INFORMATION
{
	DWORD		Attributes;
	ACCESS_MASK	GrantedAccess;
	DWORD		HandleCount;
	DWORD		PointerCount;
#if 0
	DWORD		Reserved[10];
#else
	DWORD		PagedPoolCharge;
	DWORD		NonPagedPoolCharge;
#if _WIN32_WINNT <= 0x0310
	DWORD		TotalNumberOfObjects;
	DWORD		TotalNumberOfHandles;
	DWORD		Reserved;
#else
	DWORD		Reserved[3];
#endif
	DWORD		NameInfoSize;
	DWORD		TypeInfoSize;
	DWORD		SecurityDescriptorSize;
	LARGE_INTEGER	CreationTime;
#endif
} OBJECT_BASIC_INFORMATION, *LPOBJECT_BASIC_INFORMATION, *POBJECT_BASIC_INFORMATION;

typedef	struct	_OBJECT_NAME_INFORMATION
{
	UNICODE_STRING	Name;
} OBJECT_NAME_INFORMATION, *LPOBJECT_NAME_INFORMATION, *POBJECT_NAME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff559707.aspx

typedef	enum	_POOL_TYPE
{
	NonPagedPool                         = 0,
	NonPagedPoolExecute                  = NonPagedPool,
	PagedPool,
	NonPagedPoolMustSucceed              = NonPagedPool + 2,
	DontUseThisType,
	NonPagedPoolCacheAligned             = NonPagedPool + 4,
	PagedPoolCacheAligned,
	NonPagedPoolCacheAlignedMustS        = NonPagedPool + 6,
	MaxPoolType,
	NonPagedPoolBase                     = 0,
	NonPagedPoolBaseMustSucceed          = NonPagedPoolBase + 2,
	NonPagedPoolBaseCacheAligned         = NonPagedPoolBase + 4,
	NonPagedPoolBaseCacheAlignedMustS    = NonPagedPoolBase + 6,
	NonPagedPoolSession                  = 32,
	PagedPoolSession                     = NonPagedPoolSession + 1,
	NonPagedPoolMustSucceedSession       = PagedPoolSession + 1,
	DontUseThisTypeSession               = NonPagedPoolMustSucceedSession + 1,
	NonPagedPoolCacheAlignedSession      = DontUseThisTypeSession + 1,
	PagedPoolCacheAlignedSession         = NonPagedPoolCacheAlignedSession + 1,
	NonPagedPoolCacheAlignedMustSSession = PagedPoolCacheAlignedSession + 1,
	NonPagedPoolNx                       = 512,
	NonPagedPoolNxCacheAligned           = NonPagedPoolNx + 4,
	NonPagedPoolSessionNx                = NonPagedPoolNx + 32
} POOL_TYPE;

// https://msdn.microsoft.com/en-us/library/ff551947.aspx

typedef	struct	_OBJECT_TYPE_INFORMATION
{
	UNICODE_STRING	TypeName;
#if 0
	DWORD		Reserved[22];
#else
	DWORD		TotalNumberOfObjects;
	DWORD		TotalNumberOfHandles;
	DWORD		TotalPagedPoolUsage;
	DWORD		TotalNonPagedPoolUsage;
	DWORD		TotalNamePoolUsage;
	DWORD		TotalHandleTableUsage;
	DWORD		HighWaterNumberOfObjects;
	DWORD		HighWaterNumberOfHandles;
	DWORD		HighWaterPagedPoolUsage;
	DWORD		HighWaterNonPagedPoolUsage;
	DWORD		HighWaterNamePoolUsage;
	DWORD		HighWaterHandleTableUsage;
	ACCESS_MASK	InvalidAttributes;
	GENERIC_MAPPING	GenericMapping;
	ACCESS_MASK	ValidAttributes;
	BOOLEAN		SecurityRequired;
	BOOLEAN		MaintainHandleCount;
	BYTE		TypeIndex;
	POOL_TYPE	PoolType;
	DWORD		DefaultPagedPoolCharge;
	DWORD		DefaultNonPagedPoolCharge;
#endif
} OBJECT_TYPE_INFORMATION, *LPOBJECT_TYPE_INFORMATION, *POBJECT_TYPE_INFORMATION;

typedef	struct	_OBJECT_TYPES_INFORMATION
{
	DWORD		NumberOfTypes;
	struct
	{
		OBJECT_TYPE_INFORMATION	TypeInformation;
		WCHAR			Name[ANYSIZE_ARRAY];
	} TypeInformation[ANYSIZE_ARRAY];
} OBJECT_TYPES_INFORMATION, *LPOBJECT_TYPES_INFORMATION, *POBJECT_TYPES_INFORMATION;

typedef	struct	_OBJECT_HANDLE_FLAG_INFORMATION
{
	BOOLEAN	Inherit;
	BOOLEAN	ProtectFromClose;
} OBJECT_HANDLE_FLAG_INFORMATION, *LPOBJECT_HANDLE_FLAG_INFORMATION, *POBJECT_HANDLE_FLAG_INFORMATION;

// https://msdn.microsoft.com/en-us/library/bb432383.aspx
// https://msdn.microsoft.com/en-us/library/ff567062.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryObject(HANDLE                   ObjectHandle,
		              OBJECT_INFORMATION_CLASS ObjectInformationClass,
		              LPVOID                   ObjectInformation,
		              DWORD                    ObjectInformationLength,
		              LPDWORD                  ReturnLength);

__declspec(dllimport)
LONG	NTAPI	NtQueryOpenSubKeys(OBJECT_ATTRIBUTES *TargetKey,
		                   LPDWORD           HandleCount);

// https://msdn.microsoft.com/en-us/library/bb432384.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryPerformanceCounter(LARGE_INTEGER *PerformanceCounter,
		                          LARGE_INTEGER *PerformanceFrequency);

// https://msdn.microsoft.com/en-us/library/ff567064.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryQuotaInformationFile(HANDLE          FileHandle,
		                            IO_STATUS_BLOCK *IoStatusBlock,
		                            LPVOID          Buffer,
		                            DWORD           BufferLength,
		                            BOOLEAN         ReturnSingleEntry,
		                            LPVOID          SidList,
		                            DWORD           SidListLength,
		                            SID             *StartSid,
		                            BOOLEAN         RestartScan);

typedef	enum	_SECTION_INFORMATION_CLASS
{
	SectionBasicInformation = 0,
	SectionImageInformation = 1
} SECTION_INFORMATION_CLASS;

typedef	struct	_SECTION_BASIC_INFORMATION
{
	DWORD		Reserved;
	DWORD		Attributes;
	LARGE_INTEGER	MaximumSize;
} SECTION_BASIC_INFORMATION, *LPSECTION_BASIC_INFORMATION, *PSECTION_BASIC_INFORMATION;

#pragma warning(disable: 4201 4214)
typedef	struct	_SECTION_IMAGE_INFORMATION
{
	LPVOID		TransferAddress;
	DWORD		ZeroBits;
	SIZE_T		MaximumStackSize;
	SIZE_T		CommittedStackSize;
	DWORD		SubSystemType;
	union
	{
		struct
		{
			WORD	SubSystemMinorVersion;
			WORD	SubSystemMajorVersion;
		};
		DWORD	SubSystemVersion;
	};
	union
	{
		struct
		{
			WORD	MajorOperatingSystemVersion;
			WORD	MinorOperatingSystemVersion;
		};
		DWORD	OperatingSystemVersion;
		DWORD	GpValue;
	};
	WORD	ImageCharacteristics;
	WORD	DllCharacteristics;
	WORD	Machine;
	BOOLEAN	ImageContainsCode;
	union
	{
		BYTE	ImageFlags;
		struct
		{
			BYTE	ComPlusNativeReady        : 1;
			BYTE	ComPlusILOnly             : 1;
			BYTE	ImageDynamicallyRelocated : 1;
			BYTE	ImageMappedFlat           : 1;
			BYTE	BaseBelow4gb              : 1;
			BYTE	ComPlusPrefer32bit        : 1;
			BYTE	Reserved                  : 2;
		};
	};
	DWORD	LoaderFlags;
	DWORD	ImageFileSize;
	DWORD	CheckSum;
} SECTION_IMAGE_INFORMATION, *LPSECTION_IMAGE_INFORMATION, *PSECTION_IMAGE_INFORMATION;
#pragma warning(default: 4201 4214)

__declspec(dllimport)
LONG	NTAPI	NtQuerySection(HANDLE                    SectionHandle,
		               SECTION_INFORMATION_CLASS SectionInformationClass,
		               LPVOID                    SectionInformation,
		               DWORD                     SectionInformationLength,
		               LPDWORD                   ReturnLength);

#if _WIN32_WINNT >= 0x0601
typedef	struct	_TOKEN_SECURITY_ATTRIBUTE_FQBN_VALUE
{
	QWORD		Version;
	UNICODE_STRING	Name;
} TOKEN_SECURITY_ATTRIBUTE_FQBN_VALUE, *LPTOKEN_SECURITY_ATTRIBUTE_FQBN_VALUE, *PTOKEN_SECURITY_ATTRIBUTE_FQBN_VALUE;

typedef	struct	_TOKEN_SECURITY_ATTRIBUTE_OCTET_STRING_VALUE
{
	LPVOID	pValue;
	DWORD	ValueLength;
} TOKEN_SECURITY_ATTRIBUTE_OCTET_STRING_VALUE, *PTOKEN_SECURITY_ATTRIBUTE_OCTET_STRING_VALUE, *PTOKEN_SECURITY_ATTRIBUTE_OCTET_STRING_VALUE;

#define TOKEN_SECURITY_ATTRIBUTE_TYPE_INVALID		0UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_INT64		1UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_UINT64		2UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_STRING		3UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_FQBN		4UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_SID		5UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_BOOLEAN		6UL
#define TOKEN_SECURITY_ATTRIBUTE_TYPE_OCTET_STRING	16UL

#define TOKEN_SECURITY_ATTRIBUTE_NON_INHERITABLE	0x00000001UL
#define TOKEN_SECURITY_ATTRIBUTE_VALUE_CASE_SENSITIVE	0x00000002UL
#define TOKEN_SECURITY_ATTRIBUTE_USE_FOR_DENY_ONLY	0x00000004UL
#define TOKEN_SECURITY_ATTRIBUTE_DISABLED_BY_DEFAULT	0x00000008UL
#define TOKEN_SECURITY_ATTRIBUTE_DISABLED		0x00000010UL
#define TOKEN_SECURITY_ATTRIBUTE_MANDATORY		0x00000020UL

typedef	struct	_TOKEN_SECURITY_ATTRIBUTE_V1
{
	UNICODE_STRING	Name;
	DWORD		ValueType;
	WORD		Reserved;
	DWORD		Flags;
	DWORD		ValueCount;
	union
	{
		LONGLONG	*pInt64;
		QWORD		*pUint64;
		UNICODE_STRING	*pString;
		TOKEN_SECURITY_ATTRIBUTE_FQBN_VALUE		*pFqbn;
		TOKEN_SECURITY_ATTRIBUTE_OCTET_STRING_VALUE	*pOctetString;
	} Values;
} TOKEN_SECURITY_ATTRIBUTE_V1, *LPTOKEN_SECURITY_ATTRIBUTE_V1, *PTOKEN_SECURITY_ATTRIBUTE_V1;

#define TOKEN_SECURITY_ATTRIBUTES_INFORMATION_VERSION		1
#define TOKEN_SECURITY_ATTRIBUTES_INFORMATION_VERSION_V1	1

typedef	struct	_TOKEN_SECURITY_ATTRIBUTES_INFORMATION
{
	WORD	Version;
	WORD	Reserved;
	DWORD	AttributeCount;
	union
	{
		TOKEN_SECURITY_ATTRIBUTE_V1	*pAttributeV1;
	} Attribute;
} TOKEN_SECURITY_ATTRIBUTES_INFORMATION, *LPTOKEN_SECURITY_ATTRIBUTES_INFORMATION, *PTOKEN_SECURITY_ATTRIBUTES_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQuerySecurityAttributesToken(HANDLE         TokenHandle,
		                               UNICODE_STRING *Attributes,
		                               DWORD          NumberOfAttributes,
		                               LPVOID         Buffer,
		                               DWORD          BufferLength,
		                               LPDWORD        ReturnLength);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff567066.aspx

__declspec(dllimport)
LONG	NTAPI	NtQuerySecurityObject(HANDLE               ObjectHandle,
		                      SECURITY_INFORMATION SecurityInformation,
		                      SECURITY_DESCRIPTOR  *SecurityDescriptor,
		                      DWORD                SecurityDescriptorLength,
		                      LPDWORD              ReturnLength);

typedef	enum	_SEMAPHORE_INFORMATION_CLASS
{
	SemaphoreBasicInformation = 0
} SEMAPHORE_INFORMATION_CLASS, *LPSEMAPHORE_INFORMATION_CLASS, *PSEMAPHORE_INFORMATION_CLASS;

typedef	struct	_SEMAPHORE_BASIC_INFORMATION
{
	DWORD	CurrentCount;
	DWORD	MaximumCount;
} SEMAPHORE_BASIC_INFORMATION, *LPSEMAPHORE_BASIC_INFORMATION, *PSEMAPHORE_BASIC_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQuerySemaphore(HANDLE                      SemaphoreHandle,
		                 SEMAPHORE_INFORMATION_CLASS SemaphoreInformationClass,
		                 LPVOID                      SemaphoreInformation,
		                 DWORD                       SemaphoreInformationLength,
		                 LPDWORD                     ReturnLength);

// https://msdn.microsoft.com/en-us/library/bb470241.aspx
// https://msdn.microsoft.com/en-us/library/ff567068.aspx

__declspec(dllimport)
LONG	NTAPI	NtQuerySymbolicLinkObject(HANDLE         LinkHandle,
		                          UNICODE_STRING *LinkTarget,
		                          LPDWORD        ReturnLength);

__declspec(dllimport)
LONG	NTAPI	NtQuerySystemEnvironmentValue(UNICODE_STRING *VariableName,
		                              LPWSTR         ValueBuffer,
		                              DWORD          ValueBufferLength,
		                              LPDWORD        ReturnLength);

typedef	enum	_SYSTEM_INFORMATION_CLASS
{
	SystemBasicInformation                                = 0,
	SystemProcessorInformation                            = 1,
	SystemPerformanceInformation                          = 2,
	SystemTimeOfDayInformation                            = 3,
	SystemPathInformation                                 = 4,
#if 0
	SystemProcessesAndThreadsInformation                  = 5,
#else
	SystemProcessInformation                              = 5,
#endif
	SystemCallCountInformation                            = 6,
	SystemDeviceInformation                               = 7,
	SystemProcessorPerformanceInformation                 = 8,
	SystemFlagsInformation                                = 9,
	SystemCallTimeInformation                             = 10,
	SystemModuleInformation                               = 11,
	SystemLocksInformation                                = 12,
	SystemStackTraceInformation                           = 13,
	SystemPagedPoolInformation                            = 14,
	SystemNonPagedPoolInformation                         = 15,
	SystemHandleInformation                               = 16,
	SystemObjectInformation                               = 17,
	SystemPageFileInformation                             = 18,
	SystemVdmInstemulInformation                          = 19,
	SystemVdmBopInformation                               = 20,
	SystemFileCacheInformation                            = 21,
	SystemPoolTagInformation                              = 22,
	SystemInterruptInformation                            = 23,
	SystemDpcBehaviourInformation                         = 24,
	SystemFullMemoryInformation                           = 25,
	SystemLoadGdiDriverInformation                        = 26,
	SystemUnloadGdiDriverInformation                      = 27,
	SystemTimeAdjustmentInformation                       = 28,
	SystemSummaryMemoryInformation                        = 29,
#if _WIN32_WINNT < 0x0500
	SystemNextEventIdInformation                          = 30,
	SystemEventIdsInformation                             = 31,
#else
	SystemMirrorMemoryInformation                         = 30,
	SystemPerformanceTraceInformation                     = 31,
#endif
#if _WIN32_WINNT < 0x0501
	SystemCrashDumpInformation                            = 32,
#else
	SystemObsolete0                                       = 32,
#endif
	SystemExceptionInformation                            = 33,
	SystemCrashDumpStateInformation                       = 34,
	SystemKernelDebuggerInformation                       = 35,
	SystemContextSwitchInformation                        = 36,
	SystemRegistryQuotaInformation                        = 37,
	SystemExtendedServiceTableInformation                 = 38,
	SystemPrioritySeparation                              = 39,
	SystemPrioritySeperation                              = 39,
#if _WIN32_WINNT < 0x0501
	SystemPlugPlayBusInformation                          = 40,
	SystemDockInformation                                 = 41,
#else
	SystemVerifierAddDriverInformation                    = 40,
	SystemVerifierRemoveDriverInformation                 = 41,
#endif
#if _WIN32_WINNT < 0x0501
	SystemPowerInformation                                = 42,
#else
	SystemProcessorIdleInformation                        = 42,
#endif
#if _WIN32_WINNT >= 0x0400
#if _WIN32_WINNT < 0x0500
	SystemProcessorSpeedInformation                       = 43,
#else
	SystemLegacyDriverInformation                         = 43,
#endif
	SystemCurrentTimeZoneInformation                      = 44,
	SystemLookasideInformation                            = 45,
#if _WIN32_WINNT >= 0x0500
	SystemTimeSlipNotification                            = 46,
	SystemSessionCreate                                   = 47,
	SystemSessionDetach                                   = 48,
	SystemSessionInformation                              = 49,
	SystemRangeStartInformation                           = 50,
	SystemVerifierInformation                             = 51,
	SystemVerifierThunkExtended                           = 52,
	SystemSessionProcessInformation                       = 53,
#if _WIN32_WINNT < 0x0501
	SystemObjectSecurityMode                              = 54,
#else
	SystemLoadGdiDriverInSystemSpace                      = 54,
#endif
#if _WIN32_WINNT >= 0x0501
	SystemNumaProcessorMap                                = 55,
	SystemPrefetcherInformation                           = 56,
	SystemExtendedProcessInformation                      = 57,
	SystemRecommendedSharedDataAlignment                  = 58,
	SystemComPlusPackage                                  = 59,
	SystemNumaAvailableMemory                             = 60,
	SystemProcessorPowerInformation                       = 61,
	SystemEmulationBasicInformation                       = 62,
	SystemEmulationProcessorInformation                   = 63,
	SystemExtendedHandleInformation                       = 64,
	SystemLostDelayedWriteInformation                     = 65,
	SystemBigPoolInformation                              = 66,
	SystemSessionPoolTagInformation                       = 67,
	SystemSessionMappedViewInformation                    = 68,
	SystemHotpatchInformation                             = 69,
	SystemObjectSecurityMode                              = 70,
	SystemWatchdogTimerHandler                            = 71,
	SystemWatchdogTimerInformation                        = 72,
	SystemLogicalProcessorInformation                     = 73,
#if _WIN32_WINNT >= 0x0502
	SystemWow64SharedInformationObsolete                  = 74,
	SystemRegisterFirmwareTableInformationHandler         = 75,
	SystemFirmwareTableInformation                        = 76,
	SystemModuleInformationEx                             = 77,
	SystemVerifierTriageInformation                       = 78,
	SystemSuperfetchInformation                           = 79,
	SystemMemoryListInformation                           = 80,
	SystemFileCacheInformationEx                          = 81,
#if _WIN32_WINNT >= 0x0600
	SystemThreadPriorityClientIdInformation               = 82,
	SystemProcessorIdleCycleTimeInformation               = 83,
	SystemVerifierCancellationInformation                 = 84,
	SystemProcessorPowerInformationEx                     = 85,
	SystemRefTraceInformation                             = 86,
	SystemSpecialPoolInformation                          = 87,
	SystemProcessIdInformation                            = 88,
	SystemErrorPortInformation                            = 89,
	SystemBootEnvironmentInformation                      = 90,
	SystemHypervisorInformation                           = 91,
	SystemVerifierInformationEx                           = 92,
	SystemTimeZoneInformation                             = 93,
	SystemImageFileExecutionOptionsInformation            = 94,
	SystemCoverageInformation                             = 95,
	SystemPrefetchPatchInformation                        = 96,
	SystemVerifierFaultsInformation                       = 97,
	SystemSystemPartitionInformation                      = 98,
	SystemSystemDiskInformation                           = 99,
	SystemProcessorPerformanceDistribution                = 100,
	SystemNumaProximityNodeInformation                    = 101,
	SystemDynamicTimeZoneInformation                      = 102,
	SystemCodeIntegrityInformation                        = 103,
	SystemProcessorMicrocodeUpdateInformation             = 104,
	SystemProcessorBrandString                            = 105,
	SystemVirtualAddressInformation                       = 106,
#if _WIN32_WINNT >= 0x0601
	SystemLogicalProcessorAndGroupInformation             = 107,
	SystemProcessorCycleTimeInformation                   = 108,
	SystemStoreInformation                                = 109,
	SystemRegistryAppendString                            = 110,
	SystemAitSamplingValue                                = 111,
	SystemVhdBootInformation                              = 112,
	SystemCpuQuotaInformation                             = 113,
	SystemNativeBasicInformation                          = 114,
	SystemErrorPortTimeouts                               = 115,
	SystemLowPriorityIoInformation                        = 116,
	SystemBootEntropyInformation                          = 117,
	SystemVerifierCountersInformation                     = 118,
	SystemPagedPoolInformationEx                          = 119,
	SystemSystemPtesInformationEx                         = 120,
	SystemNodeDistanceInformation                         = 121,
	SystemAcpiAuditInformation                            = 122,
	SystemBasicPerformanceInformation                     = 123,
	SystemQueryPerformanceCounterInformation              = 124,
#if _WIN32_WINNT >= 0x0602
	SystemSessionBigPoolInformation                       = 125,
	SystemBootGraphicsInformation                         = 126,
	SystemScrubPhysicalMemoryInformation                  = 127,
	SystemBadPageInformation                              = 128,
	SystemProcessorProfileControlArea                     = 129,
	SystemCombinePhysicalMemoryInformation                = 130,
	SystemEntropyInterruptTimingInformation               = 131,
	SystemConsoleInformation                              = 132,
	SystemPlatformBinaryInformation                       = 133,
	SystemThrottleNotificationInformation                 = 134,
	SystemPolicyInformation                               = 134,
	SystemHypervisorProcessorCountInformation             = 135,
	SystemDeviceDataInformation                           = 136,
	SystemDeviceDataEnumerationInformation                = 137,
	SystemMemoryTopologyInformation                       = 138,
	SystemMemoryChannelInformation                        = 139,
	SystemBootLogoInformation                             = 140,
	SystemProcessorPerformanceInformationEx               = 141,
	SystemSpare0                                          = 142,
	SystemSecureBootPolicyInformation                     = 143,
	SystemPageFileInformationEx                           = 144,
	SystemSecureBootInformation                           = 145,
	SystemEntropyInterruptTimingRawInformation            = 146,
	SystemPortableWorkspaceEfiLauncherInformation         = 147,
	SystemFullProcessInformation                          = 148,
#if _WIN32_WINNT >= 0x0603
	SystemKernelDebuggerInformationEx                     = 149,
	SystemBootMetadataInformation                         = 150,
	SystemSoftRebootInformation                           = 151,
	SystemElamCertificateInformation                      = 152,
	SystemOfflineDumpConfigInformation                    = 153,
	SystemProcessorFeaturesInformation                    = 154,
	SystemRegistryReconciliationInformation               = 155,
	SystemEdidInformation                                 = 156,
#if _WIN32_WINNT >= 0x0A00	// 'Threshold 1' (1507)
	SystemManufacturingInformation                        = 157,
	SystemEnergyEstimationConfigInformation               = 158,
	SystemHypervisorDetailInformation                     = 159,
	SystemProcessorCycleStatsInformation                  = 160,
	SystemVmGenerationCountInformation                    = 161,
	SystemTrustedPlatformModuleInformation                = 162,
	SystemKernelDebuggerFlags                             = 163,
	SystemCodeIntegrityPolicyInformation                  = 164,
	SystemIsolatedUserModeInformation                     = 165,
	SystemHardwareSecurityTestInterfaceResultsInformation = 166,
	SystemSingleModuleInformation                         = 167,
	SystemAllowedCpuSetsInformation                       = 168,
	SystemDmaProtectionInformation                        = 169,
	SystemInterruptCpuSetsInformation                     = 170,
	SystemSecureBootPolicyFullInformation                 = 171,
	SystemCodeIntegrityPolicyFullInformation              = 172,
	SystemAffinitizedInterruptProcessorInformation        = 173,
	SystemRootSiloInformation                             = 174,
				// 'Threshold 2' (1511)
	SystemCpuSetInformation                               = 175,
	SystemCpuSetTagInformation                            = 176,
	SystemWin32WerStartCallout                            = 177,
	SystemSecureKernelProfileInformation                  = 178,
				// 'Redstone 1' alias 'Anniversary Update' (1607)
	SystemCodeIntegrityPlatformManifestInformation        = 179,
	SystemInterruptSteeringInformation                    = 180,
	SystemSupportedProcessorArchitectures                 = 181,
	SystemMemoryUsageInformation                          = 182,
	SystemCodeIntegrityCertificateInformation             = 183,
				// 'Redstone 2' alias 'Creators Update' (1703)
	SystemPhysicalMemoryInformation                       = 184,
	SystemControlFlowTransition                           = 185,
	SystemKernelDebuggingAllowed                          = 186,
	SystemActivityModerationExeState                      = 187,
	SystemActivityModerationUserSettings                  = 188,
	SystemCodeIntegrityPoliciesFullInformation            = 189,
	SystemCodeIntegrityUnlockInformation                  = 190,
	SystemIntegrityQuotaInformation                       = 191,
	SystemFlushInformation                                = 192,
				// 'Redstone 3' alias 'Fall Creators Update' (1709)
	SystemProcessorIdleMaskInformation                    = 193,
	SystemSecureDumpEncryptionInformation                 = 194,
	SystemWriteConstraintInformation                      = 195,
				// 'Redstone 4' alias 'April 2018 Update' (1803)
	SystemKernelVaShadowInformation                       = 196,
	SystemHypervisorSharedPageInformation                 = 197,
	SystemFirmwareBootPerformanceInformation              = 198,
	SystemCodeIntegrityVerificationInformation            = 199,
	SystemFirmwarePartitionInformation                    = 200,
	SystemSpeculationControlInformation                   = 201,
	SystemDmaGuardPolicyInformation                       = 202,
	SystemEnclaveLaunchControlInformation                 = 203,
	SystemWorkloadAllowedCpuSetsInformation               = 204,
	SystemCodeIntegrityUnlockModeInformation              = 205,
	SystemLeapSecondInformation                           = 206,
	SystemFlags2Information                               = 207,
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
	MaxSystemInfoClass
} SYSTEM_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ms724509.aspx
// https://msdn.microsoft.com/en-us/library/ms725506.aspx

typedef	struct	_SYSTEM_BASIC_INFORMATION
{
#if 0
	BYTE	Reserved1[24];
	LPVOID	Reserved2[4];
#else
	DWORD		Unknown;
	DWORD		TimerResolution;
	DWORD		PhysicalPageSize;
	DWORD		NumberOfPhysicalPages;
	DWORD		LowestPhysicalPageNumber;
	DWORD		HighestPhysicalPageNumber;
	DWORD		AllocationGranularity;
	DWORD_PTR	MinimumUserModeAddress;
	DWORD_PTR	MaxiumuUserModeAddress;
	KAFFINITY	ActiveProcessorsAffinityMask;
#endif
	BYTE	NumberOfProcessors;
} SYSTEM_BASIC_INFORMATION, *LPSYSTEM_BASIC_INFORMATION, *PSYSTEM_BASIC_INFORMATION;

typedef	enum	_FIRMWARE_TYPE
{
	FirmwareTypeUnknown,
	FirmwareTypeBios,
	FirmwareTypeUefi,
	FirmwareTypeMax
} FIRMWARE_TYPE, *LPFIRMWARE_TYPE, *PFIRMWARE_TYPE;

typedef	struct	_SYSTEM_BOOT_ENVIRONMENT_INFORMATION
{
	GUID		BootIdentifier;
	FIRMWARE_TYPE	FirmwareType;
#if _WIN32_WINNT >= 0x0602
	QWORD		BootFlags;
#endif
} SYSTEM_BOOT_ENVIRONMENT_INFORMATION, *PSYSTEM_BOOT_ENVIRONMENT_INFORMATION, *PSYSTEM_BOOT_ENVIRONMENT_INFORMATION;

typedef	struct	_SYSTEM_CALL_COUNT_INFORMATION
{
	DWORD	Length;
	DWORD	NumberOfTables;
	DWORD	CallCountTables[ANYSIZE_ARRAY];
} SYSTEM_CALL_COUNT_INFORMATION, *LPSYSTEM_CALL_COUNT_INFORMATION, *PSYSTEM_CALL_COUNT_INFORMATION;

#define CODEINTEGRITY_OPTION_ENABLED				0x00000001UL
#define CODEINTEGRITY_OPTION_TESTSIGN				0x00000002UL
#define CODEINTEGRITY_OPTION_UMCI_ENABLED			0x00000004UL
#define CODEINTEGRITY_OPTION_UMCI_AUDITMODE_ENABLED		0x00000008UL
#define CODEINTEGRITY_OPTION_UMCI_EXCLUSIONPATHS_ENABLED	0x00000010UL
#define CODEINTEGRITY_OPTION_TEST_BUILD				0x00000020UL
#define CODEINTEGRITY_OPTION_PREPRODUCTION_BUILD		0x00000040UL
#define CODEINTEGRITY_OPTION_DEBUGMODE_ENABLED			0x00000080UL
#define CODEINTEGRITY_OPTION_FLIGHT_BUILD			0x00000100UL
#define CODEINTEGRITY_OPTION_FLIGHTING_ENABLED			0x00000200UL
#define CODEINTEGRITY_OPTION_HVCI_KMCI_ENABLED			0x00000400UL
#define CODEINTEGRITY_OPTION_HVCI_KMCI_AUDITMODE_ENABLED	0x00000800UL
#define CODEINTEGRITY_OPTION_HVCI_KMCI_STRICTMODE_ENABLED	0x00001000UL
#define CODEINTEGRITY_OPTION_HVCI_IUM_ENABLED			0x00002000UL

typedef	struct	_SYSTEM_CODEINTEGRITY_INFORMATION
{
	DWORD	Length;
	DWORD	CodeIntegrityOptions;
} SYSTEM_CODEINTEGRITY_INFORMATION, *LPSYSTEM_CODEINTEGRITY_INFORMATION, *PSYSTEM_CODEINTEGRITY_INFORMATION;

typedef	struct	_SYSTEM_CACHE_INFORMATION
{
	DWORD	SystemCacheWsSize;
	DWORD	SystemCacheWsPeakSize;
	DWORD	SystemCacheWsFaults;
	DWORD	SystemCacheWsMinimum;
	DWORD	SystemCacheWsMaximum;
	DWORD	TransitionSharedPages;
	DWORD	TransitionSharedPagesPeak;
	DWORD	Reserved[2];
} SYSTEM_CACHE_INFORMATION, *LPSYSTEM_CACHE_INFORMATION, *PSYSTEM_CACHE_INFORMATION;

typedef	struct	_SYSTEM_CONFIGURATION_INFORMATION
{
	DWORD	DiskCount;
	DWORD	FloppyCount;
	DWORD	CdRomCount;
	DWORD	TapeCount;
	DWORD	SerialCount;
	DWORD	ParallelCount;
} SYSTEM_CONFIGURATION_INFORMATION, *LPSYSTEM_CONFIGURATION_INFORMATION, *PSYSTEM_CONFIGURATION_INFORMATION;

typedef	struct	_SYSTEM_CONTEXT_SWITCH_INFORMATION
{
	DWORD	ContextSwitches;
	DWORD	ContextSwitchCounters[11];
} SYSTEM_CONTEXT_SWITCH_INFORMATION, *LPSYSTEM_CONTEXT_SWITCH_INFORMATION, *PSYSTEM_CONTEXT_SWITCH_INFORMATION;

#if _WIN32_WINNT <= 0x0500
typedef	struct	_SYSTEM_CRASH_DUMP_INFORMATION
{
	HANDLE	CrashDumpSection;
#if _WIN32_WINNT == 0x0500
	DWORD	Unknown;
#endif
} SYSTEM_CRASH_DUMP_INFORMATION, *LPSYSTEM_CRASH_DUMP_INFORMATION, *PSYSTEM_CRASH_DUMP_INFORMATION;
#endif

typedef	enum	_SYSTEM_CRASH_DUMP_CONFIGURATION_CLASS
{
	SystemCrashDumpDisable                = 0,
	SystemCrashDumpReconfigure            = 1,
	SystemCrashDumpInitializationComplete = 2
} SYSTEM_CRASH_DUMP_CONFIGURATION_CLASS, *LPSYSTEM_CRASH_DUMP_CONFIGURATION_CLASS, *PSYSTEM_CRASH_DUMP_CONFIGURATION_CLASS;

typedef	struct	_SYSTEM_CRASH_DUMP_STATE_INFORMATION
{
	SYSTEM_CRASH_DUMP_CONFIGURATION_CLASS	CrashDumpConfigurationClass;
} SYSTEM_CRASH_DUMP_STATE_INFORMATION, *LPSYSTEM_CRASH_DUMP_STATE_INFORMATION, *PSYSTEM_CRASH_DUMP_STATE_INFORMATION;

#if _WIN32_WINNT <= 0x0500
typedef	struct	_SYSTEM_CRASH_STATE_INFORMATION
{
	DWORD	ValidCrashDump;
#if _WIN32_WINNT == 0x0500
	DWORD	Unknown;
#endif
} SYSTEM_CRASH_STATE_INFORMATION, *LPSYSTEM_CRASH_STATE_INFORMATION, *PSYSTEM_CRASH_STATE_INFORMATION;
#endif

typedef	struct	_SYSTEM_CREATE_SESSION
{
	DWORD	Session;
} SYSTEM_CREATE_SESSION, *LPSYSTEM_CREATE_SESSION, *PSYSTEM_CREATE_SESSION;

typedef	struct	_SYSTEM_DELETE_SESSION
{
	DWORD	Session;
} SYSTEM_DELETE_SESSION, *LPSYSTEM_DELETE_SESSION, *PSYSTEM_DELETE_SESSION;

#if _WIN32_WINNT <= 0x0400
typedef	struct	_SYSTEM_DOCK_INFORMATION
{
#if _WIN32_WINNT < 0x0400
	INT	DockState;
	DWORD	DockIdLength;
	DWORD	SerialNumberOffset;
	DWORD	SerialNumberLength;
	WORD	DockId[ANYSIZE_ARRAY];
#else
	INT	DeviceBusType;
	DWORD	DeviceBusNumber;
	DWORD	SlotNumber;
#endif
} SYSTEM_DOCK_INFORMATION, *LPSYSTEM_DOCK_INFORMATION, *PSYSTEM_DOCK_INFORMATION;
#endif

typedef	struct	_SYSTEM_DPC_INFORMATION
{
	DWORD	Reserved;
	DWORD	MaximumDpcQueueDepth;
	DWORD	MinimumDpcRate;
	DWORD	AdjustDpcThreshold;
	DWORD	IdealDpcRate;
} SYSTEM_DPC_INFORMATION, *LPSYSTEM_DPC_INFORMATION, *PSYSTEM_DPC_INFORMATION;

typedef	struct	_SYSTEM_EXCEPTION_INFORMATION
{
#if 0
	BYTE	Reserved1[16];
#else
	DWORD	AlignmentFixupCount;
	DWORD	ExceptionDispatchCount;
	DWORD	FloatingEmulationCount;
	DWORD	ByteWordEmulationCount;
#endif
} SYSTEM_EXCEPTION_INFORMATION, *LPSYSTEM_EXCEPTION_INFORMATION, *PSYSTEM_EXCEPTION_INFORMATION;

// https://msdn.microsoft.com/en-us/library/cc779664.aspx
// https://msdn.microsoft.com/en-us/library/ff549596.aspx
// https://msdn.microsoft.com/en-us/library/ff549646.aspx
//
// https://msdn.microsoft.com/en-us/library/ff558817.aspx
// https://msdn.microsoft.com/en-us/library/ff556886.aspx
// https://msdn.microsoft.com/en-us/library/ff541503.aspx
// https://msdn.microsoft.com/en-us/library/ff558818.aspx
// https://msdn.microsoft.com/en-us/library/ff542921.aspx
// https://msdn.microsoft.com/en-us/library/ff542901.aspx
// https://msdn.microsoft.com/en-us/library/ff542906.aspx
// https://msdn.microsoft.com/en-us/library/ff542926.aspx
// https://msdn.microsoft.com/en-us/library/ff542875.aspx
// 
// https://msdn.microsoft.com/en-us/library/ff542945.aspx
// https://msdn.microsoft.com/en-us/library/ff542911.aspx
// https://msdn.microsoft.com/en-us/library/ff540107.aspx
// https://msdn.microsoft.com/en-us/library/ff540102.aspx
// https://msdn.microsoft.com/en-us/library/ff552125.aspx
// https://msdn.microsoft.com/en-us/library/ff542915.aspx
// https://msdn.microsoft.com/en-us/library/ff541929.aspx
// https://msdn.microsoft.com/en-us/library/ff542892.aspx
// https://msdn.microsoft.com/en-us/library/ff542932.aspx
// https://msdn.microsoft.com/en-us/library/ff541920.aspx
// https://msdn.microsoft.com/en-us/library/ff542950.aspx
// https://msdn.microsoft.com/en-us/library/ff541911.aspx
// https://msdn.microsoft.com/en-us/library/ff542887.aspx
// https://msdn.microsoft.com/en-us/library/ff542898.aspx
// https://msdn.microsoft.com/en-us/library/ff542937.aspx
// https://msdn.microsoft.com/en-us/library/ff542941.aspx
// https://msdn.microsoft.com/en-us/library/ff541722.aspx
// https://msdn.microsoft.com/en-us/library/ff538966.aspx
// https://msdn.microsoft.com/en-us/library/ff542796.aspx
// 
// https://msdn.microsoft.com/en-us/library/ff542881.aspx
// https://msdn.microsoft.com/en-us/library/ff541926.aspx

#define FLG_STOP_ON_EXCEPTION		0x00000001UL
#define FLG_SHOW_LDR_SNAPS		0x00000002UL
#define FLG_DEBUG_INITIAL_COMMAND	0x00000004UL
#define FLG_STOP_ON_HUNG_GUI		0x00000008UL
#define FLG_HEAP_ENABLE_TAIL_CHECK	0x00000010UL
#define FLG_HEAP_ENABLE_FREE_CHECK	0x00000020UL
#define FLG_HEAP_VALIDATE_PARAMETERS	0x00000040UL
#define FLG_HEAP_VALIDATE_ALL		0x00000080UL
#define FLG_APPLICATION_VERIFIER	0x00000100UL
#define FLG_MONITOR_SILENT_PROCESS_EXIT	0x00000200UL
#define FLG_POOL_ENABLE_TAGGING		0x00000400UL
#define FLG_HEAP_ENABLE_TAGGING		0x00000800UL
#define FLG_USER_STACK_TRACE_DB		0x00001000UL
#define FLG_KERNEL_STACK_TRACE_DB	0x00002000UL
#define FLG_MAINTAIN_OBJECT_TYPELIST	0x00004000UL
#define FLG_HEAP_ENABLE_TAG_BY_DLL	0x00008000UL
#define FLG_DISABLE_STACK_EXTENSION	0x00010000UL
#define FLG_ENABLE_CSRDEBUG		0x00020000UL
#define FLG_ENABLE_KDEBUG_SYMBOL_LOAD	0x00040000UL
#define FLG_DISABLE_PAGE_KERNEL_STACKS	0x00080000UL
#define FLG_ENABLE_SYSTEM_CRIT_BREAKS	0x00100000UL
#define FLG_HEAP_DISABLE_COALESCING	0x00200000UL
#define FLG_ENABLE_CLOSE_EXCEPTIONS	0x00400000UL
#define FLG_ENABLE_EXCEPTION_LOGGING	0x00800000UL
#define FLG_ENABLE_HANDLE_TYPE_TAGGING	0x01000000UL
#define FLG_HEAP_PAGE_ALLOCS		0x02000000UL
#define FLG_DEBUG_INITIAL_COMMAND_EX	0x04000000UL
#define FLG_DISABLE_DBGPRINT		0x08000000UL
#define FLG_CRITSEC_EVENT_CREATION	0x10000000UL
#if _WIN32_WINNT <= 0x0502
#define FLG_LDR_TOP_DOWN		0x20000000UL
#else
#define FLG_STOP_ON_UNHANDLED_EXCEPTION	0x20000000UL
#endif
#define FLG_ENABLE_HANDLE_EXCEPTIONS	0x40000000UL
#define FLG_DISABLE_PROTDLLS		0x80000000UL

typedef	struct	_SYSTEM_GLOBAL_FLAG
{
	DWORD	GlobalFlag;
} SYSTEM_GLOBAL_FLAG, *LPSYSTEM_GLOBAL_FLAG, *PSYSTEM_GLOBAL_FLAG;

typedef	struct	_SYSTEM_HANDLE_INFORMATION
{
	WORD		UniqueProcessId;
	WORD		CreatorBackTraceIndex;
	BYTE		ObjectTypeIndex;
	BYTE		HandleAttributes;
	WORD		HandleValue;
	LPVOID		Object;
	ACCESS_MASK	GrantedAccess;
} SYSTEM_HANDLE_INFORMATION, *LPSYSTEM_HANDLE_INFORMATION, *PSYSTEM_HANDLE_INFORMATION;

typedef	struct	_SYSTEM_INSTRUCTION_EMULATION_COUNTS
{
	DWORD	GenericInvalidOpcode;
	DWORD	TwoByteOpcode;
	DWORD	ESprefix;
	DWORD	CSprefix;
	DWORD	SSprefix;
	DWORD	DSprefix;
	DWORD	FSPrefix;
	DWORD	GSprefix;
	DWORD	OPER32prefix;
	DWORD	ADDR32prefix;
	DWORD	INSB;
	DWORD	INSW;
	DWORD	OUTSB;
	DWORD	OUTSW;
	DWORD	PUSHFD;
	DWORD	POPFD;
	DWORD	INTnn;
	DWORD	INTO;
	DWORD	IRETD;
	DWORD	FloatingPointOpcode;
	DWORD	INBimm;
	DWORD	INWimm;
	DWORD	OUTBimm;
	DWORD	OUTWimm;
	DWORD	INB;
	DWORD	INW;
	DWORD	OUTB;
	DWORD	OUTW;
	DWORD	LOCKprefix;
	DWORD	REPNEprefix;
	DWORD	REPprefix;
	DWORD	CLI;
	DWORD	STI;
	DWORD	HLT;
} SYSTEM_INSTRUCTION_EMULATION_COUNTS, *LPSYSTEM_INSTRUCTION_EMULATION_COUNTS, *PSYSTEM_INSTRUCTION_EMULATION_COUNTS;

typedef	struct	_SYSTEM_INTERRUPT_INFORMATION
{
#if 0
	BYTE	Reserved1[24];
#else
	DWORD	ContextSwitches;
	DWORD	DpcCount;
	DWORD	DpcRate;
	DWORD	TimeIncrement;
	DWORD	DpcBypassCount;
	DWORD	ApcBypassCount;
#endif
} SYSTEM_INTERRUPT_INFORMATION, *LPSYSTEM_INTERRUPT_INFORMATION, *PSYSTEM_INTERRUPT_INFORMATION;

typedef	struct	_SYSTEM_KERNEL_DEBUGGER_INFORMATION
{
	BOOLEAN	DebuggerEnabled;
	BOOLEAN	DebuggerNotPresent;
} SYSTEM_KERNEL_DEBUGGER_INFORMATION, *LPSYSTEM_KERNEL_DEBUGGER_INFORMATION, *PSYSTEM_KERNEL_DEBUGGER_INFORMATION;

#pragma warning(disable: 4201 4214)
typedef	union	_SYSTEM_KERNEL_VA_SHADOW_INFORMATION
{
	DWORD	KvaShadowFlags;
	struct
	{
		DWORD	KvaShadowEnabled                 : 1;
		DWORD	KvaShadowUserGlobal              : 1;
		DWORD	KvaShadowPcid                    : 1;
		DWORD	KvaShadowInvpcid                 : 1;
		DWORD	KvaShadowRequired                : 1;
		DWORD	KvaShadowRequiredAvailable       : 1;
		DWORD	InvalidPteBit                    : 6;
		DWORD	L1DataCacheFlushSupported        : 1;
		DWORD	L1TerminalFaultMitigationPresent : 1;
		DWORD	Reserved                         : 18;
	};
} SYSTEM_KERNEL_VA_SHADOW_INFORMATION, *LPSYSTEM_KERNEL_VA_SHADOW_INFORMATION, *PSYSTEM_KERNEL_VA_SHADOW_INFORMATION;
#pragma warning(default: 4201 4214)

typedef	struct	_SYSTEM_LEAP_SECOND_INFORMATION
{
	BOOLEAN	Enabled;
	DWORD	Flags;
} SYSTEM_LEAP_SECOND_INFORMATION, *LPSYSTEM_LEAP_SECOND_INFORMATION, *PSYSTEM_LEAP_SECOND_INFORMATION;

typedef	struct	_SYSTEM_LOAD_AND_CALL_IMAGE
{
	UNICODE_STRING	ModuleName;
} SYSTEM_LOAD_AND_CALL_IMAGE, *LPSYSTEM_LOAD_AND_CALL_IMAGE, *PSYSTEM_LOAD_AND_CALL_IMAGE;

typedef	struct	_SYSTEM_LOAD_IMAGE
{
	UNICODE_STRING	ModuleName;
	LPVOID		ModuleBase;
	LPVOID		Unknown;
	LPVOID		EntryPoint;
	LPVOID		ExportDirectory;
} SYSTEM_LOAD_IMAGE, *LPSYSTEM_LOAD_IMAGE, *PSYSTEM_LOAD_IMAGE;

typedef	struct	_SYSTEM_LOCK_INFORMATION
{
	LPVOID	Address;
	WORD	Type;
	WORD	Reserved1;
	DWORD	ExclusiveOwnerThreadId;
	DWORD	ActiveCount;
	DWORD	ContentionCount;
	DWORD	Reserved2[2];
	DWORD	NumberOfSharedWaiters;
	DWORD	NumberOfExclusiveWaiters;
} SYSTEM_LOCK_INFORMATION, *LPSYSTEM_LOCK_INFORMATION, *PSYSTEM_LOCK_INFORMATION;

typedef	struct	_SYSTEM_LOOKASIDE_INFORMATION
{
#if 0
	BYTE	Reserved1[32];
#else
	WORD		Depth;
	WORD		MaximumDepth;
	DWORD		TotalAllocates;
	DWORD		AllocateMisses;
	DWORD		TotalFrees;
	DWORD		FreeMisses;
	POOL_TYPE	Type;
	DWORD		Tag;
	DWORD		Size;
#endif
} SYSTEM_LOOKASIDE_INFORMATION, *LPSYSTEM_LOOKASIDE_INFORMATION, *PSYSTEM_LOOKASIDE_INFORMATION;

#if _WIN32_WINNT <= 0x0500
typedef	struct	_SYSTEM_MEMORY_INFO
{
	LPBYTE	StringOffset;
	WORD	ValidCount;
	WORD	TransitionCount;
	WORD	ModifiedCount;
	WORD	PageTableCount;
} SYSTEM_MEMORY_INFO, *LPSYSTEM_MEMORY_INFO, *PSYSTEM_MEMORY_INFO;

typedef	struct	_SYSTEM_MEMORY_INFORMATION
{
	DWORD			InfoSize;
	DWORD			StringStart;
	SYSTEM_MEMORY_INFO	Memory[ANYSIZE_ARRAY];
} SYSTEM_MEMORY_INFORMATION, *LPSYSTEM_MEMORY_INFORMATION, *PSYSTEM_MEMORY_INFORMATION;
#endif

#if _WIN32_WINNT >= 0x0602
typedef	struct	_SYSTEM_MEMORY_INFORMATION
{
	DWORD_PTR	PagedPoolCommitPageCount;
	DWORD_PTR	NonPagedPoolPageCount;
	DWORD_PTR	MdlPageCount;
	DWORD_PTR	CommitPageCount;
} SYSTEM_MEMORY_INFORMATION, *LPSYSTEM_MEMORY_INFORMATION, *PSYSTEM_MEMORY_INFORMATION;
#endif

typedef	struct	_SYSTEM_MODULE_INFORMATION
{
	DWORD	Reserved[2];
	LPVOID	Base;
	DWORD	Size;
	DWORD	Flags;
	WORD	Index;
	WORD	Unknown;
	WORD	LoadCount;
	WORD	ModuleNameOffset;
	CHAR	ImageName[256];
} SYSTEM_MODULE_INFORMATION, *LPSYSTEM_MODULE_INFORMATION, *PSYSTEM_MODULE_INFORMATION;

typedef	struct	_SYSTEM_OBJECT_INFORMATION
{
	DWORD			NextEntryOffset;
	LPVOID			Object;
	DWORD			CreatorProcessId;
	WORD			Unknown;
	WORD			Flags;
	DWORD			PointerCount;
	DWORD			HandleCount;
	DWORD			PagedPoolUsage;
	DWORD			NonPagedPoolUsage;
	DWORD			ExclusiveProcessId;
	SECURITY_DESCRIPTOR	*SecurityDescriptor;
	UNICODE_STRING		Name;
} SYSTEM_OBJECT_INFORMATION, *LPSYSTEM_OBJECT_INFORMATION, *PSYSTEM_OBJECT_INFORMATION;

typedef	struct	_SYSTEM_OBJECT_TYPE_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		ObjectCount;
	DWORD		HandleCount;
	DWORD		TypeNumber;
	DWORD		InvalidAttributes;
	GENERIC_MAPPING	GenericMapping;
	ACCESS_MASK	ValidAccessMask;
	POOL_TYPE	PoolType;
	BYTE		Unknown;
	UNICODE_STRING	Name;
} SYSTEM_OBJECT_TYPE_INFORMATION, *LPSYSTEM_OBJECT_TYPE_INFORMATION, *PSYSTEM_OBJECT_TYPE_INFORMATION;

typedef	struct	_SYSTEM_PAGEFILE_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		CurrentSize;
	DWORD		TotalUsed;
	DWORD		PeakUsed;
	UNICODE_STRING	FileName;
} SYSTEM_PAGEFILE_INFORMATION, *LPSYSTEM_PAGEFILE_INFORMATION, *PSYSTEM_PAGEFILE_INFORMATION;

typedef	struct	_SYSTEM_PERFORMANCE_INFORMATION
{
#if 0
	BYTE	Reserved1[312];
#else
	LARGE_INTEGER	IdleProcessTime;
	LARGE_INTEGER	IoReadTransferCount;
	LARGE_INTEGER	IoWriteTransferCount;
	LARGE_INTEGER	IoOtherTransferCount;
	DWORD		IoReadOperationCount;
	DWORD		IoWriteOperationCount;
	DWORD		IoOtherOperationCount;
	DWORD		AvailablePages;
	DWORD		CommittedPages;
	DWORD		CommitLimit;
	DWORD		PeakCommitment;
	DWORD		PageFaultCount;
	DWORD		CopyOnWriteCount;
	DWORD		TransitionCount;
	DWORD		CacheTransitionCount;
	DWORD		DemandZeroCount;
	DWORD		PageReadCount;
	DWORD		PageReadIoCount;
	DWORD		CacheReadCount;
	DWORD		CacheIoCount;
	DWORD		DirtyPagesWriteCount;
	DWORD		DirtyWriteIoCount;
	DWORD		MappedPagesWriteCount;
	DWORD		MappedWriteIoCount;
	DWORD		PagedPoolPages;
	DWORD		NonPagedPoolPages;
	DWORD		PagedPoolAllocs;
	DWORD		PagedPoolFrees;
	DWORD		NonPagedPoolAllocs;
	DWORD		NonPagedPoolFrees;
	DWORD		FreeSystemPtes;
	DWORD		ResidentSystemCodePage;
	DWORD		TotalSystemDriverPages;
	DWORD		TotalSystemCodePages;
	DWORD		NonPagedPoolLookasideHits;
	DWORD		PagedPoolLookasideHits;
	DWORD		AvailablePagedPoolPages;
	DWORD		ResidentSystemCachePage;
	DWORD		ResidentPagedPoolPage;
	DWORD		ResidentSystemDriverPage;
	DWORD		CcFastReadNoWait;
	DWORD		CcFastReadWait;
	DWORD		CcFastReadResourceMiss;
	DWORD		CcFastReadNotPossible;
	DWORD		CcFastMdlReadNoWait;
	DWORD		CcFastMdlReadWait;
	DWORD		CcFastMdlReadResourceMiss;
	DWORD		CcFastMdlReadNotPossible;
	DWORD		CcMapDataNoWait;
	DWORD		CcMapDataWait;
	DWORD		CcMapDataNoWaitMiss;
	DWORD		CcMapDataWaitMiss;
	DWORD		CcPinMappedDataCount;
	DWORD		CcPinReadNoWait;
	DWORD		CcPinReadWait;
	DWORD		CcPinReadNoWaitMiss;
	DWORD		CcPinReadWaitMiss;
	DWORD		CcCopyReadNoWait;
	DWORD		CcCopyReadWait;
	DWORD		CcCopyReadNoWaitMiss;
	DWORD		CcCopyReadWaitMiss;
	DWORD		CcMdlReadNoWait;
	DWORD		CcMdlReadWait;
	DWORD		CcMdlReadNoWaitMiss;
	DWORD		CcMdlReadWaitMiss;
	DWORD		CcReadAheadIos;
	DWORD		CcLazyWriteIos;
	DWORD		CcLazyWritePages;
	DWORD		CcDataFlushes;
	DWORD		CcDataPages;
	DWORD		ContextSwitches;
	DWORD		FirstLevelTbFills;
	DWORD		SecondLevelTbFills;
	DWORD		SystemCalls;
#endif
#if _WIN32_WINNT >= 0x0601
	QWORD		CcTotalDirtyPages;
	QWORD		CcDirtyPageThreshold;
#if _WIN32_WINNT >= 0x0602
	QWORD		ResidentAvailablePages;
	QWORD		SharedCommittedPages;
#endif
#endif
} SYSTEM_PERFORMANCE_INFORMATION, *LPSYSTEM_PERFORMANCE_INFORMATION, *PSYSTEM_PERFORMANCE_INFORMATION;

typedef	enum	_INTERFACE_TYPE
{
	InterfaceTypeUndefined = -1,
	Internal,
	Isa,
	Eisa,
	MicroChannel,
	TurboChannel,
	PCIBus,
	VMEBus,
	NuBus,
	PCMCIABus,
	CBus,
	MPIBus,
	MPSABus,
	ProcessorInternal,
	InternalPowerBus,
	PNPISABus,
	MaximumInterfaceType
} INTERFACE_TYPE, *LPINTERFACE_TYPE, *PINTERFACE_TYPE;

typedef	enum	_PLUGPLAY_BUS_CLASS
{
	SystemBus           = 0,
	PlugPlayVirtualBus  = 1,
	MaxPlugPlayBusClass = 2
} PLUGPLAY_BUS_CLASS, *LPPLUGPLAY_BUS_CLASS, *PPLUGPLAY_BUS_CLASS;

typedef	enum	_PLUGPLAY_VIRTUAL_BUS_TYPE
{
	Root                      = 0,
	MaxPlugPlayVirtualBusType = 1
} PLUGPLAY_VIRTUAL_BUS_TYPE, *LPPLUGPLAY_VIRTUAL_BUS_TYPE, *PPLUGPLAY_VIRTUAL_BUS_TYPE;

#pragma warning(disable: 4201)
typedef	struct	_PLUGPLAY_BUS_TYPE
{
	PLUGPLAY_BUS_CLASS	BusClass;
	union
	{
		INTERFACE_TYPE			SystemBusType;
		PLUGPLAY_VIRTUAL_BUS_TYPE	PlugPlayVirtualBusType;
	};
} PLUGPLAY_BUS_TYPE, *LPPLUGPLAY_BUS_TYPE, *PPLUGPLAY_BUS_TYPE;
#pragma warning(disable: 4201)

typedef	struct	_PLUGPLAY_BUS_INSTANCE
{
	PLUGPLAY_BUS_TYPE	BusType;
	DWORD			BusNumber;
	WORD			BusName[24];
} PLUGPLAY_BUS_INSTANCE, *LPPLUGPLAY_BUS_INSTANCE, *PPLUGPLAY_BUS_INSTANCE;

typedef	struct	_SYSTEM_PLUGPLAY_BUS_INFORMATION
{
	DWORD			BusCount;
	PLUGPLAY_BUS_INSTANCE	BusInstance[ANYSIZE_ARRAY];
} SYSTEM_PLUGPLAY_BUS_INFORMATION, *LPSYSTEM_PLUGPLAY_BUS_INFORMATION, *PSYSTEM_PLUGPLAY_BUS_INFORMATION;

typedef	struct	_SYSTEM_POLICY_INFORMATION
{
	LPVOID	Reserved1[2];
	DWORD	Reserved2[3];
} SYSTEM_POLICY_INFORMATION, *LPSYSTEM_POLICY_INFORMATION, *PSYSTEM_POLICY_INFORMATION;

typedef	struct	_SYSTEM_POOL_TAG_INFORMATION
{
	CHAR	Tag[4];
	DWORD	PagedPoolAllocs;
	DWORD	PagedPoolFrees;
	DWORD	PagedPoolUsage;
	DWORD	NonPagedPoolAllocs;
	DWORD	NonPagedPoolFrees;
	DWORD	NonPagedPoolUsage;
} SYSTEM_POOL_TAG_INFORMATION, *LPSYSTEM_POOL_TAG_INFORMATION, *PSYSTEM_POOL_TAG_INFORMATION;

typedef	struct	_SYSTEM_POWER_INFORMATION
{
	BOOLEAN		SystemSuspendSupported;
	BOOLEAN		SystemHibernateSupported;
	BOOLEAN		ResumeTimerSupportsSuspend;
	BOOLEAN		ResumeTimerSupportsHibernate;
	BOOLEAN		LidSupported;
	BOOLEAN		TurboSettingSupported;
	BOOLEAN		TurboMode;
	BOOLEAN		SystemAcOrDc;
	BOOLEAN		PowerDownDisabled;
	LARGE_INTEGER	SpindownDrives;
} SYSTEM_POWER_INFORMATION, *LPSYSTEM_POWER_INFORMATION, *PSYSTEM_POWER_INFORMATION;

typedef	struct	_SYSTEM_PRIORITY_SEPARATION
{
	DWORD	PrioritySeparation;
} SYSTEM_PRIORITY_SEPARATION, *LPSYSTEM_PRIORITY_SEPARATION, *PSYSTEM_PRIORITY_SEPARATION;

typedef	struct	_SYSTEM_PROCESS_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		NumberOfThreads;
#if 0
	BYTE		Reserved1[48];
	LPVOID		Reserved2[3];
#else
	LARGE_INTEGER	WorkingSetPrivateSize;
	DWORD		HardFaultCount;
	DWORD		NumberOfThreadsHighWatermark;
	QWORD		CycleTime;
	LARGE_INTEGER	CreateTime;
	LARGE_INTEGER	UserTime;
	LARGE_INTEGER	KernelTime;
	UNICODE_STRING	ImageName;
	DWORD		BasePriority;
#endif
	HANDLE		UniqueProcessId;
#if 0
	LPVOID		Reserved3;
#else
	LPVOID		InheritedFromUniqueProcessId;
#endif
	DWORD		HandleCount;
#if 0
	BYTE		Reserved4[4];
	LPVOID		Reserved5[11];
	SIZE_T		PeakPagefileUsage;
	SIZE_T		PrivatePageCount;
	LARGE_INTEGER	Reserved6[6];
#else
	DWORD		SessionId;
	DWORD_PTR	UniqueProcessKey;
	DWORD_PTR	PeakVirtualSize;
	DWORD_PTR	VirtualSize;
	DWORD		PageFaultCount;
	DWORD_PTR	PeakWorkingSetSize;
	DWORD_PTR	WorkingSetSize;
	DWORD_PTR	QuotaPeakPagedPoolUsage;
	DWORD_PTR	QuotaPagedPoolUsage;
	DWORD_PTR	QuotaPeakNonPagedPoolUsage;
	DWORD_PTR	QuotaNonPagedPoolUsage;
	DWORD_PTR	PagefileUsage;
	DWORD_PTR	PeakPagefileUsage;
	DWORD_PTR	PrivatePageCount;
	LARGE_INTEGER	ReadOperationCount;
	LARGE_INTEGER	WriteOperationCount;
	LARGE_INTEGER	OtherOperationCount;
	LARGE_INTEGER	ReadTransferCount;
	LARGE_INTEGER	WriteTransferCount;
	LARGE_INTEGER	OtherTransferCount;
#endif
} SYSTEM_PROCESS_INFORMATION, *LPSYSTEM_PROCESS_INFORMATION, *PSYSTEM_PROCESS_INFORMATION;

typedef	enum	_THREAD_STATE
{
	StateInitialized = 0,
	StateReady,
	StateRunning,
	StateStandby,
	StateTerminated,
	StateWait,
	StateTransition,
	StateUnknown
} THREAD_STATE;

typedef	enum	_KWAIT_REASON
{
	Executive         = 0,
	FreePage          = 1,
	PageIn            = 2,
	PoolAllocation    = 3,
	DelayExecution    = 4,
	Suspended         = 5,
	UserRequest       = 6,
	WrExecutive       = 7,
	WrFreePage        = 8,
	WrPageIn          = 9,
	WrPoolAllocation  = 10,
	WrDelayExecution  = 11,
	WrSuspended       = 12,
	WrUserRequest     = 13,
	WrEventPair       = 14,
	WrQueue           = 15,
	WrLpcReceive      = 16,
	WrLpcReply        = 17,
	WrVirtualMemory   = 18,
	WrPageOut         = 19,
	WrRendezvous      = 20,
	Spare2            = 21,
	Spare3            = 22,
	Spare4            = 23,
	Spare5            = 24,
	WrCalloutStack    = 25,
	WrKernel          = 26,
	WrResource        = 27,
	WrPushLock        = 28,
	WrMutex           = 29,
	WrQuantumEnd      = 30,
	WrDispatchInt     = 31,
	WrPreempted       = 32,
	WrYieldExecution  = 33,
	WrFastMutex       = 34,
	WrGuardedMutex    = 35,
	WrRundown         = 36,
	MaximumWaitReason = 37
} KWAIT_REASON;

typedef	struct	_SYSTEM_THREAD_INFORMATION
{
#if 0
	LARGE_INTEGER	Reserved1[3];
	DWORD		Reserved2;
#else
	LARGE_INTEGER	KernelTime;
	LARGE_INTEGER	UserTime;
	LARGE_INTEGER	CreateTime;
	DWORD		WaitTime;
#endif
	LPVOID		StartAddress;
	CLIENT_ID	ClientId;
	KPRIORITY	Priority;
	KPRIORITY	BasePriority;
	DWORD		ContextSwitchCount;
	THREAD_STATE	State;
	KWAIT_REASON	WaitReason;
} SYSTEM_THREAD_INFORMATION, *LPSYSTEM_THREAD_INFORMATION, *PSYSTEM_THREAD_INFORMATION;

typedef	struct	_SYSTEM_EXTENDED_THREAD_INFORMATION
{
	SYSTEM_THREAD_INFORMATION	ThreadInfo;
	LPVOID		StackBase;
	LPVOID		StackLimit;
	LPVOID		Win32StartAddress;
	LPVOID		TebAddress;
	DWORD_PTR	Reserved[3];
} SYSTEM_EXTENDED_THREAD_INFORMATION, *LPSYSTEM_EXTENDED_THREAD_INFORMATION, *PSYSTEM_EXTENDED_THREAD_INFORMATION;

#if _WIN32_WINNT >= 0x0600
typedef	struct	_SYSTEM_PROCESS_ID_INFORMATION
{
	HANDLE		ProcessId;
	UNICODE_STRING	ImageName;
} SYSTEM_PROCESS_ID_INFORMATION, *LPSYSTEM_PROCESS_ID_INFORMATION, *PSYSTEM_PROCESS_ID_INFORMATION;
#endif

#ifndef _WINNT_
typedef	struct	_SYSTEM_PROCESS_INFORMATION
{
	DWORD		NextEntryOffset;
	DWORD		NumberOfThreads;
#if 0
	BYTE		Reserved1[48];
#else
	DWORD		Reserved1[6];
	LARGE_INTEGER	CreateTime;
	LARGE_INTEGER	UserTime;
	LARGE_INTEGER	KernelTime;
#endif
	UNICODE_STRING	ImageName;
	KPRIORITY	BasePriority;
	HANDLE		UniqueProcessId;
	LPVOID		Reserved2;
	DWORD		HandleCount;
	DWORD		SessionId;
	LPVOID		Reserved3;
	SIZE_T		PeakVirtualSize;
	SIZE_T		VirtualSize;
	DWORD		Reserved4;
	SIZE_T		PeakWorkingSetSize;
	SIZE_T		WorkingSetSize;
	LPVOID		Reserved5;
	SIZE_T		QuotaPagedPoolUsage;
	LPVOID		Reserved6;
	SIZE_T		QuotaNonPagedPoolUsage;
	SIZE_T		PagefileUsage;
	SIZE_T		PeakPagefileUsage;
	SIZE_T		PrivatePageCount;
#if 0
	LARGE_INTEGER	Reserved7[6];
#else
	IO_COUNTERS	IOInfo;
#endif
#if 1
	SYSTEM_THREAD_INFORMATION	ThreadInfo[ANYSIZE_ARRAY];
#else
	SYSTEM_EXTENDED_THREAD_INFORMATION	ThreadInfo[ANYSIZE_ARRAY];
#endif
} SYSTEM_PROCESS_INFORMATION, *LPSYSTEM_PROCESS_INFORMATION, *PSYSTEM_PROCESS_INFORMATION;
#endif // _WINNT_

typedef	struct	_SYSTEM_PROCESSOR_INFORMATION
{
	WORD	ProcessorArchitecture;
	WORD	ProcessorLevel;
	WORD	ProcessorRevision;
	WORD	MaximumProcessors;
	DWORD	ProcessorFeatureBits;
} SYSTEM_PROCESSOR_INFORMATION, *LPSYSTEM_PROCESSOR_INFORMATION, *PSYSTEM_PROCESSOR_INFORMATION;

typedef	struct	_SYSTEM_PROCESSOR_FEATURES_INFORMATION
{
	ULONGLONG	ProcessorFeatureBits;
	ULONGLONG	Reserved[3];
} SYSTEM_PROCESSOR_FEATURES_INFORMATION, *LPSYSTEM_PROCESSOR_FEATURES_INFORMATION, *PSYSTEM_PROCESSOR_FEATURES_INFORMATION;

typedef	struct	_SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION
{
	LARGE_INTEGER	IdleTime;
	LARGE_INTEGER	KernelTime;
	LARGE_INTEGER	UserTime;
#if 0
	LARGE_INTEGER	Reserved1[2];
	DWORD		Reserved2;
#else
	LARGE_INTEGER	DpcTime;
	LARGE_INTEGER	InterruptTime;
	DWORD		InterruptCount;
#endif
} SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION, *PSYSTEM_PROCESSOR_PERFORMANCE_INFORMATION, *PSYSTEM_PROCESSOR_PERFORMANCE_INFORMATION;

typedef	struct	_PROCESSOR_PROFILE_CONTROL_AREA
{
	ULONGLONG	BtsBufferBase;
	ULONGLONG	BtsIndex;
	ULONGLONG	BtsAbsoluteMaximum;
	ULONGLONG	BtsInterruptThreshold;
	ULONGLONG	PebsBufferBase;
	ULONGLONG	PebsIndex;
	ULONGLONG	PebsAbsoluteMaximum;
	ULONGLONG	PebsInterruptThreshold;
	ULONGLONG	PebsCounterReset0;
	ULONGLONG	PebsCounterReset1;
	ULONGLONG	PebsCounterReset2;
	ULONGLONG	PebsCounterReset3;
} PROCESSOR_PROFILE_CONTROL_AREA, *LPPROCESSOR_PROFILE_CONTROL_AREA, *PPROCESSOR_PROFILE_CONTROL_AREA;

typedef	struct	_SYSTEM_PROCESSOR_PROFILE_CONTROL_AREA
{
	PROCESSOR_PROFILE_CONTROL_AREA	*ProcessorProfileControlArea;
	BOOLEAN				Allocate;
} SYSTEM_PROCESSOR_PROFILE_CONTROL_AREA, *LPSYSTEM_PROCESSOR_PROFILE_CONTROL_AREA, *PSYSTEM_PROCESSOR_PROFILE_CONTROL_AREA;

typedef	struct	_SYSTEM_PROCESSOR_SPEED_INFORMATION
{
	DWORD	MaximumProcessorSpeed;
	DWORD	CurrentAvailableSpeed;
	DWORD	ConfiguredSpeedLimit;
	BYTE	PowerLimit;
	BYTE	ThermalLimit;
	BYTE	TurboLimit;
} SYSTEM_PROCESSOR_SPEED_INFORMATION, *LPSYSTEM_PROCESSOR_SPEED_INFORMATION, *PSYSTEM_PROCESSOR_SPEED_INFORMATION;

typedef	struct	_SYSTEM_PROCESSOR_STATISTICS
{
	DWORD	ContextSwitches;
	DWORD	DpcCount;
	DWORD	DpcRequestRate;
	DWORD	TimeIncrement;
	DWORD	DpcBypassCount;
	DWORD	ApcBypassCount;
} SYSTEM_PROCESSOR_STATISTICS, *LPSYSTEM_PROCESSOR_STATISTICS, *PSYSTEM_PROCESSOR_STATISTICS;

typedef	struct	_SYSTEM_PROCESSOR_TIMES
{
	LARGE_INTEGER	IdleTime;
	LARGE_INTEGER	KernelTime;
	LARGE_INTEGER	UserTime;
	LARGE_INTEGER	DpcTime;
	LARGE_INTEGER	InterruptTime;
	DWORD		InterruptCount;
} SYSTEM_PROCESSOR_TIMES, *LPSYSTEM_PROCESSOR_TIMES, *PSYSTEM_PROCESSOR_TIMES;

#pragma warning(disable: 4201 4214)
typedef	union	_QUERY_PERFORMANCE_COUNTER_FLAGS
{
	DWORD	ul;
	struct
	{
		DWORD	KernelTransition : 1;
		DWORD	Reserved         : 31;
	};
} QUERY_PERFORMANCE_COUNTER_FLAGS;
#pragma warning(disable: 4201 4214)

typedef	struct	_SYSTEM_QUERY_PERFORMANCE_COUNTER_INFORMATION
{
	DWORD				Version;
	QUERY_PERFORMANCE_COUNTER_FLAGS	Flags;
	QUERY_PERFORMANCE_COUNTER_FLAGS	ValidFlags;
} SYSTEM_QUERY_PERFORMANCE_COUNTER_INFORMATION, *LPSYSTEM_QUERY_PERFORMANCE_COUNTER_INFORMATION, *PSYSTEM_QUERY_PERFORMANCE_COUNTER_INFORMATION;

typedef	struct	_SYSTEM_QUERY_TIME_ADJUSTMENT
{
	DWORD	TimeAdjustment;
	DWORD	MaximumIncrement;
	BOOLEAN	imeSynchronization;
} SYSTEM_QUERY_TIME_ADJUSTMENT, *LPSYSTEM_QUERY_TIME_ADJUSTMENT, *PSYSTEM_QUERY_TIME_ADJUSTMENT;

typedef	struct	_SYSTEM_RANGE_START_INFORMATION
{
	LPVOID	SystemRangeStart;
} SYSTEM_RANGE_START_INFORMATION, *LPSYSTEM_RANGE_START_INFORMATION, *PSYSTEM_RANGE_START_INFORMATION;

typedef	struct	_SYSTEM_REGISTRY_QUOTA_INFORMATION
{
	DWORD	RegistryQuotaAllowed;
	DWORD	RegistryQuotaUsed;
	LPVOID	Reserved1;
} SYSTEM_REGISTRY_QUOTA_INFORMATION, *LPSYSTEM_REGISTRY_QUOTA_INFORMATION, *PSYSTEM_REGISTRY_QUOTA_INFORMATION;

typedef	struct	_SYSTEM_SECUREBOOT_INFORMATION
{
	BOOLEAN	SecureBootEnabled;
	BOOLEAN	SecureBootCapable;
} SYSTEM_SECUREBOOT_INFORMATION, *LPSYSTEM_SECUREBOOT_INFORMATION, *PSYSTEM_SECUREBOOT_INFORMATION;

typedef	struct	_SYSTEM_SECUREBOOT_POLICY_INFORMATION
{
	GUID	PolicyPublisher;
	DWORD	PolicyVersion;
	DWORD	PolicyOptions;
} SYSTEM_SECUREBOOT_POLICY_INFORMATION, *LPSYSTEM_SECUREBOOT_POLICY_INFORMATION, *PSYSTEM_SECUREBOOT_POLICY_INFORMATION;

typedef	struct	_SYSTEM_SECUREBOOT_POLICY_FULL_INFORMATION
{
	SYSTEM_SECUREBOOT_POLICY_INFORMATION	PolicyInformation;
	DWORD	PolicySize;
	BYTE	Policy[ANYSIZE_ARRAY];
} SYSTEM_SECUREBOOT_POLICY_FULL_INFORMATION, *LPSYSTEM_SECUREBOOT_POLICY_FULL_INFORMATION, *PSYSTEM_SECUREBOOT_POLICY_FULL_INFORMATION;

typedef	struct	_SYSTEM_SESSION_PROCESS_INFORMATION
{
	DWORD	SessionId;
	DWORD	BufferSize;
	LPVOID	Buffer;
} SYSTEM_SESSION_PROCESS_INFORMATION, *LPSYSTEM_SESSION_PROCESS_INFORMATION, *PSYSTEM_SESSION_PROCESS_INFORMATION;

typedef	struct	_SYSTEM_SET_TIME_ADJUSTMENT
{
	DWORD	TimeAdjustment;
	BOOLEAN	TimeSynchronization;
} SYSTEM_SET_TIME_ADJUSTMENT, *LPSYSTEM_SET_TIME_ADJUSTMENT, *PSYSTEM_SET_TIME_ADJUSTMENT;

typedef	struct	_SYSTEM_SET_TIME_SLIP_EVENT
{
	HANDLE	TimeSlipEvent;
} SYSTEM_SET_TIME_SLIP_EVENT, *LPSYSTEM_SET_TIME_SLIP_EVENT, *PSYSTEM_SET_TIME_SLIP_EVENT;

#pragma warning(disable: 4214)
typedef	union	_SYSTEM_SPECULATION_CONTROL_INFORMATION
{
	DWORD	SpeculationControlFlags;
	struct
	{
		DWORD	BpbEnabled                               : 1;
		DWORD	BpbDisabledSystemPolicy                  : 1;
		DWORD	BpbDisabledNoHardwareSupport             : 1;
		DWORD	SpecCtrlEnumerated                       : 1;
		DWORD	SpecCmdEnumerated                        : 1;
		DWORD	IbrsPresent                              : 1;
		DWORD	StibpPresent                             : 1;
		DWORD	SmepPresent                              : 1;
		DWORD	SpeculativeStoreBypassDisableAvailable   : 1;
		DWORD	SpeculativeStoreBypassDisableSupported   : 1;
		DWORD	SpeculativeStoreBypassDisabledSystemWide : 1;
		DWORD	SpeculativeStoreBypassDisabledKernel     : 1;
		DWORD	SpeculativeStoreBypassDisableRequired    : 1;
		DWORD	BpbDisabledKernelToUser                  : 1;
		DWORD	SpecCtrlRetpolineEnabled                 : 1;
		DWORD	SpecCtrlImportOptimizationEnabled        : 1;
		DWORD	EnhancedIbrs                             : 1;
		DWORD	HvL1tfStatusAvailable                    : 1;
		DWORD	HvL1tfProcessorNotAffected               : 1;
		DWORD	HvL1tfMigitationEnabled                  : 1;
		DWORD	HvL1tfMigitationNotEnabled_Hardware      : 1;
		DWORD	HvL1tfMigitationNotEnabled_LoadOption    : 1;
		DWORD	HvL1tfMigitationNotEnabled_CoreScheduler : 1;
		DWORD	EnhancedIbrsReported                     : 1;
		DWORD	MdsHardwareProtected                     : 1;
		DWORD	MbClearEnabled                           : 1;
		DWORD	MbClearReported                          : 1;
		DWORD	Reserved                                 : 5;
	};
} SYSTEM_SPECULATION_CONTROL_INFORMATION, *LPSYSTEM_SPECULATION_CONTROL_INFORMATION, *PSYSTEM_SPECULATION_CONTROL_INFORMATION;
#pragma warning(default: 4214)

// https://msdn.microsoft.com/en-us/library/ff562879.aspx

typedef	struct	TIME_FIELDS
{
	WORD	Year;
	WORD	Month;
	WORD	Day;
	WORD	Hour;
	WORD	Minute;
	WORD	Second;
	WORD	Milliseconds;
	WORD	Weekday;
} TIME_FIELDS, *LPTIME_FIELDS, *PTIME_FIELDS;

typedef	struct	_SYSTEM_TIME_ZONE_INFORMATION
{
	LONG		Bias;
	WCHAR		StandardName[32];
	TIME_FIELDS	StandardDate;
	LONG		StandardBias;
	WCHAR		DaylightName[32];
	TIME_FIELDS	DaylightDate;
	LONG		DaylightBias;
} SYSTEM_TIME_ZONE_INFORMATION, *LPSYSTEM_TIME_ZONE_INFORMATION, *PSYSTEM_TIME_ZONE_INFORMATION;

typedef	struct	_SYSTEM_DYNAMIC_TIME_ZONE_INFORMATION
{
	SYSTEM_TIME_ZONE_INFORMATION	tzi;
	WCHAR				TimeZoneKeyName[128];
	BOOLEAN				DynamicDaylightTimeDisabled;
} SYSTEM_DYNAMIC_TIME_ZONE_INFORMATION, *LPSYSTEM_DYNAMIC_TIME_ZONE_INFORMATION, *PSYSTEM_DYNAMIC_TIME_ZONE_INFORMATION;

typedef	struct	_SYSTEM_TIMEOFDAY_INFORMATION
{
#if 0
	BYTE	Reserved1[48];
#else
	LARGE_INTEGER	BootTime;
	LARGE_INTEGER	CurrentTime;
	LARGE_INTEGER	TimeZoneBias;
	DWORD		CurrentTimeZoneId;
	DWORD		Reserved;
	QWORD		BootTimeBias;
	QWORD		SleepTimeBias;
#endif
} SYSTEM_TIMEOFDAY_INFORMATION, *LPSYSTEM_TIMEOFDAY_INFORMATION, *PSYSTEM_TIMEOFDAY_INFORMATION;

typedef	struct	_SYSTEM_UNLOAD_IMAGE
{
	LPVOID	ModuleBase;
} SYSTEM_UNLOAD_IMAGE, *LPSYSTEM_UNLOAD_IMAGE, *PSYSTEM_UNLOAD_IMAGE;

__declspec(dllimport)
LONG	NTAPI	NtQuerySystemInformation(SYSTEM_INFORMATION_CLASS SystemInformationClass,
		                         LPVOID                   SystemInformation,
		                         DWORD                    SystemInformationLength,
		                         LPDWORD                  ReturnLength);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	NtQuerySystemInformationEx(SYSTEM_INFORMATION_CLASS SystemInformationClass,
		                           LPVOID                   InputBuffer,
		                           DWORD                    InputBufferLength,
		                           LPVOID                   SystemInformation,
		                           DWORD                    SystemInformationLength,
		                           LPDWORD                  ReturnLength);
#endif

// https://msdn.microsoft.com/en-us/library/ms724512.aspx

__declspec(dllimport)
LONG	NTAPI	NtQuerySystemTime(LARGE_INTEGER *SystemTime);

typedef	enum	_TIMER_INFORMATION_CLASS
{
	TimerBasicInformation = 0
} TIMER_INFORMATION_CLASS, *LPTIMER_INFORMATION_CLASS, *PTIMER_INFORMATION_CLASS;

typedef	struct	_TIMER_BASIC_INFORMATION
{
	LARGE_INTEGER	RemainingTime;
	BOOLEAN		TimerState;
} TIMER_BASIC_INFORMATION, *LPTIMER_BASIC_INFORMATION, *PTIMER_BASIC_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtQueryTimer(HANDLE                  TimerHandle,
		             TIMER_INFORMATION_CLASS TimerInformationClass,
		             LPVOID                  TimerInformation,
		             DWORD                   TimerInformationLength,
		             LPDWORD                 ReturnLength);

__declspec(dllimport)
LONG	NTAPI	NtQueryTimerResolution(LPDWORD MinimumResolution,
		                       LPDWORD MaximumResolution,
		                       LPDWORD CurrentResolution);

// https://msdn.microsoft.com/en-us/library/ff556701.aspx
// https://msdn.microsoft.com/en-us/library/ff567069.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryValueKey(HANDLE                      KeyHandle,
		                UNICODE_STRING              *ValueName,
		                KEY_VALUE_INFORMATION_CLASS KeyValueInformationClass,
		                LPVOID                      KeyValueInformation,
		                DWORD                       KeyValueInformationLength,
		                LPDWORD                     ReturnLength);

// https://msdn.microsoft.com/en-us/library/dn957516.aspx

typedef	enum	_MEMORY_INFORMATION_CLASS
{
	MemoryBasicInformation = 0,
	MemoryEntryInformation
} MEMORY_INFORMATION_CLASS;

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/aa366775.aspx
// https://msdn.microsoft.com/en-us/library/dn957515.aspx

typedef	struct	_MEMORY_BASIC_INFORMATION
{
	LPVOID	BaseAddress;
	LPVOID	AllocationBase;
	DWORD	AllocationProtect;
	SIZE_T	RegionSize;
	DWORD	State;
	DWORD	Protect;
	DWORD	Type;
} MEMORY_BASIC_INFORMATION, *LPMEMORY_BASIC_INFORMATION, *PMEMORY_BASIC_INFORMATION;
#endif // _WINNT_

typedef	struct	_MEMORY_ENTRY_INFORMATION
{
	DWORD	NumberOfEntries;
	LPVOID	Entries[ANYSIZE_ARRAY];
} MEMORY_ENTRY_INFORMATION, *LPMEMORY_ENTRY_INFORMATION, *PMEMORY_ENTRY_INFORMATION;

// https://msdn.microsoft.com/en-us/library/dn957452.aspx
// https://msdn.microsoft.com/en-us/library/dn957455.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryVirtualMemory(HANDLE                   ProcessHandle,
		                     LPVOID                   BaseAddress,
		                     MEMORY_INFORMATION_CLASS MemoryInformationClass,
		                     LPVOID                   MemoryInformation,
		                     SIZE_T                   MemoryInformationLength,
		                     SIZE_T                   *ReturnLength);

#if _WIN32_WINNT >= 0x0501
// https://msdn.microsoft.com/en-us/library/ff540251.aspx

typedef	struct	_FILE_FS_ATTRIBUTE_INFORMATION
{
	DWORD	FileSystemAttributes;
	LONG	MaximumComponentNameLength;
	DWORD	FileSystemNameLength;
	WCHAR	FileSystemName[ANYSIZE_ARRAY];
} FILE_FS_ATTRIBUTE_INFORMATION, *LPFILE_FS_ATTRIBUTE_INFORMATION, *PFILE_FS_ATTRIBUTE_INFORMATION;

#define FILE_VC_QUOTA_NONE		0x00000000UL
#define FILE_VC_QUOTA_TRACK		0x00000001UL
#define FILE_VC_QUOTA_ENFORCE		0x00000002UL
#define FILE_VC_QUOTA_MASK		0x00000003UL
#define FILE_VC_QUOTAS_LOG_VIOLATIONS	0x00000004UL
#define FILE_VC_CONTENT_INDEX_DISABLED	0x00000008UL
#define FILE_VC_LOG_QUOTA_THRESHOLD	0x00000010UL
#define FILE_VC_LOG_QUOTA_LIMIT		0x00000020UL
#define FILE_VC_LOG_VOLUME_THRESHOLD	0x00000040UL
#define FILE_VC_LOG_VOLUME_LIMIT	0x00000080UL
#define FILE_VC_QUOTAS_INCOMPLETE	0x00000100UL
#define FILE_VC_QUOTAS_REBUILDING	0x00000200UL
#define FILE_VC_VALID_MASK		0x000003FFUL

// https://msdn.microsoft.com/en-us/library/ff540258.aspx

typedef	struct	_FILE_FS_CONTROL_INFORMATION
{
	LARGE_INTEGER	FreeSpaceStartFiltering;
	LARGE_INTEGER	FreeSpaceThreshold;
	LARGE_INTEGER	FreeSpaceStopFiltering;
	LARGE_INTEGER	DefaultQuotaThreshold;
	LARGE_INTEGER	DefaultQuotaLimit;
	DWORD		FileSystemControlFlags;
} FILE_FS_CONTROL_INFORMATION, *LPFILE_FS_CONTROL_INFORMATION, *PFILE_FS_CONTROL_INFORMATION;

#ifndef DEVICE_TYPE
#define DEVICE_TYPE	DWORD
#endif

// https://msdn.microsoft.com/en-us/library/ff545788.aspx

typedef	struct	_FILE_FS_DEVICE_INFORMATION
{
	DEVICE_TYPE	DeviceType;
	DWORD		Characteristics;
} FILE_FS_DEVICE_INFORMATION, *LPFILE_FS_DEVICE_INFORMATION, *PFILE_FS_DEVICE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540262.aspx

typedef	struct	_FILE_FS_DRIVER_PATH_INFORMATION
{
	BOOLEAN	DriverInPath;
	DWORD	DriverNameLength;
	WCHAR	DriverName[ANYSIZE_ARRAY];
} FILE_FS_DRIVER_PATH_INFORMATION, *LPFILE_FS_DRIVER_PATH_INFORMATION, *PFILE_FS_DRIVER_PATH_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540267.aspx

typedef	struct	_FILE_FS_FULL_SIZE_INFORMATION
{
	LARGE_INTEGER	TotalAllocationUnits;
	LARGE_INTEGER	CallerAvailableAllocationUnits;
	LARGE_INTEGER	ActualAvailableAllocationUnits;
	DWORD		SectorsPerAllocationUnit;
	DWORD		BytesPerSector;
} FILE_FS_FULL_SIZE_INFORMATION, *LPFILE_FS_FULL_SIZE_INFORMATION, *PFILE_FS_FULL_SIZE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540271.aspx

typedef	struct	_FILE_FS_LABEL_INFORMATION
{
	DWORD	VolumeLabelLength;
	WCHAR	VolumeLabel[ANYSIZE_ARRAY];
} FILE_FS_LABEL_INFORMATION, *LPFILE_FS_LABEL_INFORMATION, *PFILE_FS_LABEL_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540274.aspx

typedef	struct	_FILE_FS_OBJECTID_INFORMATION
{
	BYTE	ObjectId[16];
	BYTE	ExtendedInfo[48];
} FILE_FS_OBJECTID_INFORMATION, *LPFILE_FS_OBJECTID_INFORMATION, *PFILE_FS_OBJECTID_INFORMATION;

// https://msdn.microsoft.com/en-us/library/hh406395.aspx

typedef	struct	_FILE_FS_SECTOR_SIZE_INFORMATION
{
	DWORD	LogicalBytesPerSector;
	DWORD	PhysicalBytesPerSectorForAtomicity;
	DWORD	PhysicalBytesPerSectorForPerformance;
	DWORD	FileSystemEffectivePhysicalBytesPerSectorForAtomicity;
	DWORD	Flags;
	DWORD	ByteOffsetForSectorAlignment;
	DWORD	ByteOffsetForPartitionAlignment;
} FILE_FS_SECTOR_SIZE_INFORMATION, *LPFILE_FS_SECTOR_SIZE_INFORMATION, *PFILE_FS_SECTOR_SIZE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540282.aspx

typedef	struct	_FILE_FS_SIZE_INFORMATION
{
	LARGE_INTEGER	TotalAllocationUnits;
	LARGE_INTEGER	AvailableAllocationUnits;
	DWORD		SectorsPerAllocationUnit;
	DWORD		BytesPerSector;
} FILE_FS_SIZE_INFORMATION, *LPFILE_FS_SIZE_INFORMATION, *PFILE_FS_SIZE_INFORMATION;

// https://msdn.microsoft.com/en-us/library/ff540287.aspx

typedef	struct	_FILE_FS_VOLUME_INFORMATION
{
	LARGE_INTEGER	VolumeCreationTime;
	DWORD		VolumeSerialNumber;
	DWORD		VolumeLabelLength;
	BOOLEAN		SupportsObjects;
	WCHAR		VolumeLabel[ANYSIZE_ARRAY];
} FILE_FS_VOLUME_INFORMATION, *LPFILE_FS_VOLUME_INFORMATION, *PFILE_FS_VOLUME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/hh439282.aspx

typedef	enum	_FS_INFORMATION_CLASS
{
	FileFsVolumeInformation      = 1,
	FileFsLabelInformation       = 2,
	FileFsSizeInformation        = 3,
	FileFsDeviceInformation      = 4,
	FileFsAttributeInformation   = 5,
	FileFsControlInformation     = 6,
	FileFsFullSizeInformation    = 7,
	FileFsObjectIdInformation    = 8,
	FileFsDriverPathInformation  = 9,
	FileFsVolumeFlagsInformation = 10,
	FileFsSectorSizeInformation  = 11
} FS_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff567070.aspx

__declspec(dllimport)
LONG	NTAPI	NtQueryVolumeInformationFile(HANDLE               FileHandle,
		                             IO_STATUS_BLOCK      *IoStatusBlock,
		                             LPVOID               FsInformation,
		                             DWORD                FsInformationLength,
		                             FS_INFORMATION_CLASS FsInformationClass);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtRaiseException(EXCEPTION_RECORD *ExceptionRecord,
		                 CONTEXT          *ContextRecord,
		                 BOOLEAN          FirstChance);

#define MAXIMUM_HARDERROR_PARAMETERS	5

#define HARDERROR_PARAMETERS_FLAGSPOS	4
#define HARDERROR_FLAGS_DEFDESKTOPONLY	0x00020000UL

// NOTE: Parameters[3] holds the timeout in milliseconds!

#define HARDERROR_OVERRIDE_ERRORMODE	0x10000000UL

typedef	enum	_HARDERROR_RESPONSE_OPTION
{
	OptionAbortRetryIgnore = 0,
	OptionOk,
	OptionOkCancel,
	OptionRetryCancel,
	OptionYesNo,
	OptionYesNoCancel,
	OptionShutdownSystem,
	OptionOkNoWait,		// balloon notification!
	OptionCancelTryContinue
} HARDERROR_RESPONSE_OPTION;

typedef	enum	_HARDERROR_RESPONSE
{
	ResponseReturnToCaller = 0,
	ResponseNotHandled,
	ResponseAbort,
	ResponseCancel,
	ResponseIgnore,
	ResponseNo,
	ResponseOk,
	ResponseRetry,
	ResponseYes,
	ResponseTryAgain,
	ResponseContinue
} HARDERROR_RESPONSE;

__declspec(dllimport)
LONG	NTAPI	NtRaiseHardError(NTSTATUS                  ErrorStatus,
		                 DWORD                     NumberOfParameters,
		                 DWORD                     UnicodeStringParameterMask,
		                 DWORD_PTR                 *Parameters,
		                 HARDERROR_RESPONSE_OPTION ValidResponseOptions,
		                 HARDERROR_RESPONSE        *Response);

// https://msdn.microsoft.com/en-us/library/ff556706.aspx
// https://msdn.microsoft.com/en-us/library/mt708866.aspx
// https://msdn.microsoft.com/en-us/library/ff567072.aspx

__declspec(dllimport)
LONG	NTAPI	NtReadFile(HANDLE          FileHandle,
		           HANDLE          EventHandle,
		           IO_APC_ROUTINE  *ApcRoutine,		// optional
		           LPVOID          ApcContext,
		           IO_STATUS_BLOCK *IoStatusBlock,
		           LPVOID          Buffer,
		           DWORD           NumberOfBytesToRead,
		           LARGE_INTEGER   *ByteOffset,		// optional
		           LPDWORD         Key);		// optional

// NOTE: like KERNEL32!ReadProcessMemory

__declspec(dllimport)
LONG	NTAPI	NtReadVirtualMemory(HANDLE ProcessHandle,
		                    LPVOID BaseAddress,
		                    LPVOID Buffer,
		                    SIZE_T NumberOfBytesToRead,
		                    SIZE_T *NumberOfBytesRead);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567074.aspx

__declspec(dllimport)
LONG	NTAPI	NtReadOnlyEnlistment(HANDLE        EnlistmentHandle,
		                     LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567075.aspx

__declspec(dllimport)
LONG	NTAPI	NtRecoverEnlistment(HANDLE EnlistmentHandle,
		                    LPVOID EnlistmentKey);

// https://msdn.microsoft.com/en-us/library/ff567078.aspx

__declspec(dllimport)
LONG	NTAPI	NtRecoverResourceManager(HANDLE ResourceManagerHandle);

// https://msdn.microsoft.com/en-us/library/ff567079.aspx

__declspec(dllimport)
LONG	NTAPI	NtRecoverTransactionManager(HANDLE TransactionManagerHandle);

// https://msdn.microsoft.com/en-us/library/ff567081.aspx

__declspec(dllimport)
LONG	NTAPI	NtRollbackComplete(HANDLE        EnlistmentHandle,
		                   LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567083.aspx

__declspec(dllimport)
LONG	NTAPI	NtRollbackEnlistment(HANDLE        EnlistmentHandle,
		                     LARGE_INTEGER *TmVirtualClock);

// https://msdn.microsoft.com/en-us/library/ff567086.aspx

__declspec(dllimport)
LONG	NTAPI	NtRollbackTransaction(HANDLE  TransactionHandle,
		                      BOOLEAN Wait);

// https://msdn.microsoft.com/en-us/library/ff567089.aspx

__declspec(dllimport)
LONG	NTAPI	NtRollforwardTransactionManager(HANDLE        TransactionManagerHandle,
		                                LARGE_INTEGER *TmVirtualClock);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtRegisterThreadTerminatePort(HANDLE PortHandle);

__declspec(dllimport)
LONG	NTAPI	NtReleaseMutant(HANDLE MutantHandle,
		                LPLONG PreviousCount);

// NOTE: like KERNEL32!ReleaseSemaphore

__declspec(dllimport)
LONG	NTAPI	NtReleaseSemaphore(HANDLE  SemaphoreHandle,
		                   DWORD   ReleaseCount,
		                   LPDWORD PreviousCount);

__declspec(dllimport)
LONG	NTAPI	NtRemoveIoCompletion(HANDLE          IoCompletionPortHandle,
		                     LPDWORD         CompletionKey,
		                     LPDWORD         CompletionValue,
		                     IO_STATUS_BLOCK *IoStatusBlock,
		                     LARGE_INTEGER   *Timeout);

#if _WIN32_WINNT >= 0x0600
typedef	struct	_FILE_IO_COMPLETION_INFORMATION
{
	LPVOID		KeyContext;
	LPVOID		ApcContext;
	IO_STATUS_BLOCK	IoStatusBlock;
} FILE_IO_COMPLETION_INFORMATION, *LPFILE_IO_COMPLETION_INFORMATION, *PFILE_IO_COMPLETION_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	NtRemoveIoCompletionEx(HANDLE                         IoCompletionHandle,
		                       FILE_IO_COMPLETION_INFORMATION *IoCompletionInformation,
		                       DWORD                          Count,
		                       LPDWORD                        NumEntriesRemoved,
		                       LARGE_INTEGER                  *Timeout,
		                       BOOLEAN                        Alertable);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	NtRemoveProcessDebug(HANDLE ProcessHandle,
		                     HANDLE DebugObjectHandle);

// https://msdn.microsoft.com/en-us/library/cc512138.aspx

__declspec(dllimport)
LONG	NTAPI	NtRenameKey(HANDLE         KeyHandle,
		            UNICODE_STRING *NewName);

__declspec(dllimport)
LONG	NTAPI	NtReplaceKey(OBJECT_ATTRIBUTES *NewFile,
		             HANDLE            TargetHandle,
		             OBJECT_ATTRIBUTES *OldFile);

__declspec(dllimport)
LONG	NTAPI	NtResetEvent(HANDLE EventHandle,
		             LPLONG PreviousState);

__declspec(dllimport)
LONG	NTAPI	NtRestoreKey(HANDLE KeyHandle,
		             HANDLE FileHandle,
		             DWORD  Flags);

__declspec(dllimport)
LONG	NTAPI	NtResumeProcess(HANDLE ProcessHandle);

__declspec(dllimport)
LONG	NTAPI	NtResumeThread(HANDLE  ThreadHandle,
		               LPDWORD PreviousSuspendCount);

__declspec(dllimport)
LONG	NTAPI	NtSaveKey(HANDLE KeyHandle,
		          HANDLE FileHandle);

__declspec(dllimport)
LONG	NTAPI	NtSaveKeyEx(HANDLE KeyHandle,
		            HANDLE FileHandle,
		            DWORD  Format);

__declspec(dllimport)
LONG	NTAPI	NtSaveMergedKeys(HANDLE HighPrecedenceKeyHandle,
		                 HANDLE LowPrecedenceKeyHandle,
		                 HANDLE FileHandle);

__declspec(dllimport)
LONG	NTAPI	NtSetBootEntryOrder(LPDWORD Ids,
		                    DWORD   Count);

__declspec(dllimport)
LONG	NTAPI	NtSetBootOptions(BOOT_OPTIONS *BootOptions,
		                 DWORD        FieldsToChange);

__declspec(dllimport)
LONG	NTAPI	NtSetContextThread(HANDLE  ThreadHandle,
		                   CONTEXT *ContextRecord);

__declspec(dllimport)
LONG	NTAPI	NtSetDebugFilterState(DWORD   ComponentId,
		                      DWORD   Level,
		                      BOOLEAN State);

__declspec(dllimport)
LONG	NTAPI	NtSetDefaultLocale(BOOLEAN UserProfile,
		                   LCID    DefaultLocaleId);

__declspec(dllimport)
LONG	NTAPI	NtSetDefaultUILanguage(LANGID LanguageId);

// https://msdn.microsoft.com/en-us/library/ff961908.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetEaFile(HANDLE          FileHandle,
		            IO_STATUS_BLOCK *IoStatusBlock,
		            LPVOID          Buffer,
		            DWORD           Length);

// https://msdn.microsoft.com/en-us/library/ff567092.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetEvent(HANDLE EventHandle,
		           LPLONG PreviousState);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	NTAPI	NtSetEventBoostPriority(HANDLE EventHandle);
#endif

typedef	enum	_DEBUGOBJECTINFOCLASS
{
	DebugObjectFlags        = 1,
	MaxDebugObjectInfoClass
} DEBUGOBJECTINFOCLASS;

__declspec(dllimport)
LONG	NTAPI	NtSetInformationDebugObject(HANDLE               DebugObjectHandle,
		                            DEBUGOBJECTINFOCLASS DebugObjectInformationClass,
		                            LPVOID               DebugInformation,
		                            DWORD                DebugInformationLength,
		                            LPDWORD              ReturnLength);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567094.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetInformationEnlistment(HANDLE                       EnlistmentHandle,
		                           ENLISTMENT_INFORMATION_CLASS EnlistmentInformationClass,
		                           LPVOID                       EnlistmentInformation,
		                           DWORD                        EnlistmentInformationLength);
#endif

// https://msdn.microsoft.com/en-us/library/ff557671.aspx
// https://msdn.microsoft.com/en-us/library/ff567096.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetInformationFile(HANDLE                 FileHandle,
		                     IO_STATUS_BLOCK        *IoStatusBlock,
		                     LPVOID                 FileInformation,
		                     DWORD                  FileInformationLength,
		                     FILE_INFORMATION_CLASS FileInformationClass);

// NOTE: like KERNEL32!SetInformationJobObject

__declspec(dllimport)
LONG	NTAPI	NtSetInformationJobObject(HANDLE             JobHandle,
		                          JOBOBJECTINFOCLASS JobObjectInformationClass,
		                          LPVOID             JobObjectInformation,
		                          DWORD              JobObjectInformationLength);

// https://msdn.microsoft.com/en-us/library/ff553399.aspx

typedef	enum	_KEY_SET_INFORMATION_CLASS
{
	KeyWriteTimeInformation         = 0,
	KeyWow64FlagsInformation,
	KeyControlFlagsInformation,
	KeySetVirtualizationInformation,
	KeySetDebugInformation,
	KeySetHandleTagsInformation,
	MaxKeySetInfoClass
} KEY_SET_INFORMATION_CLASS;

// https://msdn.microsoft.com/en-us/library/ff554226.aspx

typedef	struct	_KEY_WRITE_TIME_INFORMATION
{
	LARGE_INTEGER	LastWriteTime;
} KEY_WRITE_TIME_INFORMATION, *LPKEY_WRITE_TIME_INFORMATION, *PKEY_WRITE_TIME_INFORMATION;

// https://msdn.microsoft.com/en-us/library/cc512139.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetInformationKey(HANDLE                    KeyHandle,
		                    KEY_SET_INFORMATION_CLASS KeySetInformationClass,
		                    LPVOID                    KeySetInformation,
		                    DWORD                     KeySetInformationLength);

__declspec(dllimport)
LONG	NTAPI	NtSetInformationObject(HANDLE                   ObjectHandle,
		                       OBJECT_INFORMATION_CLASS ObjectInformationClass,
		                       LPVOID                   ObjectInformation,
		                       DWORD                    ObjectInformationLength);

__declspec(dllimport)
LONG	NTAPI	NtSetInformationProcess(HANDLE           ProcessHandle,
		                        PROCESSINFOCLASS ProcessInformationClass,
		                        LPVOID           ProcessInformation,
		                        DWORD            ProcessInformationLength);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567098.aspx

__declspec(dllimport)
VOID	NTAPI	NtSetInformationResourceManager(VOID);
#endif

// https://msdn.microsoft.com/en-us/library/ff557675.aspx
// https://msdn.microsoft.com/en-us/library/ff567101.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetInformationThread(HANDLE          ThreadHandle,
		                       THREADINFOCLASS ThreadInformationClass,
		                       LPVOID          ThreadInformation,
		                       DWORD           ThreadInformationLength);

// https://msdn.microsoft.com/en-us/library/ff557678.aspx
// https://msdn.microsoft.com/en-us/library/ff567102.aspx
// NOTE: like ADVAPI32!SetTokenInformation and KERNEL32!SetTokenInformation or KERNELBASE!SetTokenInformation

__declspec(dllimport)
LONG	NTAPI	NtSetInformationToken(HANDLE                  TokenHandle,
		                      TOKEN_INFORMATION_CLASS TokenInformationClass,
		                      LPVOID                  TokenInformation,
		                      DWORD                   TokenInformationLength);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567104.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetInformationTransaction(HANDLE                        TransactionHandle,
		                            TRANSACTION_INFORMATION_CLASS TransactionInformationClass,
		                            LPVOID                        TransactionInformation,
		                            DWORD                         TransactionInformationLength);
#endif

__declspec(dllimport)
LONG	NTAPI	NtSetIntervalProfile(DWORD           Interval,
		                     KPROFILE_SOURCE ProfileSource);

__declspec(dllimport)
LONG	NTAPI	NtSetIoCompletion(HANDLE   IoCompletionPortHandle,
		                  DWORD    CompletionKey,
		                  DWORD    CompletionValue,
		                  NTSTATUS CompletionStatus,
		                  DWORD    CompletionInformation);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	NtSetIoCompletionEx(HANDLE    IoCompletionHandle,
		                    HANDLE    IoCompletionPacketHandle,
		                    LPVOID    KeyContext,
		                    LPVOID    ApcContext,
		                    NTSTATUS  IoStatus,
		                    DWORD_PTR IoStatusInformation);
#endif

// https://msdn.microsoft.com/en-us/library/ff567105.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetQuotaInformationFile(HANDLE          FileHandle,
		                          IO_STATUS_BLOCK *IoStatusBlock,
		                          LPVOID          Buffer,
		                          DWORD           Length);

// https://msdn.microsoft.com/en-us/library/ff567106.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetSecurityObject(HANDLE               ObjectHandle,
		                    SECURITY_INFORMATION SecurityInformation,
		                    SECURITY_DESCRIPTOR  *SecurityDescriptor);

__declspec(dllimport)
LONG	NTAPI	NtSetSystemEnvironmentValue(UNICODE_STRING *VariableName,
		                            UNICODE_STRING Value);

__declspec(dllimport)
LONG	NTAPI	NtSetSystemInformation(SYSTEM_INFORMATION_CLASS SystemInformationClass,
		                       LPVOID                   SystemInformation,
		                       DWORD                    SystemInformationLength);

__declspec(dllimport)
LONG	NTAPI	NtSetSystemTime(LARGE_INTEGER *SystemTime,
		                LARGE_INTEGER *PreviousSystemTime);

typedef	VOID	(NTAPI	TIMER_APC_ROUTINE) (LPVOID TimerContext,
			                    DWORD  TimerLowValue,
			                    LONG   TimerHighValue);

__declspec(dllimport)
LONG	NTAPI	NtSetTimer(HANDLE            TimerHandle,
		           LARGE_INTEGER     *DueTime,
		           TIMER_APC_ROUTINE *TimerApcRoutine,
		           LPVOID            TimerContext,
		           BOOLEAN           ResumeTimer,
		           LONG              Period,
		           BOOLEAN           *PreviousState);

__declspec(dllimport)
LONG	NTAPI	NtSetTimerResolution(DWORD   DesiredResolution,
		                     BOOLEAN Set,
		                     LPDWORD CurrentResolution);

__declspec(dllimport)
LONG	NTAPI	NtSetUuidSeed(LPBYTE Seed);

// https://msdn.microsoft.com/en-us/library/ff557688.aspx
// https://msdn.microsoft.com/en-us/library/ff567109.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetValueKey(HANDLE         KeyHandle,
		              UNICODE_STRING *ValueName,
		              DWORD          TitleIndex,
		              DWORD          Type,
		              LPVOID         Data,
		              DWORD          DataSize);

#if _WIN32_WINNT >= 0x0501
// https://msdn.microsoft.com/en-us/library/ff567112.aspx

__declspec(dllimport)
LONG	NTAPI	NtSetVolumeInformationFile(HANDLE               FileHandle,
		                           IO_STATUS_BLOCK      *IoStatusBlock,
		                           LPVOID               FsInformation,
		                           DWORD                FsInformationLength,
		                           FS_INFORMATION_CLASS FsInformationClass);
#endif

typedef	enum	_SHUTDOWN_ACTION
{
	ShutdownNoReboot = 0,
	ShutdownReboot,
	ShutdownPowerOff
} SHUTDOWN_ACTION;

__declspec(dllimport)
LONG	NTAPI	NtShutdownSystem(SHUTDOWN_ACTION Action);

__declspec(dllimport)
LONG	NTAPI	NtSignalAndWaitForSingleObject(HANDLE        ObjectToSignal,
		                               HANDLE        WaitableObject,
		                               BOOLEAN       Alertable,
		                               LARGE_INTEGER *Time);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff567113.aspx

__declspec(dllimport)
LONG	NTAPI	NtSinglePhaseReject(HANDLE        EnlistmentHandle,
		                    LARGE_INTEGER *TmVirtualClock);
#endif

__declspec(dllimport)
LONG	NTAPI	NtStartProfile(HANDLE ProfileHandle);

__declspec(dllimport)
LONG	NTAPI	NtStopProfile(HANDLE ProfileHandle);

__declspec(dllimport)
LONG	NTAPI	NtSuspendProcess(HANDLE ProcessHandle);

__declspec(dllimport)
LONG	NTAPI	NtSuspendThread(HANDLE  ThreadHandle,
		                LPDWORD PreviousSuspendCount);

typedef	enum	_SYSDBG_COMMAND
{
	SysDbgQueryModuleInformation  = 0,
	SysDbgQueryTraceInformation   = 1,
	SysDbgSetTracepoint           = 2,
	SysDbgSetSpecialCall          = 3,
	SysDbgClearSpecialCalls       = 4,
	SysDbgQuerySpecialCalls       = 5,
#if _WIN32_WINNT >= 0x0501
	SysDbgDbgBreakPointWithStatus = 6,
	SysDbgQueryVersion            = 7,
	SysDbgReadVirtual             = 8,
	SysDbgWriteVirtual            = 9,
	SysDbgReadPhysical            = 10,
	SysDbgWritePhysical           = 11,
	SysDbgReadControlSpace        = 12,
	SysDbgWriteControlSpace       = 13,
	SysDbgReadIoSpace             = 14,
	SysDbgWriteIoSpace            = 15,
	SysDbgReadMsr                 = 16,
	SysDbgWriteMsr                = 17,
	SysDbgReadBusData             = 18,
	SysDbgWriteBusData            = 19,
	SysDbgCheckLowMemory          = 20,
#if _WIN32_WINNT >= 0x0502
	SysDbgEnableKernelDebugger    = 21,
	SysDbgDisableKernelDebugger   = 22,
	SysDbgGetAutoEnableOnEvent    = 23,
	SysDbgSetAutoEnableOnEvent    = 24,
	SysDbgGetPrintBufferSize      = 25,
	SysDbgSetPrintBufferSize      = 26,
	SysDbgGetKdUmExceptionEnable  = 27,
	SysDbgSetKdUmExceptionEnable  = 28,
#if _WIN32_WINNT >= 0x0600
	SysDbgGetTriageDump           = 29,
	SysDbgGetKdBlockEnable        = 30,
	SysDbgSetKdBlockEnable        = 31,
	SysDbgRegisterForUmBreakInfo  = 32,
	SysDbgGetUmBreakPid           = 33,
	SysDbgClearUmBreakPid         = 34,
	SysDbgGetUmAttachPid          = 35,
	SysDbgClearUmAttachPid        = 36,
	SysDbgWriteLiveKernelDump     = 37,
#endif
#endif
#endif
	SysDbgMaxCommand
} SYSDBG_COMMAND, *LPSYSDBG_COMMAND, *PSYSDBG_COMMAND;

typedef	struct	_SYSDBG_VIRTUAL
{
	LPVOID	Address;
	LPVOID	Buffer;
	DWORD	Request;
} SYSDBG_VIRTUAL, *LPSYSDBG_VIRTUAL, *PSYSDBG_VIRTUAL;

typedef	LARGE_INTEGER	PHYSICAL_ADDRESS, *LPPHYSICAL_ADDRESS, *PPHYSICAL_ADDRESS;

typedef	struct	_SYSDBG_PHYSICAL
{
	PHYSICAL_ADDRESS	Address;
	LPVOID			Buffer;
	DWORD			Request;
} SYSDBG_PHYSICAL, *LPSYSDBG_PHYSICAL, *PSYSDBG_PHYSICAL;

typedef	struct	_SYSDBG_CONTROL_SPACE
{
	QWORD	Address;
	LPVOID	Buffer;
	DWORD	Request;
	DWORD	Processor;
} SYSDBG_CONTROL_SPACE, *LPSYSDBG_CONTROL_SPACE, *PSYSDBG_CONTROL_SPACE;

typedef	struct	_SYSDBG_IO_SPACE
{
	QWORD		Address;
	LPVOID		Buffer;
	DWORD		Request;
	INTERFACE_TYPE	InterfaceType;
	DWORD		BusNumber;
	DWORD		AddressSpace;
} SYSDBG_IO_SPACE, *LPSYSDBG_IO_SPACE, *PSYSDBG_IO_SPACE;

typedef	struct	_SYSDBG_MSR
{
	DWORD	Address;
	QWORD	Data;
} SYSDBG_MSR, *LPSYSDBG_MSR, *PSYSDBG_MSR;

typedef	enum	_BUS_DATA_TYPE
{
	ConfigurationSpaceUndefined = -1,
	Cmos,
	EisaConfiguration,
	Pos,
	CbusConfiguration,
	PCIConfiguration,
	VMEConfiguration,
	NuBusConfiguration,
	PCMCIAConfiguration,
	MPIConfiguration,
	MPSAConfiguration,
	PNPISAConfiguration,
	MaximumBusDataType
} BUS_DATA_TYPE, *LPBUS_DATA_TYPE, *PBUS_DATA_TYPE;

typedef	struct	_SYSDBG_BUS_DATA
{
	DWORD		Address;
	LPVOID		Buffer;
	DWORD		Request;
	BUS_DATA_TYPE	BusDataType;
	DWORD		BusNumber;
	DWORD		SlotNumber;
} SYSDBG_BUS_DATA, *LPSYSDBG_BUS_DATA, *PSYSDBG_BUS_DATA;

typedef	struct	_SYSDBG_TRIAGE_DUMP
{
	DWORD		Flags;
	DWORD		BugCheckCode;
	DWORD_PTR	BugCheckParam1;
	DWORD_PTR	BugCheckParam2;
	DWORD_PTR	BugCheckParam3;
	DWORD_PTR	BugCheckParam4;
	DWORD		ProcessHandles;
	DWORD		ThreadHandles;
	HANDLE		*Handles;
} SYSDBG_TRIAGE_DUMP, *LPSYSDBG_TRIAGE_DUMP, *PSYSDBG_TRIAGE_DUMP;

__declspec(dllimport)
LONG	NTAPI	NtSystemDebugControl(SYSDBG_COMMAND Command,
		                     LPVOID         InputBuffer,
		                     DWORD          InputBufferLength,
		                     LPVOID         OutputBuffer,
		                     DWORD          OutputBufferLength,
		                     LPDWORD        ReturnLength);

__declspec(dllimport)
LONG	NTAPI	NtTerminateJobObject(HANDLE   JobHandle,
		                     NTSTATUS ExitStatus);

// https://msdn.microsoft.com/en-us/library/ff557699.aspx
// https://msdn.microsoft.com/en-us/library/ff567115.aspx

__declspec(dllimport)
LONG	NTAPI	NtTerminateProcess(HANDLE   ProcessHandle,
		                   NTSTATUS ExitStatus);

__declspec(dllimport)
LONG	NTAPI	NtTerminateThread(HANDLE   ThreadHandle,
		                  NTSTATUS ExitStatus);

__declspec(dllimport)
LONG	NTAPI	NtTestAlert(VOID);

#if _WIN32_WINNT >= 0x0600
typedef	enum	_ETW_FUNCTION_CODE
{
	EtwStartLogger                        = 1,
	EtwStopLogger                         = 2,
	EtwQueryLogger                        = 3,
	EtwUpdateLogger                       = 4,
	EtwFlushLogger                        = 5,
	EtwIncrementTraceFile                 = 6,
#if _WIN32_WINNT >= 0x0603
	EtwRealtimeConnect                    = 11,
	EtwWdiDispatchControl                 = 13,
	EtwRealtimeDisconnectConsumerByHandle = 14,
	EtwRegisterUMGuid                     = 15,
	EtwReceiveNotification                = 16,
//	EtwSendNotification                   = 17,
	EtwSendDataBlock                      = 17,
	EtwSendReplyDataBlock                 = 18,
	EtwReceiveReplyDataBlock              = 19,
	EtwWdiUpdateSem                       = 20,
	EtwGetTraceGuidList                   = 21,
	EtwGetTraceGuidInfo                   = 22,
	EtwEnumerateTraceGuids                = 23,
	EtwQueryReferenceTime                 = 25,
	EtwTrackProviderBinary                = 26,
	EtwAddNotificationEvent               = 27,
	EtwUpdateDisallowList                 = 28,
	EtwUseDescriptorTypeUm                = 31,
	EtwGetTraceGroupList                  = 32,
	EtwGetTraceGroupInfo                  = 33,
	EtwGetDisallowList                    = 34,
	EtwSetCompressionSettings             = 35,
	EtwGetCompressionSettings             = 36,
	EtwUpdatePeriodicCaptureState         = 37,
	EtwGetPrivateSessionTraceHandle       = 38,
	EtwRegisterPrivateSession             = 39,
	EtwQuerySessionDemuxObject            = 40,
	EtwSetProviderBinaryTracking          = 41,
#endif
	MaxEtwFunction
} ETW_FUNCTION_CODE;

__declspec(dllimport)
LONG	NTAPI	NtTraceControl(DWORD   FunctionCode,
		               LPVOID  InputBuffer,
		               DWORD   InputBufferLength,
		               LPVOID  OutputBuffer,
		               DWORD   OutputBufferLength,
		               LPDWORD ReturnLength);
#endif

#define ETW_SYSTEM_EVENT_VERSION_MASK	0x000000FFUL
#define ETW_NT_TRACE_TYPE_MASK		0x0000FF00UL
#define ETW_USER_FRAMES_TO_SKIP_MASK	0x000F0000UL
#define ETW_NT_FLAGS_USE_NATIVE_HEADER	0x40000000UL
#define ETW_NT_FLAGS_WOW64_CALL		0x80000000UL

__declspec(dllimport)
LONG	NTAPI	NtTraceEvent(HANDLE TraceHandle,
		             DWORD  Flags,
		             DWORD  FieldSize,
		             LPVOID Fields);

__declspec(dllimport)
LONG	NTAPI	NtTranslateFilePath(FILE_PATH *InputFilePath,
		                    DWORD     OutputType,
		                    FILE_PATH *OutputFilePath,
		                    LPDWORD   OutputFilePathLength);

// https://msdn.microsoft.com/en-us/library/ff557701.aspx
// https://msdn.microsoft.com/en-us/library/ff567117.aspx

__declspec(dllimport)
LONG	NTAPI	NtUnloadDriver(UNICODE_STRING *DriverServiceName);

__declspec(dllimport)
LONG	NTAPI	NtUnloadKey(OBJECT_ATTRIBUTES *TargetKey);

__declspec(dllimport)
LONG	NTAPI	NtUnloadKeyEx(OBJECT_ATTRIBUTES *TargetKey,
		              HANDLE            EventHandle);

// https://msdn.microsoft.com/en-us/library/ff557707.aspx
// https://msdn.microsoft.com/en-us/library/ff567118.aspx

__declspec(dllimport)
LONG	NTAPI	NtUnlockFile(HANDLE          FileHandle,
		             IO_STATUS_BLOCK *IoStatusBlock,
		             LARGE_INTEGER   *ByteOffset,
		             LARGE_INTEGER   *Length,
		             DWORD           Key);

__declspec(dllimport)
LONG	NTAPI	NtUnlockVirtualMemory(HANDLE ProcessHandle,
		                      LPVOID *BaseAddress,
		                      SIZE_T *RegionSize,
		                      DWORD  MapType);

// https://msdn.microsoft.com/en-us/library/ff557711.aspx
// https://msdn.microsoft.com/en-us/library/ff567119.aspx

__declspec(dllimport)
LONG	NTAPI	NtUnmapViewOfSection(HANDLE ProcessHandle,
		                     LPVOID BaseAddress);

__declspec(dllimport)
LONG	NTAPI	NtWaitForDebugEvent(HANDLE                  DebugObjectHandle,
		                    BOOLEAN                 Alertable,
		                    LARGE_INTEGER           *Timeout,
		                    DBGUI_WAIT_STATE_CHANGE *WaitStateChange);

typedef	enum	_OBJECT_WAIT_TYPE
{
	WaitAllObject = 0,
	WaitAnyObject
} OBJECT_WAIT_TYPE;

__declspec(dllimport)
LONG	NTAPI	NtWaitForMultipleObjects(DWORD            ObjectCount,
		                         HANDLE           *ObjectHandles,
		                         OBJECT_WAIT_TYPE WaitType,
		                         BOOLEAN          Alertable,
		                         LARGE_INTEGER    *Timeout);

// https://msdn.microsoft.com/en-us/library/ff557712.aspx
// https://msdn.microsoft.com/en-us/library/ms648412.aspx
// https://msdn.microsoft.com/en-us/library/ff567120.aspx

__declspec(dllimport)
LONG	NTAPI	NtWaitForSingleObject(HANDLE        ObjectHandle,
		                      BOOLEAN       Alertable,
		                      LARGE_INTEGER *Timeout);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/dd405505.aspx

typedef	struct	_PROCESSOR_NUMBER
{
	WORD	Group;
	BYTE	Number;
	BYTE	Reserved;
} PROCESSOR_NUMBER, *LPPROCESSOR_NUMBER, *PPROCESSOR_NUMBER;
#endif

#ifndef _WIN64
#if _WIN32_WINNT >= 0x0602
__declspec(dllimport)
LONG	NTAPI	NtWow64AllocateVirtualMemory64(HANDLE   ProcessHandle,
		                               QWORD    BaseAddress,
		                               QWORD    ZeroBits,
		                               QWORD    *RegionSize,
		                               DWORD    AllocationType,
		                               DWORD    Protect);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtWow64CallFunction64(DWORD   FunctionIndex,
		                      DWORD   Flags,
		                      DWORD   InputLength,
		                      LPVOID  InputBuffer,
		                      DWORD   OutputLength,
		                      LPVOID  OutputBuffer,
		                      LPDWORD ReturnStatus);
#endif

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	NtWow64DebuggerCall(DWORD     ServiceClass,
		                    DWORD_PTR Argument1,
		                    DWORD_PTR Argument2,
		                    DWORD_PTR Argument3,
		                    DWORD_PTR Argument4);
#endif

#if _WIN32_WINNT == 0x0600
__declspec(deprecated("obsolete function, may be altered or unavailable in future versions"))
__declspec(dllimport)
DWORD	NTAPI	NtWow64GetCurrentProcessorNumber(VOID);
#endif

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	NtWow64GetCurrentProcessorNumberEx(PROCESSOR_NUMBER *ProcessorNumber);
#endif

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	NtWow64GetNativeSystemInformation(SYSTEM_INFORMATION_CLASS SystemInformationClass,
		                                  LPVOID                   SystemInformation,
		                                  DWORD                    SystemInformationLength,
		                                  LPDWORD                  ReturnLength);
#endif

#if _WIN32_WINNT == 0x0601
__declspec(dllimport)
LONG	NTAPI	NtWow64InterlockedPopEntrySList(SLIST_HEADER *ListHead);
#endif

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	NtWow64QueryInformationProcess64(HANDLE           ProcessHandle,
		                                 PROCESSINFOCLASS ProcessInformationClass,
		                                 LPVOID           ProcessInformation64,
		                                 DWORD            ProcessInformationLength,
		                                 LPDWORD          ReturnLength);
#endif

#if (_WIN32_WINNT >= 0x0502) && (_WIN32_WINNT <= 0x0603)
__declspec(dllimport)
LONG	NTAPI	NtWow64QueryVirtualMemory64(HANDLE                   ProcessHandle,
		                            QWORD                    BaseAddress,
		                            MEMORY_INFORMATION_CLASS MemoryInformationClass,
		                            LPVOID                   MemoryInformation,
		                            QWORD                    MemoryInformationLength,
		                            LPQWORD                  ReturnLength);
#endif

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	NtWow64ReadVirtualMemory64(HANDLE    ProcessHandle,
		                           QWORD     BaseAddress,
		                           LPVOID    Buffer,
		                           QWORD     BufferLength,
		                           LPQWORD   ReturnLength);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	NtWow64WriteVirtualMemory64(HANDLE    ProcessHandle,
		                            QWORD     BaseAddress,
		                            LPVOID    Buffer,
		                            QWORD     BufferLength,
		                            LPQWORD   ReturnLength);
#endif
#endif // _WIN64

// https://msdn.microsoft.com/en-us/library/ff557714.aspx
// https://msdn.microsoft.com/en-us/library/ff567121.aspx

__declspec(dllimport)
LONG	NTAPI	NtWriteFile(HANDLE          FileHandle,
		            HANDLE          EventHandle,
		            IO_APC_ROUTINE  *ApcRoutine,	// optional
		            LPVOID          ApcContext,
		            IO_STATUS_BLOCK *IoStatusBlock,
		            LPVOID          Buffer,
		            DWORD           NumberOfBytesToWrite,
		            LARGE_INTEGER   *ByteOffset,	// optional
		            LPDWORD         Key);		// optional

// NOTE: like KERNEL32!WriteProcessMemory

__declspec(dllimport)
LONG	NTAPI	NtWriteVirtualMemory(HANDLE ProcessHandle,
		                     LPVOID BaseAddress,
		                     LPVOID Buffer,
		                     SIZE_T NumberOfBytesToWrite,
		                     SIZE_T *NumberOfBytesWritten);

// NOTE: like KERNEL32!SwitchToThread

__declspec(dllimport)
LONG	NTAPI	NtYieldExecution(VOID);

typedef	struct	_RTL_RXACT_LOG
{
	DWORD	OperationCount;
	DWORD	LogSize;
	DWORD	LogSizeInUse;
#ifdef _WIN64
	DWORD	Alignment;
#endif
//	UCHAR	LogData[ANYSIZE_ARRAY];
} RTL_RXACT_LOG, *LPRTL_RXACT_LOG, *PRTL_RXACT_LOG;

typedef	struct	_RTL_RXACT_CONTEXT
{
	HANDLE		RootRegistryKey;
	HANDLE		RXactKey;
	BOOLEAN		HandlesValid;
	RTL_RXACT_LOG	*RXactLog;
} RTL_RXACT_CONTEXT, *LPRTL_RXACT_CONTEXT, *PRTL_RXACT_CONTEXT;

__declspec(dllimport)
LONG	NTAPI	RtlAbortRXact(RTL_RXACT_CONTEXT *RXactContext);

// https://msdn.microsoft.com/en-us/library/ff553354.aspx

// https://msdn.microsoft.com/en-us/library/ff552089.aspx
// NOTE: like ADVAPI32!MakeSelfRelativeSD

__declspec(dllimport)
LONG	NTAPI	RtlAbsoluteToSelfRelativeSD(SECURITY_DESCRIPTOR *AbsoluteSecurityDescriptor,
		                            SECURITY_DESCRIPTOR *SelfRelativeSecurityDescriptor,
		                            LPDWORD             BufferLength);

__declspec(dllimport)
VOID	NTAPI	RtlAcquirePebLock(VOID);

#define RTL_ACQUIRE_PRIVILEGE_REVERT	0x00000001UL
#define RTL_ACQUIRE_PRIVILEGE_PROCESS	0x00000002UL

__declspec(dllimport)
LONG	NTAPI	RtlAcquirePrivilege(DWORD  Privilege[],
		                    DWORD  PrivilegeCount,
		                    DWORD  Flags,
		                    LPVOID *ReturnedState);

__declspec(dllimport)
VOID	NTAPI	RtlReleasePrivilege(LPVOID ReturnedState);

#ifndef _WINNT_
typedef	struct	_RTL_CRITICAL_SECTION_DEBUG
{
	WORD		Type;
	WORD		CreatorBackTraceIndex;
	struct	_RTL_CRITICAL_SECTION	*CriticalSection;
	LIST_ENTRY	ProcessLocksList;
	DWORD		EntryCount;
	DWORD		ContentionCount;
	DWORD		Spare[2];
} RTL_CRITICAL_SECTION_DEBUG, *LPRTL_CRITICAL_SECTION_DEBUG, *PRTL_CRITICAL_SECTION_DEBUG, RTL_RESOURCE_DEBUG, *LPRTL_RESOURCE_DEBUG, *PRTL_RESOURCE_DEBUG;
#endif // _WINNT_

#define RTL_RESOURCE_FLAG_LONG_TERM	0x00000001UL

typedef	struct	_RTL_RESOURCE
{
	RTL_CRITICAL_SECTION	Lock;
	HANDLE			SharedSemaphore;
	DWORD			NumberOfWaitingShared;
	HANDLE			ExclusiveSemaphore;
	DWORD			NumberOfWaitingExclusive;
							// <0L: acquired exclusive
	LONG			NumberOfActive;		// =0L: available
							// >0L: acquired shared
	HANDLE			ExclusiveOwnerThread;
	DWORD			Flags;
	RTL_RESOURCE_DEBUG	*DebugInfo;
} RTL_RESOURCE, *LPRTL_RESOURCE, *PRTL_RESOURCE;

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAcquireResourceExclusive(RTL_RESOURCE *Resource,
		                            BOOLEAN      Wait);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAcquireResourceShared(RTL_RESOURCE *Resource,
		                         BOOLEAN      Wait);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/ms681930.aspx
// NOTE: forwarded from KERNEL32!AcquireSRWLockExclusive and KERNELBASE!AcquireSRWLockExclusive

__declspec(dllimport)
VOID	NTAPI	RtlAcquireSRWLockExclusive(RTL_SRWLOCK *SRWLock);

// https://msdn.microsoft.com/en-us/library/ms681934.aspx
// NOTE: forwarded from KERNEL32!AcquireSRWLockShared and KERNELBASE!AcquireSRWLockShared

__declspec(dllimport)
VOID	NTAPI	RtlAcquireSRWLockShared(RTL_SRWLOCK *SRWLock);
#endif // _WIN32_WINNT

#define ACTCTX_PROCESS_DEFAULT	((LPVOID) NULL)
#define ACTCTX_EMPTY		((LPVOID) (LONG_PTR) -3)
#define ACTCTX_SYSTEM_DEFAULT	((LPVOID) (LONG_PTR) -4)

#define INVALID_ACTIVATION_CONTEXT_ACTIVATION_COOKIE	((DWORD_PTR) 0)

__declspec(dllimport)
LONG	NTAPI	RtlActivateActivationContext(DWORD              Flags,
		                             ACTIVATION_CONTEXT *ActivationContext,
		                             DWORD_PTR          Cookie);

__declspec(dllimport)
LONG	NTAPI	RtlActivateActivationContextEx(DWORD              Flags,
		                               TEB                *Teb,
		                               ACTIVATION_CONTEXT *ActivationContext,
		                               DWORD_PTR          *Cookie);

typedef	struct	_RTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME
{
	SIZE_T	Size;
	DWORD	Format;
	RTL_ACTIVATION_CONTEXT_STACK_FRAME	Frame;
} RTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME, *LPRTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME, *PRTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME;

__declspec(dllimport)
VOID	FASTCALL	RtlActivateActivationContextUnsafeFast(RTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME *Frame,
			                                       ACTIVATION_CONTEXT                                  *ActivationContext);

// https://msdn.microsoft.com/en-us/library/ff552092.aspx
// NOTE: like ADVAPI32!AddAccessAllowedAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessAllowedAce(ACL         *Acl,
		                       DWORD       AceRevision,
		                       ACCESS_MASK AccessMask,
		                       SID         *Sid);

// https://msdn.microsoft.com/en-us/library/ff552095.aspx
// NOTE: like ADVAPI32!AddAccessAllowedAceEx

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessAllowedAceEx(ACL         *Acl,
		                         DWORD       AceRevision,
		                         DWORD       AceFlags,
		                         ACCESS_MASK AccessMask,
		                         SID         *Sid);

// NOTE: like ADVAPI32!AddAccessAllowedObjectAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessAllowedObjectAce(ACL         *Acl,
		                             DWORD       AceRevision,
		                             DWORD       AceFlags,
		                             ACCESS_MASK AccessMask,
		                             GUID        *ObjectTypeGuid,
		                             GUID        *InheritedObjectTypeGuid,
		                             SID         *Sid);

// NOTE: like ADVAPI32!AddAccessDeniedAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessDeniedAce(ACL         *Acl,
		                      DWORD       AceRevision,
		                      ACCESS_MASK AccessMask,
		                      SID         *Sid);

// NOTE: like ADVAPI32!AddAccessDeniedAceEx

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessDeniedAceEx(ACL         *Acl,
		                        DWORD       AceRevision,
		                        DWORD       AceFlags,
		                        ACCESS_MASK AccessMask,
		                        SID         *Sid);

// NOTE: like ADVAPI32!AddAccessDeniedObjectAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAccessDeniedObjectAce(ACL         *Acl,
		                            DWORD       AceRevision,
		                            DWORD       AceFlags,
		                            ACCESS_MASK AccessMask,
		                            GUID        *ObjectTypeGuid,
		                            GUID        *InheritedObjectTypeGuid,
		                            SID         *Sid);

// https://msdn.microsoft.com/en-us/library/ff552101.aspx
// NOTE: like ADVAPI32!AddAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAce(ACL    *Acl,
		          DWORD  AceRevision,
		          DWORD  StartingAceIndex,
		          LPVOID AceList,
		          DWORD  AceListLength);

typedef	enum	_RTL_RXACT_OPERATION
{
	RtlRXactOperationDelete       = 1,
	RtlRXactOperationSetValue,
	RtlRXactOperationDelAttribute,
	RtlRXactOperationSetAttribute
} RTL_RXACT_OPERATION, *LPRTL_RXACT_OPERATION, *PRTL_RXACT_OPERATION;

__declspec(dllimport)
LONG	NTAPI	RtlAddActionToRXact(RTL_RXACT_CONTEXT   *RXactContext,
		                    RTL_RXACT_OPERATION Operation,
		                    UNICODE_STRING      *SubKeyName,
		                    DWORD               NewKeyValueType,
		                    LPVOID              NewKeyValue,
		                    DWORD               NewKeyValueLength);

typedef	union	_RTL_HANDLE_TABLE_ENTRY
{
	DWORD	Flags;					// allocated entries have low bit set
	union	_RTL_HANDLE_TABLE_ENTRY	*NextFree;	// free entries use this pointer
} RTL_HANDLE_TABLE_ENTRY, *LPRTL_HANDLE_TABLE_ENTRY, *PRTL_HANDLE_TABLE_ENTRY;

typedef	struct	_RTL_HANDLE_TABLE
{
	DWORD			MaximumNumberOfHandles;
	DWORD			SizeOfHandleTableEntry;
	DWORD			Reserved[2];
	RTL_HANDLE_TABLE_ENTRY	*FreeHandles;
	RTL_HANDLE_TABLE_ENTRY	*CommittedHandles;
	RTL_HANDLE_TABLE_ENTRY	*UnCommittedHandles;
	RTL_HANDLE_TABLE_ENTRY	*MaxReservedHandles;
} RTL_HANDLE_TABLE, *LPRTL_HANDLE_TABLE, *PRTL_HANDLE_TABLE;

#define RTL_ATOM_PINNED	1

typedef	struct	_RTL_ATOM_TABLE_ENTRY
{
	struct	_RTL_ATOM_TABLE_ENTRY	*HashLink;
	WORD	HandleIndex;
	WORD	Atom;
	WORD	ReferenceCount;
	BYTE	Flags;
	BYTE	NameLength;
	WCHAR	Name[ANYSIZE_ARRAY];
} RTL_ATOM_TABLE_ENTRY, *LPRTL_ATOM_TABLE_ENTRY, *PRTL_ATOM_TABLE_ENTRY;

typedef	struct	_RTL_ATOM_TABLE
{
	DWORD			Signature;
	RTL_CRITICAL_SECTION	CriticalSection;
	RTL_HANDLE_TABLE	RtlHandleTable;
	DWORD			NumberOfBuckets;
	RTL_ATOM_TABLE_ENTRY	*Buckets[ANYSIZE_ARRAY];
} RTL_ATOM_TABLE, *LPRTL_ATOM_TABLE, *PRTL_ATOM_TABLE;

__declspec(dllimport)
LONG	NTAPI	RtlAddAtomToAtomTable(RTL_ATOM_TABLE *AtomTable,
		                      LPCWSTR        AtomName,
		                      RTL_ATOM       *Atom);

__declspec(dllimport)
LONG	NTAPI	RtlAddAttributeActionToRXact(RTL_RXACT_CONTEXT   *RXactContext,
		                             RTL_RXACT_OPERATION Operation,
		                             UNICODE_STRING      *SubKeyName,
		                             HANDLE              KeyHandle,
		                             UNICODE_STRING      *AttributeName,
		                             DWORD               NewValueType,
		                             LPVOID              NewValue,
		                             DWORD               NewValueLength);

// NOTE: like ADVAPI32!AddAuditAccessAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAuditAccessAce(ACL         *Acl,
		                     DWORD       AceRevision,
		                     ACCESS_MASK AccessMask,
		                     SID         *Sid,
		                     BOOLEAN     AuditSuccess,
		                     BOOLEAN     AuditFailure);

// NOTE: like ADVAPI32!AddAuditAccessAceEx

__declspec(dllimport)
LONG	NTAPI	RtlAddAuditAccessAceEx(ACL         *Acl,
		                       DWORD       AceRevision,
		                       DWORD       AceFlags,
		                       ACCESS_MASK AccessMask,
		                       SID         *Sid,
		                       BOOLEAN     AuditSuccess,
		                       BOOLEAN     AuditFailure);

// NOTE: like ADVAPI32!AddAuditAccessObjectAce

__declspec(dllimport)
LONG	NTAPI	RtlAddAuditAccessObjectAce(ACL         *Acl,
		                           DWORD       AceRevision,
		                           DWORD       AceFlags,
		                           ACCESS_MASK AccessMask,
		                           GUID        *ObjectTypeGuid,
		                           GUID        *InheritedObjectTypeGuid,
		                           SID         *Sid,
		                           BOOLEAN     AuditSuccess,
		                           BOOLEAN     AuditFailure);

__declspec(dllimport)
LONG	NTAPI	RtlAddCompoundAce(ACL         *Acl,
		                  DWORD       AceRevision,
		                  BYTE        CompoundAceType,
		                  ACCESS_MASK AccessMask,
		                  SID         *ServerSid,
		                  SID         *ClientSid);

#ifdef _WIN64
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms680588.aspx
// NOTE: forwarded from KERNEL32!RtlAddFunctionTable

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAddFunctionTable(RUNTIME_FUNCTION *FunctionTable,
		                    DWORD            EntryCount,
		                    QWORD            BaseAddress);
#endif

#if _WIN32_WINNT >= 0x0602
// https://msdn.microsoft.com/en-us/library/hh405352.aspx

__declspec(dllimport)
LONG	NTAPI	RtlAddGrowableFunctionTable(LPVOID           *DynamicTable,
		                            RUNTIME_FUNCTION *FunctionTable,
		                            DWORD            EntryCount,
		                            DWORD            MaximumEntryCount,
		                            DWORD_PTR        RangeBase,
		                            DWORD_PTR        RangeEnd);
#endif
#endif // _WIN64

#if _WIN32_WINNT >= 0x0601
// NOTE: like KERNEL32!AddIntegrityLabelToBoundaryDescriptor

__declspec(dllimport)
LONG	NTAPI	RtlAddIntegrityLabelToBoundaryDescriptor(LPVOID *BoundaryDescriptor,
		                                         SID    *IntegrityLabel);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlAddMandatoryAce(ACL   *Acl,
		                   DWORD AceRevision,
		                   DWORD AceFlags,
		                   DWORD MandatoryFlags,
		                   BYTE  AceType,
		                   SID   *LabelSid);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlAddRefActivationContext(ACTIVATION_CONTEXT *ActivationContext);

typedef	struct	_IMalloc	IMalloc;
typedef	struct	_IStream	IStream;
typedef	struct	_IStreamVtbl	IStreamVtbl;
typedef	struct	_STATSTG	STATSTG;

typedef	struct	_RTL_MEMORY_STREAM_WITH_VTABLE
{
	const	IStreamVtbl	*StreamVTable;
	LONG	ReferenceCount;
	DWORD	Flags;
	LPBYTE	Current;
	LPBYTE	Start;
	LPBYTE	End;
	VOID	(NTAPI *FinalRelease) (struct _RTL_MEMORY_STREAM_WITH_VTABLE *);
	IMalloc	*ReservedForMalloc;
} RTL_MEMORY_STREAM_WITH_VTABLE, *LPRTL_MEMORY_STREAM_WITH_VTABLE, *PRTL_MEMORY_STREAM_WITH_VTABLE;

typedef	struct	_RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE
{
	const	IStreamVtbl	*StreamVTable;
	LONG	ReferenceCount;
	DWORD	Flags;
	LPBYTE	Current;
	LPBYTE	Start;
	LPBYTE	End;
	VOID	(NTAPI *FinalRelease) (struct _RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE *);
	IMalloc	*ReservedForMalloc;
	HANDLE	ProcessHandle;
} RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE, *LPRTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE, *PRTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE;

__declspec(dllimport)
DWORD	NTAPI	RtlAddRefMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

__declspec(dllimport)
LPVOID	NTAPI	RtlAddressInSectionTable(IMAGE_NT_HEADERS *NtHeaders,
		                         IMAGE_DOS_HEADER *ImageBase,
		                         DWORD            VirtualAddress);

#if _WIN32_WINNT >= 0x0600
// NOTE: like KERNEL32!AddSIDToBoundaryDescriptor

__declspec(dllimport)
LONG	NTAPI	RtlAddSIDToBoundaryDescriptor(LPVOID *BoundaryDescriptor,
		                              SID    *Sid);
#endif

// https://msdn.microsoft.com/en-us/library/ms679273.aspx
// NOTE: forwarded from KERNEL32!AddVectoredContinueHandler and KERNELBASE!AddVectoredContinueHandler

__declspec(dllimport)
LPVOID	NTAPI	RtlAddVectoredContinueHandler(DWORD                       FirstHandler,
		                              PVECTORED_EXCEPTION_HANDLER VectoredHandler);

// https://msdn.microsoft.com/en-us/library/ms679274.aspx
// NOTE: forwarded from KERNEL32!AddVectoredExceptionHandler and KERNELBASE!AddVectoredExceptionHandler

__declspec(dllimport)
LPVOID	NTAPI	RtlAddVectoredExceptionHandler(DWORD                       FirstHandler,
		                               PVECTORED_EXCEPTION_HANDLER VectoredHandler);

#define SE_MIN_WELL_KNOWN_PRIVILEGE			2UL
#define SE_CREATE_TOKEN_PRIVILEGE			2UL	// "SeCreateTokenPrivilege"
#define SE_ASSIGNPRIMARYTOKEN_PRIVILEGE			3UL	// "SeAssignPrimaryTokenPrivilege"
#define SE_LOCK_MEMORY_PRIVILEGE			4UL	// "SeLockMemoryPrivilege"
#define SE_INCREASE_QUOTA_PRIVILEGE			5UL	// "SeIncreaseQuotaPrivilege"
//      SE_UNSOLICITED_INPUT_PRIVILEGE			6UL	// "SeUnsolicitedInputPrivilege"
#define SE_MACHINE_ACCOUNT_PRIVILEGE			6UL	// "SeMachineAccountPrivilege"
#define SE_TCB_PRIVILEGE				7UL	// "SeTcbPrivilege"
#define SE_SECURITY_PRIVILEGE				8UL	// "SeSecurityPrivilege"
#define SE_TAKE_OWNERSHIP_PRIVILEGE			9UL	// "SeTakeOwnershipPrivilege"
#define SE_LOAD_DRIVER_PRIVILEGE			10UL	// "SeLoadDriverPrivilege"
#define SE_SYSTEM_PROFILE_PRIVILEGE			11UL	// "SeSystemProfilePrivilege"
#define SE_SYSTEMTIME_PRIVILEGE				12UL	// "SeSystemtimePrivilege"
#define SE_PROF_SINGLE_PROCESS_PRIVILEGE		13UL	// "SeProfileSingleProcessPrivilege"
#define SE_INC_BASE_PRIORITY_PRIVILEGE			14UL	// "SeIncreaseBasePriorityPrivilege"
#define SE_CREATE_PAGEFILE_PRIVILEGE			15UL	// "SeCreatePagefilePrivilege"
#define SE_CREATE_PERMANENT_PRIVILEGE			16UL	// "SeCreatePermanentPrivilege"
#define SE_BACKUP_PRIVILEGE				17UL	// "SeBackupPrivilege"
#define SE_RESTORE_PRIVILEGE				18UL	// "SeRestorePrivilege"
#define SE_SHUTDOWN_PRIVILEGE				19UL	// "SeShutdownPrivilege"
#define SE_DEBUG_PRIVILEGE				20UL	// "SeDebugPrivilege"
#define SE_AUDIT_PRIVILEGE				21UL	// "SeAuditPrivilege"
#define SE_SYSTEM_ENVIRONMENT_PRIVILEGE			22UL	// "SeSystemEnvironmentPrivilege"
#define SE_CHANGE_NOTIFY_PRIVILEGE			23UL	// "SeChangeNotifyPrivilege"
#define SE_REMOTE_SHUTDOWN_PRIVILEGE			24UL	// "SeRemoteShutdownPrivilege"
#define SE_UNDOCK_PRIVILEGE				25UL	// "SeUndockPrivilege"
#define SE_SYNC_AGENT_PRIVILEGE				26UL	// "SeSyncAgentPrivilege"
#define SE_ENABLE_DELEGATION_PRIVILEGE			27UL	// "SeEnableDelegationPrivilege"
#define SE_MANAGE_VOLUME_PRIVILEGE			28UL	// "SeManageVolumePrivilege"
#define SE_IMPERSONATE_PRIVILEGE			29UL	// "SeImpersonatePrivilege"
#define SE_CREATE_GLOBAL_PRIVILEGE			30UL	// "SeCreateGlobalPrivilege"
#define SE_TRUSTED_CREDMAN_ACCESS_PRIVILEGE		31UL	// "SeTrustedCredManAccessPrivilege"
#define SE_RELABEL_PRIVILEGE				32UL	// "SeRelabelPrivilege"
#define SE_INCREASE_WORKING_SET_PRIVILEGE		33UL	// "SeIncreaseWorkingSetPrivilege"
#define SE_TIME_ZONE_PRIVILEGE				34UL	// "SeTimeZonePrivilege"
#define SE_CREATE_SYMBOLIC_LINK_PRIVILEGE		35UL	// "SeCreateSymbolicLinkPrivilege"
#define SE_DELEGATE_SESSION_USER_IMPERSONATE_PRIVILEGE	36UL	// "SeDelegateSessionUserImpersonatePrivilege"
#define SE_MAX_WELL_KNOWN_PRIVILEGE			36UL

__declspec(dllimport)
LONG	NTAPI	RtlAdjustPrivilege(DWORD   Privilege,
		                   BOOLEAN Enable,
		                   BOOLEAN CurrentThread,
		                   BOOLEAN *Enabled);

__declspec(dllimport)
LONG	NTAPI	RtlAllocateActivationContextStack(ACTIVATION_CONTEXT_STACK *ActivationContextStack);

// https://msdn.microsoft.com/en-us/library/ff552103.aspx
// NOTE: like ADVAPI32!AllocateAndInitializeSid

__declspec(dllimport)
LONG	NTAPI	RtlAllocateAndInitializeSid(SID_IDENTIFIER_AUTHORITY *IdentifierAuthority,
		                            BYTE                     SubAuthorityCount,
		                            DWORD                    SubAuthority0,
		                            DWORD                    SubAuthority1,
		                            DWORD                    SubAuthority2,
		                            DWORD                    SubAuthority3,
		                            DWORD                    SubAuthority4,
		                            DWORD                    SubAuthority5,
		                            DWORD                    SubAuthority6,
		                            DWORD                    SubAuthority7,
		                            SID                      **Sid);

__declspec(dllimport)
RTL_HANDLE_TABLE_ENTRY *	NTAPI	RtlAllocateHandle(RTL_HANDLE_TABLE *HandleTable,
					                  LPDWORD          HandleIndex);

// https://msdn.microsoft.com/en-us/library/aa366597.aspx
// https://msdn.microsoft.com/en-us/library/ff552108.aspx
// NOTE: forwarded from KERNEL32!HeapAlloc and KERNELBASE!HeapAlloc

__declspec(dllimport)
LPVOID	NTAPI	RtlAllocateHeap(LPCVOID HeapHandle,
		                DWORD   Flags,
		                SIZE_T  Size);

// https://msdn.microsoft.com/en-us/library/ff561132.aspx

__declspec(dllimport)
WCHAR	NTAPI	RtlAnsiCharToUnicodeChar(UCHAR **SourceCharacter);

__declspec(dllimport)
LONG	NTAPI	RtlxAnsiStringToUnicodeSize(ANSI_STRING *AnsiString);

// https://msdn.microsoft.com/en-us/library/ff561725.aspx

#if 1
__declspec(dllimport)
DWORD	NTAPI	RtlAnsiStringToUnicodeSize(ANSI_STRING *AnsiString);
#else
__inline
DWORD	NTAPI	RtlAnsiStringToUnicodeSize(ANSI_STRING *AnsiString)
{
	if (*NlsMbCodePageTag)
		return RtlxAnsiStringToUnicodeSize(AnsiString);

	return (AnsiString->Length + 1) * sizeof(WCHAR);
}
#endif

// https://msdn.microsoft.com/en-us/library/ff561729.aspx
// https://msdn.microsoft.com/en-us/library/ms648413.aspx

__declspec(dllimport)
LONG	NTAPI	RtlAnsiStringToUnicodeString(UNICODE_STRING *DestinationString,
		                             ANSI_STRING    *SourceString,
		                             BOOLEAN        AllocateDestinationString);

__declspec(dllimport)
LONG	NTAPI	RtlAppendAsciizToString(STRING *DestinationString,
		                        LPCSTR SourceString);

typedef	struct	_RTL_BUFFER
{
	UCHAR	*Buffer;
	UCHAR	*StaticBuffer;
	SIZE_T	Size;
	SIZE_T	StaticSize;
	SIZE_T	ReservedForAllocatedSize;
	LPVOID	ReservedForIMalloc;
} RTL_BUFFER, *LPRTL_BUFFER, *PRTL_BUFFER;

typedef	struct	_RTL_UNICODE_STRING_BUFFER
{
	UNICODE_STRING	String;
	RTL_BUFFER	ByteBuffer;
	UCHAR		MinimumStaticBufferForTerminalNul[sizeof(WCHAR)];
} RTL_UNICODE_STRING_BUFFER, *LPRTL_UNICODE_STRING_BUFFER, *PRTL_UNICODE_STRING_BUFFER;

#define RTL_APPEND_PATH_ELEMENT_ONLY_BACKSLASH_IS_SEPARATOR							1UL
#define RTL_APPEND_PATH_ELEMENT_BUGFIX_CHECK_FIRST_THREE_CHARS_FOR_SLASH_TAKE_FOUND_SLASH_INSTEAD_OF_FIRST_CHAR	2UL

__declspec(dllimport)
LONG	NTAPI	RtlAppendPathElement(DWORD                     Flags,
		                     RTL_UNICODE_STRING_BUFFER *Path,
		                     UNICODE_STRING            *Element);

// https://msdn.microsoft.com/en-us/library/ff552113.aspx

__declspec(dllimport)
LONG	NTAPI	RtlAppendStringToString(STRING *DestinationString,
		                        STRING *SourceString);

// https://msdn.microsoft.com/en-us/library/ff561736.aspx

__declspec(dllimport)
LONG	NTAPI	RtlAppendUnicodeStringToString(UNICODE_STRING *DestinationString,
		                               UNICODE_STRING *SourceString);

// https://msdn.microsoft.com/en-us/library/ff561738.aspx

__declspec(dllimport)
LONG	NTAPI	RtlAppendUnicodeToString(UNICODE_STRING *DestinationString,
		                         LPCWSTR        SourceString);

#ifndef DLL_PROCESS_VERIFIER
#define DLL_PROCESS_VERIFIER	4

typedef	struct	_RTL_VERIFIER_THUNK_DESCRIPTOR
{
	LPSTR	ThunkName;
	LPVOID	ThunkOldAddress;
	LPVOID	ThunkNewAddress;
} RTL_VERIFIER_THUNK_DESCRIPTOR, *LPRTL_VERIFIER_THUNK_DESCRIPTOR, *PRTL_VERIFIER_THUNK_DESCRIPTOR;

typedef	struct	_RTL_VERIFIER_DLL_DESCRIPTOR
{
	LPWSTR	DllName;
	DWORD	DllFlags;
	LPVOID	DllAddress;
	RTL_VERIFIER_THUNK_DESCRIPTOR	*DllThunks;
} RTL_VERIFIER_DLL_DESCRIPTOR, *LPRTL_VERIFIER_DLL_DESCRIPTOR, *PRTL_VERIFIER_DLL_DESCRIPTOR;

typedef	VOID	(NTAPI	RTL_VERIFIER_DLL_LOAD_CALLBACK) (LPWSTR DllName,
			                                 LPVOID DllBase,
			                                 SIZE_T DllSize,
			                                 LPVOID Reserved);

typedef	VOID	(NTAPI	RTL_VERIFIER_DLL_UNLOAD_CALLBACK) (LPWSTR DllName,
			                                   LPVOID DllBase,
			                                   SIZE_T DllSize,
			                                   LPVOID Reserved);

typedef	VOID	(NTAPI	RTL_VERIFIER_NTDLLHEAPFREE_CALLBACK) (LPVOID AllocationBase,
			                                      SIZE_T AllocationSize);

#define RTL_VRF_FLG_FULL_PAGE_HEAP			0x00000001UL
#define RTL_VRF_FLG_RESERVED_DONOTUSE			0x00000002UL	// old RTL_VRF_FLG_LOCK_CHECKS
#define RTL_VRF_FLG_HANDLE_CHECKS			0x00000004UL
#define RTL_VRF_FLG_STACK_CHECKS			0x00000008UL
#define RTL_VRF_FLG_APPCOMPAT_CHECKS			0x00000010UL
#define RTL_VRF_FLG_TLS_CHECKS				0x00000020UL
#define RTL_VRF_FLG_DIRTY_STACKS			0x00000040UL
#define RTL_VRF_FLG_RPC_CHECKS				0x00000080UL
#define RTL_VRF_FLG_COM_CHECKS				0x00000100UL
#define RTL_VRF_FLG_DANGEROUS_APIS			0x00000200UL
#define RTL_VRF_FLG_RACE_CHECKS				0x00000400UL
#define RTL_VRF_FLG_DEADLOCK_CHECKS			0x00000800UL
#define RTL_VRF_FLG_FIRST_CHANCE_EXCEPTION_CHECKS	0x00001000UL
#define RTL_VRF_FLG_VIRTUAL_MEM_CHECKS			0x00002000UL
#define RTL_VRF_FLG_ENABLE_LOGGING			0x00004000UL
#define RTL_VRF_FLG_FAST_FILL_HEAP			0x00008000UL
#define RTL_VRF_FLG_VIRTUAL_SPACE_TRACKING		0x00010000UL
#define RTL_VRF_FLG_ENABLED_SYSTEM_WIDE			0x00020000UL
#define RTL_VRF_FLG_MISCELLANEOUS_CHECKS		0x00020000UL
#define RTL_VRF_FLG_LOCK_CHECKS				0x00040000UL

#define RTL_VRF_DBG_SHOW_SNAPS				0x00000001UL
#define RTL_VRF_DBG_SHOW_FOUND_EXPORTS			0x00000002UL
#define RTL_VRF_DBG_SHOW_VERIFIED_EXPORTS		0x00000004UL
#define RTL_VRF_DBG_LIST_PROVIDERS			0x00000008UL
#define RTL_VRF_DBG_SHOW_CHAINING			0x00000010UL
#define RTL_VRF_DBG_SHOW_CHAINING_DEBUG			0x00000020UL
#define RTL_VRF_DBG_CS_SPLAYTREE			0x00000200UL
#define RTL_VRF_DBG_CS_DUMP_SPLAYTREE			0x00000400UL
#define RTL_VRF_DBG_CS_CREATE_DELETE			0x00000800UL
#define RTL_VRF_DBG_VERIFIER_LOG_CALLS			0x00004000UL
#define RTL_VRF_DBG_VERIFIER_SHOW_DYN_THUNKS		0x00008000UL
#define RTL_VRF_DBG_ENTRYPOINT_HOOKS			0x00010000UL
#define RTL_VRF_DBG_ENTRYPOINT_CALLS			0x00020000UL

typedef	struct	_RTL_VERIFIER_PROVIDER_DESCRIPTOR
{
	// to be filled by verifier provider DLL
	DWORD					Length;
	RTL_VERIFIER_DLL_DESCRIPTOR		*ProviderDlls;
	RTL_VERIFIER_DLL_LOAD_CALLBACK		*ProviderDllLoadCallback;
	RTL_VERIFIER_DLL_UNLOAD_CALLBACK	*ProviderDllUnloadCallback;

	// filled by verifier engine
	LPWSTR	VerifierImage;
	DWORD	VerifierFlags;
	DWORD	VerifierDebug;

	LPVOID	RtlpGetStackTraceAddress;
	LPVOID	RtlpDebugPageHeapCreate;
	LPVOID	RtlpDebugPageHeapDestroy;

	// to be filled by verifier provider DLL
	RTL_VERIFIER_NTDLLHEAPFREE_CALLBACK	*ProviderNtdllHeapFreeCallback;
} RTL_VERIFIER_PROVIDER_DESCRIPTOR, *LPRTL_VERIFIER_PROVIDER_DESCRIPTOR, *PRTL_VERIFIER_PROVIDER_DESCRIPTOR;

#define APPLICATION_VERIFIER_CONTINUABLE_BREAK	0x10000000UL
#define APPLICATION_VERIFIER_NO_BREAK		0x20000000UL
#define APPLICATION_VERIFIER_INTERNAL_WARNING	0x40000000UL
#define APPLICATION_VERIFIER_INTERNAL_ERROR	0x80000000UL

#define APPLICATION_VERIFIER_UNKNOWN_ERROR						0x00000001UL
#define APPLICATION_VERIFIER_ACCESS_VIOLATION						0x00000002UL
#define APPLICATION_VERIFIER_UNSYNCHRONIZED_ACCESS					0x00000003UL
#define APPLICATION_VERIFIER_EXTREME_SIZE_REQUEST					0x00000004UL
#define APPLICATION_VERIFIER_BAD_HEAP_HANDLE						0x00000005UL
#define APPLICATION_VERIFIER_SWITCHED_HEAP_HANDLE					0x00000006UL
#define APPLICATION_VERIFIER_DOUBLE_FREE						0x00000007UL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK					0x00000008UL
#define APPLICATION_VERIFIER_DESTROY_PROCESS_HEAP					0x00000009UL
#define APPLICATION_VERIFIER_UNEXPECTED_EXCEPTION					0x0000000AUL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_EXCEPTION_RAISED_FOR_HEADER		0x0000000BUL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_EXCEPTION_RAISED_FOR_PROBING		0x0000000CUL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_HEADER				0x0000000DUL
#define APPLICATION_VERIFIER_CORRUPTED_FREED_HEAP_BLOCK					0x0000000EUL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_SUFFIX				0x0000000FUL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_START_STAMP				0x00000010UL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_END_STAMP				0x00000011UL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_BLOCK_PREFIX				0x00000012UL
#define APPLICATION_VERIFIER_FIRST_CHANCE_ACCESS_VIOLATION				0x00000013UL
#define APPLICATION_VERIFIER_CORRUPTED_HEAP_LIST					0x00000014UL

#define APPLICATION_VERIFIER_TERMINATE_THREAD_CALL		0x00000100UL
#define APPLICATION_VERIFIER_STACK_OVERFLOW			0x00000101UL
#define APPLICATION_VERIFIER_INVALID_EXIT_PROCESS_CALL		0x00000102UL

#define APPLICATION_VERIFIER_EXIT_THREAD_OWNS_LOCK		0x00000200UL
#define APPLICATION_VERIFIER_LOCK_IN_UNLOADED_DLL		0x00000201UL
#define APPLICATION_VERIFIER_LOCK_IN_FREED_HEAP			0x00000202UL
#define APPLICATION_VERIFIER_LOCK_DOUBLE_INITIALIZE		0x00000203UL
#define APPLICATION_VERIFIER_LOCK_IN_FREED_MEMORY		0x00000204UL
#define APPLICATION_VERIFIER_LOCK_CORRUPTED			0x00000205UL
#define APPLICATION_VERIFIER_LOCK_INVALID_OWNER			0x00000206UL
#define APPLICATION_VERIFIER_LOCK_INVALID_RECURSION_COUNT	0x00000207UL
#define APPLICATION_VERIFIER_LOCK_INVALID_LOCK_COUNT		0x00000208UL
#define APPLICATION_VERIFIER_LOCK_OVER_RELEASED			0x00000209UL
#define APPLICATION_VERIFIER_LOCK_NOT_INITIALIZED		0x00000210UL
#define APPLICATION_VERIFIER_LOCK_ALREADY_INITIALIZED		0x00000211UL
#define APPLICATION_VERIFIER_LOCK_IN_FREED_VMEM			0x00000212UL
#define APPLICATION_VERIFIER_LOCK_IN_UNMAPPED_MEM		0x00000213UL
#define APPLICATION_VERIFIER_THREAD_NOT_LOCK_OWNER		0x00000214UL

#define APPLICATION_VERIFIER_INVALID_HANDLE			0x00000300UL
#define APPLICATION_VERIFIER_INVALID_TLS_VALUE			0x00000301UL
#define APPLICATION_VERIFIER_INCORRECT_WAIT_CALL		0x00000302UL
#define APPLICATION_VERIFIER_NULL_HANDLE			0x00000303UL
#define APPLICATION_VERIFIER_WAIT_IN_DLLMAIN			0x00000304UL

#define APPLICATION_VERIFIER_COM_ERROR				0x00000400UL
#define APPLICATION_VERIFIER_COM_API_IN_DLLMAIN			0x00000401UL
#define APPLICATION_VERIFIER_COM_UNHANDLED_EXCEPTION		0x00000402UL
#define APPLICATION_VERIFIER_COM_UNBALANCED_COINIT		0x00000403UL
#define APPLICATION_VERIFIER_COM_UNBALANCED_OLEINIT		0x00000404UL
#define APPLICATION_VERIFIER_COM_UNBALANCED_SWC			0x00000405UL
#define APPLICATION_VERIFIER_COM_NULL_DACL			0x00000406UL
#define APPLICATION_VERIFIER_COM_UNSAFE_IMPERSONATION		0x00000407UL
#define APPLICATION_VERIFIER_COM_SMUGGLED_WRAPPER		0x00000408UL
#define APPLICATION_VERIFIER_COM_SMUGGLED_PROXY			0x00000409UL
#define APPLICATION_VERIFIER_COM_CF_SUCCESS_WITH_NULL		0x0000040AUL
#define APPLICATION_VERIFIER_COM_GCO_SUCCESS_WITH_NULL		0x0000040BUL
#define APPLICATION_VERIFIER_COM_OBJECT_IN_FREED_MEMORY		0x0000040CUL
#define APPLICATION_VERIFIER_COM_OBJECT_IN_UNLOADED_DLL		0x0000040DUL
#define APPLICATION_VERIFIER_COM_VTBL_IN_FREED_MEMORY		0x0000040EUL
#define APPLICATION_VERIFIER_COM_VTBL_IN_UNLOADED_DLL		0x0000040FUL
#define APPLICATION_VERIFIER_COM_HOLDING_LOCKS_ON_CALL		0x00000410UL

#define APPLICATION_VERIFIER_RPC_ERROR				0x00000500UL

#define APPLICATION_VERIFIER_INVALID_FREEMEM			0x00000600UL
#define APPLICATION_VERIFIER_INVALID_ALLOCMEM			0x00000601UL
#define APPLICATION_VERIFIER_INVALID_MAPVIEW			0x00000602UL
#define APPLICATION_VERIFIER_PROBE_INVALID_ADDRESS		0x00000603UL
#define APPLICATION_VERIFIER_PROBE_FREE_MEM			0x00000604UL
#define APPLICATION_VERIFIER_PROBE_GUARD_PAGE			0x00000605UL
#define APPLICATION_VERIFIER_PROBE_NULL				0x00000606UL
#define APPLICATION_VERIFIER_PROBE_INVALID_START_OR_SIZE	0x00000607UL
#define APPLICATION_VERIFIER_SIZE_HEAP_UNEXPECTED_EXCEPTION	0x00000618UL
#endif // DLL_PROCESS_VERIFIER

__declspec(dllimport)
VOID	NTAPI	RtlApplicationVerifierStop(DWORD_PTR Code,
		                           LPSTR     Message,
		                           DWORD_PTR Parameter1,
		                           LPSTR     Description1,
		                           DWORD_PTR Parameter2,
		                           LPSTR     Description2,
		                           DWORD_PTR Parameter3,
		                           LPSTR     Description3,
		                           DWORD_PTR Parameter4,
		                           LPSTR     Description4);

__declspec(dllimport)
LONG	NTAPI	RtlApplyRXact(RTL_RXACT_CONTEXT *RXactContext);

__declspec(dllimport)
LONG	NTAPI	RtlApplyRXactNoFlush(RTL_RXACT_CONTEXT *RXactContext);

// https://msdn.microsoft.com/en-us/library/ff563614.aspx

typedef	struct	_RTL_BITMAP
{
	DWORD	SizeOfBitMap;
	LPDWORD	Buffer;
} RTL_BITMAP, *LPRTL_BITMAP, *PRTL_BITMAP;

typedef	struct	_RTL_BITMAP_RUN
{
	DWORD	StartingIndex;
	DWORD	NumberOfBits;
} RTL_BITMAP_RUN, *LPRTL_BITMAP_RUN, *PRTL_BITMAP_RUN;

// NOTE: like ADVAPI32!AreAllAccessesGranted

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAreAllAccessesGranted(ACCESS_MASK GrantedAccess,
		                         ACCESS_MASK DesiredAccess);

// NOTE: like ADVAPI32!AreAnyAccessesGranted

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAreAnyAccessesGranted(ACCESS_MASK GrantedAccess,
		                         ACCESS_MASK DesiredAccess);

// https://msdn.microsoft.com/en-us/library/ff561742.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAreBitsClear(RTL_BITMAP *BitMapHeader,
		                DWORD      StartingIndex,
		                DWORD      Length);

// https://msdn.microsoft.com/en-us/library/ff561745.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlAreBitsSet(RTL_BITMAP *BitMapHeader,
		              DWORD      StartingIndex,
		              DWORD      Length);

__declspec(dllimport)
VOID	NTAPI	RtlAssert(LPCSTR FailedAssertion,
		          LPCSTR FileName,
		          DWORD  LineNumber,
		          LPCSTR MutableMessage);

// NOTE: like KERNEL32!CancelTimerQueueTimer

__declspec(deprecated("obsolete function, do not use"))
__declspec(dllimport)
VOID	NTAPI	RtlCancelTimer(HANDLE TimerQueueHandle,
		               HANDLE TimerToCancel);

// https://msdn.microsoft.com/en-us/library/ff552116.aspx
// https://msdn.microsoft.com/en-us/library/ms680591.aspx
// NOTE: forwarded from KERNEL32!RtlCaptureContext

__declspec(dllimport)
VOID	NTAPI	RtlCaptureContext(CONTEXT *ContextRecord);

// https://msdn.microsoft.com/en-us/library/bb204633.aspx
// https://msdn.microsoft.com/en-us/library/ff552119.aspx
// NOTE: forwarded from KERNEL32!RtlCaptureStackBackTrace

__declspec(dllimport)
WORD	NTAPI	RtlCaptureStackBackTrace(DWORD   FramesToSkip,
		                         DWORD   FramesToCapture,
		                         LPVOID  *BackTrace,
		                         LPDWORD BackTraceHash);

typedef	struct	_RTL_STACK_CONTEXT_ENTRY
{
	DWORD_PTR Address;	// stack address
	DWORD_PTR Data;		// stack contents
} RTL_STACK_CONTEXT_ENTRY, *LPRTL_STACK_CONTEXT_ENTRY, *PRTL_STACK_CONTEXT_ENTRY;

typedef	struct	_RTL_STACK_CONTEXT
{
	DWORD			NumberOfEntries;
	RTL_STACK_CONTEXT_ENTRY	Entry[ANYSIZE_ARRAY];
} RTL_STACK_CONTEXT, *LPRTL_STACK_CONTEXT, *PRTL_STACK_CONTEXT;

__declspec(dllimport)
DWORD	NTAPI	RtlCaptureStackContext(DWORD_PTR         *Callers,
		                       RTL_STACK_CONTEXT *Context,
		                       DWORD             Limit);

// https://msdn.microsoft.com/en-us/library/ff561748.aspx
// https://msdn.microsoft.com/en-us/library/ms648414.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCharToInteger(LPCSTR  String,
		                 DWORD   Base,	// 0, 2, 8, 10, 16
		                 LPDWORD Value);

__declspec(dllimport)
VOID	NTAPI	RtlCheckForOrphanedCriticalSections(HANDLE ThreadHandle);

#if _WIN32_WINNT <= 0x0502
__declspec(dllimport)
LONG	NTAPI	RtlCheckProcessParameters(LPVOID,
		                          LPWSTR,
		                          LPDWORD,
		                          DWORD);
#endif

// https://msdn.microsoft.com/en-us/library/ff560916.aspx

#define RTL_REGISTRY_ABSOLUTE	0UL
#define RTL_REGISTRY_SERVICES	1UL	// relative to \Registry\Machine\System\CurrentControlSet\Services
#define RTL_REGISTRY_CONTROL	2UL	// relative to \Registry\Machine\System\CurrentControlSet\Control
#define RTL_REGISTRY_WINDOWS_NT	3UL	// relative to \Registry\Machine\Software\Microsoft\Windows NT\CurrentVersion
#define RTL_REGISTRY_DEVICEMAP	4UL	// relative to \Registry\Machine\Hardware\DeviceMap
#define RTL_REGISTRY_USER	5UL	// relative to \Registry\User\CurrentUser
#define RTL_REGISTRY_MAXIMUM	6UL
#define RTL_REGISTRY_HANDLE	0x40000000UL
#define RTL_REGISTRY_OPTIONAL	0x80000000UL

// https://msdn.microsoft.com/en-us/library/ff561754.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCheckRegistryKey(DWORD   RelativeTo,
		                    LPCWSTR Path);


#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlCheckSandboxedToken(HANDLE  TokenHandle,
		                       BOOLEAN *IsSandboxed);
#endif

#if _WIN32_WINNT >= 0x0602
// NOTE: like ADVAPI32!CheckTokenMembership

__declspec(dllimport)
LONG	NTAPI	RtlCheckTokenMembership(HANDLE TokenHandle,
		                        SID    *Sid,
		                        BOOL   *IsMember);

// NOTE: like ADVAPI32!CheckTokenMembershipEx

__declspec(dllimport)
LONG	NTAPI	RtlCheckTokenMembershipEx(HANDLE TokenHandle,
		                          SID    *Sid,
		                          DWORD  Flags,
		                          BOOL   *IsMember);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff561755.aspx

__declspec(dllimport)
VOID	NTAPI	RtlClearAllBits(RTL_BITMAP *BitMapHeader);

// https://msdn.microsoft.com/en-us/library/ff561763.aspx

__declspec(dllimport)
VOID	NTAPI	RtlClearBits(RTL_BITMAP *BitMapHeader,
		             DWORD      StartingIndex,
		             DWORD      NumberToClear);

__declspec(dllimport)
HRESULT	NTAPI	RtlCloneMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                     IStream                       **NewStream);

typedef	struct	_RTL_USER_PROCESS_INFORMATION
{
	DWORD				Length;
	HANDLE				Process;
	HANDLE				Thread;
	CLIENT_ID			ClientId;
	SECTION_IMAGE_INFORMATION	ImageInformation;
} RTL_USER_PROCESS_INFORMATION, *LPRTL_USER_PROCESS_INFORMATION, *PRTL_USER_PROCESS_INFORMATION;

#if _WIN32_WINNT >= 0x0600
#define RTL_CLONE_PROCESS_FLAGS_CREATE_SUSPENDED	0x00000001UL
#define RTL_CLONE_PROCESS_FLAGS_INHERIT_HANDLES		0x00000002UL
#define RTL_CLONE_PROCESS_FLAGS_NO_SYNCHRONIZE		0x00000004UL

#ifndef STATUS_PROCESS_CLONED
#define STATUS_PROCESS_CLONED		0x00000129UL
#endif

__declspec(dllimport)
LONG	NTAPI	RtlCloneUserProcess(DWORD                        ProcessFlags,
		                    SECURITY_DESCRIPTOR          *ProcessSecurityDescriptor,	// optional
		                    SECURITY_DESCRIPTOR          *ThreadSecurityDescriptor,	// optional
		                    HANDLE                       DebugPort,			// optional
		                    RTL_USER_PROCESS_INFORMATION *ProcessInformation);
#endif

typedef	struct	_RTL_PROCESS_MODULE_INFORMATION_EX
{
	WORD				NextOffset;
	RTL_PROCESS_MODULE_INFORMATION	BaseInfo;
	DWORD				ImageChecksum;
	DWORD				TimeDateStamp;
	LPVOID				DefaultBase;
} RTL_PROCESS_MODULE_INFORMATION_EX, *LPRTL_PROCESS_MODULE_INFORMATION_EX, *PRTL_PROCESS_MODULE_INFORMATION_EX;

typedef	struct	_RTL_PROCESS_BACKTRACE_INFORMATION
{
	LPSTR	SymbolicBackTrace;
	DWORD	TraceCount;
	WORD	Index;
	WORD	Depth;
	LPVOID	BackTrace[32];
} RTL_PROCESS_BACKTRACE_INFORMATION, *LPRTL_PROCESS_BACKTRACE_INFORMATION, *PRTL_PROCESS_BACKTRACE_INFORMATION;

typedef	struct	_RTL_PROCESS_BACKTRACES
{
	DWORD_PTR				CommittedMemory;
	DWORD_PTR				ReservedMemory;
	DWORD					NumberOfBackTraceLookups;
	DWORD					NumberOfBackTraces;
	RTL_PROCESS_BACKTRACE_INFORMATION	BackTraces[ANYSIZE_ARRAY];
} RTL_PROCESS_BACKTRACES, *LPRTL_PROCESS_BACKTRACES, *PRTL_PROCESS_BACKTRACES;

typedef	struct	_RTL_HEAP_TAG
{
	DWORD	NumberOfAllocations;
	DWORD	NumberOfFrees;
	SIZE_T	BytesAllocated;
	WORD	TagIndex;
	WORD	CreatorBackTraceIndex;
	WCHAR	TagName[24];
} RTL_HEAP_TAG, *LPRTL_HEAP_TAG, *PRTL_HEAP_TAG;

typedef	struct	_RTL_HEAP_ENTRY
{
	SIZE_T	Size;
	WORD	Flags;
	WORD	AllocatorBackTraceIndex;
	union
	{
		struct
		{
			SIZE_T	Settable;
			DWORD	Tag;
		} s1;
		struct
		{
			SIZE_T	CommittedSize;
			LPVOID	FirstBlock;
		} s2;
	} u;
} RTL_HEAP_ENTRY, *LPRTL_HEAP_ENTRY, *PRTL_HEAP_ENTRY;

typedef	struct	_RTL_HEAP_INFORMATION
{
	LPVOID		BaseAddress;
	DWORD		Flags;
	WORD		EntryOverhead;
	WORD		CreatorBackTraceIndex;
	SIZE_T		BytesAllocated;
	SIZE_T		BytesCommitted;
	DWORD		NumberOfTags;
	DWORD		NumberOfEntries;
	DWORD		NumberOfPseudoTags;
	DWORD		PseudoTagGranularity;
	DWORD		Reserved[5];
	RTL_HEAP_TAG	*Tags;
	RTL_HEAP_ENTRY	*Entries;
} RTL_HEAP_INFORMATION, *LPRTL_HEAP_INFORMATION, *PRTL_HEAP_INFORMATION;

typedef	struct	_RTL_PROCESS_HEAPS
{
	DWORD	NumberOfHeaps;
	RTL_HEAP_INFORMATION	Heaps[ANYSIZE_ARRAY];
} RTL_PROCESS_HEAPS, *LPRTL_PROCESS_HEAPS, *PRTL_PROCESS_HEAPS;

typedef	struct	_RTL_PROCESS_LOCK_INFORMATION
{
	LPVOID	Address;
	WORD	Type;
	WORD	CreatorBackTraceIndex;
	DWORD	OwnerThreadId;
	DWORD	ActiveCount;
	DWORD	ContentionCount;
	DWORD	EntryCount;
	DWORD	RecursionCount;
	DWORD	NumberOfSharedWaiters;
	DWORD	NumberOfExclusiveWaiters;
} RTL_PROCESS_LOCK_INFORMATION, *LPRTL_PROCESS_LOCK_INFORMATION, *PRTL_PROCESS_LOCK_INFORMATION;

typedef	struct	_RTL_PROCESS_LOCKS
{
	DWORD				NumberOfLocks;
	RTL_PROCESS_LOCK_INFORMATION	Locks[ANYSIZE_ARRAY];
} RTL_PROCESS_LOCKS, *LPRTL_PROCESS_LOCKS, *PRTL_PROCESS_LOCKS;

typedef	struct	_RTL_PROCESS_VERIFIER_OPTIONS
{
	DWORD	SizeStruct;
	DWORD	Option;
	BYTE	OptionData[ANYSIZE_ARRAY];
} RTL_PROCESS_VERIFIER_OPTIONS, *LPRTL_PROCESS_VERIFIER_OPTIONS, *PRTL_PROCESS_VERIFIER_OPTIONS;

#pragma warning(disable: 4201)
typedef	struct	_RTL_DEBUG_INFORMATION
{
	HANDLE	SectionHandleClient;
	LPVOID	ViewBaseClient;
	LPVOID	ViewBaseTarget;
	DWORD	ViewBaseDelta;
	HANDLE	EventPairClient;
	HANDLE	EventPairTarget;
	HANDLE	TargetProcessId;
	HANDLE	TargetThreadHandle;
	DWORD	Flags;
	DWORD	OffsetFree;
	DWORD	CommitSize;
	DWORD	ViewSize;
	union
	{
		RTL_PROCESS_MODULES			*Modules;
		RTL_PROCESS_MODULE_INFORMATION_EX	*ModulesEx;
	};
	RTL_PROCESS_BACKTRACES		*BackTraces;
	RTL_PROCESS_HEAPS		*Heaps;
	RTL_PROCESS_LOCKS		*Locks;
	HANDLE				SpecificHeap;
	HANDLE				TargetProcessHandle;
	RTL_PROCESS_VERIFIER_OPTIONS	VerifierOptions;
	HANDLE				ProcessHeap;
	HANDLE				CriticalSectionHandle;
	HANDLE				CriticalSectionOwnerThread;
	LPVOID				Reserved[4];
} RTL_DEBUG_INFORMATION, *LPRTL_DEBUG_INFORMATION, *PRTL_DEBUG_INFORMATION;
#pragma warning(default: 4201)

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LPVOID	NTAPI	RtlCommitDebugInfo(RTL_DEBUG_INFORMATION *Buffer,
		                   SIZE_T                Size);
#endif

__declspec(dllimport)
HRESULT	NTAPI	RtlCommitMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                      DWORD                         Flags);

__declspec(dllimport)
LONG	NTAPI	RtlCompactHeap(LPCVOID HeapHandle,
		               DWORD   Flags);

// https://msdn.microsoft.com/en-us/library/ff561778.aspx

__declspec(dllimport)
SIZE_T	NTAPI	RtlCompareMemory(LPCVOID Source1,
		                 LPCVOID Source2,
		                 SIZE_T  Length);

// https://msdn.microsoft.com/en-us/library/ff552123.aspx

__declspec(dllimport)
SIZE_T	NTAPI	RtlCompareMemoryUlong(LPVOID Source,
		                      SIZE_T Length,
		                      ULONG  Pattern);

// https://msdn.microsoft.com/en-us/library/ff561781.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCompareString(STRING  *String1,
		                 STRING  *String2,
		                 BOOLEAN CaseInSensitive);

// https://msdn.microsoft.com/en-us/library/ff561782.aspx
// https://msdn.microsoft.com/en-us/library/mt708867.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCompareUnicodeString(UNICODE_STRING *String1,
		                        UNICODE_STRING *String2,
		                        BOOLEAN        CaseInSensitive);

#ifndef COMPRESSION_FORMAT_NONE
#define COMPRESSION_FORMAT_NONE		0x0000
#endif

#ifndef COMPRESSION_FORMAT_DEFAULT
#define COMPRESSION_FORMAT_DEFAULT	0x0001
#endif

#ifndef COMPRESSION_FORMAT_LZNT1
#define COMPRESSION_FORMAT_LZNT1	0x0002
#endif

#ifndef COMPRESSION_FORMAT_XPRESS
#define COMPRESSION_FORMAT_XPRESS	0x0003
#endif

#ifndef COMPRESSION_FORMAT_XPRESS_HUFF
#define COMPRESSION_FORMAT_XPRESS_HUFF	0x0004
#endif

#ifndef COMPRESSION_ENGINE_STANDARD
#define COMPRESSION_ENGINE_STANDARD	0x0000
#endif

#ifndef COMPRESSION_ENGINE_MAXIMUM
#define COMPRESSION_ENGINE_MAXIMUM	0x0100
#endif

#ifndef COMPRESSION_ENGINE_HIBER
#define COMPRESSION_ENGINE_HIBER	0x0200
#endif

// https://msdn.microsoft.com/en-us/library/ff552127.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCompressBuffer(WORD    CompressionFormatAndEngine,
		                  LPBYTE  UncompressedBuffer,
		                  DWORD   UncompressedBufferSize,
		                  LPBYTE  CompressedBuffer,
		                  DWORD   CompressedBufferSize,
		                  DWORD   UncompressedChunkSize,
		                  LPDWORD FinalCompressedSize,
		                  LPVOID  WorkSpace);

__declspec(dllimport)
DWORD	NTAPI	RtlComputeCrc32(DWORD   PartialCrc,
		                LPCVOID Buffer,
		                DWORD   Length);

__declspec(dllimport)
LONG	NTAPI	RtlComputeImportTableHash(HANDLE FileHandle,
		                          LPBYTE Hash,
		                          DWORD  ImportTableHashRevision);

__declspec(dllimport)
LONG	NTAPI	RtlComputePrivatizedDllName_U(UNICODE_STRING *DllName,
		                              UNICODE_STRING *NewDllNameUnderImageDir,
		                              UNICODE_STRING *NewDllNameUnderLocalDir);

__declspec(dllimport)
LONG	NTAPI	RtlConsoleMultiByteToUnicodeN(LPWSTR  UnicodeString,
		                              DWORD   MaxBytesInUnicodeString,
		                              LPDWORD BytesInUnicodeString,
		                              LPCSTR  MultiByteString,
		                              DWORD   BytesInMultiByteString,
		                              LPDWORD SpecialChar);

#if _WIN32_WINNT >= 0x0601
typedef	struct	_RTL_DYNAMIC_HASH_TABLE
{
	DWORD	Flags;
	DWORD	Shift;
	DWORD	TableSize;
	DWORD	Pivot;
	DWORD	DivisorMask;
	DWORD	NumEntries;
	DWORD	NonEmptyBuckets;
	DWORD	NumEnumerators;
	LPVOID	Directory;
} RTL_DYNAMIC_HASH_TABLE, *LPRTL_DYNAMIC_HASH_TABLE, *PRTL_DYNAMIC_HASH_TABLE;

__declspec(dllimport)
BOOLEAN	NTAPI	RtlContractHashTable(RTL_DYNAMIC_HASH_TABLE *HashTable);
#endif // _WIN32_WINNT

__declspec(dllimport)
VOID	NTAPI	RtlConvertExclusiveToShared(RTL_RESOURCE *Resource);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlConvertLCIDToString(LCID   LocaleId,
		                       DWORD  Base,	// 0, 2, 8, 10, 16
		                       DWORD  Padding,
		                       LPWSTR LocaleString,
		                       DWORD  MaxBytesInLocaleString);
#endif

#ifndef _WIN64
// https://msdn.microsoft.com/en-us/library/ff561786.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlConvertLongToLargeInteger(LONG SignedInteger);
#endif

#if 0	// NOTE: not exported from NTDLL.DLL!
// https://msdn.microsoft.com/en-us/library/ff561790.aspx

__declspec(dllimport)
LUID	NTAPI	RtlConvertLongToLuid(LONG Long);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlConvertSharedToExclusive(RTL_RESOURCE *Resource);

// https://msdn.microsoft.com/en-us/library/aa379320.aspx
// https://msdn.microsoft.com/en-us/library/ff552134.aspx

__declspec(dllimport)
LONG	NTAPI	RtlConvertSidToUnicodeString(UNICODE_STRING *UnicodeString,
		                             SID            *Sid,
		                             BOOLEAN        AllocateDestinationString);

__declspec(dllimport)
LONG	NTAPI	RtlConvertToAutoInheritSecurityObject(SECURITY_DESCRIPTOR *ParentDescriptor,
		                                      SECURITY_DESCRIPTOR *CurrentSecurityDescriptor,
		                                      SECURITY_DESCRIPTOR **NewSecurityDescriptor,
		                                      GUID                *ObjectType,
		                                      BOOLEAN             IsDirectoryObject,
		                                      GENERIC_MAPPING     *GenericMapping);

#if _WIN32_WINNT <= 0x0602
__declspec(dllimport)
LONG	NTAPI	RtlConvertUiListToApiList(UNICODE_STRING *UiList,
		                          UNICODE_STRING *ApiList,
		                          BOOLEAN        BlankIsDelimiter);
#endif

#ifndef _WIN64
// https://msdn.microsoft.com/en-us/library/ff561796.aspx

__declspec(dllimport)
ULARGE_INTEGER	NTAPI	RtlConvertUlongToLargeInteger(ULONG UnsignedInteger);
#endif

#if 0	// NOTE: not exported from NTDLL.DLL!
// https://msdn.microsoft.com/en-us/library/ff561798.aspx

__declspec(dllimport)
LUID	NTAPI	RtlConvertUlongToLuid(ULONG UnsignedLong);
#endif

#if _WIN32_WINNT >= 0x0601
#ifndef _WINNT_
__declspec(dllimport)
DWORD	NTAPI	RtlCopyExtendedContext(CONTEXT_EX *Destination,
		                       DWORD      ContextFlags,
		                       CONTEXT_EX *Source);
#endif
#endif

// https://msdn.microsoft.com/en-us/library/ff552140.aspx

__declspec(dllimport)
VOID	NTAPI	RtlCopyLuid(LUID *DestinationLuid,
		            LUID *SourceLuid);

__declspec(dllimport)
VOID	NTAPI	RtlCopyLuidAndAttributesArray(DWORD               Count,
		                              LUID_AND_ATTRIBUTES *SourceArray,
		                              LUID_AND_ATTRIBUTES *DestinationArray);

#if 1
#pragma intrinsic(memcpy)
#else
#if _WIN32_WINNT >= 0x0601
#undef RtlCopyMemory

// https://msdn.microsoft.com/en-us/library/ff561808.aspx

__declspec(dllimport)
VOID	NTAPI	RtlCopyMemory(LPVOID  Destination,
		              LPCVOID Source,
		              SIZE_T  Length);
#endif
#endif

__declspec(dllimport)
HRESULT	NTAPI	RtlCopyMemoryStreamTo(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                      IStream                       *AnotherStream,
		                      ULARGE_INTEGER                BytesToCopy,
		                      ULARGE_INTEGER                *BytesRead,
		                      ULARGE_INTEGER                *BytesWritten);

__declspec(dllimport)
HRESULT	NTAPI	RtlCopyOutOfProcessMemoryStreamTo(RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                                  IStream                                      *AnotherStream,
		                                  ULARGE_INTEGER                               BytesToCopy,
		                                  ULARGE_INTEGER                               *BytesRead,
		                                  ULARGE_INTEGER                               *BytesWritten);

__declspec(dllimport)
LONG	NTAPI	RtlCopySecurityDescriptor(SECURITY_DESCRIPTOR *InputSecurityDescriptor,
		                          SECURITY_DESCRIPTOR **OutputSecurityDescriptor);

// https://msdn.microsoft.com/en-us/library/ff552146.aspx
// NOTE: like ADVAPI32!CopySid

__declspec(dllimport)
LONG	NTAPI	RtlCopySid(DWORD DestinationSidLength,
		           SID   *DestinationSid,
		           SID   *SourceSid);

__declspec(dllimport)
LONG	NTAPI	RtlCopySidAndAttributesArray(DWORD              ArrayLength,
		                             SID_AND_ATTRIBUTES *Source,
		                             DWORD              TargetSidBufferSize,
		                             SID_AND_ATTRIBUTES *TargetArrayElement,
		                             SID                *TargetSid,
		                             SID                **NextTargetSid,
		                             LPDWORD            RemainingTargetSidSize);

// https://msdn.microsoft.com/en-us/library/ff561814.aspx

__declspec(dllimport)
VOID	NTAPI	RtlCopyString(STRING *DestinationString,
		              STRING *SourceString);

// https://msdn.microsoft.com/en-us/library/ff561817.aspx

__declspec(dllimport)
VOID	NTAPI	RtlCopyUnicodeString(UNICODE_STRING *DestinationString,
		                     UNICODE_STRING *SourceString);

#if _WIN32_WINNT >= 0x0602
__declspec(dllimport)
DWORD	NTAPI	RtlCrc32(LPCVOID Buffer,
		         SIZE_T  Size,
		         DWORD   InitialCrc);

__declspec(dllimport)
QWORD	NTAPI	RtlCrc64(LPCVOID Buffer,
		         SIZE_T  Size,
		         QWORD   InitialCrc);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff552151.aspx
// NOTE: like ADVAPI32!InitializeAcl

__declspec(dllimport)
LONG	NTAPI	RtlCreateAcl(ACL   *Acl,
		             DWORD AclLength,
		             DWORD AceRevision);

#define ACTIVATION_CONTEXT_NOTIFICATION_DESTROY	1UL
#define ACTIVATION_CONTEXT_NOTIFICATION_ZOMBIFY	2UL
#define ACTIVATION_CONTEXT_NOTIFICATION_USED	3UL

typedef	struct	_ACTIVATION_CONTEXT_DATA	ACTIVATION_CONTEXT_DATA, *LPACTIVATION_CONTEXT_DATA, *PACTIVATION_CONTEXT_DATA;

typedef	VOID	(NTAPI	ACTIVATION_CONTEXT_NOTIFY_ROUTINE) (DWORD                   NotificationType,
			                                    ACTIVATION_CONTEXT      *ActivationContext,
			                                    ACTIVATION_CONTEXT_DATA *ActivationContextData,
			                                    LPVOID                  NotificationContext,
			                                    LPVOID                  NotificationData,
			                                    BOOLEAN                 *DisableThisNotification);

__declspec(dllimport)
LONG	NTAPI	RtlCreateActivationContext(DWORD                             Flags,
		                           ACTIVATION_CONTEXT_DATA           *ActivationContextData,
		                           DWORD                             ExtraBytes,
		                           ACTIVATION_CONTEXT_NOTIFY_ROUTINE *NotificationRoutine,
		                           LPVOID                            NotificationContext,
		                           ACTIVATION_CONTEXT                **ActivationContext);

typedef	struct	_RTL_ACE_DATA
{
	BYTE		AceType;
	BYTE		InheritFlags;
	BYTE		AceFlags;
	ACCESS_MASK	Mask;
	SID		**Sid;
} RTL_ACE_DATA, *LPRTL_ACE_DATA, *PRTL_ACE_DATA;

__declspec(dllimport)
LONG	NTAPI	RtlCreateAndSetSD(RTL_ACE_DATA        *AceData,
		                  DWORD               AceCount,
		                  SID                 *OwnerSid,
		                  SID                 *GroupSid,
		                  SECURITY_DESCRIPTOR **NewDescriptor);

__declspec(dllimport)
LONG	NTAPI	RtlCreateAtomTable(DWORD          NumberOfBuckets,
		                   RTL_ATOM_TABLE **AtomTable);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	NTAPI	RtlCreateBootStatusDataFile(VOID);
#endif

#if _WIN32_WINNT >= 0x0600
// NOTE: like KERNEL32!CreateBoundaryDescriptor

__declspec(dllimport)
LPVOID	NTAPI	RtlCreateBoundaryDescriptor(UNICODE_STRING *Name,
		                            DWORD          Flags);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlCreateEnvironment(BOOLEAN CloneCurrentEnvironment,
		                     LPVOID  *Environment);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlCreateHashTable(RTL_DYNAMIC_HASH_TABLE **HashTable,
		                   DWORD                  Shift,
		                   DWORD                  Flags);
#endif

#ifndef _WINNT_
#define HEAP_NO_SERIALIZE		0x00000001UL
#define HEAP_GROWABLE			0x00000002UL
#define HEAP_GENERATE_EXCEPTIONS	0x00000004UL
#define HEAP_ZERO_MEMORY		0x00000008UL
#define HEAP_REALLOC_IN_PLACE_ONLY	0x00000010UL
#define HEAP_TAIL_CHECKING_ENABLED	0x00000020UL
#define HEAP_FREE_CHECKING_ENABLED	0x00000040UL
#define HEAP_DISABLE_COALESCE_ON_FREE	0x00000080UL
#define HEAP_CREATE_ALIGN_16		0x00010000UL
#define HEAP_CREATE_ENABLE_TRACING	0x00020000UL
#define HEAP_CREATE_ENABLE_EXECUTE	0x00040000UL
#endif

// https://msdn.microsoft.com/en-us/library/ff552159.aspx

typedef	LONG	(NTAPI	RTL_HEAP_COMMIT_ROUTINE) (LPVOID Base,
			                          LPVOID *CommitAddress,
			                          SIZE_T *CommitSize);

typedef	struct	_RTL_HEAP_PARAMETERS
{
	DWORD			Length;
	SIZE_T			SegmentReserve;
	SIZE_T			SegmentCommit;
	SIZE_T			DeCommitFreeBlockThreshold;
	SIZE_T			DeCommitTotalFreeThreshold;
	SIZE_T			MaximumAllocationSize;
	SIZE_T			VirtualMemoryThreshold;
	SIZE_T			InitialCommit;
	SIZE_T			InitialReserve;
	RTL_HEAP_COMMIT_ROUTINE	*CommitRoutine;
	SIZE_T			Reserved[2];
} RTL_HEAP_PARAMETERS, *LPRTL_HEAP_PARAMETERS, *PRTL_HEAP_PARAMETERS;

__declspec(dllimport)
LPVOID	NTAPI	RtlCreateHeap(DWORD               Flags,
		              LPVOID              HeapBase,
		              SIZE_T              ReserveSize,
		              SIZE_T              CommitSize,
		              LPVOID              Lock,
		              RTL_HEAP_PARAMETERS *Parameters);

__declspec(dllimport)
LONG	NTAPI	RtlCreateProcessParameters(RTL_USER_PROCESS_PARAMETERS **ProcessParameters,
		                           UNICODE_STRING              *ImagePathName,
		                           UNICODE_STRING              *DllPath,
		                           UNICODE_STRING              *CurrentDirectory,
		                           UNICODE_STRING              *CommandLine,
		                           LPVOID                      Environment,
		                           UNICODE_STRING              *WindowTitle,
		                           UNICODE_STRING              *DesktopInfo,
		                           UNICODE_STRING              *ShellInfo,
		                           UNICODE_STRING              *RuntimeData);

#if _WIN32_WINNT >= 0x0600
#define RTL_USER_PROC_PARAMETERS_NORMALIZED	0x00000001UL
#define RTL_USER_PROC_PROFILE_USER		0x00000002UL
#define RTL_USER_PROC_PROFILE_KERNEL		0x00000004UL
#define RTL_USER_PROC_PROFILE_SERVER		0x00000008UL
#define RTL_USER_PROC_RESERVE_1MB		0x00000020UL
#define RTL_USER_PROC_RESERVE_16MB		0x00000040UL
#define RTL_USER_PROC_CASE_SENSITIVE		0x00000080UL
#define RTL_USER_PROC_DISABLE_HEAP_DECOMMIT	0x00000100UL
#define RTL_USER_PROC_DLL_REDIRECTION_LOCAL	0x00001000UL
#define RTL_USER_PROC_APP_MANIFEST_PRESENT	0x00002000UL
#define RTL_USER_PROC_IMAGE_KEY_MISSING		0x00004000UL
#define RTL_USER_PROC_OPTIN_PROCESS		0x00020000UL

__declspec(dllimport)
LONG	NTAPI	RtlCreateProcessParametersEx(RTL_USER_PROCESS_PARAMETERS **ProcessParameters,
		                             UNICODE_STRING              *ImagePathName,
		                             UNICODE_STRING              *DllPath,
		                             UNICODE_STRING              *CurrentDirectory,
		                             UNICODE_STRING              *CommandLine,
		                             LPVOID                      Environment,
		                             UNICODE_STRING              *WindowTitle,
		                             UNICODE_STRING              *DesktopInfo,
		                             UNICODE_STRING              *ShellInfo,
		                             UNICODE_STRING              *RuntimeData,
		                             DWORD                       Flags);
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0601
typedef	struct	_RTL_PROCESS_REFLECTION_INFORMATION
{
	HANDLE		Process;
	HANDLE		Thread;
	CLIENT_ID	ClientId;
} RTL_PROCESS_REFLECTION_INFORMATION, *LPRTL_PROCESS_REFLECTION_INFORMATION, *PRTL_PROCESS_REFLECTION_INFORMATION;

__declspec(dllimport)
LONG	NTAPI	RtlCreateProcessReflection(HANDLE                             ProcessHandle,
		                           DWORD                              Flags,
		                           LPVOID                             StartRoutine,
		                           LPVOID                             StartContext,
		                           HANDLE                             EventHandle,
		                           RTL_PROCESS_REFLECTION_INFORMATION *ReflectionInformation);
#endif // _WIN32_WINNT

__declspec(dllimport)
RTL_DEBUG_INFORMATION *	NTAPI	RtlCreateQueryDebugBuffer(DWORD   Size,
				                          BOOLEAN EventPair);

// https://msdn.microsoft.com/en-us/library/ff561822.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCreateRegistryKey(DWORD   RelativeTo,
		                     LPCWSTR Path);

// https://msdn.microsoft.com/en-us/library/ff561827.aspx
// NOTE: like ADVAPI32!InitializeSecurityDescriptor

__declspec(dllimport)
LONG	NTAPI	RtlCreateSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                            DWORD               Revision);

// https://msdn.microsoft.com/en-us/library/ff552165.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCreateSecurityDescriptorRelative(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                                    DWORD               Revision);

__declspec(dllimport)
LONG	NTAPI	RtlCreateServiceSid(UNICODE_STRING *ServiceName,
		                    SID            *ServiceSid,
		                    LPDWORD        ServiceSidLength);

// https://msdn.microsoft.com/en-us/library/ff552175.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCreateSystemVolumeInformationFolder(UNICODE_STRING *VolumeRootPath);

__declspec(dllimport)
DWORD	NTAPI	RtlCreateTagHeap(LPCVOID HeapHandle,
		                 DWORD   Flags,
		                 LPWSTR  TagPrefix,
		                 LPWSTR  TagNames);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms687066.aspx

typedef	VOID	(NTAPI	*WAITORTIMERCALLBACKFUNC) (LPVOID  Parameter,
			                           BOOLEAN TimerOrWaitFired);
#endif

// NOTE: like KERNEL32!CreateTimerQueueTimer

__declspec(dllimport)
LONG	NTAPI	RtlCreateTimer(HANDLE                  TimerQueueHandle,
		               HANDLE                  *TimerHandle,
		               WAITORTIMERCALLBACKFUNC Function,
		               LPVOID                  Context,
		               DWORD                   DueTime,
		               DWORD                   Period,
		               DWORD                   Flags);

// NOTE: like KERNEL32!CreateTimerQueue

__declspec(dllimport)
LONG	NTAPI	RtlCreateTimerQueue(HANDLE *TimerQueueHandle);

// https://msdn.microsoft.com/en-us/library/ff552182.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlCreateUnicodeString(UNICODE_STRING *DestinationString,
		                       LPCWSTR        SourceString);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlCreateUnicodeStringFromAsciiz(UNICODE_STRING *DestinationString,
		                                 LPCSTR         SourceString);

__declspec(dllimport)
LONG	NTAPI	RtlCreateUserProcess(UNICODE_STRING               *NtImagePathName,
		                     DWORD                        Attributes,
		                     RTL_USER_PROCESS_PARAMETERS  *ProcessParameters,
		                     SECURITY_DESCRIPTOR          *ProcessSecurityDescriptor,
		                     SECURITY_DESCRIPTOR          *ThreadSecurityDescriptor,
		                     HANDLE                       ParentProcessHandle,
		                     BOOLEAN                      InheritHandles,
		                     HANDLE                       DebugPort,
		                     HANDLE                       ExceptionPort,
		                     RTL_USER_PROCESS_INFORMATION *ProcessInformation);


__declspec(dllimport)
LONG	NTAPI	RtlCreateUserSecurityObject(RTL_ACE_DATA        *AceData,
		                            DWORD               AceCount,
		                            SID                 *OwnerSid,
		                            SID                 *GroupSid,
		                            BOOLEAN             IsDirectoryObject,
		                            GENERIC_MAPPING     *GenericMapping,
		                            SECURITY_DESCRIPTOR **NewDescriptor);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlCreateUserStack(SIZE_T      CommittedStackSize,
		                   SIZE_T      MaximumStackSize,
		                   DWORD_PTR   ZeroBits,
		                   SIZE_T      PageSize,
		                   DWORD_PTR   ReserveAlignment,
		                   INITIAL_TEB *InitialTeb);
#endif

typedef	LONG	(NTAPI	USER_THREAD_START_ROUTINE) (LPVOID ThreadParameter);

__declspec(dllimport)
LONG	NTAPI	RtlCreateUserThread(HANDLE                    ProcessHandle,
		                    SECURITY_DESCRIPTOR       *ThreadSecurityDescriptor,
		                    BOOLEAN                   CreateSuspended,
		                    DWORD                     ZeroBits,
		                    SIZE_T                    MaximumStackSize,
		                    SIZE_T                    CommittedStackSize,
		                    USER_THREAD_START_ROUTINE *StartAddress,
		                    LPVOID                    Parameter,
		                    HANDLE                    *ThreadHandle,
		                    CLIENT_ID                 *ClientId);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlCreateVirtualAccountSid(UNICODE_STRING *Name,
		                           DWORD          BaseSubAuthority,
		                           SID            *Sid,
		                           LPDWORD        SidLength);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
BOOLEAN	NTAPI	RtlCultureNameToLCID(UNICODE_STRING *CultureName,
		                     LCID           *LocaleId);
#endif

#define MAXIMUM_LEADBYTES	12

typedef	struct	_CPTABLEINFO
{
	WORD	CodePage;
	WORD	MaximumCharacterSize;	// 1 = SBCS, 2 = DBCS
	WORD	DefaultChar;		// Default multi-byte character for CP to Unicode conversion
	WORD	UniDefaultChar;		// Default Unicode character for CP to Unicode conversion
	WORD	TransDefaultChar;	// Default multi-byte character for Unicode to CP conversion
	WORD	TransUniDefaultChar;	// Default Unicode character for Unicode to CP conversion
	WORD	DBCSCodePage;
	BYTE	LeadByte[MAXIMUM_LEADBYTES];
	LPWORD	MultiByteTable;		// Table for CP to Unicode conversion
	LPVOID	WideCharTable;		// Table for Unicode to CP conversion
	LPWORD	DBCSRanges;
	LPWORD	DBCSOffsets;
} CPTABLEINFO, *LPCPTABLEINFO, *PCPTABLEINFO;

// https://msdn.microsoft.com/en-us/library/ff552187.aspx

__declspec(dllimport)
LONG	NTAPI	RtlCustomCPToUnicodeN(CPTABLEINFO *CustomCP,
		                      LPWSTR      UnicodeString,
		                      DWORD       MaxBytesInUnicodeString,
		                      LPDWORD     BytesInUnicodeString,
		                      LPCSTR      CustomString,
		                      DWORD       BytesInCustomString);

__declspec(dllimport)
VOID	NTAPI	RtlDeactivateActivationContext(DWORD     Flags,
		                               DWORD_PTR Cookie);

__declspec(dllimport)
VOID	FASTCALL	RtlDeactivateActivationContextUnsafeFast(RTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME *Frame);

__declspec(deprecated("obsolete function, do not use"))
__declspec(dllimport)
VOID	NTAPI	RtlDebugPrintTimes(VOID);

// https://msdn.microsoft.com/en-us/library/bb432242.aspx
// NOTE: forwarded from KERNEL32!DecodePointer and KERNELBASE!DecodePointer

__declspec(dllimport)
LPVOID	NTAPI	RtlDecodePointer(LPVOID Pointer);

#if _WIN32_WINNT >= 0x0A00
// https://msdn.microsoft.com/en-us/library/dn877133.aspx
// NOTE: forwarded from KERNELBASE!DecodeRemotePointer

__declspec(dllimport)
LONG	NTAPI	RtlDecodeRemotePointer(HANDLE ProcessHandle,
		                       LPVOID Pointer,
		                       LPVOID *DecodedPointer);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/bb432243.aspx
// NOTE: forwarded from KERNEL32!DecodeSystemPointer and KERNELBASE!DecodeSystemPointer

__declspec(dllimport)
LPVOID	NTAPI	RtlDecodeSystemPointer(LPVOID Pointer);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
VOID	NTAPI	RtlDeCommitDebugInfo(RTL_DEBUG_INFORMATION Buffer,
		                     LPVOID                p,
		                     SIZE_T                Size);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff552191.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDecompressBuffer(WORD    CompressionFormat,
		                    LPBYTE  UncompressedBuffer,
		                    DWORD   UncompressedBufferSize,
		                    LPBYTE  CompressedBuffer,
		                    DWORD   CompressedBufferSize,
		                    LPDWORD FinalUncompressedSize);

// https://msdn.microsoft.com/en-us/library/hh439511.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDecompressBufferEx(WORD    CompressionFormat,
		                      LPBYTE  UncompressedBuffer,
		                      DWORD   UncompressedBufferSize,
		                      LPBYTE  CompressedBuffer,
		                      DWORD   CompressedBufferSize,
		                      LPDWORD FinalUncompressedSize,
		                      LPVOID  WorkSpace);

// https://msdn.microsoft.com/en-us/library/ff552197.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDecompressFragment(WORD    CompressionFormat,
		                      LPBYTE  UncompressedFragment,
		                      DWORD   UncompressedFragmentSize,
		                      LPBYTE  CompressedBuffer,
		                      DWORD   CompressedBufferSize,
		                      DWORD   FragmentOffset,
		                      LPDWORD FinalUncompressedSize,
		                      LPVOID  WorkSpace);

// https://msdn.microsoft.com/en-us/library/mt426738.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDecompressFragmentEx(DWORD   CompressionFormat,
		                        LPBYTE  UncompressedFragment,
		                        DWORD   UncompressedFragmentSize,
		                        LPBYTE  CompressedBuffer,
		                        DWORD   CompressedBufferSize,
		                        DWORD   FragmentOffset,
		                        DWORD   UncompressedChunkSize,
		                        LPDWORD FinalUncompressedSize,
		                        LPVOID  WorkSpace);

__declspec(dllimport)
LONG	NTAPI	RtlDefaultNpAcl(ACL **Acl);

// https://msdn.microsoft.com/en-us/library/ff553351.aspx

typedef	struct	_RTL_SPLAY_LINKS
{
	struct	_RTL_SPLAY_LINKS	*Parent;
	struct	_RTL_SPLAY_LINKS	*LeftChild;
	struct	_RTL_SPLAY_LINKS	*RightChild;
} RTL_SPLAY_LINKS, *LPRTL_SPLAY_LINKS, *PRTL_SPLAY_LINKS;

// https://msdn.microsoft.com/en-us/library/ff552201.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlDelete(RTL_SPLAY_LINKS *Links);

// https://msdn.microsoft.com/en-us/library/ff552206.aspx
// NOTE: like ADVAPI32!DeleteAce

__declspec(dllimport)
LONG	NTAPI	RtlDeleteAce(ACL   *Acl,
		             DWORD AceIndex);

__declspec(dllimport)
LONG	NTAPI	RtlDeleteAtomFromAtomTable(RTL_ATOM_TABLE *AtomTable,
		                           RTL_ATOM       Atom);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms682549.aspx
// NOTE: forwarded from KERNEL32!DeleteBoundaryDescriptor

__declspec(dllimport)
VOID	NTAPI	RtlDeleteBoundaryDescriptor(LPVOID BoundaryDescriptor);
#endif

// https://msdn.microsoft.com/en-us/library/ms682552.aspx
// NOTE: forwarded from KERNEL32!DeleteCriticalSection and KERNELBASE!DeleteCriticalSection

__declspec(dllimport)
LONG	NTAPI	RtlDeleteCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

typedef	enum	_RTL_GENERIC_COMPARE_RESULTS
{
	GenericLessThan    = 0,
	GenericGreaterThan,
	GenericEqual
} RTL_GENERIC_COMPARE_RESULTS;

typedef	struct	_RTL_GENERIC_TABLE	RTL_GENERIC_TABLE, *LPRTL_GENERIC_TABLE, *PRTL_GENERIC_TABLE;

// https://msdn.microsoft.com/en-us/library/ff552989.aspx

typedef	RTL_GENERIC_COMPARE_RESULTS	(NTAPI	RTL_GENERIC_COMPARE_ROUTINE) (RTL_GENERIC_TABLE *Table,
						                              LPVOID            FirstStruct,
						                              LPVOID            SecondStruct);

typedef	LPVOID	(NTAPI	RTL_GENERIC_ALLOCATE_ROUTINE) (RTL_GENERIC_TABLE *Table,
			                               DWORD             ByteSize);

typedef	VOID	(NTAPI	RTL_GENERIC_FREE_ROUTINE) (RTL_GENERIC_TABLE *Table,
			                           LPVOID            Buffer);

// https://msdn.microsoft.com/en-us/library/ff553345.aspx

typedef	struct	_RTL_GENERIC_TABLE
{
	RTL_SPLAY_LINKS			*TableRoot;
	LIST_ENTRY			InsertOrderList;
	LIST_ENTRY			*OrderedPointer;
	DWORD				WhichOrderedElement;
	DWORD				NumberGenericTableElements;
	RTL_GENERIC_COMPARE_ROUTINE	*CompareRoutine;
	RTL_GENERIC_ALLOCATE_ROUTINE	*AllocateRoutine;
	RTL_GENERIC_FREE_ROUTINE	*FreeRoutine;
	LPVOID				TableContext;
} RTL_GENERIC_TABLE, *LPRTL_GENERIC_TABLE, *PRTL_GENERIC_TABLE;

// https://msdn.microsoft.com/en-us/library/ff552209.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDeleteElementGenericTable(RTL_GENERIC_TABLE *Table,
		                             LPVOID            Buffer);

typedef	struct	_RTL_AVL_TABLE	RTL_AVL_TABLE, *LPRTL_AVL_TABLE, *PRTL_AVL_TABLE;

// https://msdn.microsoft.com/en-us/library/hh406465.aspx

typedef	RTL_GENERIC_COMPARE_RESULTS	(NTAPI	RTL_AVL_COMPARE_ROUTINE) (RTL_AVL_TABLE *Table,
						                          LPVOID        FirstStruct,
						                          LPVOID        SecondStruct);

typedef	LPVOID	(NTAPI	RTL_AVL_ALLOCATE_ROUTINE) (RTL_AVL_TABLE *Table,
			                           DWORD         ByteSize);

typedef	VOID	(NTAPI	RTL_AVL_FREE_ROUTINE) (RTL_AVL_TABLE *Table,
			                       LPVOID        Buffer);

typedef	struct	_RTL_BALANCED_LINKS
{
	struct	_RTL_BALANCED_LINKS	*Parent;
	struct	_RTL_BALANCED_LINKS	*LeftChild;
	struct	_RTL_BALANCED_LINKS	*RightChild;
	CHAR	Balance;
	UCHAR	Reserved[3];
} RTL_BALANCED_LINKS, *LPRTL_BALANCED_LINKS, *PRTL_BALANCED_LINKS;

// https://msdn.microsoft.com/en-us/library/ff553327.aspx

typedef	struct	_RTL_AVL_TABLE
{
	RTL_BALANCED_LINKS       BalancedRoot;
	LPVOID                   OrderedPointer;
	DWORD                    WhichOrderedElement;
	DWORD                    NumberGenericTableElements;
	DWORD                    DepthOfTree;
	RTL_BALANCED_LINKS       *RestartKey;
	DWORD                    DeleteCount;
	RTL_AVL_COMPARE_ROUTINE  *CompareRoutine;
	RTL_AVL_ALLOCATE_ROUTINE *AllocateRoutine;
	RTL_AVL_FREE_ROUTINE     *FreeRoutine;
	LPVOID                   TableContext;
} RTL_AVL_TABLE, *LPRTL_AVL_TABLE, *PRTL_AVL_TABLE;

// https://msdn.microsoft.com/en-us/library/hh439514.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDeleteElementGenericTableAvl(RTL_AVL_TABLE *Table,
		                                LPVOID        Buffer);

#ifdef _WIN64
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms680593.aspx
// NOTE: forwarded from KERNEL32!RtlDeleteFunctionTable

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDeleteFunctionTable(RUNTIME_FUNCTION *FunctionTable);
#endif

#if _WIN32_WINNT >= 0x0602
// https://msdn.microsoft.com/en-us/library/hh405353.aspx

__declspec(dllimport)
VOID	NTAPI	RtlDeleteGrowableFunctionTable(LPVOID DynamicTable);
#endif
#endif // _WIN64

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
VOID	NTAPI	RtlDeleteHashTable(RTL_DYNAMIC_HASH_TABLE *HashTable);
#endif

// https://msdn.microsoft.com/en-us/library/ff552214.aspx

__declspec(dllimport)
VOID	NTAPI	RtlDeleteNoSplay(RTL_SPLAY_LINKS *Links,
		                 RTL_SPLAY_LINKS **Root);

// https://msdn.microsoft.com/en-us/library/ff561829.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDeleteRegistryValue(DWORD   RelativeTo,
		                       LPCWSTR Path,
		                       LPCWSTR ValueName);

__declspec(dllimport)
VOID	NTAPI	RtlDeleteResource(RTL_RESOURCE *Resource);

__declspec(dllimport)
LONG	NTAPI	RtlDeleteSecurityObject(SECURITY_DESCRIPTOR **ObjectDescriptor);

// NOTE: like KERNEL32!DeleteTimerQueueTimer

__declspec(dllimport)
LONG	NTAPI	RtlDeleteTimer(HANDLE TimerQueueHandle,
		               HANDLE TimerToCancel,
		               HANDLE CompletionEvent);

// NOTE: like KERNEL32!DeleteTimerQueue

__declspec(dllimport)
LONG	NTAPI	RtlDeleteTimerQueue(HANDLE TimerQueueHandle);

// NOTE: like KERNEL32!DeleteTimerQueueEx

__declspec(dllimport)
LONG	NTAPI	RtlDeleteTimerQueueEx(HANDLE TimerQueueHandle,
		                      HANDLE CompletionEvent);

__declspec(dllimport)
RTL_USER_PROCESS_PARAMETERS *	NTAPI	RtlDeNormalizeProcessParams(RTL_USER_PROCESS_PARAMETERS *ProcessParameters);

__declspec(dllimport)
LONG	NTAPI	RtlDestroyAtomTable(RTL_ATOM_TABLE *AtomTable);

__declspec(dllimport)
LONG	NTAPI	RtlDestroyEnvironment(LPVOID Environment);

__declspec(dllimport)
LONG	NTAPI	RtlDestroyHandleTable(RTL_HANDLE_TABLE *HandleTable);

// https://msdn.microsoft.com/en-us/library/ff552233.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlDestroyHeap(LPCVOID HeapHandle);

__declspec(dllimport)
LONG	NTAPI	RtlDestroyProcessParameters(RTL_USER_PROCESS_PARAMETERS *ProcessParameters);

__declspec(dllimport)
LONG	NTAPI	RtlDestroyQueryDebugBuffer(RTL_DEBUG_INFORMATION *Buffer);

typedef	enum	_RTL_PATH_TYPE
{
	RtlPathTypeUnknown         = 0,
	RtlPathTypeUncAbsolute,
	RtlPathTypeDriveAbsolute,
	RtlPathTypeDriveRelative,
	RtlPathTypeRooted,
	RtlPathTypeRelative,
	RtlPathTypeLocalDevice,
	RtlPathTypeRootLocalDevice
} RTL_PATH_TYPE;

__declspec(dllimport)
LONG	NTAPI	RtlDetermineDosPathNameType_U(LPCWSTR PathName);

#if _WIN32_WINNT >= 0x0601
#if 0	// NOTE: only exported as ordinal 8!
__declspec(dllimport)
LONG	NTAPI	RtlDispatchAPC(PS_APC_ROUTINE     *ApcRoutine,
		               VOID               *Data,
		               ACTIVATION_CONTEXT *ActivationContext);
#endif
#endif

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDispatchException(EXCEPTION_RECORD *ExceptionRecord,
		                     CONTEXT          *ContextRecord);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDllShutdownInProgress(VOID);

__declspec(dllimport)
LONG	NTAPI	RtlDnsHostNameToComputerName(UNICODE_STRING *ComputerNameString,
		                             UNICODE_STRING *DnsHostNameString,
		                             BOOLEAN        AllocateComputerNameString);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDoesFileExists_U(LPCWSTR PathName);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDoesFileExists_UEx(LPCWSTR PathName,
		                      BOOLEAN SucceedIfBusy);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDoesFileExists_UStr(UNICODE_STRING *PathName);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDoesFileExists_UstrEx(UNICODE_STRING *PathName,
		                         BOOLEAN        SucceedIfBusy);

#define RTL_DOS_APPLY_FILE_REDIRECTION_USTR_FLAG_RESPECT_DOT_LOCAL	1UL

#define RTL_DOS_APPLY_FILE_REDIRECTION_USTR_OUTFLAG_DOT_LOCAL_REDIRECT	1UL
#define RTL_DOS_APPLY_FILE_REDIRECTION_USTR_OUTFLAG_ACTCTX_REDIRECT	2UL

__declspec(dllimport)
LONG	NTAPI	RtlDosApplyFileIsolationRedirection_Ustr(DWORD          Flags,
		                                         UNICODE_STRING *FileName,
		                                         UNICODE_STRING *DefaultExtension,
		                                         UNICODE_STRING *PreAllocatedString,
		                                         UNICODE_STRING *DynamicallyAllocatedString,
		                                         UNICODE_STRING **FullPath,
		                                         LPDWORD        OutFlags,
		                                         SIZE_T         *FilePartPrefixCch,
		                                         SIZE_T         *BytesRequired);

typedef	struct	_RTLP_CURDIR_REF
{
	LONG	RefCount;
	HANDLE	Handle;
} RTLP_CURDIR_REF, *LPRTLP_CURDIR_REF, *PRTLP_CURDIR_REF;

typedef	struct	_RTL_RELATIVE_NAME_U
{
	UNICODE_STRING	RelativeName;
	HANDLE		ContainingDirectory;
	RTLP_CURDIR_REF	*CurDirRef;
} RTL_RELATIVE_NAME_U, *LPRTL_RELATIVE_NAME_U, *PRTL_RELATIVE_NAME_U;

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDosPathNameToNtPathName_U(LPCWSTR             DosName,
		                             UNICODE_STRING      *NtName,
		                             LPCWSTR             *PartName,
		                             RTL_RELATIVE_NAME_U *RelativeName);

__declspec(dllimport)
LONG	NTAPI	RtlDosPathNameToNtPathName_U_WithStatus(LPCWSTR             DosName,
		                                        UNICODE_STRING      *NtName,
		                                        LPCWSTR             *PartName,
		                                        RTL_RELATIVE_NAME_U *RelativeName);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlDosPathNameToRelativeNtPathName_U(LPCWSTR             DosName,
		                                     UNICODE_STRING      *NtName,
		                                     LPCWSTR             *PartName,
		                                     RTL_RELATIVE_NAME_U *RelativeName);

__declspec(dllimport)
LONG	NTAPI	RtlDosPathNameToRelativeNtPathName_U_WithStatus(LPCWSTR             DosName,
		                                                UNICODE_STRING      *NtName,
		                                                LPCWSTR             *PartName,
		                                                RTL_RELATIVE_NAME_U *RelativeName);

__declspec(dllimport)
LONG	NTAPI	RtlDosSearchPath_U(LPCWSTR Path,
		                   LPCWSTR FileName,
		                   LPCWSTR Extension,
		                   DWORD   BufferSize,
		                   LPWSTR  Buffer,
		                   LPWSTR  *PartName);

__declspec(dllimport)
LONG	NTAPI	RtlDosSearchPath_Ustr(DWORD          Flags,
		                      UNICODE_STRING *Path,
		                      UNICODE_STRING *FileName,
		                      UNICODE_STRING *DefaultExtension,
		                      UNICODE_STRING *StaticString,
		                      UNICODE_STRING *DynamicString,
		                      UNICODE_STRING **FullFileNameOut,
		                      SIZE_T         *FilePartSize,
		                      SIZE_T         *BytesRequired);

// https://msdn.microsoft.com/en-us/library/ff561834.aspx

__declspec(dllimport)
WCHAR	NTAPI	RtlDowncaseUnicodeChar(WCHAR SourceCharacter);

// https://msdn.microsoft.com/en-us/library/ff552237.aspx

__declspec(dllimport)
LONG	NTAPI	RtlDowncaseUnicodeString(UNICODE_STRING *DestinationString,
		                         UNICODE_STRING *SourceString,
		                         BOOLEAN        AllocateDestinationString);

__declspec(dllimport)
VOID	NTAPI	RtlDumpResource(RTL_RESOURCE *Resource);

#define RTL_DUPLICATE_UNICODE_STRING_NULL_TERMINATE		1UL
#define RTL_DUPLICATE_UNICODE_STRING_ALLOCATE_NULL_STRING	2UL

__declspec(dllimport)
LONG	NTAPI	RtlDuplicateUnicodeString(DWORD          AppendNul,
		                          UNICODE_STRING *SourceString,
		                          UNICODE_STRING *DestinationString);

__declspec(dllimport)
LONG	NTAPI	RtlEmptyAtomTable(RTL_ATOM_TABLE *AtomTable,
		                  BOOLEAN        IncludePinnedAtoms);

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
VOID	NTAPI	RtlEnableEarlyCriticalSectionEventCreation(VOID);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/bb432254.aspx
// NOTE: forwarded from KERNEL32!EncodePointer and KERNELBASE!EncodePointer

__declspec(dllimport)
LPVOID	NTAPI	RtlEncodePointer(LPVOID Pointer);

#if _WIN32_WINNT >= 0x0A00
// https://msdn.microsoft.com/en-us/library/dn877135.aspx
// NOTE: forwarded from KERNELBASE!EncodeRemotePointer

__declspec(dllimport)
LONG	NTAPI	RtlEncodeRemotePointer(HANDLE ProcessHandle,
		                       LPVOID Pointer,
		                       LPVOID *EncodedPointer);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/bb432255.aspx
// NOTE: forwarded from KERNEL32!EncodeSystemPointer and KERNELBASE!EncodeSystemPointer

__declspec(dllimport)
LPVOID	NTAPI	RtlEncodeSystemPointer(LPVOID Pointer);

#if _WIN32_WINNT >= 0x0601
typedef	struct	_RTL_DYNAMIC_HASH_TABLE_ENTRY
{
	LIST_ENTRY	Linkage;
	DWORD_PTR	Signature;
} RTL_DYNAMIC_HASH_TABLE_ENTRY, *LPRTL_DYNAMIC_HASH_TABLE_ENTRY, *PRTL_DYNAMIC_HASH_TABLE_ENTRY;

typedef	struct	_RTL_DYNAMIC_HASH_TABLE_ENUMERATOR
{
	RTL_DYNAMIC_HASH_TABLE_ENTRY	HashEntry;
	LIST_ENTRY			*ChainHead;
	DWORD				BucketIndex;
} RTL_DYNAMIC_HASH_TABLE_ENUMERATOR, *LPRTL_DYNAMIC_HASH_TABLE_ENUMERATOR, *PRTL_DYNAMIC_HASH_TABLE_ENUMERATOR;

__declspec(dllimport)
VOID	NTAPI	RtlEndEnumerationHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
		                           RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);

__declspec(dllimport)
VOID	NTAPI	RtlEndWeakEnumerationHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
		                               RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff561837.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlEnlargedIntegerMultiply(LONG	Multiplicand,
			                           LONG Multiplier);

// https://msdn.microsoft.com/en-us/library/ff561838.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlEnlargedUnsignedDivide(ULARGE_INTEGER Dividend,
		                          DWORD          Divisor,
		                          LPDWORD        Remainder);		// optional

// https://msdn.microsoft.com/en-us/library/ff561838.aspx

__declspec(dllimport)
ULARGE_INTEGER	NTAPI	RtlEnlargedUnsignedMultiply(DWORD Multiplicand,
			                            DWORD Multiplier);

// https://msdn.microsoft.com/en-us/library/ms682608.aspx
// NOTE: forwarded from KERNEL32!EnterCriticalSection and KERNELBASE!EnterCriticalSection

__declspec(dllimport)
LONG	NTAPI	RtlEnterCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
RTL_DYNAMIC_HASH_TABLE_ENTRY *	NTAPI	RtlEnumerateEntryHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
					                           RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff552243.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlEnumerateGenericTable(RTL_GENERIC_TABLE *Table,
		                         BOOLEAN           Restart);

// https://msdn.microsoft.com/en-us/library/hh406458.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlEnumerateGenericTableAvl(RTL_AVL_TABLE *Table,
		                            BOOLEAN       Restart);

typedef	LONG	(NTAPI	RTL_AVL_MATCH_FUNCTION) (RTL_AVL_TABLE *Table,
			                         LPVOID        UserData,
			                         LPVOID        MatchData);

// https://msdn.microsoft.com/en-us/library/ff552245.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlEnumerateGenericTableLikeADirectory(RTL_AVL_TABLE          *Table,
		                                       RTL_AVL_MATCH_FUNCTION *MatchFunction,
		                                       LPVOID                 MatchData,
		                                       DWORD                  NextFlag,
		                                       LPVOID                 *RestartKey,
		                                       LPDWORD                DeleteCount,
		                                       LPVOID                 Buffer);

// https://msdn.microsoft.com/en-us/library/ff552247.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlEnumerateGenericTableWithoutSplaying(RTL_GENERIC_TABLE *Table,
		                                        LPVOID            *RestartKey);

// https://msdn.microsoft.com/en-us/library/hh406462.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlEnumerateGenericTableWithoutSplayingAvl(RTL_AVL_TABLE *Table,
		                                           LPVOID        *RestartKey);

typedef	LONG	(NTAPI	HEAP_ENUMERATION_ROUTINE) (LPCVOID HeapHandle,
			                           LPCVOID UserParam);

__declspec(dllimport)
LONG	NTAPI	RtlEnumProcessHeaps(HEAP_ENUMERATION_ROUTINE *HeapEnumerationRoutine,
		                    LPCVOID                  Param);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualComputerName(UNICODE_STRING *String1,
		                     UNICODE_STRING *String2);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualDomainName(UNICODE_STRING *String1,
		                   UNICODE_STRING *String2);

// https://msdn.microsoft.com/en-us/library/ff561842.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualLuid(LUID *Luid1,
		             LUID *Luid2);

#ifdef RtlEqualMemory
#undef RtlEqualMemory
#endif

#define RtlEqualMemory(Destination, Source, Length)	(RtlCompareMemory((Destination), (Source), (Length)) == (Length))

// https://msdn.microsoft.com/en-us/library/ff552256.aspx
// NOTE: like ADVAPI32!EqualPrefixSid

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualPrefixSid(SID *Sid1,
		                  SID *Sid2);

// https://msdn.microsoft.com/en-us/library/ff552260.aspx
// NOTE: like ADVAPI32!EqualSid

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualSid(SID *Sid1,
		            SID *Sid2);

// https://msdn.microsoft.com/en-us/library/ff561852.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualString(STRING  *String1,
		               STRING  *String2,
		               BOOLEAN CaseInSensitive);

// https://msdn.microsoft.com/en-us/library/ff561854.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlEqualUnicodeString(UNICODE_STRING *UnicodeString1,
		                      UNICODE_STRING *UnicodeString2,
		                      BOOLEAN        CaseInSensitive);

__declspec(dllimport)
VOID	NTAPI	RtlEraseUnicodeString(UNICODE_STRING *UnicodeString);

__forceinline
VOID	NTAPI	RtlEraseAnsiString(ANSI_STRING *AnsiString)
{
	RtlEraseUnicodeString((UNICODE_STRING *) AnsiString);
}

__forceinline
VOID	NTAPI	RtlEraseOEMString(OEM_STRING *OemString)
{
	RtlEraseUnicodeString((UNICODE_STRING *) OemString);
}

__forceinline
VOID	NTAPI	RtlEraseString(STRING *String)
{
	RtlEraseUnicodeString((UNICODE_STRING *) String);
}

#if _WIN32_WINNT >= 0x0600
#pragma warning(disable: 4201 4214)
typedef	union	_DL_OUI		// Organizationally Unique Identifier
{
	BYTE	Byte[3];
	struct
	{
		BYTE	Group : 1;
		BYTE	Local : 1;
	};
} DL_OUI, *LPDL_OUI, *PDL_OUI;
#pragma warning(default: 4201 4214)

typedef	union	_DL_EI48	// Extension Identifier
{
	BYTE	Byte[3];
} DL_EI48, *LPDL_EI48, *PDL_EI48;

#pragma warning(disable: 4201)
typedef	union	_DL_EUI48
{
	BYTE	Byte[6];
	struct
	{
		DL_OUI	Oui;
		DL_EI48	Ei48;
	};
} DL_EUI48, *LPDL_EUI48, *PDL_EUI48;
#pragma warning(default: 4201)

// https://msdn.microsoft.com/en-us/library/dn550721.aspx

__declspec(dllimport)
LPSTR	NTAPI	RtlEthernetAddressToStringA(DL_EUI48 *Addr,
		                            LPSTR    S);

__declspec(dllimport)
LPWSTR	NTAPI	RtlEthernetAddressToStringW(DL_EUI48 *Addr,
		                            LPWSTR   S);

// https://msdn.microsoft.com/en-us/library/dn550722.aspx

__declspec(dllimport)
LONG	NTAPI	RtlEthernetStringToAddressA(LPCSTR   S,
		                            LPCSTR   *Terminator,
		                            DL_EUI48 *Addr);

__declspec(dllimport)
LONG	NTAPI	RtlEthernetStringToAddressW(LPCWSTR  S,
		                            LPCWSTR  *Terminator,
		                            DL_EUI48 *Addr);

// https://msdn.microsoft.com/en-us/library/ms682658.aspx
// NOTE: forwarded from KERNEL32!ExitProcess and KERNELBASE!ExitProcess

__declspec(dllimport)
VOID	NTAPI	RtlExitUserProcess(NTSTATUS ExitStatus);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms682659.aspx
// NOTE: forwarded from KERNEL32!ExitThread and KERNELBASE!ExitThread

__declspec(dllimport)
VOID	NTAPI	RtlExitUserThread(NTSTATUS ExitStatus);

__declspec(dllimport)
LONG	NTAPI	RtlExpandEnvironmentStrings_U(LPVOID         Environment,
		                              UNICODE_STRING *Source,
		                              UNICODE_STRING *Destination,
		                              LPDWORD        ReturnedLength);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlExpandHashTable(RTL_DYNAMIC_HASH_TABLE *HashTable);
#endif

// https://msdn.microsoft.com/en-us/library/bb432425.aspx
// https://msdn.microsoft.com/en-us/library/ff561857.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlExtendedIntegerMultiply(LARGE_INTEGER Multiplicand,
			                           LONG          Multiplier);

// https://msdn.microsoft.com/en-us/library/bb432426.aspx
// https://msdn.microsoft.com/en-us/library/ff561859.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlExtendedLargeIntegerDivide(LARGE_INTEGER Dividend,
			                              DWORD         Divisor,
			                              LPDWORD       Remainder);	// optional

// https://msdn.microsoft.com/en-us/library/ff561860.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlExtendedMagicDivide(LARGE_INTEGER Dividend,
			                       LARGE_INTEGER MagicDivisor,
			                       BYTE          ShiftCount);

#if _WIN32_WINNT <= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlExtendHeap(LPCVOID HeapHandle,
		              DWORD   Flags,
		              LPVOID  Base,
		              SIZE_T  Size);
#endif

#if _MSC_VER >= 1610
__declspec(noreturn)
__forceinline
VOID	NTAPI	RtlFailFast(DWORD Code)
{
	__fastfail(Code);
}
#endif

#if 1
#pragma intrinsic(memset)
#else
#undef RtlFillMemory

// https://msdn.microsoft.com/en-us/library/ff561870.aspx
// NOTE: forwarded from KERNEL32!RtlFillMemory

__declspec(dllimport)
VOID	NTAPI	RtlFillMemory(LPVOID *Destination,
		              SIZE_T Length,
		              BYTE   Fill);
#endif

// https://msdn.microsoft.com/en-us/library/ff552263.aspx

__declspec(dllimport)
VOID	NTAPI	RtlFillMemoryUlong(LPVOID *Destination,
		                   SIZE_T Length,
		                   ULONG  Pattern);

// https://msdn.microsoft.com/en-us/library/ff552267.aspx

__declspec(dllimport)
VOID	NTAPI	RtlFillMemoryUlonglong(LPVOID    *Destination,
		                       SIZE_T    Length,
		                       ULONGLONG Pattern);

__declspec(dllimport)
VOID	NTAPI	RtlFinalReleaseOutOfProcessMemoryStream(RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LPVOID	NTAPI	RtlFindAceByType(ACL     *Acl,
		                 BYTE    AceType,
		                 LPDWORD AceIndex);
#endif

#if _WIN32_WINNT >= 0x0500
#define FIND_ACTIVATION_CONTEXT_SECTION_KEY_RETURN_ACTIVATION_CONTEXT	0x00000001UL
#define FIND_ACTIVATION_CONTEXT_SECTION_KEY_RETURN_FLAGS		0x00000002UL
#define FIND_ACTIVATION_CONTEXT_SECTION_KEY_RETURN_ASSEMBLY_METADATA	0x00000004UL

#define FIND_ACTIVATION_CONTEXT_SECTION_OUTFLAG_FOUND_IN_PROCESS_DEFAULT	0x00000001UL
#define FIND_ACTIVATION_CONTEXT_SECTION_OUTFLAG_FOUND_IN_SYSTEM_DEFAULT		0x00000002UL

__declspec(dllimport)
LONG	NTAPI	RtlFindActivationContextSectionGuid(DWORD                     Flags,
		                                    GUID                      *ExtensionGuid,
		                                    DWORD                     SectionId,
		                                    GUID                      *GuidToFind,
		                                    ACTCTX_SECTION_KEYED_DATA *ReturnedData);

__declspec(dllimport)
LONG	NTAPI	RtlFindActivationContextSectionString(DWORD                     Flags,
		                                      GUID                      *ExtensionGuid,
		                                      DWORD                     SectionId,
		                                      UNICODE_STRING            *StringToFind,
		                                      ACTCTX_SECTION_KEYED_DATA *ReturnedData);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff561873.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindClearBits(RTL_BITMAP *BitMapHeader,
		                 DWORD      NumberToFind,
		                 DWORD      HintIndex);

// https://msdn.microsoft.com/en-us/library/ff561874.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindClearBitsAndSet(RTL_BITMAP *BitMapHeader,
		                       DWORD      NumberToFind,
		                       DWORD      HintIndex);

// https://msdn.microsoft.com/en-us/library/ff561875.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindClearRuns(RTL_BITMAP     *BitMapHeader,
		                 RTL_BITMAP_RUN *RunArray,
		                 DWORD          SizeOfRunArray,
		                 BOOLEAN        LocateLongestRuns);

// https://msdn.microsoft.com/en-us/library/ff561877.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindFirstRunClear(RTL_BITMAP *BitMapHeader,
		                     LPDWORD    StartingIndex);

// https://msdn.microsoft.com/en-us/library/ff561879.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindLastBackwardRunClear(RTL_BITMAP *BitMapHeader,
		                            DWORD      FromIndex,
		                            LPDWORD    StartingRunIndex);

// https://msdn.microsoft.com/en-us/library/ff561880.aspx

__declspec(dllimport)
CCHAR	NTAPI	RtlFindLeastSignificantBit(QWORD Set);

// https://msdn.microsoft.com/en-us/library/ff561881.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindLongestRunClear(RTL_BITMAP *BitMapHeader,
		                       LPDWORD    StartingIndex);

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms648020.aspx

typedef	struct	_MESSAGE_RESOURCE_BLOCK
{
	DWORD	LowId;
	DWORD	HighId;
	DWORD	OffsetToEntries;
} MESSAGE_RESOURCE_BLOCK, *LPMESSAGE_RESOURCE_BLOCK, *PMESSAGE_RESOURCE_BLOCK;

// https://msdn.microsoft.com/en-us/library/ms648021.aspx

typedef	struct	_MESSAGE_RESOURCE_DATA
{
	DWORD			NumberOfBlocks;
	MESSAGE_RESOURCE_BLOCK	Blocks[ANYSIZE_ARRAY];
} MESSAGE_RESOURCE_DATA, *LPMESSAGE_RESOURCE_DATA, *PMESSAGE_RESOURCE_DATA;

// https://msdn.microsoft.com/en-us/library/ms648022.aspx

#define MESSAGE_RESOURCE_ANSI		0x0000
#define MESSAGE_RESOURCE_UNICODE	0x0001

typedef	struct	_MESSAGE_RESOURCE_ENTRY
{
	WORD	Length;
	WORD	Flags;
	BYTE	Text[ANYSIZE_ARRAY];
} MESSAGE_RESOURCE_ENTRY, *LPMESSAGE_RESOURCE_ENTRY, *PMESSAGE_RESOURCE_ENTRY;
#endif // _WINNT_

#define MESSAGE_RESOURCE_TABLE		11UL

__declspec(dllimport)
LONG	NTAPI	RtlFindMessage(LPCVOID                DllHandle,
		               DWORD                  MessageTableId,
		               DWORD                  MessageLanguageId,
		               DWORD                  MessageId,
		               MESSAGE_RESOURCE_ENTRY **MessageEntry);

// https://msdn.microsoft.com/en-us/library/ff561883.aspx

__declspec(dllimport)
CCHAR	NTAPI	RtlFindMostSignificantBit(QWORD Set);

// https://msdn.microsoft.com/en-us/library/ff561885.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindNextForwardRunClear(RTL_BITMAP *BitMapHeader,
		                           DWORD      FromIndex,
		                           LPDWORD    StartingRunIndex);

// https://msdn.microsoft.com/en-us/library/ff561890.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindSetBits(RTL_BITMAP *BitMapHeader,
		               DWORD      NumberToFind,
		               DWORD      HintIndex);

// https://msdn.microsoft.com/en-us/library/ff561895.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlFindSetBitsAndClear(RTL_BITMAP *BitMapHeader,
		                       DWORD      NumberToFind,
		                       DWORD      HintIndex);

typedef	struct	_UNICODE_PREFIX_TABLE_ENTRY
{
	WORD					NodeTypeCode;
	WORD					NameLength;
	struct	_UNICODE_PREFIX_TABLE_ENTRY	*NextPrefixTree;
	struct	_UNICODE_PREFIX_TABLE_ENTRY	*CaseMatch;
	RTL_SPLAY_LINKS				Links;
	UNICODE_STRING				*Prefix;
} UNICODE_PREFIX_TABLE_ENTRY, *LPUNICODE_PREFIX_TABLE_ENTRY, *PUNICODE_PREFIX_TABLE_ENTRY;

typedef	struct	_UNICODE_PREFIX_TABLE
{
	WORD				NodeTypeCode;
	WORD				NameLength;
	UNICODE_PREFIX_TABLE_ENTRY	*NextPrefixTree;
	UNICODE_PREFIX_TABLE_ENTRY	*LastNextEntry;
} UNICODE_PREFIX_TABLE, *LPUNICODE_PREFIX_TABLE, *PUNICODE_PREFIX_TABLE;

// https://msdn.microsoft.com/en-us/library/ff552272.aspx

__declspec(dllimport)
UNICODE_PREFIX_TABLE_ENTRY *	NTAPI	RtlFindUnicodePrefix(UNICODE_PREFIX_TABLE *PrefixTable,
					                     UNICODE_STRING       FullName,
					                     DWORD                CaseInsensitiveIndex);

#if _WIN32_WINNT < 0x0600
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms686309.aspx
// https://msdn.microsoft.com/en-us/library/ff563805.aspx

typedef	struct	_SLIST_ENTRY
{
	struct	_SLIST_ENTRY	*Next;
} SLIST_ENTRY, *LPSLIST_ENTRY, *PSLIST_ENTRY;

#if _M_IX86_
typedef	union	_SLIST_HEADER
{
	QWORD	Alignment;
	struct
	{
		SLIST_ENTRY	Next;
		WORD		Depth;
		WORD		CpuId;
	};
} SLIST_HEADER, *LPSLIST_HEADER, *PSLIST_HEADER;
#elif _M_AMD64_
typedef	union	__declspec(align(16))	_SLIST_HEADER
{
	struct
	{
		QWORD	Alignment;
		QWORD	Region;
	};
	struct
	{				// x64 16-byte header
		QWORD	Depth     : 16;
		QWORD	Sequence  : 48;
		QWORD	Reserved  : 4;
		QWORD	NextEntry : 60;	// last 4 bits are always 0
	} HeaderX64;
} SLIST_HEADER, *LPSLIST_HEADER, *PSLIST_HEADER;
#else
#error Only I386 and AMD64 processor architectures supported!
#endif
#endif // _WINNT_

// https://msdn.microsoft.com/en-us/library/aa904931.aspx

__declspec(dllimport)
SLIST_ENTRY *	NTAPI	RtlFirstEntrySList(SLIST_HEADER *ListHead);
#endif // _WIN32_WINNT

// NOTE: like ADVAPI32!FindFirstFreeAce

__declspec(dllimport)
BOOLEAN	NTAPI	RtlFirstFreeAce(ACL    *Acl,
		                LPVOID *FirstFree);

#if _WIN32_WINNT >= 0x0600
#ifndef _WINNT_
typedef	VOID	(NTAPI	*PFLS_CALLBACK_FUNCTION) (LPVOID lpFlsData);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlFlsAlloc(PFLS_CALLBACK_FUNCTION Callback,
		            LPDWORD                FlsIndex);

__declspec(dllimport)
LONG	NTAPI	RtlFlsFree(DWORD FlsIndex);
#endif // _WIN32_WINNT

__declspec(dllimport)
BOOLEAN	NTAPI	RtlFreeHandle(RTL_HANDLE_TABLE       *HandleTable,
		              RTL_HANDLE_TABLE_ENTRY *Handle);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlFlushSecureMemoryCache(LPVOID MemoryCache,
		                          SIZE_T MemoryLength);

// https://msdn.microsoft.com/en-us/library/ff899322.aspx

__declspec(dllimport)
LONG	NTAPI	RtlFormatCurrentUserKeyPath(UNICODE_STRING *CurrentUserKeyPath);

__declspec(dllimport)
LONG	NTAPI	RtlFormatMessage(LPCWSTR MessageFormat,
		                 DWORD   MaximumWidth,
		                 BOOLEAN IgnoreInserts,
		                 BOOLEAN ArgumentsAreAnsi,
		                 BOOLEAN ArgumentsAreAnArray,
		                 va_list *Arguments,
		                 LPWSTR  Buffer,
		                 DWORD   BufferLength,
		                 LPDWORD ReturnLength);

#if _WIN32_WINNT >= 0x0502
typedef	struct	_PARSE_MESSAGE_CONTEXT
{
	DWORD	fFlags;
	DWORD	cwSavColumn;
	SIZE_T	iwSrc;
	SIZE_T	iwDst;
	SIZE_T	iwDstSpace;
	va_list	lpvArgStart;
} PARSE_MESSAGE_CONTEXT, *LPPARSE_MESSAGE_CONTEXT, *PPARSE_MESSAGE_CONTEXT;

__declspec(dllimport)
LONG	NTAPI	RtlFormatMessageEx(LPCWSTR               MessageFormat,
		                   DWORD                 MaximumWidth,
		                   BOOLEAN               IgnoreInserts,
		                   BOOLEAN               ArgumentsAreAnsi,
		                   BOOLEAN               ArgumentsAreAnArray,
		                   va_list               *Arguments,
		                   LPWSTR                Buffer,
		                   DWORD                 BufferLength,
		                   LPDWORD               ReturnLength,
		                   PARSE_MESSAGE_CONTEXT *ParseContext);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlFreeActivationContextStack(ACTIVATION_CONTEXT_STACK *ActivationContextStack);

// https://msdn.microsoft.com/en-us/library/ff561899.aspx
// https://msdn.microsoft.com/en-us/library/ms648416.aspx

__declspec(dllimport)
VOID	NTAPI	RtlFreeAnsiString(ANSI_STRING *AnsiString);

// https://msdn.microsoft.com/en-us/library/aa366701.aspx
// https://msdn.microsoft.com/en-us/library/ff552276.aspx
// https://msdn.microsoft.com/en-us/library/mt723417.aspx
// NOTE: forwarded from KERNEL32!HeapFree and KERNELBASE!HeapFree

__declspec(dllimport)
BOOL	NTAPI	RtlFreeHeap(LPCVOID HeapHandle,
		            DWORD   Flags,
		            LPCVOID HeapBase);

// https://msdn.microsoft.com/en-us/library/ms648417.aspx

__declspec(dllimport)
VOID	NTAPI	RtlFreeOemString(OEM_STRING *OemString);

// https://msdn.microsoft.com/en-us/library/ff552283.aspx
// NOTE: like ADVAPI32!FreeSid

__declspec(dllimport)
BOOL	NTAPI	RtlFreeSid(SID *Sid);

__declspec(dllimport)
VOID	NTAPI	RtlFreeThreadActivationContextStack(VOID);

// https://msdn.microsoft.com/en-us/library/ff561903.aspx
// https://msdn.microsoft.com/en-us/library/ms648418.aspx

__declspec(dllimport)
VOID	NTAPI	RtlFreeUnicodeString(UNICODE_STRING *UnicodeString);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlFreeUserStack(LPCVOID AllocationBase);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlFreeUserThreadStack(HANDLE ProcessHandle,
		                       HANDLE ThreadHandle);

// https://msdn.microsoft.com/en-us/library/ff552285.aspx

typedef	struct	_GENERATE_NAME_CONTEXT
{
	WORD	Checksum;
	BOOLEAN	ChecksumInserted;
	BYTE	NameLength;		// filename without extension,
	WCHAR	NameBuffer[8];		// e.g. L"ntoskrnl"
	DWORD	ExtensionLength;	// extension including dot,
	WCHAR	ExtensionBuffer[4];	// e.g. L".exe"
	DWORD	LastIndexValue;
} GENERATE_NAME_CONTEXT;

__declspec(dllimport)
LONG	NTAPI	RtlGenerate8dot3Name(UNICODE_STRING        *Name,
		                     BOOLEAN               AllowExtendedCharacters,
		                     GENERATE_NAME_CONTEXT *Context,
		                     UNICODE_STRING        *Name8dot3);

// https://msdn.microsoft.com/en-us/library/ff552288.aspx
// NOTE: like ADVAPI32!GetAce

__declspec(dllimport)
LONG	NTAPI	RtlGetAce(ACL    *Acl,
		          DWORD  AceIndex,
		          LPVOID *Ace);

__declspec(dllimport)
LONG	NTAPI	RtlGetActiveActivationContext(ACTIVATION_CONTEXT **ActivationContext);

__declspec(dllimport)
VOID	NTAPI	RtlGetCallersAddress(LPVOID *CallersAddress,
		                     LPVOID *CallersCaller);

// https://msdn.microsoft.com/en-us/library/ff552291.aspx

__declspec(dllimport)
LONG	NTAPI	RtlGetCompressionWorkSpaceSize(WORD    CompressionFormatAndEngine,
		                               LPDWORD CompressBufferWorkSpaceSize,
		                               LPDWORD CompressFragmentWorkSpaceSize);

// NOTE: like ADVAPI32!GetSecurityDescriptorControl

__declspec(dllimport)
LONG	NTAPI	RtlGetControlSecurityDescriptor(SECURITY_DESCRIPTOR         *SecurityDescriptor,
		                                SECURITY_DESCRIPTOR_CONTROL *Control,
		                                LPDWORD                     Revision);

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
LONG	NTAPI	RtlGetCriticalSectionRecursionCount(RTL_CRITICAL_SECTION *CriticalSection);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlGetCurrentDirectory_U(DWORD  BufferLength,
		                         LPWSTR UnicodeString);

__declspec(dllimport)
PEB *	NTAPI	RtlGetCurrentPeb(VOID);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms683181.aspx
// NOTE: forwarded from KERNEL32!GetCurrentProcessorNumber

__declspec(deprecated("obsolete function, may be altered or unavailable in future versions"))
__declspec(dllimport)
DWORD	NTAPI	RtlGetCurrentProcessorNumber(VOID);

// https://msdn.microsoft.com/en-us/library/dd405487.aspx
// NOTE: forwarded from KERNEL32!GetCurrentProcessorNumberEx

__declspec(dllimport)
VOID	NTAPI	RtlGetCurrentProcessorNumberEx(PROCESSOR_NUMBER *ProcessorNumber);
#endif // _WIN32_WINNT

#if 1
#define RtlGetCurrentProcessId()	(NtCurrentTeb()->ClientId.UniqueProcess)
#else
__forceinline
LPVOID	NTAPI	RtlGetCurrentProcessId(VOID)
{
	return NtCurrentTeb()->ClientId.UniqueProcess;
}
#endif

#if 1
#define RtlGetCurrentThreadId()		(NtCurrentTeb()->ClientId.UniqueThread)
#else
__forceinline
LPVOID	NTAPI	RtlGetCurrentThreadId(VOID)
{
	return NtCurrentTeb()->ClientId.UniqueThread;
}
#endif

#if _WIN32_WINNT >= 0x0A00
__declspec(dllimport)
DWORD	NTAPI	RtlGetCurrentServiceSessionId(VOID);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
HANDLE	NTAPI	RtlGetCurrentTransaction(VOID);
#endif

// https://msdn.microsoft.com/en-us/library/ff552292.aspx
// NOTE: like ADVAPI32!GetSecurityDescriptorDacl

__declspec(dllimport)
LONG	NTAPI	RtlGetDaclSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                             BOOLEAN             *DaclPresent,
		                             ACL                 **Dacl,
		                             BOOLEAN             *DaclDefaulted);

// https://msdn.microsoft.com/en-us/library/ff552297.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlGetElementGenericTable(RTL_GENERIC_TABLE *Table,
		                          DWORD             Index);

// https://msdn.microsoft.com/en-us/library/ff552301.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlGetElementGenericTableAvl(RTL_AVL_TABLE *Table,
		                             DWORD         Index);

#if _WIN32_WINNT >= 0x0601
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ff561905.aspx

__declspec(dllimport)
QWORD	NTAPI	RtlGetEnabledExtendedFeatures(DWORD64 FeatureMask);

__declspec(dllimport)
DWORD	NTAPI	RtlGetExtendedContextLength(DWORD   ContextFlags,
		                            LPDWORD ContextLength);

__declspec(dllimport)
QWORD	NTAPI	RtlGetExtendedFeaturesMask(CONTEXT_EX *ContextEx);
#endif // _WINNT_
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0603
__declspec(dllimport)
DWORD	NTAPI	RtlGetExtendedContextLength2(DWORD   ContextFlags,
		                             LPDWORD ContextLength,
		                             DWORD64 XStateCompactionMask);
#endif

typedef	struct	_TEB_ACTIVE_FRAME_CONTEXT
{
	DWORD	Flags;
	LPCSTR	FrameName;
} TEB_ACTIVE_FRAME_CONTEXT, *LPTEB_ACTIVE_FRAME_CONTEXT, *PTEB_ACTIVE_FRAME_CONTEXT;

typedef	struct	_TEB_ACTIVE_FRAME
{
	DWORD				Flags;
	struct	_TEB_ACTIVE_FRAME	*Previous;
	TEB_ACTIVE_FRAME_CONTEXT	*Context;
} TEB_ACTIVE_FRAME, *LPTEB_ACTIVE_FRAME, *PTEB_ACTIVE_FRAME;

__declspec(dllimport)
TEB_ACTIVE_FRAME *	NTAPI	RtlGetFrame(VOID);

#ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/bb432427.aspx

__declspec(dllimport)
LIST_ENTRY *	NTAPI	RtlGetFunctionTableListHead(VOID);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlGetFullPathName_U(LPCWSTR FileName,
		                     DWORD   BufferLength,
		                     LPWSTR  Buffer,
		                     LPWSTR  *FilePart);

__declspec(dllimport)
LONG	NTAPI	RtlGetFullPathName_UEx(LPCWSTR       FileName,
		                       DWORD         BufferLength,
		                       LPWSTR        Buffer,
		                       LPWSTR        *FilePart,
		                       RTL_PATH_TYPE *PathType);

__declspec(dllimport)
LONG	NTAPI	RtlGetFullPathName_UstrEx(UNICODE_STRING *FileName,
		                          UNICODE_STRING *StaticString,
		                          UNICODE_STRING *DynamicString,
		                          UNICODE_STRING **StringUsed,
		                          SIZE_T         *FilePartSize,
		                          BOOLEAN        *NameInvalid,
		                          RTL_PATH_TYPE  *PathType,
		                          SIZE_T         *BytesRequired);

// https://msdn.microsoft.com/en-us/library/ff552305.aspx
// NOTE: like ADVAPI32!GetSecurityDescriptorGroup

__declspec(dllimport)
LONG	NTAPI	RtlGetGroupSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                              SID                 **Group,
		                              BOOLEAN             *GroupDefaulted);

__declspec(dllimport)
LONG	NTAPI	RtlGetLastNtStatus(VOID);

// https://msdn.microsoft.com/en-us/library/ms679360.aspx
// NOTE: forwarded from KERNEL32!GetLastError

__declspec(dllimport)
DWORD	NTAPI	RtlGetLastWin32Error(VOID);

__declspec(dllimport)
LONG	NTAPI	RtlGetLengthWithoutLastFullDosOrNtPathElement(DWORD          Flags,
		                                              UNICODE_STRING *Path,
		                                              LPDWORD        Length);

#define RtlGetLengthWithoutTrailingPathSeparators	RtlGetLengthWithoutTrailingPathSeperators

__declspec(dllimport)
LONG	NTAPI	RtlGetLengthWithoutTrailingPathSeperators(DWORD          Flags,
		                                          UNICODE_STRING *Path,
		                                          LPDWORD        Length);

__declspec(dllimport)
DWORD	NTAPI	RtlGetLongestNtPathLength(VOID);

__declspec(dllimport)
LONG	NTAPI	RtlGetNativeSystemInformation(DWORD   SystemInformationClass,
		                              LPVOID  NativeSystemInformation,
		                              DWORD   InformationLength,
		                              LPDWORD ReturnLength);

#if _WIN32_WINNT >= 0x0601
typedef	struct	_RTL_DYNAMIC_HASH_TABLE_CONTEXT
{
	LIST_ENTRY	*ChainHead;
	LIST_ENTRY	*PrevLinkage;
	DWORD_PTR	Signature;
} RTL_DYNAMIC_HASH_TABLE_CONTEXT, *LPRTL_DYNAMIC_HASH_TABLE_CONTEXT, *PRTL_DYNAMIC_HASH_TABLE_CONTEXT;

__declspec(dllimport)
RTL_DYNAMIC_HASH_TABLE_ENTRY *	NTAPI	RtlGetNextEntryHashTable(RTL_DYNAMIC_HASH_TABLE         *HashTable,
					                         RTL_DYNAMIC_HASH_TABLE_CONTEXT *Context);
#endif // _WIN32_WINNT

__declspec(dllimport)
DWORD	NTAPI	RtlGetNtGlobalFlags(VOID);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlGetNtProductType(LPDWORD ProductType);

#if _WIN32_WINNT >= 0x0A00
__declspec(dllimport)
LPCWSTR	NTAPI	RtlGetNtSystemRoot(VOID);
#endif

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
VOID	NTAPI	RtlGetNtVersionNumbers(LPDWORD MajorVersion,
		                       LPDWORD MinorVersion,
		                       LPDWORD BuildNumber);
#endif

// https://msdn.microsoft.com/en-us/library/ff552976.aspx
// NOTE: like ADVAPI32!GetSecurityDescriptorOwner

__declspec(dllimport)
LONG	NTAPI	RtlGetOwnerSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                              SID                 **Owner,
		                              BOOLEAN             *OwnerDefaulted);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlGetParentLocaleName(LPCWSTR        LocaleName,
		                       UNICODE_STRING *ParentLocaleName,
		                       DWORD          Flags,
		                       BOOLEAN        AllocateDestinationString);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlGetProcessHeaps(DWORD  MaxNumberOfHeaps,
		                   LPVOID *HeapArray);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlGetProcessPreferredUILanguages(DWORD   Flags,
		                                  LPCWSTR LanguagesBuffer,
		                                  DWORD   LanguagesBufferLength,
		                                  LPDWORD ReturnedLength);
#endif

#if _WIN32_WINNT >= 0x0600
// NOTE: like KERNEL32!GetProductInfo

__declspec(dllimport)
BOOLEAN	NTAPI	RtlGetProductInfo(DWORD   OSMajorVersion,
		                  DWORD   OSMinorVersion,
		                  DWORD   SPMajorVersion,
		                  DWORD   SPMinorVersion,
		                  LPDWORD ProductType);
#endif

// https://msdn.microsoft.com/en-us/library/ff552980.aspx
// NOTE: like ADVAPI32!GetSecurityDescriptorSacl

__declspec(dllimport)
LONG	NTAPI	RtlGetSaclSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                             BOOLEAN             *SaclPresent,
		                             ACL                 **Sacl,
		                             BOOLEAN             *SaclDefaulted);

// NOTE: like ADVAPI32!GetSecurityDescriptorRMControl

__declspec(dllimport)
BOOLEAN	NTAPI	RtlGetSecurityDescriptorRMControl(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                                  LPBYTE              RMControl);

#if _WIN32_WINNT >= 0x0501
typedef	enum	_RTL_BSD_ITEM_TYPE
{
	RtlBsdItemVersionNumber = 0,
	RtlBsdItemProductType,
	RtlBsdItemAabEnabled,
	RtlBsdItemAabTimeout,
	RtlBsdItemBootGood,
	RtlBsdItemBootShutdown,
	RtlBsdItemMax
} RTL_BSD_ITEM_TYPE;

__declspec(dllimport)
LONG	NTAPI	RtlGetSetBootStatusData(HANDLE            Handle,
		                        BOOLEAN           Get,
		                        RTL_BSD_ITEM_TYPE DataItem,
		                        LPVOID            DataBuffer,
		                        DWORD             DataBufferLength,
		                        LPDWORD           ByteRead);
#endif

#if 0 // _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlGetSystemPreferredUILanguages(DWORD   Flags,
		                                 VOID,
		                                 LPCWSTR LanguagesBuffer,
		                                 DWORD   LanguagesBufferLength,
		                                 LPDWORD ReturnedLength);
#endif

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
DWORD	NTAPI	RtlGetThreadErrorMode(VOID);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlGetThreadPreferredUILanguages(DWORD   Flags,
		                                 LPCWSTR LanguagesBuffer,
		                                 DWORD   LanguagesBufferLength,
		                                 LPDWORD ReturnedLength);
#endif

#if _WIN32_WINNT >= 0x0502
#define RtlGetTickCount	NtGetTickCount
#else
__declspec(dllimport)
DWORD	NTAPI	RtlGetTickCount(VOID);
#endif

// https://msdn.microsoft.com/en-us/library/bb432428.aspx

typedef	struct	_RTL_UNLOAD_EVENT_TRACE
{
	LPVOID	BaseAddress;	// Base address of DLL
	SIZE_T	SizeOfImage;	// Size of image
	DWORD	Sequence;	// Sequence number for this event
	DWORD	TimeDateStamp;	// Time and date of image
	DWORD	CheckSum;	// Image checksum
	WCHAR	ImageName[32];	// Image name
} RTL_UNLOAD_EVENT_TRACE, *LPRTL_UNLOAD_EVENT_TRACE, *PRTL_UNLOAD_EVENT_TRACE;

__declspec(dllimport)
RTL_UNLOAD_EVENT_TRACE *	NTAPI	RtlGetUnloadEventTrace(VOID);

// https://msdn.microsoft.com/en-us/library/cc678403.aspx

__declspec(dllimport)
VOID	NTAPI	RtlGetUnloadEventTraceEx(LPDWORD *ElementSize,
		                         LPDWORD *ElementCount,
		                         LPVOID  *EventTrace);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlGetUserInfoHeap(LPCVOID HeapHandle,
		                   DWORD   Flags,
		                   LPVOID  BaseAddress,
		                   LPVOID  *UserValue,
		                   LPDWORD UserFlags);

#if 0 // _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlGetUserPreferredUILanguages(DWORD   Flags,
		                               VOID,
		                               LPCWSTR LanguagesBuffer,
		                               DWORD   LanguagesBufferLength,
		                               LPDWORD ReturnedLength);
#endif

// https://msdn.microsoft.com/en-us/library/ff561910.aspx
// https://msdn.microsoft.com/en-us/library/mt723418.aspx

__declspec(dllimport)
LONG	NTAPI	RtlGetVersion(RTL_OSVERSIONINFOEXW *VersionInfo);

#ifdef _WIN64
#if _WIN32_WINNT >= 0x0602
// https://msdn.microsoft.com/en-us/library/hh405354.aspx

__declspec(dllimport)
VOID	NTAPI	RtlGrowFunctionTable(LPVOID DynamicTable,
		                     DWORD  NewEntryCount);
#endif
#endif

// https://msdn.microsoft.com/en-us/library/ff561913.aspx

__declspec(dllimport)
LONG	NTAPI	RtlGUIDFromString(UNICODE_STRING *GuidString,
		                  GUID           *Guid);

// https://msdn.microsoft.com/en-us/library/ff561915.aspx

__declspec(dllimport)
LONG	NTAPI	RtlHashUnicodeString(UNICODE_STRING *String,
		                     BOOLEAN        CaseInSensitive,
		                     DWORD          HashAlgorithm,
		                     LPDWORD        HashValue);

// https://msdn.microsoft.com/en-us/library/ff552983.aspx
// NOTE: like ADVAPI32!GetSidIdentifierAuthority

__declspec(dllimport)
SID_IDENTIFIER_AUTHORITY *	NTAPI	RtlIdentifierAuthoritySid(SID *Sid);

// NOTE: like DBGHELP!ImageDirectoryEntryToData and IMAGEHLP!ImageDirectoryEntryToData

__declspec(dllimport)
LPVOID	NTAPI	RtlImageDirectoryEntryToData(LPCVOID BaseOfImage,
		                             BOOLEAN MappedAsImage,
		                             WORD    DirectoryEntry,
		                             LPDWORD Size);

// NOTE: like DBGHELP!ImageNtHeader and IMAGEHLP!ImageNtHeader

__declspec(dllimport)
IMAGE_NT_HEADERS *	NTAPI	RtlImageNtHeader(LPCVOID BaseOfImage);

#define RTL_IMAGE_NT_HEADER_EX_FLAG_NO_RANGE_CHECK	0x00000001UL

__declspec(dllimport)
LONG	NTAPI	RtlImageNtHeaderEx(DWORD            Flags,
		                   LPCVOID          BaseOfImage,
		                   QWORD            Size,
		                   IMAGE_NT_HEADERS **NtHeaders);

// NOTE: like DBGHELP!ImageRvaToSection and IMAGEHLP!ImageRvaToSection

__declspec(dllimport)
IMAGE_SECTION_HEADER *	NTAPI	RtlImageRvaToSection(IMAGE_NT_HEADERS *NtHeaders,
				                     LPCVOID          BaseOfImage,
				                     DWORD            Rva);

// NOTE: like DBGHELP!ImageRvaToVa and IMAGEHLP!ImageRvaToVa

__declspec(dllimport)
LPVOID	NTAPI	RtlImageRvaToVa(IMAGE_NT_HEADERS     *NtHeaders,
		                LPCVOID              BaseOfImage,
		                DWORD                Rva,
		                IMAGE_SECTION_HEADER *LastSection);

// NOTE: like ADVAPI32!ImpersonateSelf and KERNELBASE!ImpersonateSelf

__declspec(dllimport)
LONG	NTAPI	RtlImpersonateSelf(SECURITY_IMPERSONATION_LEVEL ImpersonationLevel);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlImpersonateSelfEx(SECURITY_IMPERSONATION_LEVEL ImpersonationLevel,
		                     ACCESS_MASK                  AdditionalAccess,
		                     HANDLE                       *ThreadToken);
#endif

// https://msdn.microsoft.com/en-us/library/ff561918.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitAnsiString(ANSI_STRING *DestinationString,
		                  LPCSTR      SourceString);

__declspec(dllimport)
LONG	NTAPI	RtlInitAnsiStringEx(ANSI_STRING *DestinationString,
		                    LPCSTR      SourceString);

// https://msdn.microsoft.com/en-us/library/ff561920.aspx

__inline
VOID	NTAPI	RtlInitEmptyAnsiString(ANSI_STRING *DestinationString,
		                       LPSTR       Buffer,
		                       WORD        BufferSize)
{
	// https://msdn.microsoft.com/en-us/library/ff540605.aspx

	DestinationString->Length = 0;
	DestinationString->MaximumLength = BufferSize;
	DestinationString->Buffer = Buffer;
}

__inline
VOID	NTAPI	RtlInitEmptyOEMString(OEM_STRING *DestinationString,
		                      LPSTR      Buffer,
		                      WORD       BufferSize)
{
	// https://msdn.microsoft.com/en-us/library/ff558741.aspx

	DestinationString->Length = 0;
	DestinationString->MaximumLength = BufferSize;
	DestinationString->Buffer = Buffer;
}

__inline
VOID	NTAPI	RtlInitEmptyString(STRING *DestinationString,
		                   LPSTR  Buffer,
		                   WORD   BufferSize)
{
	// https://msdn.microsoft.com/en-us/library/ms648424.aspx

	DestinationString->Length = 0;
	DestinationString->MaximumLength = BufferSize;
	DestinationString->Buffer = Buffer;
}

// https://msdn.microsoft.com/en-us/library/ff561922.aspx

__inline
VOID	NTAPI	RtlInitEmptyUnicodeString(UNICODE_STRING *DestinationString,
		                          LPWSTR         Buffer,
		                          WORD           BufferSize)
{
	// https://msdn.microsoft.com/en-us/library/ff564879.aspx

	DestinationString->Length = 0;
	DestinationString->MaximumLength = BufferSize;
	DestinationString->Buffer = Buffer;
}

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlInitEnumerationHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
		                            RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlInitializeAtomPackage(DWORD AllocationTag);

// https://msdn.microsoft.com/en-us/library/ff561925.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitializeBitMap(RTL_BITMAP *BitMapHeader,
		                    LPDWORD    BitMapBuffer,
		                    DWORD      SizeOfBitMap);

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms683469.aspx
// NOTE: forwarded from KERNEL32!InitializeConditionVariable

__declspec(dllimport)
VOID	NTAPI	RtlInitializeConditionVariable(RTL_CONDITION_VARIABLE *ConditionVariable);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlInitializeContext(HANDLE  ProcessHandle,
		                     CONTEXT *ContextRecord,
		                     LPVOID  Parameter,
		                     LPVOID  InitialPc,
		                     LPVOID  InitialSp);

// https://msdn.microsoft.com/en-us/library/ms683472.aspx
// NOTE: forwarded from KERNEL32!InitializeCriticalSection and KERNELBASE!InitializeCriticalSection

__declspec(dllimport)
LONG	NTAPI	RtlInitializeCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

__declspec(dllimport)
LONG	NTAPI	RtlInitializeCriticalSectionAndSpinCount(RTL_CRITICAL_SECTION *CriticalSection,
		                                         DWORD                SpinCount);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlInitializeCriticalSectionEx(RTL_CRITICAL_SECTION *CriticalSection,
		                               DWORD                SpinCount,
		                               DWORD                Flags);
#endif

#ifndef _WIN64
#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
VOID	NTAPI	RtlInitializeExceptionChain(LPVOID Unknown);
#endif
#endif

#if _WIN32_WINNT >= 0x0601
#ifndef _WINNT_
__declspec(dllimport)
DWORD	NTAPI	RtlInitializeExtendedContext(LPVOID     Context,
		                             DWORD      ContextFlags,
		                             CONTEXT_EX **ContextEx);
#endif
#endif

#if _WIN32_WINNT >= 0x0603
__declspec(dllimport)
DWORD	NTAPI	RtlInitializeExtendedContext2(LPVOID     Context,
		                              DWORD      ContextFlags,
		                              CONTEXT_EX **ContextEx,
		                              DWORD64    XStateCompactionMask);
#endif

// https://msdn.microsoft.com/en-us/library/ff552989.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitializeGenericTable(RTL_GENERIC_TABLE            *Table,
		                          RTL_GENERIC_COMPARE_ROUTINE  *CompareRoutine,
		                          RTL_GENERIC_ALLOCATE_ROUTINE *AllocateRoutine,
		                          RTL_GENERIC_FREE_ROUTINE     *FreeRoutine,
		                          LPVOID                       TableContext);

// https://msdn.microsoft.com/en-us/library/hh406465.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitializeGenericTableAvl(RTL_AVL_TABLE            *Table,
		                             RTL_AVL_COMPARE_ROUTINE  *CompareRoutine,
		                             RTL_AVL_ALLOCATE_ROUTINE *AllocateRoutine,
		                             RTL_AVL_FREE_ROUTINE     *FreeRoutine,
		                             LPVOID                   TableContext);

__declspec(dllimport)
VOID	NTAPI	RtlInitializeHandleTable(DWORD            MaximumNumberOfHandles,
		                         DWORD            SizeOfHandleTableEntry,
		                         RTL_HANDLE_TABLE *HandleTable);

__declspec(dllimport)
VOID	NTAPI	RtlInitializeResource(RTL_RESOURCE *Resource);

__declspec(dllimport)
LONG	NTAPI	RtlInitializeRXact(HANDLE            RootRegistryKey,
		                   BOOLEAN           CommitIfNecessary,
		                   RTL_RXACT_CONTEXT **RXactContext);

// https://msdn.microsoft.com/en-us/library/ff552998.aspx
// NOTE: like ADVAPI32!InitializeSid

__declspec(dllimport)
LONG	NTAPI	RtlInitializeSid(SID                      *Sid,
		                 SID_IDENTIFIER_AUTHORITY *IdentifierAuthority,
		                 BYTE                     SubAuthorityCount);

// https://msdn.microsoft.com/en-us/library/mt426739.aspx

__declspec(dllimport)
LONG	CDECL	RtlInitializeSidEx(SID                      *Sid,
		                   SID_IDENTIFIER_AUTHORITY *IdentifierAuthority,
		                   BYTE                     SubAuthorityCount,
		                   ...);

// https://msdn.microsoft.com/en-us/library/aa904932.aspx
// NOTE: forwarded from KERNEL32!InitializeSListHead and KERNELBASE!InitializeSListHead

__declspec(dllimport)
VOID	NTAPI	RtlInitializeSListHead(SLIST_HEADER *ListHead);

// https://msdn.microsoft.com/en-us/library/ff553010.aspx

__inline
VOID	NTAPI	RtlInitializeSplayLinks(RTL_SPLAY_LINKS *Links)
{
	Links->Parent = Links;
	Links->LeftChild = NULL;
	Links->RightChild = NULL;
}

__declspec(dllimport)
LONG	NTAPI	RtlInitializeStackTraceDataBase(LPVOID CommitBase,
		                                SIZE_T CommitSize,
		                                SIZE_T ReserveSize);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/ms683483.aspx
// NOTE: forwarded from KERNEL32!InitializeSRWLock and KERNELBASE!InitializeSRWLock

__declspec(dllimport)
VOID	NTAPI	RtlInitializeSRWLock(RTL_SRWLOCK *SRWLock);
#endif

// https://msdn.microsoft.com/en-us/library/ff553015.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitializeUnicodePrefix(UNICODE_PREFIX_TABLE *PrefixTable);

__declspec(dllimport)
VOID	NTAPI	RtlInitMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

typedef	struct	_NLSTABLEINFO
{
	CPTABLEINFO	OemTableInfo;
	CPTABLEINFO	AnsiTableInfo;
	LPWORD		UpperCaseTable;
	LPWORD		LowerCaseTable;
} NLSTABLEINFO, *LPNLSTABLEINFO, *PNLSTABLEINFO;

__declspec(dllimport)
VOID	NTAPI	RtlInitNlsTables(LPWORD       AnsiNlsBase,
		                 LPWORD       OemNlsBase,
		                 LPWORD       LanguageNlsBase,
		                 NLSTABLEINFO *TableInfo);

__declspec(dllimport)
VOID	NTAPI	RtlInitOutOfProcessMemoryStream(RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

// https://msdn.microsoft.com/en-us/library/ff561929.aspx
// https://msdn.microsoft.com/en-us/library/ms648419.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitString(STRING *DestinationString,
		              LPCSTR SourceString);

// https://msdn.microsoft.com/en-us/library/ff561934.aspx
// https://msdn.microsoft.com/en-us/library/ms648420.aspx

__declspec(dllimport)
VOID	NTAPI	RtlInitUnicodeString(UNICODE_STRING *DestinationString,
		                     LPCWSTR        SourceString);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	NTAPI	RtlInitUnicodeStringEx(UNICODE_STRING *DestinationString,
		                       LPCWSTR        SourceString);
#endif

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlInitWeakEnumerationHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
		                                RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);
#endif

// https://msdn.microsoft.com/en-us/library/ff553017.aspx

__inline
VOID	NTAPI	RtlInsertAsLeftChild(RTL_SPLAY_LINKS *ParentLinks,
		                     RTL_SPLAY_LINKS *ChildLinks)
{
//	NT_ASSERT(ParentLinks->LeftChild == NULL)
	ParentLinks->LeftChild = ChildLinks;

//	NT_ASSERT(ChildLinks->Parent == ChildLinks)
	ChildLinks->Parent = ParentLinks;
}

// https://msdn.microsoft.com/en-us/library/ff553022.aspx

__inline
VOID	NTAPI	RtlInsertAsRightChild(RTL_SPLAY_LINKS *ParentLinks,
		                      RTL_SPLAY_LINKS *ChildLinks)
{
//	NT_ASSERT(ParentLinks->RightChild == NULL)
	ParentLinks->RightChild = ChildLinks;

//	NT_ASSERT(ChildLinks->Parent == ChildLinks)
	ChildLinks->Parent = ParentLinks;
}

// https://msdn.microsoft.com/en-us/library/ff553024.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlInsertElementGenericTable(RTL_GENERIC_TABLE *Table,
		                             LPVOID            Buffer,
		                             DWORD             BufferSize,
		                             BOOLEAN           *NewElement);

// https://msdn.microsoft.com/en-us/library/hh406468.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlInsertElementGenericTableAvl(RTL_AVL_TABLE *Table,
		                                LPVOID        Buffer,
		                                DWORD         BufferSize,
		                                BOOLEAN       *NewElement);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlInsertEntryHashTable(RTL_DYNAMIC_HASH_TABLE         *HashTable,
		                        RTL_DYNAMIC_HASH_TABLE_ENTRY   *Entry,
		                        DWORD_PTR                      Signature,
		                        RTL_DYNAMIC_HASH_TABLE_CONTEXT *Context);
#endif

// https://msdn.microsoft.com/en-us/library/ff553038.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlInsertUnicodePrefix(UNICODE_PREFIX_TABLE       *PrefixTable,
		                       UNICODE_STRING             *Prefix,
		                       UNICODE_PREFIX_TABLE_ENTRY *PrefixTableEntry);

#ifdef _WIN64
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms680595.aspx
// NOTE: forwarded from KERNEL32!RtlInstallFunctionTableCallback

__declspec(dllimport)
BOOLEAN	NTAPI	RtlInstallFunctionTableCallback(QWORD                         TableIdentifier,
		                                QWORD                         BaseAddress,
		                                DWORD                         Length,
		                                GET_RUNTIME_FUNCTION_CALLBACK *Callback,
		                                LPVOID                        Context,
		                                LPCWSTR                       OutOfProcessCallbackDll);
#endif
#endif

// https://msdn.microsoft.com/en-us/library/ff561939.aspx

__declspec(dllimport)
LONG	NTAPI	RtlInt64ToUnicodeString(QWORD          Value,
		                        DWORD          Base,	// 0, 2, 8, 10, 16
		                        UNICODE_STRING *String);

// https://msdn.microsoft.com/en-us/library/ff561941.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIntegerToUnicodeString(DWORD          Value,
		                          DWORD          Base,	// 0, 2, 8, 10, 16
		                          UNICODE_STRING *String);

// https://msdn.microsoft.com/en-us/library/ff561947.aspx

#ifdef _WIN64
#define RtlIntPtrToUnicodeString(Value, Base, String)	RtlInt64ToUnicodeString(Value, Base, String)
#else
#define RtlIntPtrToUnicodeString(Value, Base, String)	RtlIntegerToUnicodeString(Value, Base, String)
#endif

// https://msdn.microsoft.com/en-us/library/ms683562.aspx
// NOTE: forwarded from KERNEL32!InterlockedCompareExchange64

__declspec(dllimport)
LONGLONG	CDECL	RtlInterlockedCompareExchange64(LONGLONG volatile *Destination,
			                                LONGLONG          Exchange,
			                                LONGLONG          Comparand);

// https://msdn.microsoft.com/en-us/library/aa904933.aspx
// https://msdn.microsoft.com/en-us/library/ms683612.aspx
// NOTE: forwarded from KERNEL32!InterlockedFlushSList and KERNELBASE!InterlockedFlushSList

__declspec(dllimport)
SLIST_ENTRY *	NTAPI	RtlInterlockedFlushSList(SLIST_HEADER *ListHead);

// https://msdn.microsoft.com/en-us/library/aa904934.aspx
// https://msdn.microsoft.com/en-us/library/ms683648.aspx
// NOTE: forwarded from KERNEL32!InterlockedPopEntrySList and KERNELBASE!InterlockedPopEntrySList

__declspec(dllimport)
SLIST_ENTRY *	NTAPI	RtlInterlockedPopEntrySList(SLIST_HEADER *ListHead);

// https://msdn.microsoft.com/en-us/library/aa904935.aspx
// https://msdn.microsoft.com/en-us/library/ms684020.aspx
// NOTE: forwarded from KERNEL32!InterlockedPushEntrySList and KERNELBASE!InterlockedPushEntrySList

__declspec(dllimport)
SLIST_ENTRY *	NTAPI	RtlInterlockedPushEntrySList(SLIST_HEADER *ListHead,
			                             SLIST_ENTRY  *ListEntry);

// https://msdn.microsoft.com/en-us/library/hh448545.aspx
// NOTE: forwarded from KERNEL32!InterlockedPushListSList and KERNELBASE!InterlockedPushListSList

__declspec(dllimport)
SLIST_ENTRY *	FASTCALL	RtlInterlockedPushListSList(SLIST_HEADER *ListHead,
				                            SLIST_ENTRY  *List,
				                            SLIST_ENTRY  *ListEnd,
				                            DWORD        Count);

#if _WIN32_WINNT >= 0x0602
__declspec(dllimport)
SLIST_ENTRY *	NTAPI	RtlInterlockedPushListSListEx(SLIST_HEADER *ListHead,
			                              SLIST_ENTRY  *List,
			                              SLIST_ENTRY  *ListEnd,
			                              DWORD        Count);
#endif

// https://msdn.microsoft.com/en-us/library/ff556972.aspx
// https://msdn.microsoft.com/en-us/library/ms738571.aspx

typedef	struct	in_addr
{
	union
	{
		struct
		{
			BYTE	s_b1;
			BYTE	s_b2;
			BYTE	s_b3;
			BYTE	s_b4;
		} S_un_b;
		struct
		{
			WORD	s_w1;
			WORD	s_w2;
		} S_un_w;
		DWORD	S_addr;
	} S_un;
} IN_ADDR, *LPIN_ADDR, *PIN_ADDR;

// https://msdn.microsoft.com/en-us/library/aa814456.aspx

__declspec(dllimport)
LPSTR	NTAPI	RtlIpv4AddressToStringA(IN_ADDR *Address,
		                        LPSTR   AddressString);

__declspec(dllimport)
LPWSTR	NTAPI	RtlIpv4AddressToStringW(IN_ADDR *Address,
		                        LPWSTR  AddressString);

// https://msdn.microsoft.com/en-us/library/aa814457.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv4AddressToStringExA(IN_ADDR *Address,
		                          WORD    Port,
		                          LPSTR   AddressString,
		                          LPDWORD AddressStringLength);

__declspec(dllimport)
LONG	NTAPI	RtlIpv4AddressToStringExW(IN_ADDR *Address,
		                          WORD    Port,
		                          LPWSTR  AddressString,
		                          LPDWORD AddressStringLength);

// https://msdn.microsoft.com/en-us/library/aa814458.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv4StringToAddressA(LPCSTR  AddressString,
		                        BOOLEAN Strict,
		                        LPCSTR  *Terminator,
		                        IN_ADDR *Address);

__declspec(dllimport)
LONG	NTAPI	RtlIpv4StringToAddressW(LPCWSTR AddressString,
		                        BOOLEAN Strict,
		                        LPCWSTR *Terminator,
		                        IN_ADDR *Address);

// https://msdn.microsoft.com/en-us/library/aa814459.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv4StringToAddressExA(LPCSTR  AddressString,
		                          BOOLEAN Strict,
		                          LPCSTR  *Terminator,
		                          IN_ADDR *Address,
		                          LPWORD  Port);

__declspec(dllimport)
LONG	NTAPI	RtlIpv4StringToAddressExW(LPCWSTR AddressString,
		                          BOOLEAN Strict,
		                          LPCWSTR *Terminator,
		                          IN_ADDR *Address,
		                          LPWORD  Port);

// https://msdn.microsoft.com/en-us/library/ff554787.aspx
// https://msdn.microsoft.com/en-us/library/ms738560.aspx

typedef	struct	in6_addr
{
	union
	{
		BYTE	Byte[16];
		WORD	Word[8];
	} u;
} IN6_ADDR, *LPIN6_ADDR, *PIN6_ADDR;

// https://msdn.microsoft.com/en-us/library/aa814460.aspx

__declspec(dllimport)
LPSTR	NTAPI	RtlIpv6AddressToStringA(IN6_ADDR *Address,
		                        LPSTR    AddressString);

__declspec(dllimport)
LPWSTR	NTAPI	RtlIpv6AddressToStringW(IN6_ADDR *Address,
		                        LPWSTR   AddressString);

// https://msdn.microsoft.com/en-us/library/aa814461.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv6AddressToStringExA(IN6_ADDR *Address,
		                          DWORD    ScopeId,
		                          WORD     Port,
		                          LPSTR    AddressString,
		                          LPDWORD  AddressStringLength);

__declspec(dllimport)
LONG	NTAPI	RtlIpv6AddressToStringExW(IN6_ADDR *Address,
		                          DWORD    ScopeId,
		                          WORD     Port,
		                          LPWSTR   AddressString,
		                          LPDWORD  AddressStringLength);

// https://msdn.microsoft.com/en-us/library/aa814462.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv6StringToAddressA(LPCSTR   AddressString,
		                        LPCSTR   *Terminator,
		                        IN6_ADDR *Address);

__declspec(dllimport)
LONG	NTAPI	RtlIpv6StringToAddressW(LPCWSTR  AddressString,
		                        LPCWSTR  *Terminator,
		                        IN6_ADDR *Address);

// https://msdn.microsoft.com/en-us/library/aa814463.aspx

__declspec(dllimport)
LONG	NTAPI	RtlIpv6StringToAddressExA(LPCSTR   AddressString,
		                          IN6_ADDR *Address,
		                          LPDWORD  ScopeId,
		                          LPWORD   Port);

__declspec(dllimport)
LONG	NTAPI	RtlIpv6StringToAddressExW(LPCWSTR  AddressString,
		                          IN6_ADDR *Address,
		                          LPDWORD  ScopeId,
		                          LPWORD   Port);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsActivationContextActive(ACTIVATION_CONTEXT *ActivationContext);

#if _WIN32_WINNT >= 0x0502
__declspec(dllimport)
BOOL	NTAPI	RtlIsCriticalSectionLocked(RTL_CRITICAL_SECTION *CriticalSection);

__declspec(dllimport)
BOOL	NTAPI	RtlIsCriticalSectionLockedByThread(RTL_CRITICAL_SECTION *CriticalSection);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlIsDosDeviceName_U(LPCWSTR PathName);

// https://msdn.microsoft.com/en-us/library/ff553046.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsGenericTableEmpty(RTL_GENERIC_TABLE *Table);

// https://msdn.microsoft.com/en-us/library/hh406472.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsGenericTableEmptyAvl(RTL_AVL_TABLE *Table);

// https://msdn.microsoft.com/en-us/library/ff553054.aspx

__forceinline
BOOL	NTAPI	RtlIsLeftChild(RTL_SPLAY_LINKS *Links)
{
#if 0
//	if (Links->Parent == Links)
//		return FALSE;

	if (Links->Parent->LeftChild == Links)
		return TRUE;

	return FALSE;
#else
	return (Links->Parent->LeftChild == Links);
#endif
}

// https://msdn.microsoft.com/en-us/library/hh551132.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsNameInExpression(UNICODE_STRING *Expression,
		                      UNICODE_STRING *Name,
		                      BOOLEAN        IgnoreCase,
		                      LPWSTR         UpcaseTable);

// https://msdn.microsoft.com/en-us/library/aa365524.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsNameLegalDOS8Dot3(UNICODE_STRING *Name,
		                       OEM_STRING     *OemName,
		                       BOOLEAN        *NameContainsSpaces);

// https://msdn.microsoft.com/en-us/library/ff553066.aspx

__forceinline
BOOL	NTAPI	RtlIsRightChild(RTL_SPLAY_LINKS *Links)
{
#if 0
//	if (Links->Parent == Links)
//		return FALSE;

	if (Links->Parent->RightChild == Links)
		return TRUE;

	return FALSE;
#else
	return (Links->Parent->RightChild == Links);
#endif
}

// https://msdn.microsoft.com/en-us/library/ff553072.aspx

__forceinline
BOOL	NTAPI	RtlIsRoot(RTL_SPLAY_LINKS *Links)
{
#if 0
	if (Links->Parent == Links)
		return TRUE;

	return FALSE;
#else
	return (Links->Parent == Links);
#endif
}

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsTextUnicode(LPVOID  *Buffer,
		                 DWORD   BufferSize,
		                 LPDWORD Result);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsThreadWithinLoaderCallout(VOID);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsValidHandle(RTL_HANDLE_TABLE       *HandleTable,
		                 RTL_HANDLE_TABLE_ENTRY *Handle);

#ifdef _WIN64
__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsValidHandler(EXCEPTION_ROUTINE *Handler);
#endif

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsValidIndexHandle(RTL_HANDLE_TABLE       *HandleTable,
		                      DWORD                  HandleIndex,
		                      RTL_HANDLE_TABLE_ENTRY **Handle);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsValidLocaleName(LPCWSTR LocaleName,
		                     DWORD   Flags);
#endif

// https://msdn.microsoft.com/en-us/library/ff553076.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlIsValidOemCharacter(WCHAR *Char);

// https://msdn.microsoft.com/en-us/library/ff561958.aspx

__forceinline
BOOL	NTAPI	RtlIsZeroLuid(LUID *Luid)
{
#if 0
	// https://msdn.microsoft.com/en-us/library/aa379261.aspx

	static	const	LUID	ZeroLuid = {0L, 0UL};

	return RtlEqualLuid(Luid, &ZeroLuid);
#else
	return (Luid->HighPart == 0L)
	    && (Luid->LowPart == 0UL);
#endif
}

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlKnownExceptionFilter(EXCEPTION_POINTERS *ExceptionInfo);
#endif

// https://msdn.microsoft.com/en-us/library/ff561961.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerAdd(LARGE_INTEGER Summand1,
			                   LARGE_INTEGER Summand2);

// https://msdn.microsoft.com/en-us/library/ff561966.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerArithmeticShift(LARGE_INTEGER LargeInteger,
			                               BYTE          ShiftCount);

// https://msdn.microsoft.com/en-us/library/ff561969.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerDivide(LARGE_INTEGER Dividend,
			                      LARGE_INTEGER Divisor,
			                      LARGE_INTEGER *Remainder);	// optional

// https://msdn.microsoft.com/en-us/library/ff561971.aspx

#if 0	// NOTE: not exported from NTDLL.DLL!
__declspec(dllimport)
BOOLEAN	NTAPI	RtlLargeIntegerEqualTo(LARGE_INTEGER Operand1,
		                       LARGE_INTEGER Operand2);
#else
__forceinline
BOOL	NTAPI	RtlLargeIntegerEqualTo(LARGE_INTEGER Operand1,
		                       LARGE_INTEGER Operand2)
{
#if 0
	return (Operand1.HighPart == Operand2.HighPart)
	    && (Operand1.LowPart == Operand2.LowPart);
#else
	return (Operand1.QuadPart == Operand2.QuadPart);
#endif
}
#endif

// https://msdn.microsoft.com/en-us/library/ff562002.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerNegate(LARGE_INTEGER Subtrahend);

// https://msdn.microsoft.com/en-us/library/ff562014.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerShiftLeft(LARGE_INTEGER LargeInteger,
			                         BYTE          ShiftCount);

// https://msdn.microsoft.com/en-us/library/ff562017.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerShiftRight(LARGE_INTEGER LargeInteger,
			                          BYTE          ShiftCount);

// https://msdn.microsoft.com/en-us/library/ff562020.aspx

__declspec(dllimport)
LARGE_INTEGER	NTAPI	RtlLargeIntegerSubtract(LARGE_INTEGER Minuend,
			                        LARGE_INTEGER Subtrahend);

__declspec(dllimport)
LONG	NTAPI	RtlLargeIntegerToChar(LARGE_INTEGER *Value,
		                      DWORD         Base,	// 0, 2, 8, 10, 16
		                      DWORD         OutputLength,
		                      LPSTR         String);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
BOOLEAN	NTAPI	RtlLCIDToCultureName(LCID           LocaleId,
		                     UNICODE_STRING *CultureName);

__declspec(dllimport)
LONG	NTAPI	RtlLcidToLocaleName(LCID           LocaleId,
		                    UNICODE_STRING *LocaleName,
		                    DWORD          Flags,
		                    BOOLEAN        AllocateDestinationString);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms684169.aspx
// NOTE: forwarded from KERNEL32!LeaveCriticalSection and KERNELBASE!LeaveCriticalSection

__declspec(dllimport)
LONG	NTAPI	RtlLeaveCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

// https://msdn.microsoft.com/en-us/library/ff553080.aspx

__forceinline
RTL_SPLAY_LINKS *	NTAPI	RtlLeftChild(RTL_SPLAY_LINKS *Links)
{
	return Links->LeftChild;
}

// https://msdn.microsoft.com/en-us/library/ff553083.aspx
// NOTE: like ADVAPI32!GetSidLengthRequired

__declspec(dllimport)
DWORD	NTAPI	RtlLengthRequiredSid(DWORD SubAuthorityCount);

// https://msdn.microsoft.com/en-us/library/ff562025.aspx
// NOTE: like ADVAPI32!GetSecurityDescriptorLength

__declspec(dllimport)
DWORD	NTAPI	RtlLengthSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor);

// https://msdn.microsoft.com/en-us/library/ff553085.aspx
// NOTE: like ADVAPI32!GetLengthSid

__declspec(dllimport)
DWORD	NTAPI	RtlLengthSid(SID *Sid);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlLocaleNameToLcid(LPCWSTR LocaleName,
		                    LCID    *LocaleId,
		                    DWORD   Flags);
#endif

// https://msdn.microsoft.com/en-us/library/ms724927.aspx

__declspec(dllimport)
LONG	NTAPI	RtlLocalTimeToSystemTime(LARGE_INTEGER *LocalTime,
		                         LARGE_INTEGER *SystemTime);

#if _WIN32_WINNT >= 0x0601
#ifndef _WINNT_
__declspec(dllimport)
LPVOID	NTAPI	RtlLocateExtendedFeature(CONTEXT_EX *ContextEx,
		                         DWORD      FeatureId,
		                         LPDWORD    Length);

__declspec(dllimport)
CONTEXT *	NTAPI	RtlLocateLegacyContext(CONTEXT_EX *ContextEx,
			                       LPDWORD    Length);
#endif // _WINNT_
#endif // _WIN32_WINNT

#if _WIN32_WINNT >= 0x0603
__declspec(dllimport)
LPVOID	NTAPI	RtlLocateExtendedFeature2(CONTEXT_EX           *ContextEx,
		                          DWORD                FeatureId,
		                          XSTATE_CONFIGURATION *Configuration,
		                          LPDWORD              Length);
#endif

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	NTAPI	RtlLockBootStatusData(HANDLE *BootStatusDataHandle);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlLockCurrentThread(VOID);
#endif

__declspec(dllimport)
BOOLEAN	NTAPI	RtlLockHeap(LPCVOID HeapHandle);

__declspec(dllimport)
HRESULT	NTAPI	RtlLockMemoryStreamRegion(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                          ULARGE_INTEGER                Offset,
		                          ULARGE_INTEGER                Bytes,
		                          DWORD                         LockType);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlLockModuleSection(LPVOID Address);
#endif

__declspec(dllimport)
WORD	NTAPI	RtlLogStackBackTrace(VOID);

__declspec(dllimport)
LONG	NTAPI	RtlLookupAtomInAtomTable(RTL_ATOM_TABLE *AtomTable,
		                         LPCWSTR        AtomName,
		                         RTL_ATOM       *Atom);

// https://msdn.microsoft.com/en-us/library/ff553091.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlLookupElementGenericTable(RTL_GENERIC_TABLE *Table,
		                             LPVOID            Buffer);

// https://msdn.microsoft.com/en-us/library/hh406476.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlLookupElementGenericTableAvl(RTL_AVL_TABLE *Table,
		                                LPVOID        Buffer);

typedef	enum	_TABLE_SEARCH_RESULT
{
	TableEmptyTree     = 0,
	TableFoundNode,
	TableInsertAsLeft,
	TableInsertAsRight
} TABLE_SEARCH_RESULT;

// https://msdn.microsoft.com/en-us/library/ff553092.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlLookupElementGenericTableFullAvl(RTL_AVL_TABLE       *Table,
		                                    LPVOID              Buffer,
		                                    LPVOID              *NodeOrParent,
		                                    TABLE_SEARCH_RESULT *SearchResult);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
RTL_DYNAMIC_HASH_TABLE_ENTRY *	NTAPI	RtlLookupEntryHashTable(RTL_DYNAMIC_HASH_TABLE         *HashTable,
					                        DWORD_PTR                      Signature,
					                        RTL_DYNAMIC_HASH_TABLE_CONTEXT *Context);
#endif

// https://msdn.microsoft.com/en-us/library/ff553106.aspx

__declspec(dllimport)
LPVOID	NTAPI	RtlLookupFirstMatchingElementGenericTableAvl(RTL_AVL_TABLE *Table,
		                                             LPVOID        Buffer,
		                                             LPVOID        *RestartKey);

#ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/ms680597.aspx

#ifndef UNWIND_HISTORY_TABLE_SIZE
#define UNWIND_HISTORY_TABLE_SIZE	12

typedef	struct	_UNWIND_HISTORY_TABLE_ENTRY
{
	LPVOID			ImageBase;
	RUNTIME_FUNCTION	*FunctionEntry;
} UNWIND_HISTORY_TABLE_ENTRY, *LPUNWIND_HISTORY_TABLE_ENTRY, *PUNWIND_HISTORY_TABLE_ENTRY;

typedef	struct	_UNWIND_HISTORY_TABLE
{
	DWORD	Count;
	UCHAR	Search;
	QWORD	LowAddress;
	QWORD	HighAddress;
	UNWIND_HISTORY_TABLE_ENTRY	Entry[UNWIND_HISTORY_TABLE_SIZE];
} UNWIND_HISTORY_TABLE, *LPUNWIND_HISTORY_TABLE, *PUNWIND_HISTORY_TABLE;

// NOTE: forwarded from KERNEL32!RtlLookupFunctionEntry

__declspec(dllimport)
RUNTIME_FUNCTION *	NTAPI	RtlLookupFunctionEntry(QWORD                ControlPc,
				                       LPVOID               *ImageBase,
				                       UNWIND_HISTORY_TABLE *HistoryTable);

// NOTE: forwarded from KERNEL32!RtlLookupFunctionTable

__declspec(dllimport)
RUNTIME_FUNCTION *	NTAPI	RtlLookupFunctionTable(QWORD   ControlPc,
				                       LPVOID  *ImageBase,
				                       LPDWORD Length);
#endif // UNWIND_HISTORY_TABLE_SIZE
#endif // _WIN64

// NOTE: like ADVAPI32!MakeSelfRelativeSD

__declspec(dllimport)
LONG	NTAPI	RtlMakeSelfRelativeSD(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                      SECURITY_DESCRIPTOR *SelfRelativeSecurityDescriptor,
		                      LPDWORD             BufferLength);

// https://msdn.microsoft.com/en-us/library/ff562027.aspx
// NOTE: like ADVAPI32!MapGenericMask

__declspec(dllimport)
VOID	NTAPI	RtlMapGenericMask(ACCESS_MASK     *AccessMask,
		                  GENERIC_MAPPING *GenericMapping);

__declspec(dllimport)
LONG	NTAPI	RtlMapSecurityErrorToNtStatus(SECURITY_STATUS Error);

#if 1
#pragma intrinsic(memmove)
#else
#undef RtlMoveMemory

// https://msdn.microsoft.com/en-us/library/ff562030.aspx
// https://msdn.microsoft.com/en-us/library/mt723419.aspx
// NOTE: forwarded from KERNEL32!RtlMoveMemory

__declspec(dllimport)
VOID	NTAPI	RtlMoveMemory(LPVOID Destination,
		              LPVOID Source,
		              SIZE_T Length);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlMultiAppendUnicodeStringBuffer(RTL_UNICODE_STRING_BUFFER *UnicodeStringBuffer,
		                                  DWORD                     StringCount,
		                                  UNICODE_STRING            *Strings);

// https://msdn.microsoft.com/en-us/library/ff553113.aspx

__declspec(dllimport)
LONG	NTAPI	RtlMultiByteToUnicodeN(LPWSTR  UnicodeString,
		                       DWORD   MaxBytesInUnicodeString,
		                       LPDWORD BytesInUnicodeString,
		                       LPCSTR  MultiByteString,
		                       DWORD   BytesInMultiByteString);

// https://msdn.microsoft.com/en-us/library/ff553121.aspx

__declspec(dllimport)
LONG	NTAPI	RtlMultiByteToUnicodeSize(LPDWORD BytesInUnicodeString,
		                          LPCSTR  MultiByteString,
		                          DWORD   BytesInMultiByteString);

__declspec(dllimport)
LONG	NTAPI	RtlNewInstanceSecurityObject(BOOLEAN             ParentDescriptorChanged,
		                             BOOLEAN             CreatorDescriptorChanged,
		                             LUID                *OldClientTokenModifiedId,
		                             LUID                *NewClientTokenModifiedId,
		                             SECURITY_DESCRIPTOR *ParentDescriptor,
		                             SECURITY_DESCRIPTOR *CreatorDescriptor,
		                             SECURITY_DESCRIPTOR **NewDescriptor,
		                             BOOLEAN             IsDirectoryObject,
		                             HANDLE              TokenHandle,
		                             GENERIC_MAPPING     *GenericMapping);

__declspec(dllimport)
LONG	NTAPI	RtlNewSecurityObject(SECURITY_DESCRIPTOR *ParentDescriptor,
		                     SECURITY_DESCRIPTOR *CreatorDescriptor,
		                     SECURITY_DESCRIPTOR **NewDescriptor,
		                     BOOLEAN             IsDirectoryObject,
		                     HANDLE              TokenHandle,
		                     GENERIC_MAPPING     GenericMapping);

__declspec(dllimport)
LONG	NTAPI	RtlNewSecurityObjectEx(SECURITY_DESCRIPTOR *ParentDescriptor,
		                       SECURITY_DESCRIPTOR *CreatorDescriptor,
		                       SECURITY_DESCRIPTOR **NewDescriptor,
		                       GUID                *ObjectType,
		                       BOOLEAN             IsDirectoryObject,
		                       DWORD               AutoInheritFlags,
		                       HANDLE              TokenHandle,
		                       GENERIC_MAPPING     *GenericMapping);

__declspec(dllimport)
LONG	NTAPI	RtlNewSecurityObjectWithMultipleInheritance(SECURITY_DESCRIPTOR *ParentDescriptor,
		                                            SECURITY_DESCRIPTOR *CreatorDescriptor,
		                                            SECURITY_DESCRIPTOR **NewDescriptor,
		                                            GUID                **ObjectType,
		                                            DWORD               GuidCount,
		                                            BOOLEAN             IsDirectoryObject,
		                                            DWORD               AutoInheritFlags,
		                                            HANDLE              TokenHandle,
		                                            GENERIC_MAPPING     *GenericMapping);

// https://msdn.microsoft.com/en-us/library/ff553123.aspx

__declspec(dllimport)
UNICODE_PREFIX_TABLE_ENTRY *	NTAPI	RtlNextUnicodePrefix(UNICODE_PREFIX_TABLE *PrefixTable,
					                     BOOLEAN              Restart);

__declspec(dllimport)
RTL_USER_PROCESS_PARAMETERS *	NTAPI	RtlNormalizeProcessParams(RTL_USER_PROCESS_PARAMETERS *ProcessParameters);

#define RTL_NT_PATH_NAME_TO_DOS_PATH_NAME_AMBIGUOUS	1UL
#define RTL_NT_PATH_NAME_TO_DOS_PATH_NAME_UNC		2UL
#define RTL_NT_PATH_NAME_TO_DOS_PATH_NAME_DRIVE		3UL
#define RTL_NT_PATH_NAME_TO_DOS_PATH_NAME_ALREADY_DOS	4UL

__declspec(dllimport)
LONG	NTAPI	RtlNtPathNameToDosPathName(DWORD                     Flags,
		                           RTL_UNICODE_STRING_BUFFER *Path,
		                           LPDWORD                   Disposition,
		                           LPWSTR                    *FilePart);

// https://msdn.microsoft.com/en-us/library/ms680600.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNtStatusToDosError(NTSTATUS Status);

// https://msdn.microsoft.com/en-us/library/dn934204.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNtStatusToDosErrorNoTeb(NTSTATUS Status);

// https://msdn.microsoft.com/en-us/library/ff553134.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNumberGenericTableElements(RTL_GENERIC_TABLE *Table);

// https://msdn.microsoft.com/en-us/library/hh406522.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNumberGenericTableElementsAvl(RTL_AVL_TABLE *Table);

// https://msdn.microsoft.com/en-us/library/ff562034.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNumberOfClearBits(RTL_BITMAP *BitMapHeader);

// https://msdn.microsoft.com/en-us/library/ff562037.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlNumberOfSetBits(RTL_BITMAP *BitMapHeader);

// https://msdn.microsoft.com/en-us/library/ff553136.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlOemStringToCountedUnicodeSize(OEM_STRING *OemString);

// https://msdn.microsoft.com/en-us/library/ff553142.aspx

__declspec(dllimport)
LONG	NTAPI	RtlOemStringToCountedUnicodeString(UNICODE_STRING *DestinationString,
		                                   OEM_STRING     *SourceString,
		                                   BOOLEAN        AllocateDestinationString);

__declspec(dllimport)
LONG	NTAPI	RtlxOemStringToUnicodeSize(OEM_STRING *OemString);

// https://msdn.microsoft.com/en-us/library/ff553147.aspx

#if 1
__declspec(dllimport)
DWORD	NTAPI	RtlOemStringToUnicodeSize(OEM_STRING *OemString);
#else
__inline
DWORD	NTAPI	RtlOemStringToUnicodeSize(OEM_STRING *OemString)
{
	if (*NlsMbOemCodePageTag)
		return RtlxOemStringToUnicodeSize(OemString);

	return (OemString->Length + 1) * sizeof(WCHAR);
}
#endif

// https://msdn.microsoft.com/en-us/library/ff553153.aspx

__declspec(dllimport)
LONG	NTAPI	RtlOemStringToUnicodeString(UNICODE_STRING *DestinationString,
		                            OEM_STRING     *SourceString,
		                            BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff553157.aspx

__declspec(dllimport)
LONG	NTAPI	RtlOemToUnicodeN(LPWSTR  UnicodeString,
		                 DWORD   MaxBytesInUnicodeString,
		                 LPDWORD BytesInUnicodeString,
		                 LPCSTR  OemString,
		                 DWORD   BytesInOemString);

// https://msdn.microsoft.com/en-us/library/ff553160.aspx

#if 0
#define RtlOffsetToPointer(Base, Offset)	((LPBYTE) ((LPBYTE) (Base) + (DWORD_PTR) (Offset)))
#else
__forceinline
LPBYTE	NTAPI	RtlOffsetToPointer(LPVOID Base,
		                   DWORD  Offset)
{
	return (LPBYTE) Base + (DWORD_PTR) Offset;
}
#endif

__declspec(dllimport)
LONG	NTAPI	RtlOpenCurrentUser(ACCESS_MASK DesiredAccess,
		                   HANDLE      *CurrentUserKey);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlOwnerAcesPresent(ACL *Acl);
#endif

// https://msdn.microsoft.com/en-us/library/ff553165.aspx

__forceinline
RTL_SPLAY_LINKS *	NTAPI	RtlParent(RTL_SPLAY_LINKS *Links)
{
	return Links->Parent;
}

#if _WIN32_WINNT < 0x0600
// https://msdn.microsoft.com/en-us/library/ms680603.aspx
// NOTE: forwarded from KERNEL32!RtlPcToFileHeader

__declspec(dllimport)
LPVOID	NTAPI	RtlPcToFileHeader(LPCVOID PcValue,
		                  LPVOID  *BaseOfImage);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlPinAtomInAtomTable(RTL_ATOM_TABLE *AtomTable,
		                      RTL_ATOM       Atom);

// https://msdn.microsoft.com/en-us/library/ff553171.aspx

#if 0
#define RtlPointerToOffset(Base, Pointer)	((DWORD) ((LPBYTE) (Pointer) - (LPBYTE) (Base)))
#else
__forceinline
DWORD	NTAPI	RtlPointerToOffset(LPVOID Base,
		                   LPVOID Pointer)
{
	return (DWORD) ((LPBYTE) Pointer - (LPBYTE) Base);
}
#endif

__declspec(dllimport)
VOID	NTAPI	RtlPopFrame(TEB_ACTIVE_FRAME *Frame);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlPrefixString(STRING  *String1,
		                STRING  *String2,
		                BOOLEAN CaseInSensitive);

// https://msdn.microsoft.com/en-us/library/mt708868.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlPrefixUnicodeString(UNICODE_STRING *UnicodeString1,
		                       UNICODE_STRING *UnicodeString2,
		                       BOOLEAN        CaseInSensitive);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlProcessFlsData(LPVOID Unknown);
#endif

#if 1
#define RtlProcessHeap()	(NtCurrentPeb()->ProcessHeap)
#else
__forceinline
HANDLE	NTAPI	RtlProcessHeap(VOID)
{
	return NtCurrentPeb()->ProcessHeap;
}
#endif

__declspec(dllimport)
LONG	NTAPI	RtlProtectHeap(LPCVOID HeapHandle,
		               BOOLEAN Protect);

__declspec(dllimport)
VOID	NTAPI	RtlPushFrame(TEB_ACTIVE_FRAME *Frame);

#ifdef _WIN64
__declspec(dllimport)
LONG	NTAPI	RtlQueueApcWow64Thread(HANDLE          ThreadHandle,
		                       KNORMAL_ROUTINE *ApcRoutine,
		                       LPVOID          NormalContext,
		                       LPVOID          SystemArgument1,
		                       LPVOID          SystemArgument2);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlQueryActivationContextApplicationSettings(DWORD              Flags,
		                                             ACTIVATION_CONTEXT *ActivationContext,
		                                             LPCWSTR            NameSpace,
		                                             LPCWSTR            Name,
		                                             LPVOID             Buffer,
		                                             SIZE_T             BufferLength,
		                                             SIZE_T             *ReturnedLength);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlQueryAtomInAtomTable(RTL_ATOM_TABLE *AtomTable,
		                        RTL_ATOM       Atom,
		                        LPDWORD        RefCount,
		                        LPDWORD        PinCount,
		                        LPWSTR         AtomName,
		                        LPDWORD        NameLength);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
HANDLE	NTAPI	RtlQueryCriticalSectionOwner(HANDLE EventHandle);
#endif

// https://msdn.microsoft.com/en-us/library/aa904936.aspx
// NOTE: forwarded from KERNEL32!QueryDepthSList and KERNELBASE!QueryDepthSList

__declspec(dllimport)
WORD	NTAPI	RtlQueryDepthSList(SLIST_HEADER *ListHead);

#if _WIN32_WINNT >= 0x0600
#define ELEVATION_UAC_ENABLED			1UL
#define ELEVATION_VIRTUALIZATION_ENABLED	2UL
#define ELEVATION_INSTALLER_DETECTION_ENABLED	4UL

__declspec(dllimport)
LONG	NTAPI	RtlQueryElevationFlags(LPDWORD ElevationFlags);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlQueryEnvironmentVariable_U(LPVOID         Environment,
		                              UNICODE_STRING *Name,
		                              UNICODE_STRING *Value);

// NOTE: like ADVAPI32!GetAclInformation

__declspec(dllimport)
LONG	NTAPI	RtlQueryInformationAcl(ACL                   *Acl,
		                       LPVOID                AclInformation,
		                       DWORD                 AclInformationLength,
		                       ACL_INFORMATION_CLASS AclInformationClass);

#define RTL_QUERY_INFORMATION_ACTIVATION_CONTEXT_FLAG_USE_ACTIVE_ACTIVATION_CONTEXT	0x00000001UL
#define RTL_QUERY_INFORMATION_ACTIVATION_CONTEXT_FLAG_ACTIVATION_CONTEXT_IS_MODULE	0x00000002UL
#define RTL_QUERY_INFORMATION_ACTIVATION_CONTEXT_FLAG_ACTIVATION_CONTEXT_IS_ADDRESS	0x00000004UL
#define RTL_QUERY_INFORMATION_ACTIVATION_CONTEXT_FLAG_NO_ADDREF				0x80000000UL

__declspec(dllimport)
LONG	NTAPI	RtlQueryInformationActivationContext(DWORD                         Flags,
		                                     ACTIVATION_CONTEXT            *ActivationContext,
		                                     LPVOID                        SubInstanceIndex,
		                                     ACTIVATION_CONTEXT_INFO_CLASS ActivationContextInformationClass,
		                                     LPVOID                        ActivationContextInformation,
		                                     SIZE_T                        ActivationContextInformationLength,
		                                     SIZE_T                        *ReturnedLength);

__declspec(dllimport)
LONG	NTAPI	RtlQueryInformationActiveActivationContext(ACTIVATION_CONTEXT_INFO_CLASS ActivationContextInformationClass,
		                                           LPVOID                        ActivationContextInformation,
		                                           SIZE_T                        ActivationContextInformationLength,
		                                           SIZE_T                        *ReturnLength);

__declspec(dllimport)
HRESULT	NTAPI	RtlQueryInterfaceMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                              IID                           *Interface,
		                              LPVOID                        *Object);

#if _WIN32_WINNT >= 0x0600
typedef	struct	_RTL_MODULE_BASIC_INFO
{
	LPVOID	ImageBase;
} RTL_MODULE_BASIC_INFO, *LPRTL_MODULE_BASIC_INFO, *PRTL_MODULE_BASIC_INFO;

typedef	struct	_RTL_MODULE_EXTENDED_INFO
{
	RTL_MODULE_BASIC_INFO	BasicInfo;
	DWORD			ImageSize;
	WORD			FileNameOffset;
	CHAR			FullPathName[256];
} RTL_MODULE_EXTENDED_INFO, *LPRTL_MODULE_EXTENDED_INFO, *PRTL_MODULE_EXTENDED_INFO;

__declspec(dllimport)
LONG	NTAPI	RtlQueryModuleInformation(LPDWORD InformationLength,
		                          DWORD   SizePerModule,
		                          LPVOID  InformationBuffer);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms644904.aspx
// NOTE: forwarded from KERNEL32!QueryPerformanceCounter and KERNELBASE!QueryPerformanceCounter

__declspec(dllimport)
BOOL	NTAPI	RtlQueryPerformanceCounter(LARGE_INTEGER *PerformanceCount);

// https://msdn.microsoft.com/en-us/library/ms644905.aspx
// NOTE: forwarded from KERNEL32!QueryPerformanceFrequency and KERNELBASE!QueryPerformanceFrequency

__declspec(dllimport)
BOOL	NTAPI	RtlQueryPerformanceFrequency(LARGE_INTEGER *PerformanceFrequency);

#define RTL_DEBUG_QUERY_MODULES		0x00000001UL
#define RTL_DEBUG_QUERY_BACKTRACES	0x00000002UL
#define RTL_DEBUG_QUERY_HEAPS		0x00000004UL
#define RTL_DEBUG_QUERY_HEAP_TAGS	0x00000008UL
#define RTL_DEBUG_QUERY_HEAP_BLOCKS	0x00000010UL
#define RTL_DEBUG_QUERY_LOCKS		0x00000020UL

__declspec(dllimport)
LONG	NTAPI	RtlQueryProcessDebugInformation(DWORD                 ProcessId,
		                                DWORD                 Flags,
		                                RTL_DEBUG_INFORMATION *Buffer);

__declspec(dllimport)
LONG	NTAPI	RtlQueryProcessHeapInformation(RTL_DEBUG_INFORMATION *Buffer);

// https://msdn.microsoft.com/en-us/library/ff562046.aspx

#define RTL_QUERY_REGISTRY_SUBKEY	0x00000001UL
#define RTL_QUERY_REGISTRY_TOPKEY	0x00000002UL
#define RTL_QUERY_REGISTRY_REQUIRED	0x00000004UL
#define RTL_QUERY_REGISTRY_NOVALUE	0x00000008UL
#define RTL_QUERY_REGISTRY_NOEXPAND	0x00000010UL
#define RTL_QUERY_REGISTRY_DIRECT	0x00000020UL
#define RTL_QUERY_REGISTRY_DELETE	0x00000040UL
#define RTL_QUERY_REGISTRY_TYPECHECK	0x00000100UL

typedef	LONG	(NTAPI	RTL_QUERY_REGISTRY_ROUTINE) (LPWSTR ValueName,
			                             DWORD  ValueType,
			                             LPVOID ValueData,
			                             DWORD  ValueLength,
			                             LPVOID Context,
			                             LPVOID EntryContext);

typedef	struct	_RTL_QUERY_REGISTRY_TABLE
{
	RTL_QUERY_REGISTRY_ROUTINE	*QueryRoutine;
	DWORD				Flags;
	LPWSTR				Name;
	LPVOID				EntryContext;
	DWORD				DefaultType;
	LPVOID				DefaultData;
	DWORD				DefaultLength;
} RTL_QUERY_REGISTRY_TABLE, *LPRTL_QUERY_REGISTRY_TABLE, *PRTL_QUERY_REGISTRY_TABLE;

__declspec(dllimport)
LONG	NTAPI	RtlQueryRegistryValues(DWORD                    RelativeTo,
		                       LPCWSTR                  Path,
		                       RTL_QUERY_REGISTRY_TABLE *QueryTable,
		                       LPVOID                   Context,
		                       LPVOID                   Environment);

__declspec(dllimport)
LONG	NTAPI	RtlQuerySecurityObject(SECURITY_DESCRIPTOR  *ObjectDescriptor,
		                       SECURITY_INFORMATION SecurityInformation,
		                       SECURITY_DESCRIPTOR  *ResultantDescriptor,
		                       DWORD                DescriptorLength,
		                       LPDWORD              ReturnLength);

typedef	struct	_RTL_HEAP_TAG_INFO
{
	DWORD	NumberOfAllocations;
	DWORD	NumberOfFrees;
	SIZE_T	BytesAllocated;
} RTL_HEAP_TAG_INFO, *LPRTL_HEAP_TAG_INFO, *PRTL_HEAP_TAG_INFO;

__declspec(dllimport)
LPWSTR	NTAPI	RtlQueryTagHeap(LPCVOID           HeapHandle,
		                DWORD             Flags,
		                WORD              TagIndex,
		                BOOLEAN           ResetCounters,
		                RTL_HEAP_TAG_INFO *TagInfo);

// https://msdn.microsoft.com/en-us/library/hh829883.aspx

__declspec(dllimport)
__declspec(noreturn)
VOID	NTAPI	RtlRaiseException(EXCEPTION_RECORD *ExceptionRecord);

__declspec(dllimport)
VOID	NTAPI	RtlRaiseStatus(NTSTATUS Status);

// https://msdn.microsoft.com/en-us/library/ff553177.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlRandom(LPDWORD Seed);

// https://msdn.microsoft.com/en-us/library/ff553181.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlRandomEx(LPDWORD Seed);

#if _WIN32_WINNT >= 0x0602
typedef	struct	_RTL_BALANCED_NODE
{
	union
	{
		struct	_RTL_BALANCED_NODE	*Children[2];
		struct
		{
			struct	_RTL_BALANCED_NODE	*Left;
			struct	_RTL_BALANCED_NODE	*Right;
		};
	};
	union
	{
		BYTE		Red     : 1;
		BYTE		Balance : 2;
		DWORD_PTR	ParentValue;
	};
} RTL_BALANCED_NODE, *LPRTL_BALANCED_NODE, *PRTL_BALANCED_NODE;

typedef	struct	_RTL_AVL_TREE
{
	RTL_BALANCED_NODE	*Root;
} RTL_AVL_TREE, *LPRTL_AVL_TREE, *PRTL_AVL_TREE;

typedef	struct	_RTL_RB_TREE
{
	RTL_BALANCED_NODE	*Root;
	RTL_BALANCED_NODE	*Min;
} RTL_RB_TREE, *LPRTL_RB_TREE, *PRTL_RB_TREE;

__declspec(dllimport)
VOID	NTAPI	RtlRbInsertNodeEx(RTL_RB_TREE       *Tree,
		                  RTL_BALANCED_NODE *Parent,
		                  BOOLEAN           Right,
		                  RTL_BALANCED_NODE *Node);
#endif // _WIN32_WINNT

__declspec(dllimport)
HRESULT	NTAPI	RtlReadMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                    LPVOID                        Buffer,
		                    DWORD                         BytesToRead,
		                    LPDWORD                       BytesRead);

__declspec(dllimport)
HRESULT	NTAPI	RtlReadOutOfProcessMemoryStream(RTL_OUT_OF_PROCESS_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                                LPVOID                                       Buffer,
		                                DWORD                                        BytesToRead,
		                                LPDWORD                                      BytesRead);

// https://msdn.microsoft.com/en-us/library/aa366704.aspx
// NOTE: forwarded from KERNEL32!HeapReAlloc and KERNELBASE!HeapReAlloc

__declspec(dllimport)
LPVOID	NTAPI	RtlReAllocateHeap(LPCVOID HeapHandle,
		                  DWORD   Flags,
		                  LPCVOID HeapBase,
		                  SIZE_T  Size);

// https://msdn.microsoft.com/en-us/library/ff553186.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlRealPredecessor(RTL_SPLAY_LINKS *Links);

// https://msdn.microsoft.com/en-us/library/ff553188.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlRealSuccessor(RTL_SPLAY_LINKS *Links);

__declspec(dllimport)
VOID	NTAPI	RtlReleaseActivationContext(ACTIVATION_CONTEXT *ActivationContext);

__declspec(dllimport)
DWORD	NTAPI	RtlReleaseMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

__declspec(dllimport)
VOID	NTAPI	RtlReleasePebLock(VOID);

__declspec(dllimport)
VOID	NTAPI	RtlReleaseRelativeName(RTL_RELATIVE_NAME_U *RelativeName);

__declspec(dllimport)
VOID	NTAPI	RtlReleaseResource(RTL_RESOURCE *Resource);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/ms681930.aspx
// NOTE: forwarded from KERNEL32!ReleaseSRWLockExclusive and KERNELBASE!ReleaseSRWLockExclusive

__declspec(dllimport)
VOID	NTAPI	RtlReleaseSRWLockExclusive(RTL_SRWLOCK *SRWLock);

// https://msdn.microsoft.com/en-us/library/ms685080.aspx
// NOTE: forwarded from KERNEL32!ReleaseSRWLockShared and KERNELBASE!ReleaseSRWLockShared

__declspec(dllimport)
VOID	NTAPI	RtlReleaseSRWLockShared(RTL_SRWLOCK *SRWLock);
#endif // _WIN32_WINNT

__declspec(dllimport)
LONG	NTAPI	RtlRemoteCall(HANDLE    Process,
		              HANDLE    Thread,
		              LPVOID    CallSite,
		              DWORD     ArgumentCount,
		              DWORD_PTR *Arguments,
		              BOOLEAN   PassContext,
		              BOOLEAN   AlreadySuspended);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
BOOLEAN	NTAPI	RtlRemoveEntryHashTable(RTL_DYNAMIC_HASH_TABLE         *HashTable,
		                        RTL_DYNAMIC_HASH_TABLE_ENTRY   *Entry,
		                        RTL_DYNAMIC_HASH_TABLE_CONTEXT *Context);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlRemovePrivileges(HANDLE TokenHandle,
		                    DWORD  PrivilegesToKeep[],
		                    DWORD  PrivilegeCount);
#endif

// https://msdn.microsoft.com/en-us/library/ff553193.aspx

__declspec(dllimport)
VOID	NTAPI	RtlRemoveUnicodePrefix(UNICODE_PREFIX_TABLE       *PrefixTable,
		                       UNICODE_PREFIX_TABLE_ENTRY *PrefixTableEntry);

// https://msdn.microsoft.com/en-us/library/ms680567.aspx
// NOTE: forwarded from KERNEL32!RemoveVectoredContinueHandler and KERNELBASE!RemoveVectoredContinueHandler

__declspec(dllimport)
LONG	NTAPI	RtlRemoveVectoredContinueHandler(LPVOID VectoredHandlerHandle);

// https://msdn.microsoft.com/en-us/library/ms680571.aspx
// NOTE: forwarded from KERNEL32!RemoveVectoredExceptionHandler and KERNELBASE!RemoveVectoredExceptionHandler

__declspec(dllimport)
LONG	NTAPI	RtlRemoveVectoredExceptionHandler(LPVOID VectoredHandlerHandle);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlReplaceSidInSd(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                  SID                 *OldSid,
		                  SID                 *NewSid,
		                  LPDWORD             NumChanges);
#endif

#if 0
?	NTAPI	RtlReportCriticalFailure(?,
		                         ?);
#endif

#if _WIN32_WINNT >= 0x0600
// NOTE: like KERNEL32!RaiseFailFastException and KERNEL32!RtlRaiseException

__declspec(dllimport)
LONG	NTAPI	RtlReportException(EXCEPTION_RECORD *ExceptionRecord,
		                   CONTEXT          *ContextRecord,
		                   DWORD            Flags);
#endif

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	RtlReportSilentProcessExit(HANDLE   ProcessHandle,
		                           NTSTATUS ExitStatus);
#endif

__declspec(dllimport)
VOID	NTAPI	RtlResetRtlTranslations(NLSTABLEINFO *TableInfo);

#ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/ms680605.aspx
// NOTE: like KERNEL32!RtlRestoreContext

__declspec(dllimport)
VOID	NTAPI	RtlRestoreContext(CONTEXT          *ContextRecord,
		                  EXCEPTION_RECORD *ExceptionRecord);
#endif

// NOTE: forwarded from KERNEL32!RestoreLastError

__declspec(dllimport)
VOID	NTAPI	RtlRestoreLastWin32Error(DWORD Error);

__declspec(dllimport)
HRESULT	NTAPI	RtlRevertMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream);

// https://msdn.microsoft.com/en-us/library/ff553200.aspx

__forceinline
RTL_SPLAY_LINKS *	NTAPI	RtlRightChild(RTL_SPLAY_LINKS *Links)
{
	return Links->RightChild;
}

__inline
RTL_SPLAY_LINKS *	NTAPI	RtlRoot(RTL_SPLAY_LINKS *Links)
{
	while (Links->Parent != Links)
		Links = Links->Parent;

	return Links;
}

__declspec(dllimport)
VOID	NTAPI	RtlRunDecodeUnicodeString(BYTE           Seed,
		                          UNICODE_STRING *String);

__forceinline
VOID	NTAPI	RtlRunDecodeString(BYTE   Seed,
		                   STRING *String)
{
	RtlRunDecodeUnicodeString(Seed, (UNICODE_STRING *) String);
}

__declspec(dllimport)
VOID	NTAPI	RtlRunEncodeUnicodeString(LPBYTE         Seed,
		                          UNICODE_STRING *String);

__forceinline
VOID	NTAPI	RtlRunEncodeString(LPBYTE Seed,
		                   STRING *String)
{
	RtlRunEncodeUnicodeString(Seed, (UNICODE_STRING *) String);
}

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ff562759.aspx

__declspec(dllimport)
LONG	NTAPI	RtlRunOnceBeginInitialize(RTL_RUN_ONCE *RunOnce,
		                          DWORD        Flags,
		                          LPVOID       *Context);

// https://msdn.microsoft.com/en-us/library/ff562763.aspx

__declspec(dllimport)
LONG	NTAPI	RtlRunOnceComplete(RTL_RUN_ONCE *RunOnce,
		                   DWORD        Flags,
		                   LPVOID       Context);

// https://msdn.microsoft.com/en-us/library/ff562765.aspx

__declspec(dllimport)
LONG	NTAPI	RtlRunOnceExecuteOnce(RTL_RUN_ONCE         *RunOnce,
		                      RTL_RUN_ONCE_INIT_FN *InitFn,
		                      LPVOID               Parameter,
		                      LPVOID               *Context);

// https://msdn.microsoft.com/en-us/library/ms683495.aspx
// NOTE: forwarded from KERNEL32!InitOnceInitialize

__declspec(dllimport)
VOID	NTAPI	RtlRunOnceInitialize(RTL_RUN_ONCE *RunOnce);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff553204.aspx

__declspec(dllimport)
VOID	NTAPI	RtlSecondsSince1970ToTime(DWORD         ElapsedSeconds,
		                          LARGE_INTEGER *Time);

// https://msdn.microsoft.com/en-us/library/ff553210.aspx

__declspec(dllimport)
VOID	NTAPI	RtlSecondsSince1980ToTime(DWORD         ElapsedSeconds,
		                          LARGE_INTEGER *Time);

// https://msdn.microsoft.com/en-us/library/aa366877.aspx

#ifndef _WINNT_
__inline
LPVOID	NTAPI	RtlSecureZeroMemory(LPVOID Memory,
		                    SIZE_T Length)
{
#if defined(_M_IX86) || defined(_M_X64) || defined(_M_AMD64)
	// https://msdn.microsoft.com/en-us/library/ywwcatsa.aspx

	__stosb((volatile char *volatile) Memory, 0, Length);
#else
	volatile char *volatile Pointer = (volatile char *volatile) Memory;

	while (Length--)
		*Pointer++ = 0;
#endif
	return Memory;
}
#endif // _WINNT_

__declspec(dllimport)
HRESULT	NTAPI	RtlSeekMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                    LARGE_INTEGER                 Distance,
		                    DWORD                         Origin,
		                    ULARGE_INTEGER                *NewPosition);

// https://msdn.microsoft.com/en-us/library/ff553214.aspx
// NOTE: like ADVAPI32!MakeAbsoluteSD

__declspec(dllimport)
LONG	NTAPI	RtlSelfRelativeToAbsoluteSD(SECURITY_DESCRIPTOR *SelfRelativeSecurityDescriptor,
		                            SECURITY_DESCRIPTOR *AbsoluteSecurityDescriptor,
		                            LPDWORD             AbsoluteSecurityDescriptorSize,
		                            ACL                 *Dacl,
		                            LPDWORD             DaclSize,
		                            ACL                 *Sacl,
		                            LPDWORD             SaclSize,
		                            SID                 *Owner,
		                            LPDWORD             OwnerSize,
		                            SID                 *PrimaryGroup,
		                            LPDWORD             PrimaryGroupSize);

__declspec(dllimport)
LONG	NTAPI	RtlSelfRelativeToAbsoluteSD2(SECURITY_DESCRIPTOR *SelfRelativeSecurityDescriptor,
		                             LPDWORD             BufferSize);

// https://msdn.microsoft.com/en-us/library/ff562770.aspx

__declspec(dllimport)
VOID	NTAPI	RtlSetAllBits(RTL_BITMAP *BitMapHeader);

__declspec(dllimport)
LONG	NTAPI	RtlSetAttributesSecurityDescriptor(SECURITY_DESCRIPTOR         *SecurityDescriptor,
		                                   SECURITY_DESCRIPTOR_CONTROL Control,
		                                   LPDWORD                     Revision);

// https://msdn.microsoft.com/en-us/library/ff562775.aspx

__declspec(dllimport)
VOID	NTAPI	RtlSetBits(RTL_BITMAP *BitMapHeader,
		           DWORD      StartingIndex,
		           DWORD      NumberToSet);

// NOTE: like ADVAPI32!SetSecurityDescriptorControl

__declspec(dllimport)
LONG	NTAPI	RtlSetControlSecurityDescriptor(SECURITY_DESCRIPTOR         *SecurityDescriptor,
		                                SECURITY_DESCRIPTOR_CONTROL ControlBitsOfInterest,
		                                SECURITY_DESCRIPTOR_CONTROL ControlBitsToSet);

// https://msdn.microsoft.com/en-us/library/ms686197.aspx
// NOTE: forwarded from KERNEL32!SetCriticalSectionSpinCount and KERNELBASE!SetCriticalSectionSpinCount

__declspec(dllimport)
DWORD	NTAPI	RtlSetCriticalSectionSpinCount(RTL_CRITICAL_SECTION *CriticalSection,
		                               DWORD                SpinCount);

__declspec(dllimport)
LONG	NTAPI	RtlSetCurrentDirectory_U(UNICODE_STRING *PathName);

__declspec(dllimport)
LONG	NTAPI	RtlSetCurrentEnvironment(LPVOID Environment,
		                         LPVOID *PreviousEnvironment);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
BOOLEAN	NTAPI	RtlSetCurrentTransaction(HANDLE TransactionHandle);
#endif

// https://msdn.microsoft.com/en-us/library/ff562781.aspx
// NOTE: like ADVAPI32!SetSecurityDescriptorDacl

__declspec(dllimport)
LONG	NTAPI	RtlSetDaclSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                             BOOLEAN             DaclPresent,
		                             ACL                 *Dacl,
		                             BOOLEAN             DaclDefaulted);

// https://msdn.microsoft.com/en-us/library/ff553218.aspx
// NOTE: like ADVAPI32!SetSecurityDescriptorGroup

__declspec(dllimport)
LONG	NTAPI	RtlSetGroupSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                              SID                 *Group,
		                              BOOLEAN             GroupDefaulted);

__declspec(dllimport)
LONG	NTAPI	RtlSetEnvironmentVariable(LPVOID         *Environment,
		                          UNICODE_STRING *Name,
		                          UNICODE_STRING *Value);

#if _WIN32_WINNT >= 0x0601
#ifndef _WINNT_
__declspec(dllimport)
VOID	NTAPI	RtlSetExtendedFeaturesMask(CONTEXT_EX *ContextEx,
		                           DWORD64    FeatureMask);
#endif
#endif

// NOTE: like ADVAPI32!SetAclInformation

__declspec(dllimport)
LONG	NTAPI	RtlSetInformationAcl(ACL                   *Acl,
		                     LPVOID                AclInformation,
		                     DWORD                 AclInformationLength,
		                     ACL_INFORMATION_CLASS AclInformationClass);

#ifndef _WINNT_
typedef	VOID	(NTAPI	*APC_CALLBACK_FUNCTION) (NTSTATUS, LPVOID, LPVOID);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlSetIoCompletionCallback(HANDLE                FileHandle,
		                           APC_CALLBACK_FUNCTION CompletionProc,
		                           DWORD                 Flags);

// https://msdn.microsoft.com/en-us/library/ms680627.aspx
// NOTE: forwarded from KERNEL32!SetLastError and KERNELBASE!SetLastError

__declspec(dllimport)
VOID	NTAPI	RtlSetLastWin32Error(DWORD Error);

__declspec(dllimport)
VOID	NTAPI	RtlSetLastWin32ErrorAndNtStatusFromNtStatus(NTSTATUS Status);

__declspec(dllimport)
HRESULT	NTAPI	RtlSetMemoryStreamSize(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                       ULARGE_INTEGER                NewSize);

// https://msdn.microsoft.com/en-us/library/ff553220.aspx
// NOTE: like ADVAPI32!SetSecurityDescriptorOwner

__declspec(dllimport)
LONG	NTAPI	RtlSetOwnerSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                              SID                 *Owner,
		                              BOOLEAN             OwnerDefaulted);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	CDECL	RtlSetProcessIsCritical(BOOLEAN NewValue,
		                        BOOLEAN *OldValue,
		                        BOOLEAN NeedSystemCriticalBreaks);
#endif

#if _WIN32_WINNT >= 0x0601
// NOTE: like KERNEL32!SetProcessPreferredUILanguages

__declspec(dllimport)
LONG	NTAPI	RtlSetProcessPreferredUILanguages(DWORD   Flags,
		                                  LPCWSTR LanguagesBuffer,
		                                  LPDWORD NumberOfLanguagesSet);
#endif

// NOTE: like ADVAPI32!SetSecurityDescriptorSacl

__declspec(dllimport)
LONG	NTAPI	RtlSetSaclSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                             BOOLEAN             SaclPresent,
		                             ACL                 *Sacl,
		                             BOOLEAN             SaclDefaulted);

// NOTE: like ADVAPI32!SetSecurityDescriptorRMControl

__declspec(dllimport)
VOID	NTAPI	RtlSetSecurityDescriptorRMControl(SECURITY_DESCRIPTOR *SecurityDescriptor,
		                                  LPBYTE              RMControl);

__declspec(dllimport)
LONG	NTAPI	RtlSetSecurityObject(SECURITY_INFORMATION SecurityInformation,
		                     SECURITY_DESCRIPTOR  *ModificationDescriptor,
		                     SECURITY_DESCRIPTOR  **ObjectsSecurityDescriptor,
		                     GENERIC_MAPPING      *GenericMapping,
		                     HANDLE               TokenHandle);

__declspec(dllimport)
LONG	NTAPI	RtlSetSecurityObjectEx(SECURITY_INFORMATION SecurityInformation,
		                       SECURITY_DESCRIPTOR  *ModificationDescriptor,
		                       SECURITY_DESCRIPTOR  **ObjectsSecurityDescriptor,
		                       DWORD                AutoInheritFlags,
		                       GENERIC_MAPPING      *GenericMapping,
		                       HANDLE               TokenHandle);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	CDECL	RtlSetThreadIsCritical(BOOLEAN NewValue,
		                       BOOLEAN *OldValue,
		                       BOOLEAN NeedSystemCriticalBreaks);
#endif

#if _WIN32_WINNT >= 0x0601
#define RTL_ERRORMODE_FAILCRITICALERRORS	0x00000010UL
#define RTL_ERRORMODE_NOGPFAULTERRORBOX		0x00000020UL
#define RTL_ERRORMODE_NOOPENFILEERRORBOX	0x00000040UL

__declspec(dllimport)
LONG	NTAPI	RtlSetThreadErrorMode(DWORD   NewMode,
		                      LPDWORD OldMode);
#endif

typedef	LONG	(NTAPI	RTL_START_POOL_THREAD) (THREAD_START_ROUTINE *StartAddress,
			                        LPVOID               Parameter,
			                        HANDLE               *ThreadHandle);

typedef	LONG	(NTAPI	RTL_EXIT_POOL_THREAD) (NTSTATUS ExitStatus);

__declspec(dllimport)
LONG	NTAPI	RtlSetThreadPoolStartFunc(RTL_START_POOL_THREAD *StartPoolThread,
		                          RTL_EXIT_POOL_THREAD  *ExitPoolThread);

#if _WIN32_WINNT >= 0x0600
// NOTE: like KERNEL32!GetThreadPreferredUILanguages

__declspec(dllimport)
LONG	NTAPI	RtlSetThreadPreferredUILanguages(DWORD   Flags,
		                                 LPCWSTR LanguagesBuffer,
		                                 LPDWORD NumberOfLanguagesSet);
#endif

// NOTE: like KERNEL32!SetTimerQueueTimer

__declspec(deprecated("obsolete function, do not use"))
__declspec(dllimport)
VOID	NTAPI	RtlSetTimer(HANDLE                  TimerQueueHandle,
		            HANDLE                  *TimerHandle,
		            WAITORTIMERCALLBACKFUNC Function,
		            LPVOID                  Context,
		            DWORD                   DueTime,
		            DWORD                   Period,
		            DWORD                   Flags);

typedef	DWORD	(NTAPI	RTLP_UNHANDLED_EXCEPTION_FILTER) (EXCEPTION_POINTERS *ExceptionInfo);

__declspec(dllimport)
VOID	NTAPI	RtlSetUnhandledExceptionFilter(RTLP_UNHANDLED_EXCEPTION_FILTER *UnhandledExceptionFilter);

#ifndef _WIN64
#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlSetUserCallbackExceptionFilter(LPVOID);
#endif
#endif

__declspec(dllimport)
BOOLEAN	NTAPI	RtlSetUserFlagsHeap(LPCVOID HeapHandle,
		                    DWORD   Flags,
		                    LPVOID  BaseAddress,
		                    DWORD   UserFlagsReset,
		                    DWORD   UserFlagsSet);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlSetUserValueHeap(LPCVOID HeapHandle,
		                    DWORD   Flags,
		                    LPVOID  BaseAddress,
		                    LPVOID  UserValue);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlSidDominates(SID     *Sid1,
		                SID     *Sid2,
		                BOOLEAN *Dominates);

__declspec(dllimport)
LONG	NTAPI	RtlSidEqualLevel(SID     *Sid1,
		                 SID     *Sid2,
		                 BOOLEAN *EqualLevel);

__declspec(dllimport)
LONG	NTAPI	RtlSidHashInitialize(SID_AND_ATTRIBUTES      *SidAttr,
		                     DWORD                   SidCount,
		                     SID_AND_ATTRIBUTES_HASH *SidAttrHash);

__declspec(dllimport)
SID_AND_ATTRIBUTES *	NTAPI	RtlSidHashLookup(SID_AND_ATTRIBUTES_HASH *SidAttrHash,
				                 SID                     *Sid);

__declspec(dllimport)
LONG	NTAPI	RtlSidIsHigherLevel(SID     *Sid1,
		                    SID     *Sid2,
		                    BOOLEAN *HigherLevel);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/aa366706.aspx
// NOTE: forwarded from KERNEL32!HeapSize and KERNELBASE!HeapSize

__declspec(dllimport)
SIZE_T	NTAPI	RtlSizeHeap(LPCVOID HeapHandle,
		            DWORD   Flags,
		            LPCVOID HeapBase);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlSleepConditionVariableCS(RTL_CONDITION_VARIABLE *ConditionVariable,
		                            RTL_CRITICAL_SECTION   *CriticalSection,
		                            LARGE_INTEGER          *TimeOut);

__declspec(dllimport)
LONG	NTAPI	RtlSleepConditionVariableSRW(RTL_CONDITION_VARIABLE *ConditionVariable,
		                             RTL_SRWLOCK            *SRWLock,
		                             LARGE_INTEGER          *TimeOut,
		                             DWORD                  Flags);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ff553226.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlSplay(RTL_SPLAY_LINKS *Links);

__declspec(dllimport)
LONG	NTAPI	RtlStartRXact(RTL_RXACT_CONTEXT *RXactContext);

__declspec(dllimport)
HRESULT	NTAPI	RtlStatMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                    STATSTG                       *StatusInformation,
		                    DWORD                         Flags);

// https://msdn.microsoft.com/en-us/library/ff562871.aspx

__declspec(dllimport)
LONG	NTAPI	RtlStringFromGUID(GUID           *Guid,
		                  UNICODE_STRING *GuidString);

// https://msdn.microsoft.com/en-us/library/ff553230.aspx
// NOTE: like ADVAPI32!GetSidSubAuthorityCount

__declspec(dllimport)
LPBYTE	NTAPI	RtlSubAuthorityCountSid(SID *Sid);

// https://msdn.microsoft.com/en-us/library/ff553236.aspx
// NOTE: like ADVAPI32!GetSidSubAuthority

__declspec(dllimport)
LPDWORD	NTAPI	RtlSubAuthoritySid(SID   *Sid,
		                   DWORD SubAuthorityCount);

__inline
RTL_SPLAY_LINKS *	NTAPI	RtlSubtreeLargest(RTL_SPLAY_LINKS *Links)
{
	while (Links->LeftChild != NULL)
		Links = Links->LeftChild;

	return Links;
}

// https://msdn.microsoft.com/en-us/library/ff553237.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlSubtreePredecessor(RTL_SPLAY_LINKS *Links);

__inline
RTL_SPLAY_LINKS *	NTAPI	RtlSubtreeSmallest(RTL_SPLAY_LINKS *Links)
{
	while (Links->RightChild != NULL)
		Links = Links->RightChild;

	return Links;
}

// https://msdn.microsoft.com/en-us/library/ff553239.aspx

__declspec(dllimport)
RTL_SPLAY_LINKS *	NTAPI	RtlSubtreeSuccessor(RTL_SPLAY_LINKS *Links);

__declspec(dllimport)
LONG	NTAPI	RtlSystemTimeToLocalTime(LARGE_INTEGER *SystemTime,
		                         LARGE_INTEGER *LocalTime);

// https://msdn.microsoft.com/en-us/library/ff562879.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTimeFieldsToTime(TIME_FIELDS   *TimeFields,
		                    LARGE_INTEGER *Time);

__declspec(dllimport)
VOID	NTAPI	RtlTimeToElapsedTimeFields(LARGE_INTEGER *Time,
		                           TIME_FIELDS   *TimeFields);

// https://msdn.microsoft.com/en-us/library/ff553241.aspx
// https://msdn.microsoft.com/en-us/library/ms724928.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTimeToSecondsSince1970(LARGE_INTEGER *Time,
		                          LPDWORD       ElapsedSeconds);

// https://msdn.microsoft.com/en-us/library/ff553244.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTimeToSecondsSince1980(LARGE_INTEGER *Time,
		                          LPDWORD       ElapsedSeconds);

// https://msdn.microsoft.com/en-us/library/ff562884.aspx

__declspec(dllimport)
VOID	NTAPI	RtlTimeToTimeFields(LARGE_INTEGER *Time,
		                    TIME_FIELDS   *TimeFields);

#define RTL_TRACE_BLOCK_MAGIC		0xABCDAAAAUL
#define RTL_TRACE_SEGMENT_MAGIC		0xABCDBBBBUL
#define RTL_TRACE_DATABASE_MAGIC	0xABCDCCCCUL

typedef	DWORD	(NTAPI	RTL_TRACE_HASH_FUNCTION) (DWORD  Count,
			                          LPVOID *Trace);

typedef	struct	_RTL_TRACE_BLOCK
{
	DWORD	Magic;
	DWORD	Count;
	DWORD	Size;
	DWORD	UserCount;
	DWORD	UserSize;
	LPVOID	UserContext;
	struct	_RTL_TRACE_BLOCK	*Next;
	LPVOID	*Trace;
} RTL_TRACE_BLOCK, *LPRTL_TRACE_BLOCK, *PRTL_TRACE_BLOCK;

typedef	struct	_RTL_TRACE_SEGMENT	RTL_TRACE_SEGMENT;

typedef	struct	_RTL_TRACE_DATABASE
{
	DWORD			Magic;
	DWORD			Flags;
	DWORD			Tag;
	RTL_TRACE_SEGMENT	*SegmentList;
	SIZE_T			MaximumSize;
	SIZE_T			CurrentSize;
	LPVOID			Owner;
	RTL_CRITICAL_SECTION	Lock;
	DWORD			NoOfBuckets;
	RTL_TRACE_BLOCK		**Buckets;
	RTL_TRACE_HASH_FUNCTION	*HashFunction;
	SIZE_T			NoOfTraces;
	SIZE_T			NoOfHits;
	DWORD			HashCounter[16];
} RTL_TRACE_DATABASE, *LPRTL_TRACE_DATABASE, *PRTL_TRACE_DATABASE;

typedef	struct	_RTL_TRACE_ENUMERATE
{
	RTL_TRACE_DATABASE	*Database;
	DWORD	Index;
	RTL_TRACE_BLOCK	*Block;
} RTL_TRACE_ENUMERATE, *LPRTL_TRACE_ENUMERATE, *PRTL_TRACE_ENUMERATE;

typedef	struct	_RTL_TRACE_SEGMENT
{
	DWORD			Magic;
	RTL_TRACE_DATABASE	*Database;
	RTL_TRACE_SEGMENT	*NextSegment;
	SIZE_T			TotalSize;
	CHAR			*SegmentStart;
	CHAR			*SegmentEnd;
	CHAR			*SegmentFree;
} RTL_TRACE_SEGMENT, *LPRTL_TRACE_SEGMENT, *PRTL_TRACE_SEGMENT;

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseAdd(RTL_TRACE_DATABASE *Database,
		                    DWORD              Count,
		                    LPVOID             *Trace,
		                    RTL_TRACE_BLOCK    **TraceBlock);	// optional

__declspec(dllimport)
RTL_TRACE_DATABASE *	NTAPI	RtlTraceDatabaseCreate(DWORD                   Buckets,
				                       SIZE_T                  MaximumSize,
				                       DWORD                   Flags,
				                       DWORD                   Tag,
				                       RTL_TRACE_HASH_FUNCTION *HashFunction);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseDestroy(RTL_TRACE_DATABASE *Database);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseEnumerate(RTL_TRACE_DATABASE  *Database,
		                          RTL_TRACE_ENUMERATE *TraceEnumerate,
		                          RTL_TRACE_BLOCK     **TraceBlock);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseFind(RTL_TRACE_DATABASE *Database,
		                     DWORD              Count,
		                     LPVOID             *Trace,
		                     RTL_TRACE_BLOCK    **TraceBlock);	// optional

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseLock(RTL_TRACE_DATABASE *Database);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseUnlock(RTL_TRACE_DATABASE *Database);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTraceDatabaseValidate(RTL_TRACE_DATABASE *Database);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/dd405523.aspx
// NOTE: forwarded from KERNEL32!TryAcquireSRWLockExclusive and KERNELBASE!TryAcquireSRWLockExclusive

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTryAcquireSRWLockExclusive(RTL_SRWLOCK *SRWLock);

// https://msdn.microsoft.com/en-us/library/dd405524.aspx
// NOTE: forwarded from KERNEL32!TryAcquireSRWLockShared and KERNELBASE!TryAcquireSRWLockShared

__declspec(dllimport)
BOOLEAN	NTAPI	RtlTryAcquireSRWLockShared(RTL_SRWLOCK *SRWLock);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms686857.aspx
// NOTE: forwarded from KERNEL32!TryEnterCriticalSection and KERNELBASE!TryEnterCriticalSection

__declspec(dllimport)
BOOL	NTAPI	RtlTryEnterCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

__declspec(dllimport)
DWORD	FASTCALL	RtlUlongByteSwap(ULONG Source);

__declspec(dllimport)
QWORD	FASTCALL	RtlUlonglongByteSwap(ULONGLONG Source);

__declspec(dllimport)
LONG	NTAPI	RtlUnhandledExceptionFilter(EXCEPTION_POINTERS *ExceptionInfo);

__declspec(dllimport)
LONG	NTAPI	RtlUnhandledExceptionFilter2(EXCEPTION_POINTERS *ExceptionInfo,
		                             LPCSTR             Function);

__declspec(dllimport)
LONG	NTAPI	RtlxUnicodeStringToAnsiSize(UNICODE_STRING *UnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553248.aspx

#if 1
__declspec(dllimport)
DWORD	NTAPI	RtlUnicodeStringToAnsiSize(UNICODE_STRING *UnicodeString);
#else
__inline
DWORD	NTAPI	RtlUnicodeStringToAnsiSize(UNICODE_STRING *UnicodeString)
{
	if (*NlsMbCodePageTag)
		return RtlxUnicodeStringToAnsiSize(UnicodeString);

	return UnicodeString->Length / sizeof(WCHAR) + 1;
}
#endif

// https://msdn.microsoft.com/en-us/library/ff562969.aspx
// https://msdn.microsoft.com/en-us/library/ms648421.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeStringToAnsiString(ANSI_STRING    *DestinationString,
		                             UNICODE_STRING *SourceString,
		                             BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff562973.aspx
// NOTE: when 0, base is inferred from (the absence of) a case-sensitive "0b", "0o" or "0x" prefix!

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeStringToInteger(UNICODE_STRING *String,
		                          DWORD          Base,	// 0, 2, 8, 10, 16
		                          LPDWORD        Value);

// https://msdn.microsoft.com/en-us/library/ff553251.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeStringToCountedOemString(OEM_STRING     *DestinationString,
		                                   UNICODE_STRING *SourceString,
		                                   BOOLEAN        AllocateDestinationString);

__declspec(dllimport)
LONG	NTAPI	RtlxUnicodeStringToOemSize(UNICODE_STRING *UnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553254.aspx

#if 1
__declspec(dllimport)
DWORD	NTAPI	RtlUnicodeStringToOemSize(UNICODE_STRING *UnicodeString);
#else
__inline
DWORD	NTAPI	RtlUnicodeStringToOemSize(UNICODE_STRING *UnicodeString)
{
	if (*NlsMbOemCodePageTag)
		return RtlxUnicodeStringToOemSize(UnicodeString);

	return UnicodeString->Length / sizeof(WCHAR) + 1;
}
#endif

// https://msdn.microsoft.com/en-us/library/ms648422.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeStringToOemString(OEM_STRING     *DestinationString,
		                            UNICODE_STRING *SourceString,
		                            BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff553256.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeToCustomCPN(CPTABLEINFO *CustomCP,
		                      LPSTR       CustomString,
		                      DWORD       MaxBytesInCustomString,
		                      LPDWORD     BytesInCustomString,
		                      LPCWSTR     UnicodeString,
		                      DWORD       BytesInUnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553261.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeToMultiByteN(LPSTR   MultiByteString,
		                       DWORD   MaxBytesInMultiByteString,
		                       LPDWORD BytesInMultiByteString,
		                       LPCWSTR UnicodeString,
		                       DWORD   BytesInUnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553266.aspx
// https://msdn.microsoft.com/en-us/library/ms648423.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeToMultiByteSize(LPDWORD BytesInMultiByteString,
		                          LPCWSTR UnicodeString,
		                          DWORD   BytesInUnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553272.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeToOemN(LPSTR   OemString,
		                 DWORD   MaxBytesInOemString,
		                 LPDWORD BytesInOemString,
		                 LPCWSTR UnicodeString,
		                 DWORD   BytesInUnicodeString);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/ee453688.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUnicodeToUTF8N(LPSTR   UTF8StringDestination,
		                  DWORD   UTF8StringMaxByteCount,
		                  LPDWORD UTF8StringActualByteCount,
		                  LPCWSTR UnicodeStringSource,
		                  DWORD   UnicodeStringWCharCount);
#endif

// https://msdn.microsoft.com/en-us/library/bb432429.aspx

__declspec(dllimport)
DWORD	NTAPI	RtlUniform(LPDWORD Seed);

#if _WIN32_WINNT >= 0x0501
__declspec(dllimport)
LONG	NTAPI	RtlUnlockBootStatusData(HANDLE *BootStatusDataHandle);
#endif

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlUnlockCurrentThread(VOID);
#endif

__declspec(dllimport)
BOOLEAN	NTAPI	RtlUnlockHeap(LPCVOID HeapHandle);

__declspec(dllimport)
HRESULT	NTAPI	RtlUnlockMemoryStreamRegion(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                            ULARGE_INTEGER                Offset,
		                            ULARGE_INTEGER                Bytes,
		                            DWORD                         LockType);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	RtlUnlockModuleSection(LPVOID Address);
#endif

#if _WIN32_WINNT < 0x0600
// https://msdn.microsoft.com/en-us/library/ms680609.aspx
// NOTE: forwarded from KERNEL32!RtlUnwind

__declspec(dllimport)
VOID	NTAPI	RtlUnwind(LPCVOID          TargetFrame,
		          LPCVOID          TargetIp,
		          EXCEPTION_RECORD *ExceptionRecord,
		          LPCVOID          ReturnValue);
#endif

#ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/ms680615.aspx
// NOTE: forwarded from KERNEL32!RtlUnwindEx

__declspec(dllimport)
VOID	NTAPI	RtlUnwindEx(LPCVOID              TargetFrame,
		            LPCVOID              TargetIp,
		            EXCEPTION_RECORD     *ExceptionRecord,
		            LPCVOID              ReturnValue,
		            CONTEXT              *OriginalContext,
		            UNWIND_HISTORY_TABLE *HistoryTable);
#endif

// https://msdn.microsoft.com/en-us/library/ff563003.aspx

__declspec(dllimport)
WCHAR	NTAPI	RtlUpcaseUnicodeChar(WCHAR SourceCharacter);

// https://msdn.microsoft.com/en-us/library/ff563005.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeString(UNICODE_STRING *DestinationString,
		                       UNICODE_STRING *SourceString,
		                       BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff553277.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeStringToCountedOemString(OEM_STRING     *DestinationString,
		                                         UNICODE_STRING *SourceString,
		                                         BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff553282.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeStringToOemString(OEM_STRING     *DestinationString,
		                                  UNICODE_STRING *SourceString,
		                                  BOOLEAN        AllocateDestinationString);

// https://msdn.microsoft.com/en-us/library/ff553285.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeToCustomCPN(CPTABLEINFO *CustomCP,
		                            LPSTR       CustomString,
		                            DWORD       MaxBytesInCustomString,
		                            LPDWORD     BytesInCustomString,
		                            LPCWSTR     UnicodeString,
		                            DWORD       BytesInUnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553296.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeToMultiByteN(LPSTR   MultiByteString,
		                             DWORD   MaxBytesInMultiByteString,
		                             LPDWORD BytesInMultiByteString,
		                             LPCWSTR UnicodeString,
		                             DWORD   BytesInUnicodeString);

// https://msdn.microsoft.com/en-us/library/ff553309.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUpcaseUnicodeToOemN(LPSTR   OemString,
		                       DWORD   MaxBytesInOemString,
		                       LPDWORD BytesInOemString,
		                       LPCWSTR UnicodeString,
		                       DWORD   BytesInUnicodeString);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
VOID	NTAPI	RtlUpdateClonedCriticalSection(RTL_CRITICAL_SECTION *CriticalSection);

__declspec(dllimport)
VOID	NTAPI	RtlUpdateClonedSRWLock(RTL_SRWLOCK *SRWLock,
		                       BOOL        Shared);
#endif // _WIN32_WINNT

// NOTE: like KERNEL32!ChangeTimerQueueTimer

__declspec(dllimport)
LONG	NTAPI	RtlUpdateTimer(HANDLE TimerQueueHandle,
		               HANDLE TimerHandle,
		               DWORD  DueTime,
		               DWORD  Period);

// https://msdn.microsoft.com/en-us/library/ff563009.aspx

__declspec(dllimport)
CHAR	NTAPI	RtlUpperChar(CHAR Character);

// https://msdn.microsoft.com/en-us/library/ff563013.aspx

__declspec(dllimport)
VOID	NTAPI	RtlUpperString(STRING *DestinationString,
		               STRING *SourceString);

typedef	struct	_RTL_HEAP_USAGE_ENTRY
{
	struct	_RTL_HEAP_USAGE_ENTRY	*Next;
	LPVOID	Address;
	SIZE_T	Size;
	WORD	AllocatorBackTraceIndex;
	WORD	TagIndex;
} RTL_HEAP_USAGE_ENTRY, *LPRTL_HEAP_USAGE_ENTRY, *PRTL_HEAP_USAGE_ENTRY;

typedef	struct	_RTL_HEAP_USAGE
{
	DWORD			Length;
	SIZE_T			BytesAllocated;
	SIZE_T			BytesCommitted;
	SIZE_T			BytesReserved;
	SIZE_T			BytesReservedMaximum;
	RTL_HEAP_USAGE_ENTRY	*Entries;
	RTL_HEAP_USAGE_ENTRY	*AddedEntries;
	RTL_HEAP_USAGE_ENTRY	*RemovedEntries;
	DWORD_PTR		Reserved[8];
} RTL_HEAP_USAGE, *LPRTL_HEAP_USAGE, *PRTL_HEAP_USAGE;

#define HEAP_USAGE_ALLOCATED_BLOCKS	HEAP_REALLOC_IN_PLACE_ONLY
#define HEAP_USAGE_FREE_BUFFER		HEAP_ZERO_MEMORY

__declspec(dllimport)
LONG	NTAPI	RtlUsageHeap(LPCVOID        HeapHandle,
		             DWORD          Flags,
		             RTL_HEAP_USAGE *Usage);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
VOID	NTAPI	RtlUserThreadStart(USER_THREAD_START_ROUTINE *StartAddress,
		                   LPVOID                    Argument);
#endif

__declspec(dllimport)
WORD	FASTCALL	RtlUshortByteSwap(WORD Source);

#if _WIN32_WINNT >= 0x0601
// https://msdn.microsoft.com/en-us/library/ee453688.aspx

__declspec(dllimport)
LONG	NTAPI	RtlUTF8ToUnicodeN(LPWSTR  UnicodeStringDestination,
		                  DWORD   UnicodeStringMaxWCharCount,
		                  LPDWORD UnicodeStringActualWCharCount,
		                  LPCSTR  UTF8StringSource,
		                  DWORD   UTF8StringByteCount);
#endif

// NOTE: like ADVAPI32!IsValidAcl

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidAcl(ACL *Acl);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidateHeap(LPCVOID HeapHandle,
		                DWORD   Flags,
		                LPCVOID AddressToValidate);

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidateProcessHeaps(VOID);

__declspec(dllimport)
LONG	NTAPI	RtlValidateUnicodeString(DWORD          Flags,
		                         UNICODE_STRING *UnicodeString);

// https://msdn.microsoft.com/en-us/library/ff563019.aspx

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidRelativeSecurityDescriptor(SECURITY_DESCRIPTOR  *SecurityDescriptorInput,
		                                   DWORD                SecurityDescriptorLength,
		                                   SECURITY_INFORMATION RequiredInformation);

// https://msdn.microsoft.com/en-us/library/ff563024.aspx
// NOTE: like ADVAPI32!IsValidSecurityDescriptor

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidSecurityDescriptor(SECURITY_DESCRIPTOR *SecurityDescriptorInput);

// https://msdn.microsoft.com/en-us/library/ff553314.aspx
// NOTE: like ADVAPI32!IsValidSid

__declspec(dllimport)
BOOLEAN	NTAPI	RtlValidSid(SID *Sid);

// https://msdn.microsoft.com/en-us/library/ff563026.aspx

__declspec(dllimport)
LONG	NTAPI	RtlVerifyVersionInfo(RTL_OSVERSIONINFOEXW *VersionInfo,
		                     DWORD                TypeMask,
		                     QWORD                ConditionMask);

#ifdef _WIN64
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/ms680612.aspx
// https://msdn.microsoft.com/en-us/library/ms680615.aspx

typedef	struct	_FRAME_POINTERS
{
	QWORD	MemoryStackFp;
	QWORD	BackingStoreFp;
} FRAME_POINTERS, *LPFRAME_POINTERS, *PFRAME_POINTERS;

// https://msdn.microsoft.com/en-us/library/ms680617.aspx
// NOTE: forwarded from KERNEL32!RtlVirtualUnwind

typedef	struct	_KNONVOLATILE_CONTEXT_POINTERS	KNONVOLATILE_CONTEXT_POINTERS, *LPKNONVOLATILE_CONTEXT_POINTERS, *PKNONVOLATILE_CONTEXT_POINTERS;

__declspec(dllimport)
EXCEPTION_ROUTINE *	NTAPI	RtlVirtualUnwind(DWORD                         HandlerType,
				                 QWORD                         ImageBase,
				                 QWORD                         ControlPc,
				                 RUNTIME_FUNCTION              *FunctionEntry,
				                 CONTEXT                       *ContextRecord,
				                 LPVOID                        *HandlerData,
				                 FRAME_POINTERS                *EstablisherFrame,
				                 KNONVOLATILE_CONTEXT_POINTERS *ContextPointers);
#endif // _WINNT_
#endif // _WIN64

#if _WIN32_WINNT >= 0x0600
// https://msdn.microsoft.com/en-us/library/ms687076.aspx
// NOTE: forwarded from KERNEL32!WakeAllConditionVariable

__declspec(dllimport)
VOID	NTAPI	RtlWakeAllConditionVariable(RTL_CONDITION_VARIABLE *ConditionVariable);

// https://msdn.microsoft.com/en-us/library/ms687080.aspx
// NOTE: forwarded from KERNEL32!WakeConditionVariable

__declspec(dllimport)
VOID	NTAPI	RtlWakeConditionVariable(RTL_CONDITION_VARIABLE *ConditionVariable);
#endif // _WIN32_WINNT

__declspec(dllimport)
DWORD	NTAPI	RtlWalkFrameChain(LPVOID *Callers,
		                  DWORD  Count,
		                  DWORD  Flags);

// NOTE: like KERNEL32!HeapWalk

__declspec(dllimport)
LONG	NTAPI	RtlWalkHeap(LPCVOID            HeapHandle,
		            PROCESS_HEAP_ENTRY *ProcessHeapEntry);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
RTL_DYNAMIC_HASH_TABLE_ENTRY *	NTAPI	RtlWeaklyEnumerateEntryHashTable(RTL_DYNAMIC_HASH_TABLE            *HashTable,
					                                 RTL_DYNAMIC_HASH_TABLE_ENUMERATOR *Enumerator);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlWow64EnableFsRedirection(BOOLEAN Wow64FsEnableRedirection);

__declspec(dllimport)
LONG	NTAPI	RtlWow64EnableFsRedirectionEx(LPVOID Wow64FsEnableRedirection,
		                              LPVOID *OldFsRedirectionLevel);

#ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/ms681670.aspx

typedef	struct	_WOW64_CONTEXT	WOW64_CONTEXT, *LPWOW64_CONTEXT, *PWOW64_CONTEXT;

__declspec(dllimport)
LONG	NTAPI	RtlWow64GetThreadContext(HANDLE        ThreadHandle,
		                         WOW64_CONTEXT *ThreadContext);

__declspec(dllimport)
LONG	NTAPI	RtlWow64SetThreadContext(HANDLE        ThreadHandle,
		                         WOW64_CONTEXT *ThreadContext);
#endif // _WIN64

__declspec(dllimport)
HRESULT	NTAPI	RtlWriteMemoryStream(RTL_MEMORY_STREAM_WITH_VTABLE *MemoryStream,
		                     LPCVOID                       Buffer,
		                     DWORD                         BytesToWrite,
		                     LPDWORD                       BytesWritten);

// https://msdn.microsoft.com/en-us/library/ff563034.aspx

__declspec(dllimport)
LONG	NTAPI	RtlWriteRegistryValue(DWORD   RelativeTo,
		                      LPCWSTR Path,
		                      LPCWSTR ValueName,
		                      DWORD   ValueType,
		                      LPVOID  ValueData,
		                      DWORD   ValueLength);

#if 1
#pragma intrinsic(memset)
#else
#undef RtlZeroMemory

// https://msdn.microsoft.com/en-us/library/ff563610.aspx
// NOTE: forwarded from KERNEL32!RtlZeroMemory

__declspec(dllimport)
VOID	NTAPI	RtlZeroMemory(LPVOID *Destination,
		              SIZE_T Length);
#endif

__declspec(dllimport)
LONG	NTAPI	RtlZombifyActivationContext(ACTIVATION_CONTEXT *ActivationContext);

__declspec(dllimport)
LONG	NTAPI	RtlpApplyLengthFunction(DWORD    Flags,
		                        SIZE_T   SizeOfStruct,
		                        LPVOID   UnicodeStringOrUnicodeStringBuffer,
		                        NTSTATUS (NTAPI *LengthFunction) (DWORD, UNICODE_STRING *, DWORD *));

#define RtlRemoveLastNtPathElement(Flags, Path)			(RtlpApplyLengthFunction((Flags), sizeof(*Path), (Path), RtlGetLengthWithoutLastNtPathElement))
#define RtlRemoveLastFullDosPathElement(Flags, Path)		(RtlpApplyLengthFunction((Flags), sizeof(*Path), (Path), RtlGetLengthWithoutLastFullDosPathElement))
#define RtlRemoveLastFullDosOrNtPathElement(Flags, Path)	(RtlpApplyLengthFunction((Flags), sizeof(*Path), (Path), RtlGetLengthWithoutLastFullDosOrNtPathElement))
#define RtlRemoveTrailingPathSeparators(Flags, Path)		(RtlpApplyLengthFunction((Flags), sizeof(*Path), (Path), RtlGetLengthWithoutTrailingPathSeperators))

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LONG	NTAPI	TpAllocCleanupGroup(PTP_CLEANUP_GROUP *CleanupGroupReturn);

typedef	VOID	(NTAPI	PTP_IO_CALLBACK) (PTP_CALLBACK_INSTANCE Instance,
			                  LPVOID                Context,
			                  LPVOID                ApcContext,
			                  IO_STATUS_BLOCK       *IoStatusBlock,
			                  PTP_IO                Io);

__declspec(dllimport)
LONG	NTAPI	TpAllocIoCompletion(PTP_IO               *IoReturn,
		                    HANDLE               File,
		                    PTP_IO_CALLBACK      *Callback,
		                    LPVOID               Context,
		                    PTP_CALLBACK_ENVIRON CallbackEnviron);

__declspec(dllimport)
LONG	NTAPI	TpAllocPool(PTP_POOL *PoolReturn,
		            LPVOID   Reserved);

__declspec(dllimport)
LONG	NTAPI	TpAllocTimer(PTP_TIMER            *Timer,
		             PTP_TIMER_CALLBACK   Callback,
		             LPVOID               Context,
		             PTP_CALLBACK_ENVIRON CallbackEnviron);

__declspec(dllimport)
LONG	NTAPI	TpAllocWait(PTP_WAIT             *WaitReturn,
		            PTP_WAIT_CALLBACK    Callback,
		            LPVOID               Context,
		            PTP_CALLBACK_ENVIRON CallbackEnviron);

__declspec(dllimport)
LONG	NTAPI	TpAllocWork(PTP_WORK             *WorkReturn,
		            PTP_WORK_CALLBACK    Callback,
		            LPVOID               Context,
		            PTP_CALLBACK_ENVIRON CallbackEnviron);

// https://msdn.microsoft.com/en-us/library/ms684171.aspx
// NOTE: forwarded from KERNEL32!LeaveCriticalSectionWhenCallbackReturns and KERNELBASE!LeaveCriticalSectionWhenCallbackReturns

__declspec(dllimport)
VOID	NTAPI	TpCallbackLeaveCriticalSectionOnCompletion(TP_CALLBACK_INSTANCE *pci,
		                                           CRITICAL_SECTION     *pcs);

// https://msdn.microsoft.com/en-us/library/ms685070.aspx
// NOTE: forwarded from KERNEL32!ReleaseMutexWhenCallbackReturns and KERNELBASE!ReleaseMutexWhenCallbackReturns

__declspec(dllimport)
VOID	NTAPI	TpCallbackReleaseMutexOnCompletion(TP_CALLBACK_INSTANCE *pci,
		                                   HANDLE               mut);

// https://msdn.microsoft.com/en-us/library/ms685073.aspx
// NOTE: forwarded from KERNEL32!ReleaseSemaphoreWhenCallbackReturns and KERNELBASE!ReleaseSemaphoreWhenCallbackReturns

__declspec(dllimport)
VOID	NTAPI	TpCallbackReleaseSemaphoreOnCompletion(TP_CALLBACK_INSTANCE *pci,
		                                       HANDLE               sem,
		                                       DWORD                crel);

// https://msdn.microsoft.com/en-us/library/ms686214.aspx
// NOTE: forwarded from KERNEL32!SetEventWhenCallbackReturns and KERNELBASE!SetEventWhenCallbackReturns

__declspec(dllimport)
VOID	NTAPI	TpCallbackSetEventOnCompletion(TP_CALLBACK_INSTANCE *pci,
		                               HANDLE               evt);

// https://msdn.microsoft.com/en-us/library/ms683154.aspx
// NOTE: forwarded from KERNEL32!FreeLibraryWhenCallbackReturns and KERNELBASE!FreeLibraryWhenCallbackReturns

__declspec(dllimport)
VOID	NTAPI	TpCallbackUnloadDllOnCompletion(TP_CALLBACK_INSTANCE *pci,
		                                HMODULE              mod);

// https://msdn.microsoft.com/en-us/library/ms681983.aspx
// NOTE: forwarded from KERNEL32!CancelThreadpoolIo and KERNELBASE!CancelThreadpoolIo

__declspec(dllimport)
VOID	NTAPI	TpCancelAsyncIoOperation(TP_IO *pio);

// https://msdn.microsoft.com/en-us/library/ms682581.aspx
// NOTE: forwarded from KERNEL32!DisassociateCurrentThreadFromCallback and KERNELBASE!DisassociateCurrentThreadFromCallback

__declspec(dllimport)
VOID	NTAPI	TpDisassociateCallback(TP_CALLBACK_INSTANCE *pci);

// https://msdn.microsoft.com/en-us/library/ms684133.aspx
// NOTE: forwarded from KERNEL32!IsThreadpoolTimerSet and KERNELBASE!IsThreadpoolTimerSet

__declspec(dllimport)
BOOL	NTAPI	TpIsTimerSet(TP_TIMER *pti);

// https://msdn.microsoft.com/en-us/library/ms686338.aspx
// NOTE: forwarded from KERNEL32!SubmitThreadpoolWork and KERNELBASE!SubmitThreadpoolWork

__declspec(dllimport)
VOID	NTAPI	TpPostWork(TP_WORK *pwk);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	TpQueryPoolStackInformation(PTP_POOL                   Pool,
		                            PTP_POOL_STACK_INFORMATION PoolStackInformation);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms682030.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolCleanupGroup and KERNELBASE!CloseThreadpoolCleanupGroup

__declspec(dllimport)
VOID	NTAPI	TpReleaseCleanupGroup(TP_POOL *ptpp);

// https://msdn.microsoft.com/en-us/library/ms682036.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolCleanupGroupMembers and KERNELBASE!CloseThreadpoolCleanupGroupMembers

__declspec(dllimport)
VOID	NTAPI	TpReleaseCleanupGroupMembers(TP_CLEANUP_GROUP *ptpcg,
		                             BOOL             fCancelPendingCallbacks,
		                             LPVOID           pvCleanupContext);

// https://msdn.microsoft.com/en-us/library/ms682038.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolIo and KERNELBASE!CloseThreadpoolIo

__declspec(dllimport)
VOID	NTAPI	TpReleaseIoCompletion(TP_IO *pio);

// https://msdn.microsoft.com/en-us/library/ms682030.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpool and KERNELBASE!CloseThreadpool

__declspec(dllimport)
VOID	NTAPI	TpReleasePool(TP_POOL *ptpp);

// https://msdn.microsoft.com/en-us/library/ms682040.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolTimer and KERNELBASE!CloseThreadpoolTimer

__declspec(dllimport)
VOID	NTAPI	TpReleaseTimer(TP_TIMER *pti);

// https://msdn.microsoft.com/en-us/library/ms682042.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolWait and KERNELBASE!CloseThreadpoolWait

__declspec(dllimport)
VOID	NTAPI	TpReleaseWait(TP_WAIT *pwa);

// https://msdn.microsoft.com/en-us/library/ms682043.aspx
// NOTE: forwarded from KERNEL32!CloseThreadpoolWork and KERNELBASE!CloseThreadpoolWork

__declspec(dllimport)
VOID	NTAPI	TpReleaseWork(TP_WORK *pwk);

// https://msdn.microsoft.com/en-us/library/ms686266.aspx
// NOTE: forwarded from KERNEL32!SetThreadpoolThreadMaximum and KERNELBASE!SetThreadpoolThreadMaximum

__declspec(dllimport)
VOID	NTAPI	TpSetPoolMaxThreads(TP_POOL *ptpp,
		                    DWORD   cthrdMost);

__declspec(dllimport)
LONG	NTAPI	TpSetPoolMinThreads(TP_POOL *ptpp,
		                    DWORD   cthrdLowest);

#if _WIN32_WINNT >= 0x0601
__declspec(dllimport)
LONG	NTAPI	TpSetPoolStackInformation(PTP_POOL                   Pool,
		                          PTP_POOL_STACK_INFORMATION PoolStackInformation);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms686271.aspx
// NOTE: forwarded from KERNEL32!SetThreadpoolTimer and KERNELBASE!SetThreadpoolTimer

__declspec(dllimport)
VOID	NTAPI	TpSetTimer(TP_TIMER *pti,
		           FILETIME *pftDueTime,
		           DWORD    msPeriod,
		           DWORD    msWindowLength);

// https://msdn.microsoft.com/en-us/library/ms686273.aspx
// NOTE: forwarded from KERNEL32!SetThreadpoolWait and KERNELBASE!SetThreadpoolWait

__declspec(dllimport)
VOID	NTAPI	TpSetWait(TP_WAIT  *pwa,
		          HANDLE   h,
		          FILETIME *pftTimeout);

// https://msdn.microsoft.com/en-us/library/ms686326.aspx
// NOTE: forwarded from KERNEL32!StartThreadpoolIo and KERNELBASE!StartThreadpoolIo

__declspec(dllimport)
VOID	NTAPI	TpStartAsyncIoOperation(TP_IO *pio);

// https://msdn.microsoft.com/en-us/library/ms687038.aspx
// NOTE: forwarded from KERNEL32!WaitForThreadpoolIoCallbacks and KERNELBASE!WaitForThreadpoolIoCallbacks

__declspec(dllimport)
VOID	NTAPI	TpWaitForIoCompletion(TP_IO *pio,
		                      BOOL  fCancelPendingCallbacks);

// https://msdn.microsoft.com/en-us/library/ms687042.aspx
// NOTE: forwarded from KERNEL32!WaitForThreadpoolTimerCallbacks and KERNELBASE!WaitForThreadpoolTimerCallbacks

__declspec(dllimport)
VOID	NTAPI	TpWaitForTimer(TP_TIMER *pti,
		               BOOL     fCancelPendingCallbacks);

// https://msdn.microsoft.com/en-us/library/ms687047.aspx
// NOTE: forwarded from KERNEL32!WaitForThreadpoolWaitCallbacks and KERNELBASE!WaitForThreadpoolWaitCallbacks

__declspec(dllimport)
VOID	NTAPI	TpWaitForWait(TP_WAIT *pwa,
		              BOOL    fCancelPendingCallbacks);

// https://msdn.microsoft.com/en-us/library/ms687053.aspx
// NOTE: forwarded from KERNEL32!WaitForThreadpoolWorkCallbacks and KERNELBASE!WaitForThreadpoolWorkCallbacks

__declspec(dllimport)
VOID	NTAPI	TpWaitForWork(TP_WORK *pwk,
		              BOOL    fCancelPendingCallbacks);
#endif // _WIN32_WINNT

// https://msdn.microsoft.com/en-us/library/ms725493.aspx
// NOTE: forwarded from KERNEL32!VerSetConditionMask

__declspec(dllimport)
QWORD	NTAPI	VerSetConditionMask(QWORD dwlConditionMask,
		                    DWORD dwTypeBitMask,
		                    BYTE  dwConditionMask);

#if 0 // #ifdef _WIN64
// https://msdn.microsoft.com/en-us/library/mt732293.aspx
// NOTE: forwarded from KERNEL32!__C_specific_handler and KERNELBASE!__C_specific_handler

__declspec(dllimport)
VOID	CDECL	__C_specific_handler(EXCEPTION_RECORD   *ExceptionRecord,
		                     LPVOID             *EstablisherFrame,
		                     CONTEXT            *ContextRecord,
		                     DISPATCHER_CONTEXT *DispatcherContext);

// https://msdn.microsoft.com/en-us/library/ms648426.aspx
// https://msdn.microsoft.com/en-us/library/tawsa7cb.aspx
// NOTE: forwarded from KERNEL32!__chkstk and KERNELBASE!__chkstk

__declspec(dllimport)
VOID	CDECL	__chkstk(SIZE_T size);

// NOTE: forwarded from KERNEL32!__misaligned_access and KERNELBASE!__misaligned_access

__declspec(dllimport)
???	CDECL	__misaligned_access(???);

// NOTE: forwarded from KERNEL32!_local_unwind and KERNELBASE!_local_unwind

__declspec(dllimport)
VOID	CDECL	_local_unwind(LPCVOID frame, LPCVOID target);
#if 0
{
	RtlUnwind(frame, target, NULL, NULL);
}
#endif
#endif // _WIN64

#if 0	// NOTE: see CTYPE.H
// https://msdn.microsoft.com/en-us/library/5wcd48xw.aspx

__declspec(dllimport)
LONG	CDECL	__isascii(CHAR chr);

// https://msdn.microsoft.com/en-us/library/9aw53yek.aspx

__declspec(dllimport)
LONG	CDECL	__iscsym(CHAR chr);

// https://msdn.microsoft.com/en-us/library/9aw53yek.aspx

__declspec(dllimport)
LONG	CDECL	__iscsymf(CHAR chr);

// https://msdn.microsoft.com/en-us/library/sf55c429.aspx

__declspec(dllimport)
CHAR	CDECL	__toascii(CHAR chr);
#endif

#if 0	// NOTE: only for internal use of MSC compiler!
#ifndef _WIN64
// https://msdn.microsoft.com/en-us/library/ms648425.aspx

__declspec(dllimport)
LONGLONG	CDECL	_alldiv(LONGLONG Dividend, LONGLONG Divisor);

__declspec(dllimport)
LONGLONG	CDECL	_alldvrm(LONGLONG Dividend, LONGLONG Divisor);

__declspec(dllimport)
LONGLONG	CDECL	_allmul(LONGLONG Multiplicand, LONGLONG Multiplier);

__declspec(dllimport)
LONGLONG	CDECL	_allrem(LONGLONG Dividend, LONGLONG Divisor);

__declspec(dllimport)
LONGLONG	CDECL	_allshl(LONGLONG Value, LONG ShiftCount);

__declspec(dllimport)
LONGLONG	CDECL	_allshr(LONGLONG Value, LONG ShiftCount);

// https://msdn.microsoft.com/en-us/library/mt703294.aspx

__declspec(dllimport)
ULONGLONG	CDECL	_aulldiv(ULONGLONG Dividend, ULONGLONG Divisor);

__declspec(dllimport)
ULONGLONG	CDECL	_aulldvrm(ULONGLONG Dividend, ULONGLONG Divisor);

__declspec(dllimport)
ULONGLONG	CDECL	_aullrem(ULONGLONG Dividend, ULONGLONG Divisor);

__declspec(dllimport)
ULONGLONG	CDECL	_aullshr(ULONGLONG Value, LONG ShiftCount);
#endif // _WIN64

__declspec(dllimport)
LPVOID	CDECL	_alloca_probe(SIZE_T size);

#if _WIN32_WINNT >= 0x0600
__declspec(dllimport)
LPVOID	CDECL	_alloca_probe_8(SIZE_T size);

__declspec(dllimport)
LPVOID	CDECL	_alloca_probe_16(SIZE_T size);
#endif

// https://msdn.microsoft.com/en-us/library/ms648426.aspx

__declspec(dllimport)
LPVOID	CDECL	_chkstk(SIZE_T size);

__declspec(dllimport)
DWORD	_fltused;

#ifndef _WIN64
// https://msdn.microsoft.com/en-us/library/ff770589.aspx

__declspec(dllimport)
DOUBLE	CDECL	_CIcos(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/ff770580.aspx

__declspec(dllimport)
DOUBLE	CDECL	_CIlog(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/ff770583.aspx

__declspec(dllimport)
DOUBLE	CDECL	_CIpow(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/ff770597.aspx

__declspec(dllimport)
DOUBLE	CDECL	_CIsin(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/ff770588.aspx

__declspec(dllimport)
DOUBLE	CDECL	_CIsqrt(DOUBLE dbl);
#endif // _WIN64

__declspec(dllimport)
LONGLONG	CDECL	_ftol(DOUBLE dbl);
#endif

// https://msdn.microsoft.com/en-us/library/a6xkkxfz.aspx

__declspec(dllimport)
LPVOID	CDECL	_lfind(LPCVOID key, LPCVOID base, LPDWORD count, DWORD width, LONG (CDECL *compare) (LPCVOID left, LPCVOID right));

// https://msdn.microsoft.com/en-us/library/e737s6tf.aspx

__declspec(dllimport)
VOID	CDECL	_splitpath(LPCSTR path, LPSTR drive, LPSTR directory, LPSTR filename, LPSTR extension);

// https://msdn.microsoft.com/en-us/library/w0k41tbs.aspx

__declspec(dllimport)
LPVOID	CDECL	bsearch(LPCVOID key, LPCVOID base, SIZE_T count, SIZE_T width, LONG (CDECL *compare) (LPCVOID left, LPCVOID right));

#if 0	// NOTE: see CTYPE.H
// https://msdn.microsoft.com/en-us/library/k84c0490.aspx

__declspec(dllimport)
LONG	CDECL	isalnum(CHAR chr);

// https://msdn.microsoft.com/en-us/library/xt82b8z8.aspx

__declspec(dllimport)
LONG	CDECL	isalpha(CHAR chr);

// https://msdn.microsoft.com/en-us/library/1ad6a790.aspx

__declspec(dllimport)
LONG	CDECL	iscntrl(CHAR chr);

// https://msdn.microsoft.com/en-us/library/fcc4ksh8.aspx

__declspec(dllimport)
LONG	CDECL	isdigit(CHAR chr);

// https://msdn.microsoft.com/en-us/library/wsfaff19.aspx

__declspec(dllimport)
LONG	CDECL	isgraph(CHAR chr);

// https://msdn.microsoft.com/en-us/library/1z2s6by9.aspx

__declspec(dllimport)
LONG	CDECL	islower(CHAR chr);

// https://msdn.microsoft.com/en-us/library/ewx8s4kw.aspx

__declspec(dllimport)
LONG	CDECL	isprint(CHAR chr);

// https://msdn.microsoft.com/en-us/library/0ce6f2zc.aspx

__declspec(dllimport)
LONG	CDECL	ispunct(CHAR chr);

// https://msdn.microsoft.com/en-us/library/y13z34da.aspx

__declspec(dllimport)
LONG	CDECL	isspace(CHAR chr);

// https://msdn.microsoft.com/en-us/library/htd1kwta.aspx

__declspec(dllimport)
LONG	CDECL	isupper(CHAR chr);

// https://msdn.microsoft.com/en-us/library/t6h81hwd.aspx

__declspec(dllimport)
LONG	CDECL	isxdigit(CHAR chr);

// https://msdn.microsoft.com/en-us/library/xt82b8z8.aspx

__declspec(dllimport)
LONG	CDECL	iswalpha(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/czece495.aspx

__declspec(dllimport)
LONG	CDECL	iswctype(WCHAR chr, LONG desc);

// https://msdn.microsoft.com/en-us/library/fcc4ksh8.aspx

__declspec(dllimport)
LONG	CDECL	iswdigit(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/1z2s6by9.aspx

__declspec(dllimport)
LONG	CDECL	iswlower(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/y13z34da.aspx

__declspec(dllimport)
LONG	CDECL	iswspace(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/t6h81hwd.aspx

__declspec(dllimport)
LONG	CDECL	iswxdigit(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/2337tekx.aspx

__declspec(dllimport)
LPVOID	CDECL	_memccpy(LPVOID destination, LPCVOID source, CHAR chr, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/28e13y37.aspx

__declspec(dllimport)
LONG	CDECL	_memicmp(LPCVOID left, LPCVOID right, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

__declspec(dllimport)
LPVOID	CDECL	memchr(LPCVOID buffer, CHAR chr, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

__declspec(dllimport)
LONG	CDECL	memcmp(LPCVOID left, LPCVOID right, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

__declspec(dllimport)
LPVOID	CDECL	memcpy(LPVOID destination, LPCVOID source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/8k35d1fx.aspx

__declspec(dllimport)
LPVOID	CDECL	memmove(LPVOID destination, LPCVOID source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

__declspec(dllimport)
LPVOID	CDECL	memset(LPVOID destination, CHAR chr, SIZE_T count);
#endif

// https://msdn.microsoft.com/en-us/library/kw1ee768.aspx

__declspec(dllimport)
INT	CDECL	abs(INT integer);

// https://msdn.microsoft.com/en-us/library/88c36t42.aspx

__declspec(dllimport)
DOUBLE	CDECL	atan(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/atdhw2dx.aspx

__declspec(dllimport)
DOUBLE	CDECL	ceil(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/ydcbat90.aspx

__declspec(dllimport)
DOUBLE	CDECL	cos(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/18z15bk0.aspx

__declspec(dllimport)
DOUBLE	CDECL	fabs(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/x39715t6.aspx

__declspec(dllimport)
DOUBLE	CDECL	floor(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/kw1ee768.aspx

__declspec(dllimport)
LONG	CDECL	labs(LONG lng);

// https://msdn.microsoft.com/en-us/library/t63833dz.aspx

__declspec(dllimport)
DOUBLE	CDECL	log(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/dt5dakze.aspx

__declspec(dllimport)
DOUBLE	CDECL	pow(DOUBLE base, DOUBLE exponent);

// https://msdn.microsoft.com/en-us/library/wkbss70y.aspx

__declspec(dllimport)
DOUBLE	CDECL	sin(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/f1xa99e6.aspx

__declspec(dllimport)
DOUBLE	CDECL	sqrt(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/twws9xxw.aspx

__declspec(dllimport)
DOUBLE	CDECL	tan(DOUBLE dbl);

// https://msdn.microsoft.com/en-us/library/zes7xw0h.aspx

__declspec(dllimport)
VOID	CDECL	qsort(LPVOID base, SIZE_T count, SIZE_T width, LONG (CDECL *compare) (LPCVOID left, LPCVOID right));

#if 0	// NOTE: see CTYPE.H
// https://msdn.microsoft.com/en-us/library/8h19t214.aspx

__declspec(dllimport)
CHAR	CDECL	_tolower(CHAR chr);

// https://msdn.microsoft.com/en-us/library/45119yx3.aspx

__declspec(dllimport)
CHAR	CDECL	_toupper(CHAR chr);

// https://msdn.microsoft.com/en-us/library/8h19t214.aspx

__declspec(dllimport)
CHAR	CDECL	tolower(CHAR chr);

// https://msdn.microsoft.com/en-us/library/45119yx3.aspx

__declspec(dllimport)
CHAR	CDECL	toupper(CHAR chr);

// https://msdn.microsoft.com/en-us/library/8h19t214.aspx

__declspec(dllimport)
WCHAR	CDECL	towlower(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/45119yx3.aspx

__declspec(dllimport)
WCHAR	CDECL	towupper(WCHAR chr);

// https://msdn.microsoft.com/en-us/library/k59z8dwe.aspx
#endif

#if 0	// NOTE: see STRING.H
__declspec(dllimport)
LONG	CDECL	_strcmpi(LPCSTR left, LPCSTR right);

// https://msdn.microsoft.com/en-us/library/k59z8dwe.aspx

__declspec(dllimport)
LONG	CDECL	_stricmp(LPCSTR left, LPCSTR right);

// https://msdn.microsoft.com/en-us/library/hkxwh33z.aspx

__declspec(dllimport)
LPSTR	CDECL	_strlwr(LPSTR string);

// https://msdn.microsoft.com/en-us/library/chd90w8e.aspx

__declspec(dllimport)
LONG	CDECL	_strnicmp(LPCSTR left, LPCSTR right, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/sch3dy08.aspx

__declspec(dllimport)
LPSTR	CDECL	_strupr(LPSTR string);

// https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

__declspec(dllimport)
LPSTR	CDECL	strcat(LPSTR destination, LPCSTR source);

// https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

__declspec(dllimport)
LPSTR	CDECL	strchr(LPCSTR string, CHAR chr);

// https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

__declspec(dllimport)
LONG	CDECL	strcmp(LPCSTR left, LPCSTR right);

// https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

__declspec(dllimport)
LPSTR	CDECL	strcpy(LPSTR destination, LPCSTR source);

// https://msdn.microsoft.com/en-us/library/xe8sk0x7.aspx

__declspec(dllimport)
SIZE_T	CDECL	strcspn(LPCSTR string, LPCSTR set);

// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

__declspec(dllimport)
SIZE_T	CDECL	strlen(LPCSTR string);

// https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

__declspec(dllimport)
LPSTR	CDECL	strncat(LPSTR destination, LPCSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

__declspec(dllimport)
LONG	CDECL	strncmp(LPCSTR left, LPCSTR right, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

__declspec(dllimport)
LPSTR	CDECL	strncpy(LPSTR destination, LPCSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/217yyhy9.aspx

__declspec(dllimport)
LPSTR	CDECL	strpbrk(LPCSTR string, LPCSTR set);

// https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

__declspec(dllimport)
LPSTR	CDECL	strrchr(LPCSTR string, CHAR chr);

// https://msdn.microsoft.com/en-us/library/kthzzay9.aspx

__declspec(dllimport)
SIZE_T	CDECL	strspn(LPCSTR string, LPCSTR set);

// https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

__declspec(dllimport)
LPSTR	CDECL	strstr(LPCSTR string, LPCSTR substring);
#endif

// https://msdn.microsoft.com/en-us/library/w4z2wdyc.aspx

__declspec(dllimport)
LONG	CDECL	strtol(LPCSTR string, LPSTR *, LONG base);

// https://msdn.microsoft.com/en-us/library/5k9xb7x1.aspx

__declspec(dllimport)
DWORD	CDECL	strtoul(LPCSTR string, LPSTR *, LONG base);

// https://msdn.microsoft.com/en-us/library/2ts7cx93.aspx

__declspec(dllimport)
LONG	CDECL	_snprintf(LPSTR buffer, SIZE_T count, LPCSTR format, ...);

// https://msdn.microsoft.com/en-us/library/ybk95axf.aspx

__declspec(dllimport)
LONG	CDECL	sprintf(LPSTR buffer, LPCSTR format, ...);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPSTR	CDECL	_itoa(INT value, LPSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPWSTR	CDECL	_itow(INT value, LPWSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/6y51bcy2.aspx

__declspec(dllimport)
LPSTR	CDECL	_ltoa(LONG value, LPSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/6y51bcy2.aspx

__declspec(dllimport)
LPWSTR	CDECL	_ltow(LONG value, LPWSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/d5t0h1wd.aspx

__declspec(dllimport)
LPSTR	CDECL	_ultoa(ULONG value, LPSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/d5t0h1wd.aspx

__declspec(dllimport)
LPWSTR	CDECL	_ultow(ULONG value, LPWSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPSTR	CDECL	_i64toa(LONGLONG value, LPSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPWSTR	CDECL	_i64tow(LONGLONG value, LPWSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPSTR	CDECL	_ui64toa(ULONGLONG value, LPSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yakksftt.aspx

__declspec(dllimport)
LPWSTR	CDECL	_ui64tow(ULONGLONG value, LPWSTR buffer, LONG radix);

// https://msdn.microsoft.com/en-us/library/yd5xkb5c.aspx

__declspec(dllimport)
INT	CDECL	atoi(LPCSTR string);

// https://msdn.microsoft.com/en-us/library/ts4a5f28.aspx

__declspec(dllimport)
LONG	CDECL	atol(LPCSTR string);

// https://msdn.microsoft.com/en-us/library/czcad93k.aspx

__declspec(dllimport)
LONGLONG	CDECL	_atoi64(LPCSTR string);

// https://msdn.microsoft.com/en-us/library/yd5xkb5c.aspx

__declspec(dllimport)
INT	CDECL	_wtoi(LPCWSTR string);

// https://msdn.microsoft.com/en-us/library/czcad93k.aspx

__declspec(dllimport)
LONGLONG	CDECL	_wtoi64(LPCWSTR string);

// https://msdn.microsoft.com/en-us/library/ts4a5f28.aspx

__declspec(dllimport)
LONG	CDECL	_wtol(LPCWSTR string);

// https://msdn.microsoft.com/en-us/library/zkx076cy.aspx

__declspec(dllimport)
LONG	CDECL	sscanf(LPCSTR buffer, LPCSTR format, ...);

// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

__declspec(dllimport)
LONG	CDECL	_vsnprintf(LPSTR buffer, SIZE_T count, LPCSTR format, va_list argptr);

// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

__declspec(dllimport)
LONG	CDECL	vsprintf(LPSTR buffer, LPCSTR format, va_list argptr);

#if 0	// NOTE: see STRING.H
// https://msdn.microsoft.com/en-us/library/k59z8dwe.aspx

__declspec(dllimport)
LONG	CDECL	_wcsicmp(LPCWSTR left, LPCWSTR right);

// https://msdn.microsoft.com/en-us/library/hkxwh33z.aspx

__declspec(dllimport)
LPWSTR	CDECL	_wcslwr(LPWSTR string);

// https://msdn.microsoft.com/en-us/library/chd90w8e.aspx

__declspec(dllimport)
LONG	CDECL	_wcsnicmp(LPCWSTR left, LPCWSTR right, SIZE_T count);
#endif

// https://msdn.microsoft.com/en-us/library/85zk715d.aspx

__declspec(dllimport)
QWORD	CDECL	_wcstoui64(LPCWSTR string, LPWSTR *, LONG base);

#if 0	// NOTE: see STRING.H
// https://msdn.microsoft.com/en-us/library/sch3dy08.aspx

__declspec(dllimport)
LPWSTR	CDECL	_wcsupr(LPWSTR string);

// https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcscat(LPWSTR destination, LPCWSTR source);

// https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcschr(LPCWSTR string, WCHAR chr);

// https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

__declspec(dllimport)
LONG	CDECL	wcscmp(LPCWSTR left, LPCWSTR right);

// https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcscpy(LPWSTR destination, LPCWSTR source);

// https://msdn.microsoft.com/en-us/library/xe8sk0x7.aspx

__declspec(dllimport)
SIZE_T	CDECL	wcscspn(LPCWSTR string, LPCWSTR set);

// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

__declspec(dllimport)
SIZE_T	CDECL	wcslen(LPCWSTR string);

// https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcsncat(LPWSTR destination, LPCWSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

__declspec(dllimport)
LONG	CDECL	wcsncmp(LPCWSTR left, LPCWSTR right, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcsncpy(LPWSTR destination, LPCWSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/217yyhy9.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcspbrk(LPCWSTR string, LPCWSTR set);

// https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcsrchr(LPCWSTR string, WCHAR chr);

// https://msdn.microsoft.com/en-us/library/kthzzay9.aspx

__declspec(dllimport)
SIZE_T	CDECL	wcsspn(LPCWSTR string, LPCWSTR set);

// https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

__declspec(dllimport)
LPWSTR	CDECL	wcsstr(LPCWSTR string, LPCWSTR substring);
#endif

// https://msdn.microsoft.com/en-us/library/k1f9b8cy.aspx

__declspec(dllimport)
SIZE_T	CDECL	mbstowcs(LPWSTR destination, LPCSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/5d7tc9zw.aspx

__declspec(dllimport)
SIZE_T	CDECL	wcstombs(LPSTR destination, LPCWSTR source, SIZE_T count);

// https://msdn.microsoft.com/en-us/library/w4z2wdyc.aspx

__declspec(dllimport)
LONG	CDECL	wcstol(LPCWSTR string, LPWSTR *, LONG base);

// https://msdn.microsoft.com/en-us/library/5k9xb7x1.aspx

__declspec(dllimport)
ULONG	CDECL	wcstoul(LPCWSTR string, LPWSTR *, LONG base);

// https://msdn.microsoft.com/en-us/library/2ts7cx93.aspx

__declspec(dllimport)
LONG	CDECL	_snwprintf(LPWSTR buffer, SIZE_T count, LPCWSTR format, ...);

// https://msdn.microsoft.com/en-us/library/ybk95axf.aspx

__declspec(dllimport)
LONG	CDECL	_swprintf(LPWSTR buffer, LPCWSTR format, ...);

// https://msdn.microsoft.com/en-us/library/ybk95axf.aspx

__declspec(dllimport)
LONG	CDECL	swprintf(LPWSTR buffer, LPCWSTR format, ...);

// https://msdn.microsoft.com/en-us/library/w05tbk72.aspx

__declspec(dllimport)
LONG	CDECL	_vscwprintf(LPCWSTR format, va_list argptr);

// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

__declspec(dllimport)
LONG	CDECL	_vsnwprintf(LPWSTR buffer, SIZE_T count, LPCWSTR format, va_list argptr);

// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

__declspec(dllimport)
LONG	CDECL	_vswprintf(LPWSTR buffer, LPCWSTR format, va_list argptr);

#define PROCESSOR_FEATURE_MAX		64
#define MAX_WOW64_SHARED_ENTRIES	16

#pragma pack(show)
#pragma pack(push, 4)

typedef	struct	_KSYSTEM_TIME
{
	DWORD	LowPart;
	LONG	High1Time;
	LONG	High2Time;
} KSYSTEM_TIME;

typedef	enum	_NT_PRODUCT_TYPE
{
	NtProductWinNt    = 1,
	NtProductLanManNt = 2,
	NtProductServer   = 3
} NT_PRODUCT_TYPE;

typedef	enum	_ALTERNATIVE_ARCHITECTURE_TYPE
{
	StandardDesign  = 0,
	NEC98x86        = 1,
	EndAlternatives = 2
} ALTERNATIVE_ARCHITECTURE_TYPE;

typedef	struct	_KUSER_SHARED_DATA		// located at address 0x7FFE0000
{
	DWORD		TickCountLowDeprecated;
	DWORD		TickCountMultiplier;
	volatile	KSYSTEM_TIME	InterruptTime;
	volatile	KSYSTEM_TIME	SystemTime;
	volatile	KSYSTEM_TIME	TimeZoneBias;
	WORD		ImageNumberLow;
	WORD		ImageNumberHigh;
	WCHAR		NtSystemRoot[MAX_PATH];
	DWORD		MaxStackTraceDepth;
	DWORD		CryptoExponent;
	DWORD		TimeZoneId;
	DWORD		LargePageMinimum;
#if _WIN32_WINNT < 0x0602
	DWORD		Reserved2[7];
#else
	DWORD		AitSamplingValue;
	DWORD		AppCompatFlag;
	QWORD		RNGSeedVersion;
	DWORD		GlobalValidationRunLevel;
	volatile	DWORD		TimeZoneBiasStamp;
#if _WIN32_WINNT < 0x0A00
	DWORD		Reserved2[1];
#else
	DWORD		NtBuildNumber;
#endif
#endif
	NT_PRODUCT_TYPE	NtProductType;
	BOOLEAN		ProductTypeIsValid;
	BYTE		Reserved0;
	WORD		NativeProcessorArchitecture;
	DWORD		NtMajorVersion;
	DWORD		NtMinorVersion;
	BOOLEAN		ProcessorFeatures[PROCESSOR_FEATURE_MAX];
	DWORD		Reserved1;
	DWORD		Reserved3;
	volatile	DWORD		TimeSlip;
	ALTERNATIVE_ARCHITECTURE_TYPE	AlternativeArchitecture;
#if _WIN32_WINNT < 0x0A00
	DWORD		AltArchitecturePad[1];
#else
	DWORD		BootId;
#endif
	LARGE_INTEGER	SystemExpirationDate;
	DWORD		SuiteMask;
	BOOLEAN		KdDebuggerEnabled;
	BYTE		NXSupportPolicy;
//	BYTE		Padding2[2];
	volatile	DWORD		ActiveConsoleId;
	volatile	DWORD		DismountCount;
	DWORD		ComPlusPackage;
	DWORD		LastSystemRITEventTickCount;
	DWORD		NumberOfPhysicalPages;
	BOOLEAN		SafeBootMode;
	BYTE		TscQpcData;
	BYTE		TscQpcPad[2];
#if _WIN32_WINNT < 0x0600
	DWORD		TraceLogging;
#else
	DWORD		SharedDataFlags;
#endif
	DWORD		DataFlagsPad[1];
	QWORD		TestRetInstruction;
	DWORD		SystemCall;
	DWORD		SystemCallReturn;
	QWORD		SystemCallPad[3];
#pragma warning(disable: 4201)
	union		// offset = 0x320
	{
		volatile	KSYSTEM_TIME	TickCount;
		volatile	QWORD	TickCountQuad;
		struct
		{
			DWORD	ReservedTickCountOverlay[3];
			DWORD	TickCountPad[1];
		};
	};
#pragma warning(default: 4201)
	DWORD		Cookie;
	DWORD		CookiePad[1];
	QWORD		ConsoleSessionForegroundProcessId;
	DWORD		Wow64SharedInformation[MAX_WOW64_SHARED_ENTRIES];
#if _WIN32_WINNT >= 0x0600
	WORD		UserModeGlobalLogger[16];
	DWORD		ImageFileExecutionOptions;
	DWORD		LangGenerationCount;
	QWORD		Reserved5;
	volatile	QWORD		InterruptTimeBias;
#if _WIN32_WINNT >= 0x0601
	volatile	QWORD		TscQpcBias;
	volatile	DWORD		ActiveProcessorCount;
	volatile	WORD		ActiveGroupCount;
	WORD		Reserved4;
	volatile	DWORD		AitSamplingValue;
	volatile	DWORD		AppCompatFlag;
	QWORD		SystemDllNativeRelocation;
	DWORD		SystemDllWowRelocation;
	DWORD		XStatePad[1];
	XSTATE_CONFIGURATION	XState;
#endif
#endif
} KUSER_SHARED_DATA;

#pragma pack(pop)

LPVOID	NTAPI	NtDllFunctionAddress(LPCSTR FunctionName)
{
	IMAGE_DOS_HEADER	*DosHeader;

	// https://msdn.microsoft.com/en-us/library/ms680336.aspx

	IMAGE_NT_HEADERS	*NtHeader;

	// https://msdn.microsoft.com/en-us/library/ms680305.aspx

	IMAGE_DATA_DIRECTORY	*DataDirectory;

	IMAGE_EXPORT_DIRECTORY	*ExportDirectory;

	// https://msdn.microsoft.com/en-us/library/aa813708.aspx

	LDR_DATA_TABLE_ENTRY	*DataTableEntry;

	// https://msdn.microsoft.com/en-us/library/ff554296.aspx

	LIST_ENTRY	*Head, *Link;

	LPBYTE	DllBase;
	LPWORD	Ordinals;
	LPDWORD	Functions;
	LPDWORD	Names;
	DWORD	Index;
#ifndef LINEAR
	DWORD	Low, High;
	INT	Delta;
#endif
	// https://msdn.microsoft.com/en-us/library/hh285210.aspx
	// https://msdn.microsoft.com/en-us/library/ms686708.aspx
	// https://msdn.microsoft.com/en-us/library/aa813708.aspx

	for (Head = &NtCurrentTeb()->ProcessEnvironmentBlock->Ldr->InLoadOrderModuleList,
	     Link = Head->Flink; Link != Head; Link = Link->Flink)
	{
		DataTableEntry = (LDR_DATA_TABLE_ENTRY *) Link;

		if (DataTableEntry->BaseDllName.Length != sizeof(L"ntdll.dll") - sizeof(L""))
			continue;

		Index = 0UL;

		do
			if (DataTableEntry->BaseDllName.Buffer[Index] != L"ntdll.dll"[Index])
				break;
		while (++Index < sizeof("ntdll.dll") - sizeof(""));

		if (Index != sizeof("ntdll.dll") - sizeof(""))
			continue;

		DllBase = (LPBYTE) DataTableEntry->DllBase;
		DosHeader = (IMAGE_DOS_HEADER *) DllBase;
		NtHeader = (IMAGE_NT_HEADERS *) (DllBase + DosHeader->e_lfanew);
		DataDirectory = (IMAGE_DATA_DIRECTORY *) &NtHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT];
		ExportDirectory = (IMAGE_EXPORT_DIRECTORY *) (DllBase + DataDirectory->VirtualAddress);

		Functions = (LPDWORD) (DllBase + ExportDirectory->AddressOfFunctions);
		Ordinals = (LPWORD) (DllBase + ExportDirectory->AddressOfNameOrdinals);
		Names = (LPDWORD) (DllBase + ExportDirectory->AddressOfNames);
#ifdef LINEAR
		Index = 0UL;

		do
			if (strcmp(FunctionName, (LPCSTR) (DllBase + Names[Index])) == 0)
				return (LPVOID) (DllBase + Functions[Ordinals[Index]]);
		while (++Index < ExportDirectory->NumberOfNames);
#else
		High = ExportDirectory->NumberOfNames;
		Low = 0UL;

		do
		{
			Index = (Low + High) >> 1;
			Delta = strcmp(FunctionName, (LPCSTR) (DllBase + Names[Index]));

			if (Delta < 0)
				High = Index;
			else if (Delta > 0)
				Low = ++Index;
			else
				return (LPVOID) (DllBase + Functions[Ordinals[Index]]);
		}
		while (Low < High);
#endif
		break;
	}

	return NULL;
}

#if 0
#pragma comment(linker, "/ENTRY:NtProcessStartup")
#pragma comment(linker, "/SUBSYSTEM:NATIVE")

__declspec(noreturn)
VOID	NTAPI	NtProcessStartup(PEB *ProcessEnvironmentBlock)
{
	static	const	UNICODE_STRING	usImagePath = RTL_CONSTANT_STRING(L"\r\nImage Path:\t");
	static	const	UNICODE_STRING	usCmdLine = RTL_CONSTANT_STRING(L"\r\nCommand Line:\t");
	static	const	UNICODE_STRING	usCRLF = RTL_CONSTANT_STRING(L"\r\n");

	LARGE_INTEGER	li;

	LONG	ntStatus = RtlNormalizeProcessParams(ProcessEnvironmentBlock->ProcessParameters);

	if (ntStatus >= 0L)
	{
		ntStatus = NtDisplayString(&usImagePath);
		ntStatus = NtDisplayString(&ProcessEnvironmentBlock->ProcessParameters->ImagePathName);
		ntStatus = NtDisplayString(&usCmdLine);
		ntStatus = NtDisplayString(&ProcessEnvironmentBlock->ProcessParameters->CommandLine);
		ntStatus = NtDisplayString(&usCRLF);

		li.QuadPart = -120000000LL;

		ntStatus = NtDelayExecution(FALSE, &li);
	}
#if 1
	ntStatus = NtTerminateProcess(NtCurrentProcess(), ntStatus);
#else
	ntStatus = NtTerminateThread(NtCurrentThread(), ntStatus);
#endif
}
#endif
#endif // __NTDLL_H__
